    /* Autogenerated file -- do not edit
     * generated by gen_enum_to_str.py
     *
     * Copyright © 2017 Intel Corporation
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
     */

    #ifndef MESA_VK_ENUM_TO_STR_H
    #define MESA_VK_ENUM_TO_STR_H

    #include <vulkan/vulkan.h>
    #include <vulkan/vk_android_native_buffer.h>

    #ifdef __cplusplus
    extern "C" {
    #endif

    const char * vk_AccelerationStructureBuildTypeKHR_to_str(VkAccelerationStructureBuildTypeKHR input);
    const char * vk_AccelerationStructureCompatibilityKHR_to_str(VkAccelerationStructureCompatibilityKHR input);
    const char * vk_AccelerationStructureMemoryRequirementsTypeNV_to_str(VkAccelerationStructureMemoryRequirementsTypeNV input);
    const char * vk_AccelerationStructureMotionInstanceTypeNV_to_str(VkAccelerationStructureMotionInstanceTypeNV input);
    const char * vk_AccelerationStructureTypeKHR_to_str(VkAccelerationStructureTypeKHR input);
    const char * vk_AttachmentLoadOp_to_str(VkAttachmentLoadOp input);
    const char * vk_AttachmentStoreOp_to_str(VkAttachmentStoreOp input);
    const char * vk_BlendFactor_to_str(VkBlendFactor input);
    const char * vk_BlendOp_to_str(VkBlendOp input);
    const char * vk_BlendOverlapEXT_to_str(VkBlendOverlapEXT input);
    const char * vk_BlockMatchWindowCompareModeQCOM_to_str(VkBlockMatchWindowCompareModeQCOM input);
    const char * vk_BorderColor_to_str(VkBorderColor input);
    const char * vk_BuildAccelerationStructureModeKHR_to_str(VkBuildAccelerationStructureModeKHR input);
    const char * vk_BuildMicromapModeEXT_to_str(VkBuildMicromapModeEXT input);
    const char * vk_ChromaLocation_to_str(VkChromaLocation input);
    const char * vk_CoarseSampleOrderTypeNV_to_str(VkCoarseSampleOrderTypeNV input);
    const char * vk_ColorSpaceKHR_to_str(VkColorSpaceKHR input);
    const char * vk_CommandBufferLevel_to_str(VkCommandBufferLevel input);
    const char * vk_CompareOp_to_str(VkCompareOp input);
    const char * vk_ComponentSwizzle_to_str(VkComponentSwizzle input);
    const char * vk_ComponentTypeKHR_to_str(VkComponentTypeKHR input);
    const char * vk_ConservativeRasterizationModeEXT_to_str(VkConservativeRasterizationModeEXT input);
    const char * vk_CopyAccelerationStructureModeKHR_to_str(VkCopyAccelerationStructureModeKHR input);
    const char * vk_CopyMicromapModeEXT_to_str(VkCopyMicromapModeEXT input);
    const char * vk_CoverageModulationModeNV_to_str(VkCoverageModulationModeNV input);
    const char * vk_CoverageReductionModeNV_to_str(VkCoverageReductionModeNV input);
    const char * vk_CubicFilterWeightsQCOM_to_str(VkCubicFilterWeightsQCOM input);
    const char * vk_DebugReportObjectTypeEXT_to_str(VkDebugReportObjectTypeEXT input);
    const char * vk_DepthBiasRepresentationEXT_to_str(VkDepthBiasRepresentationEXT input);
    const char * vk_DescriptorType_to_str(VkDescriptorType input);
    const char * vk_DescriptorUpdateTemplateType_to_str(VkDescriptorUpdateTemplateType input);
    const char * vk_DeviceAddressBindingTypeEXT_to_str(VkDeviceAddressBindingTypeEXT input);
    const char * vk_DeviceEventTypeEXT_to_str(VkDeviceEventTypeEXT input);
    const char * vk_DeviceFaultAddressTypeEXT_to_str(VkDeviceFaultAddressTypeEXT input);
    const char * vk_DeviceFaultVendorBinaryHeaderVersionEXT_to_str(VkDeviceFaultVendorBinaryHeaderVersionEXT input);
    const char * vk_DeviceMemoryReportEventTypeEXT_to_str(VkDeviceMemoryReportEventTypeEXT input);
    const char * vk_DirectDriverLoadingModeLUNARG_to_str(VkDirectDriverLoadingModeLUNARG input);
    const char * vk_DiscardRectangleModeEXT_to_str(VkDiscardRectangleModeEXT input);
    const char * vk_DisplayEventTypeEXT_to_str(VkDisplayEventTypeEXT input);
    const char * vk_DisplayPowerStateEXT_to_str(VkDisplayPowerStateEXT input);
    const char * vk_DriverId_to_str(VkDriverId input);
    const char * vk_DynamicState_to_str(VkDynamicState input);
    const char * vk_Filter_to_str(VkFilter input);
    const char * vk_Format_to_str(VkFormat input);
    const char * vk_FragmentShadingRateCombinerOpKHR_to_str(VkFragmentShadingRateCombinerOpKHR input);
    const char * vk_FragmentShadingRateNV_to_str(VkFragmentShadingRateNV input);
    const char * vk_FragmentShadingRateTypeNV_to_str(VkFragmentShadingRateTypeNV input);
    const char * vk_FrontFace_to_str(VkFrontFace input);
#ifdef VK_USE_PLATFORM_WIN32_KHR
    const char * vk_FullScreenExclusiveEXT_to_str(VkFullScreenExclusiveEXT input);
#endif
    const char * vk_GeometryTypeKHR_to_str(VkGeometryTypeKHR input);
    const char * vk_ImageLayout_to_str(VkImageLayout input);
    const char * vk_ImageTiling_to_str(VkImageTiling input);
    const char * vk_ImageType_to_str(VkImageType input);
    const char * vk_ImageViewType_to_str(VkImageViewType input);
    const char * vk_IndexType_to_str(VkIndexType input);
    const char * vk_IndirectCommandsTokenTypeNV_to_str(VkIndirectCommandsTokenTypeNV input);
    const char * vk_InternalAllocationType_to_str(VkInternalAllocationType input);
    const char * vk_LatencyMarkerNV_to_str(VkLatencyMarkerNV input);
    const char * vk_LayerSettingTypeEXT_to_str(VkLayerSettingTypeEXT input);
    const char * vk_LayeredDriverUnderlyingApiMSFT_to_str(VkLayeredDriverUnderlyingApiMSFT input);
    const char * vk_LineRasterizationModeEXT_to_str(VkLineRasterizationModeEXT input);
    const char * vk_LogicOp_to_str(VkLogicOp input);
    const char * vk_MemoryOverallocationBehaviorAMD_to_str(VkMemoryOverallocationBehaviorAMD input);
    const char * vk_MicromapTypeEXT_to_str(VkMicromapTypeEXT input);
    const char * vk_ObjectType_to_str(VkObjectType input);
    const char * vk_OpacityMicromapFormatEXT_to_str(VkOpacityMicromapFormatEXT input);
    const char * vk_OpacityMicromapSpecialIndexEXT_to_str(VkOpacityMicromapSpecialIndexEXT input);
    const char * vk_OpticalFlowPerformanceLevelNV_to_str(VkOpticalFlowPerformanceLevelNV input);
    const char * vk_OpticalFlowSessionBindingPointNV_to_str(VkOpticalFlowSessionBindingPointNV input);
    const char * vk_OutOfBandQueueTypeNV_to_str(VkOutOfBandQueueTypeNV input);
    const char * vk_PerformanceConfigurationTypeINTEL_to_str(VkPerformanceConfigurationTypeINTEL input);
    const char * vk_PerformanceCounterScopeKHR_to_str(VkPerformanceCounterScopeKHR input);
    const char * vk_PerformanceCounterStorageKHR_to_str(VkPerformanceCounterStorageKHR input);
    const char * vk_PerformanceCounterUnitKHR_to_str(VkPerformanceCounterUnitKHR input);
    const char * vk_PerformanceOverrideTypeINTEL_to_str(VkPerformanceOverrideTypeINTEL input);
    const char * vk_PerformanceParameterTypeINTEL_to_str(VkPerformanceParameterTypeINTEL input);
    const char * vk_PerformanceValueTypeINTEL_to_str(VkPerformanceValueTypeINTEL input);
    const char * vk_PhysicalDeviceType_to_str(VkPhysicalDeviceType input);
    const char * vk_PipelineBindPoint_to_str(VkPipelineBindPoint input);
    const char * vk_PipelineCacheHeaderVersion_to_str(VkPipelineCacheHeaderVersion input);
    const char * vk_PipelineExecutableStatisticFormatKHR_to_str(VkPipelineExecutableStatisticFormatKHR input);
    const char * vk_PipelineRobustnessBufferBehaviorEXT_to_str(VkPipelineRobustnessBufferBehaviorEXT input);
    const char * vk_PipelineRobustnessImageBehaviorEXT_to_str(VkPipelineRobustnessImageBehaviorEXT input);
    const char * vk_PointClippingBehavior_to_str(VkPointClippingBehavior input);
    const char * vk_PolygonMode_to_str(VkPolygonMode input);
    const char * vk_PresentModeKHR_to_str(VkPresentModeKHR input);
    const char * vk_PrimitiveTopology_to_str(VkPrimitiveTopology input);
    const char * vk_ProvokingVertexModeEXT_to_str(VkProvokingVertexModeEXT input);
    const char * vk_QueryPoolSamplingModeINTEL_to_str(VkQueryPoolSamplingModeINTEL input);
    const char * vk_QueryResultStatusKHR_to_str(VkQueryResultStatusKHR input);
    const char * vk_QueryType_to_str(VkQueryType input);
    const char * vk_QueueGlobalPriorityKHR_to_str(VkQueueGlobalPriorityKHR input);
    const char * vk_RasterizationOrderAMD_to_str(VkRasterizationOrderAMD input);
    const char * vk_RayTracingInvocationReorderModeNV_to_str(VkRayTracingInvocationReorderModeNV input);
    const char * vk_RayTracingShaderGroupTypeKHR_to_str(VkRayTracingShaderGroupTypeKHR input);
    const char * vk_Result_to_str(VkResult input);
    const char * vk_SamplerAddressMode_to_str(VkSamplerAddressMode input);
    const char * vk_SamplerMipmapMode_to_str(VkSamplerMipmapMode input);
    const char * vk_SamplerReductionMode_to_str(VkSamplerReductionMode input);
    const char * vk_SamplerYcbcrModelConversion_to_str(VkSamplerYcbcrModelConversion input);
    const char * vk_SamplerYcbcrRange_to_str(VkSamplerYcbcrRange input);
    const char * vk_ScopeKHR_to_str(VkScopeKHR input);
    const char * vk_SemaphoreType_to_str(VkSemaphoreType input);
    const char * vk_ShaderCodeTypeEXT_to_str(VkShaderCodeTypeEXT input);
    const char * vk_ShaderFloatControlsIndependence_to_str(VkShaderFloatControlsIndependence input);
    const char * vk_ShaderGroupShaderKHR_to_str(VkShaderGroupShaderKHR input);
    const char * vk_ShaderInfoTypeAMD_to_str(VkShaderInfoTypeAMD input);
    const char * vk_ShadingRatePaletteEntryNV_to_str(VkShadingRatePaletteEntryNV input);
    const char * vk_SharingMode_to_str(VkSharingMode input);
    const char * vk_StencilOp_to_str(VkStencilOp input);
    const char * vk_StructureType_to_str(VkStructureType input);
    const char * vk_SubpassContents_to_str(VkSubpassContents input);
    const char * vk_SubpassMergeStatusEXT_to_str(VkSubpassMergeStatusEXT input);
    const char * vk_SystemAllocationScope_to_str(VkSystemAllocationScope input);
    const char * vk_TessellationDomainOrigin_to_str(VkTessellationDomainOrigin input);
    const char * vk_TimeDomainKHR_to_str(VkTimeDomainKHR input);
    const char * vk_ValidationCacheHeaderVersionEXT_to_str(VkValidationCacheHeaderVersionEXT input);
    const char * vk_ValidationCheckEXT_to_str(VkValidationCheckEXT input);
    const char * vk_ValidationFeatureDisableEXT_to_str(VkValidationFeatureDisableEXT input);
    const char * vk_ValidationFeatureEnableEXT_to_str(VkValidationFeatureEnableEXT input);
    const char * vk_VendorId_to_str(VkVendorId input);
    const char * vk_VertexInputRate_to_str(VkVertexInputRate input);
    const char * vk_VideoEncodeTuningModeKHR_to_str(VkVideoEncodeTuningModeKHR input);
    const char * vk_ViewportCoordinateSwizzleNV_to_str(VkViewportCoordinateSwizzleNV input);

    const char * vk_AccelerationStructureCreateFlagBitsKHR_to_str(VkAccelerationStructureCreateFlagBitsKHR input);
    const char * vk_AccessFlagBits_to_str(VkAccessFlagBits input);
    const char * vk_AccessFlagBits2_to_str(VkAccessFlagBits2 input);
    const char * vk_AcquireProfilingLockFlagBitsKHR_to_str(VkAcquireProfilingLockFlagBitsKHR input);
    const char * vk_AttachmentDescriptionFlagBits_to_str(VkAttachmentDescriptionFlagBits input);
    const char * vk_BufferCreateFlagBits_to_str(VkBufferCreateFlagBits input);
    const char * vk_BufferUsageFlagBits_to_str(VkBufferUsageFlagBits input);
    const char * vk_BufferUsageFlagBits2KHR_to_str(VkBufferUsageFlagBits2KHR input);
    const char * vk_BuildAccelerationStructureFlagBitsKHR_to_str(VkBuildAccelerationStructureFlagBitsKHR input);
    const char * vk_BuildMicromapFlagBitsEXT_to_str(VkBuildMicromapFlagBitsEXT input);
    const char * vk_ColorComponentFlagBits_to_str(VkColorComponentFlagBits input);
    const char * vk_CommandBufferResetFlagBits_to_str(VkCommandBufferResetFlagBits input);
    const char * vk_CommandBufferUsageFlagBits_to_str(VkCommandBufferUsageFlagBits input);
    const char * vk_CommandPoolCreateFlagBits_to_str(VkCommandPoolCreateFlagBits input);
    const char * vk_CommandPoolResetFlagBits_to_str(VkCommandPoolResetFlagBits input);
    const char * vk_CompositeAlphaFlagBitsKHR_to_str(VkCompositeAlphaFlagBitsKHR input);
    const char * vk_ConditionalRenderingFlagBitsEXT_to_str(VkConditionalRenderingFlagBitsEXT input);
    const char * vk_CullModeFlagBits_to_str(VkCullModeFlagBits input);
    const char * vk_DebugReportFlagBitsEXT_to_str(VkDebugReportFlagBitsEXT input);
    const char * vk_DebugUtilsMessageSeverityFlagBitsEXT_to_str(VkDebugUtilsMessageSeverityFlagBitsEXT input);
    const char * vk_DebugUtilsMessageTypeFlagBitsEXT_to_str(VkDebugUtilsMessageTypeFlagBitsEXT input);
    const char * vk_DependencyFlagBits_to_str(VkDependencyFlagBits input);
    const char * vk_DescriptorBindingFlagBits_to_str(VkDescriptorBindingFlagBits input);
    const char * vk_DescriptorPoolCreateFlagBits_to_str(VkDescriptorPoolCreateFlagBits input);
    const char * vk_DescriptorSetLayoutCreateFlagBits_to_str(VkDescriptorSetLayoutCreateFlagBits input);
    const char * vk_DeviceAddressBindingFlagBitsEXT_to_str(VkDeviceAddressBindingFlagBitsEXT input);
    const char * vk_DeviceDiagnosticsConfigFlagBitsNV_to_str(VkDeviceDiagnosticsConfigFlagBitsNV input);
    const char * vk_DeviceGroupPresentModeFlagBitsKHR_to_str(VkDeviceGroupPresentModeFlagBitsKHR input);
    const char * vk_DeviceQueueCreateFlagBits_to_str(VkDeviceQueueCreateFlagBits input);
    const char * vk_DisplayPlaneAlphaFlagBitsKHR_to_str(VkDisplayPlaneAlphaFlagBitsKHR input);
    const char * vk_EventCreateFlagBits_to_str(VkEventCreateFlagBits input);
#ifdef VK_USE_PLATFORM_METAL_EXT
    const char * vk_ExportMetalObjectTypeFlagBitsEXT_to_str(VkExportMetalObjectTypeFlagBitsEXT input);
#endif
    const char * vk_ExternalFenceFeatureFlagBits_to_str(VkExternalFenceFeatureFlagBits input);
    const char * vk_ExternalFenceHandleTypeFlagBits_to_str(VkExternalFenceHandleTypeFlagBits input);
    const char * vk_ExternalMemoryFeatureFlagBits_to_str(VkExternalMemoryFeatureFlagBits input);
    const char * vk_ExternalMemoryFeatureFlagBitsNV_to_str(VkExternalMemoryFeatureFlagBitsNV input);
    const char * vk_ExternalMemoryHandleTypeFlagBits_to_str(VkExternalMemoryHandleTypeFlagBits input);
    const char * vk_ExternalMemoryHandleTypeFlagBitsNV_to_str(VkExternalMemoryHandleTypeFlagBitsNV input);
    const char * vk_ExternalSemaphoreFeatureFlagBits_to_str(VkExternalSemaphoreFeatureFlagBits input);
    const char * vk_ExternalSemaphoreHandleTypeFlagBits_to_str(VkExternalSemaphoreHandleTypeFlagBits input);
    const char * vk_FenceCreateFlagBits_to_str(VkFenceCreateFlagBits input);
    const char * vk_FenceImportFlagBits_to_str(VkFenceImportFlagBits input);
    const char * vk_FormatFeatureFlagBits_to_str(VkFormatFeatureFlagBits input);
    const char * vk_FormatFeatureFlagBits2_to_str(VkFormatFeatureFlagBits2 input);
    const char * vk_FrameBoundaryFlagBitsEXT_to_str(VkFrameBoundaryFlagBitsEXT input);
    const char * vk_FramebufferCreateFlagBits_to_str(VkFramebufferCreateFlagBits input);
    const char * vk_GeometryFlagBitsKHR_to_str(VkGeometryFlagBitsKHR input);
    const char * vk_GeometryInstanceFlagBitsKHR_to_str(VkGeometryInstanceFlagBitsKHR input);
    const char * vk_GraphicsPipelineLibraryFlagBitsEXT_to_str(VkGraphicsPipelineLibraryFlagBitsEXT input);
    const char * vk_HostImageCopyFlagBitsEXT_to_str(VkHostImageCopyFlagBitsEXT input);
    const char * vk_ImageAspectFlagBits_to_str(VkImageAspectFlagBits input);
    const char * vk_ImageCompressionFixedRateFlagBitsEXT_to_str(VkImageCompressionFixedRateFlagBitsEXT input);
    const char * vk_ImageCompressionFlagBitsEXT_to_str(VkImageCompressionFlagBitsEXT input);
#ifdef VK_USE_PLATFORM_FUCHSIA
    const char * vk_ImageConstraintsInfoFlagBitsFUCHSIA_to_str(VkImageConstraintsInfoFlagBitsFUCHSIA input);
#endif
    const char * vk_ImageCreateFlagBits_to_str(VkImageCreateFlagBits input);
    const char * vk_ImageUsageFlagBits_to_str(VkImageUsageFlagBits input);
    const char * vk_ImageViewCreateFlagBits_to_str(VkImageViewCreateFlagBits input);
    const char * vk_IndirectCommandsLayoutUsageFlagBitsNV_to_str(VkIndirectCommandsLayoutUsageFlagBitsNV input);
    const char * vk_IndirectStateFlagBitsNV_to_str(VkIndirectStateFlagBitsNV input);
    const char * vk_InstanceCreateFlagBits_to_str(VkInstanceCreateFlagBits input);
    const char * vk_MemoryAllocateFlagBits_to_str(VkMemoryAllocateFlagBits input);
    const char * vk_MemoryDecompressionMethodFlagBitsNV_to_str(VkMemoryDecompressionMethodFlagBitsNV input);
    const char * vk_MemoryHeapFlagBits_to_str(VkMemoryHeapFlagBits input);
    const char * vk_MemoryPropertyFlagBits_to_str(VkMemoryPropertyFlagBits input);
    const char * vk_MicromapCreateFlagBitsEXT_to_str(VkMicromapCreateFlagBitsEXT input);
    const char * vk_OpticalFlowExecuteFlagBitsNV_to_str(VkOpticalFlowExecuteFlagBitsNV input);
    const char * vk_OpticalFlowGridSizeFlagBitsNV_to_str(VkOpticalFlowGridSizeFlagBitsNV input);
    const char * vk_OpticalFlowSessionCreateFlagBitsNV_to_str(VkOpticalFlowSessionCreateFlagBitsNV input);
    const char * vk_OpticalFlowUsageFlagBitsNV_to_str(VkOpticalFlowUsageFlagBitsNV input);
    const char * vk_PeerMemoryFeatureFlagBits_to_str(VkPeerMemoryFeatureFlagBits input);
    const char * vk_PerformanceCounterDescriptionFlagBitsKHR_to_str(VkPerformanceCounterDescriptionFlagBitsKHR input);
    const char * vk_PhysicalDeviceSchedulingControlsFlagBitsARM_to_str(VkPhysicalDeviceSchedulingControlsFlagBitsARM input);
    const char * vk_PipelineCacheCreateFlagBits_to_str(VkPipelineCacheCreateFlagBits input);
    const char * vk_PipelineColorBlendStateCreateFlagBits_to_str(VkPipelineColorBlendStateCreateFlagBits input);
    const char * vk_PipelineCompilerControlFlagBitsAMD_to_str(VkPipelineCompilerControlFlagBitsAMD input);
    const char * vk_PipelineCreateFlagBits_to_str(VkPipelineCreateFlagBits input);
    const char * vk_PipelineCreateFlagBits2KHR_to_str(VkPipelineCreateFlagBits2KHR input);
    const char * vk_PipelineCreationFeedbackFlagBits_to_str(VkPipelineCreationFeedbackFlagBits input);
    const char * vk_PipelineDepthStencilStateCreateFlagBits_to_str(VkPipelineDepthStencilStateCreateFlagBits input);
    const char * vk_PipelineLayoutCreateFlagBits_to_str(VkPipelineLayoutCreateFlagBits input);
    const char * vk_PipelineShaderStageCreateFlagBits_to_str(VkPipelineShaderStageCreateFlagBits input);
    const char * vk_PipelineStageFlagBits_to_str(VkPipelineStageFlagBits input);
    const char * vk_PipelineStageFlagBits2_to_str(VkPipelineStageFlagBits2 input);
    const char * vk_PresentGravityFlagBitsEXT_to_str(VkPresentGravityFlagBitsEXT input);
    const char * vk_PresentScalingFlagBitsEXT_to_str(VkPresentScalingFlagBitsEXT input);
    const char * vk_QueryControlFlagBits_to_str(VkQueryControlFlagBits input);
    const char * vk_QueryPipelineStatisticFlagBits_to_str(VkQueryPipelineStatisticFlagBits input);
    const char * vk_QueryResultFlagBits_to_str(VkQueryResultFlagBits input);
    const char * vk_QueueFlagBits_to_str(VkQueueFlagBits input);
    const char * vk_RenderPassCreateFlagBits_to_str(VkRenderPassCreateFlagBits input);
    const char * vk_RenderingFlagBits_to_str(VkRenderingFlagBits input);
    const char * vk_ResolveModeFlagBits_to_str(VkResolveModeFlagBits input);
    const char * vk_SampleCountFlagBits_to_str(VkSampleCountFlagBits input);
    const char * vk_SamplerCreateFlagBits_to_str(VkSamplerCreateFlagBits input);
    const char * vk_SemaphoreImportFlagBits_to_str(VkSemaphoreImportFlagBits input);
    const char * vk_SemaphoreWaitFlagBits_to_str(VkSemaphoreWaitFlagBits input);
    const char * vk_ShaderCorePropertiesFlagBitsAMD_to_str(VkShaderCorePropertiesFlagBitsAMD input);
    const char * vk_ShaderCreateFlagBitsEXT_to_str(VkShaderCreateFlagBitsEXT input);
    const char * vk_ShaderStageFlagBits_to_str(VkShaderStageFlagBits input);
    const char * vk_SparseImageFormatFlagBits_to_str(VkSparseImageFormatFlagBits input);
    const char * vk_SparseMemoryBindFlagBits_to_str(VkSparseMemoryBindFlagBits input);
    const char * vk_StencilFaceFlagBits_to_str(VkStencilFaceFlagBits input);
    const char * vk_SubgroupFeatureFlagBits_to_str(VkSubgroupFeatureFlagBits input);
    const char * vk_SubmitFlagBits_to_str(VkSubmitFlagBits input);
    const char * vk_SubpassDescriptionFlagBits_to_str(VkSubpassDescriptionFlagBits input);
    const char * vk_SurfaceCounterFlagBitsEXT_to_str(VkSurfaceCounterFlagBitsEXT input);
    const char * vk_SurfaceTransformFlagBitsKHR_to_str(VkSurfaceTransformFlagBitsKHR input);
    const char * vk_SwapchainCreateFlagBitsKHR_to_str(VkSwapchainCreateFlagBitsKHR input);
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    const char * vk_SwapchainImageUsageFlagBitsANDROID_to_str(VkSwapchainImageUsageFlagBitsANDROID input);
#endif
    const char * vk_ToolPurposeFlagBits_to_str(VkToolPurposeFlagBits input);
    const char * vk_VideoCapabilityFlagBitsKHR_to_str(VkVideoCapabilityFlagBitsKHR input);
    const char * vk_VideoChromaSubsamplingFlagBitsKHR_to_str(VkVideoChromaSubsamplingFlagBitsKHR input);
    const char * vk_VideoCodecOperationFlagBitsKHR_to_str(VkVideoCodecOperationFlagBitsKHR input);
    const char * vk_VideoCodingControlFlagBitsKHR_to_str(VkVideoCodingControlFlagBitsKHR input);
    const char * vk_VideoComponentBitDepthFlagBitsKHR_to_str(VkVideoComponentBitDepthFlagBitsKHR input);
    const char * vk_VideoDecodeCapabilityFlagBitsKHR_to_str(VkVideoDecodeCapabilityFlagBitsKHR input);
    const char * vk_VideoDecodeH264PictureLayoutFlagBitsKHR_to_str(VkVideoDecodeH264PictureLayoutFlagBitsKHR input);
    const char * vk_VideoDecodeUsageFlagBitsKHR_to_str(VkVideoDecodeUsageFlagBitsKHR input);
    const char * vk_VideoEncodeCapabilityFlagBitsKHR_to_str(VkVideoEncodeCapabilityFlagBitsKHR input);
    const char * vk_VideoEncodeContentFlagBitsKHR_to_str(VkVideoEncodeContentFlagBitsKHR input);
    const char * vk_VideoEncodeFeedbackFlagBitsKHR_to_str(VkVideoEncodeFeedbackFlagBitsKHR input);
    const char * vk_VideoEncodeH264CapabilityFlagBitsKHR_to_str(VkVideoEncodeH264CapabilityFlagBitsKHR input);
    const char * vk_VideoEncodeH264RateControlFlagBitsKHR_to_str(VkVideoEncodeH264RateControlFlagBitsKHR input);
    const char * vk_VideoEncodeH264StdFlagBitsKHR_to_str(VkVideoEncodeH264StdFlagBitsKHR input);
    const char * vk_VideoEncodeH265CapabilityFlagBitsKHR_to_str(VkVideoEncodeH265CapabilityFlagBitsKHR input);
    const char * vk_VideoEncodeH265CtbSizeFlagBitsKHR_to_str(VkVideoEncodeH265CtbSizeFlagBitsKHR input);
    const char * vk_VideoEncodeH265RateControlFlagBitsKHR_to_str(VkVideoEncodeH265RateControlFlagBitsKHR input);
    const char * vk_VideoEncodeH265StdFlagBitsKHR_to_str(VkVideoEncodeH265StdFlagBitsKHR input);
    const char * vk_VideoEncodeH265TransformBlockSizeFlagBitsKHR_to_str(VkVideoEncodeH265TransformBlockSizeFlagBitsKHR input);
    const char * vk_VideoEncodeRateControlModeFlagBitsKHR_to_str(VkVideoEncodeRateControlModeFlagBitsKHR input);
    const char * vk_VideoEncodeUsageFlagBitsKHR_to_str(VkVideoEncodeUsageFlagBitsKHR input);
    const char * vk_VideoSessionCreateFlagBitsKHR_to_str(VkVideoSessionCreateFlagBitsKHR input);

    size_t vk_structure_type_size(const struct VkBaseInStructure *item);

    const char * vk_ObjectType_to_ObjectName(VkObjectType type);

    #ifdef __cplusplus
    } /* extern "C" */
    #endif

    #endif