    /* Autogenerated file -- do not edit
     * generated by gen_enum_to_str.py
     *
     * Copyright © 2017 Intel Corporation
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
     */

    #include <string.h>
    #include <vulkan/vulkan_core.h>
    #include <vulkan/vk_android_native_buffer.h>
    #include <vulkan/vk_layer.h>
    #include "util/macros.h"
    #include "vk_enum_to_str.h"


    const char *
    vk_AccelerationStructureBuildTypeKHR_to_str(VkAccelerationStructureBuildTypeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCELERATION_STRUCTURE_BUILD_TYPE_HOST_KHR";
        case 1:
            return "VK_ACCELERATION_STRUCTURE_BUILD_TYPE_DEVICE_KHR";
        case 2:
            return "VK_ACCELERATION_STRUCTURE_BUILD_TYPE_HOST_OR_DEVICE_KHR";
        case VK_ACCELERATION_STRUCTURE_BUILD_TYPE_MAX_ENUM_KHR: return "VK_ACCELERATION_STRUCTURE_BUILD_TYPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkAccelerationStructureBuildTypeKHR value.";
        }
    }


    const char *
    vk_AccelerationStructureCompatibilityKHR_to_str(VkAccelerationStructureCompatibilityKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCELERATION_STRUCTURE_COMPATIBILITY_COMPATIBLE_KHR";
        case 1:
            return "VK_ACCELERATION_STRUCTURE_COMPATIBILITY_INCOMPATIBLE_KHR";
        case VK_ACCELERATION_STRUCTURE_COMPATIBILITY_MAX_ENUM_KHR: return "VK_ACCELERATION_STRUCTURE_COMPATIBILITY_MAX_ENUM_KHR";
        default:
            return "Unknown VkAccelerationStructureCompatibilityKHR value.";
        }
    }


    const char *
    vk_AccelerationStructureMemoryRequirementsTypeNV_to_str(VkAccelerationStructureMemoryRequirementsTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_OBJECT_NV";
        case 1:
            return "VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_BUILD_SCRATCH_NV";
        case 2:
            return "VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_UPDATE_SCRATCH_NV";
        case VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_MAX_ENUM_NV: return "VK_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkAccelerationStructureMemoryRequirementsTypeNV value.";
        }
    }


    const char *
    vk_AccelerationStructureMotionInstanceTypeNV_to_str(VkAccelerationStructureMotionInstanceTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_STATIC_NV";
        case 1:
            return "VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_MATRIX_MOTION_NV";
        case 2:
            return "VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_SRT_MOTION_NV";
        case VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_MAX_ENUM_NV: return "VK_ACCELERATION_STRUCTURE_MOTION_INSTANCE_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkAccelerationStructureMotionInstanceTypeNV value.";
        }
    }


    const char *
    vk_AccelerationStructureTypeKHR_to_str(VkAccelerationStructureTypeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_NV";
        case 1:
            return "VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_NV";
        case 2:
            return "VK_ACCELERATION_STRUCTURE_TYPE_GENERIC_KHR";
        case VK_ACCELERATION_STRUCTURE_TYPE_MAX_ENUM_KHR: return "VK_ACCELERATION_STRUCTURE_TYPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkAccelerationStructureTypeKHR value.";
        }
    }


    const char *
    vk_AttachmentLoadOp_to_str(VkAttachmentLoadOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ATTACHMENT_LOAD_OP_LOAD";
        case 1:
            return "VK_ATTACHMENT_LOAD_OP_CLEAR";
        case 2:
            return "VK_ATTACHMENT_LOAD_OP_DONT_CARE";
        case 1000400000:
            return "VK_ATTACHMENT_LOAD_OP_NONE_EXT";
        case VK_ATTACHMENT_LOAD_OP_MAX_ENUM: return "VK_ATTACHMENT_LOAD_OP_MAX_ENUM";
        default:
            return "Unknown VkAttachmentLoadOp value.";
        }
    }


    const char *
    vk_AttachmentStoreOp_to_str(VkAttachmentStoreOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ATTACHMENT_STORE_OP_STORE";
        case 1:
            return "VK_ATTACHMENT_STORE_OP_DONT_CARE";
        case 1000301000:
            return "VK_ATTACHMENT_STORE_OP_NONE";
        case VK_ATTACHMENT_STORE_OP_MAX_ENUM: return "VK_ATTACHMENT_STORE_OP_MAX_ENUM";
        default:
            return "Unknown VkAttachmentStoreOp value.";
        }
    }


    const char *
    vk_BlendFactor_to_str(VkBlendFactor input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BLEND_FACTOR_ZERO";
        case 1:
            return "VK_BLEND_FACTOR_ONE";
        case 2:
            return "VK_BLEND_FACTOR_SRC_COLOR";
        case 3:
            return "VK_BLEND_FACTOR_ONE_MINUS_SRC_COLOR";
        case 4:
            return "VK_BLEND_FACTOR_DST_COLOR";
        case 5:
            return "VK_BLEND_FACTOR_ONE_MINUS_DST_COLOR";
        case 6:
            return "VK_BLEND_FACTOR_SRC_ALPHA";
        case 7:
            return "VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA";
        case 8:
            return "VK_BLEND_FACTOR_DST_ALPHA";
        case 9:
            return "VK_BLEND_FACTOR_ONE_MINUS_DST_ALPHA";
        case 10:
            return "VK_BLEND_FACTOR_CONSTANT_COLOR";
        case 11:
            return "VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_COLOR";
        case 12:
            return "VK_BLEND_FACTOR_CONSTANT_ALPHA";
        case 13:
            return "VK_BLEND_FACTOR_ONE_MINUS_CONSTANT_ALPHA";
        case 14:
            return "VK_BLEND_FACTOR_SRC_ALPHA_SATURATE";
        case 15:
            return "VK_BLEND_FACTOR_SRC1_COLOR";
        case 16:
            return "VK_BLEND_FACTOR_ONE_MINUS_SRC1_COLOR";
        case 17:
            return "VK_BLEND_FACTOR_SRC1_ALPHA";
        case 18:
            return "VK_BLEND_FACTOR_ONE_MINUS_SRC1_ALPHA";
        case VK_BLEND_FACTOR_MAX_ENUM: return "VK_BLEND_FACTOR_MAX_ENUM";
        default:
            return "Unknown VkBlendFactor value.";
        }
    }


    const char *
    vk_BlendOp_to_str(VkBlendOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BLEND_OP_ADD";
        case 1:
            return "VK_BLEND_OP_SUBTRACT";
        case 2:
            return "VK_BLEND_OP_REVERSE_SUBTRACT";
        case 3:
            return "VK_BLEND_OP_MIN";
        case 4:
            return "VK_BLEND_OP_MAX";
        case 1000148000:
            return "VK_BLEND_OP_ZERO_EXT";
        case 1000148001:
            return "VK_BLEND_OP_SRC_EXT";
        case 1000148002:
            return "VK_BLEND_OP_DST_EXT";
        case 1000148003:
            return "VK_BLEND_OP_SRC_OVER_EXT";
        case 1000148004:
            return "VK_BLEND_OP_DST_OVER_EXT";
        case 1000148005:
            return "VK_BLEND_OP_SRC_IN_EXT";
        case 1000148006:
            return "VK_BLEND_OP_DST_IN_EXT";
        case 1000148007:
            return "VK_BLEND_OP_SRC_OUT_EXT";
        case 1000148008:
            return "VK_BLEND_OP_DST_OUT_EXT";
        case 1000148009:
            return "VK_BLEND_OP_SRC_ATOP_EXT";
        case 1000148010:
            return "VK_BLEND_OP_DST_ATOP_EXT";
        case 1000148011:
            return "VK_BLEND_OP_XOR_EXT";
        case 1000148012:
            return "VK_BLEND_OP_MULTIPLY_EXT";
        case 1000148013:
            return "VK_BLEND_OP_SCREEN_EXT";
        case 1000148014:
            return "VK_BLEND_OP_OVERLAY_EXT";
        case 1000148015:
            return "VK_BLEND_OP_DARKEN_EXT";
        case 1000148016:
            return "VK_BLEND_OP_LIGHTEN_EXT";
        case 1000148017:
            return "VK_BLEND_OP_COLORDODGE_EXT";
        case 1000148018:
            return "VK_BLEND_OP_COLORBURN_EXT";
        case 1000148019:
            return "VK_BLEND_OP_HARDLIGHT_EXT";
        case 1000148020:
            return "VK_BLEND_OP_SOFTLIGHT_EXT";
        case 1000148021:
            return "VK_BLEND_OP_DIFFERENCE_EXT";
        case 1000148022:
            return "VK_BLEND_OP_EXCLUSION_EXT";
        case 1000148023:
            return "VK_BLEND_OP_INVERT_EXT";
        case 1000148024:
            return "VK_BLEND_OP_INVERT_RGB_EXT";
        case 1000148025:
            return "VK_BLEND_OP_LINEARDODGE_EXT";
        case 1000148026:
            return "VK_BLEND_OP_LINEARBURN_EXT";
        case 1000148027:
            return "VK_BLEND_OP_VIVIDLIGHT_EXT";
        case 1000148028:
            return "VK_BLEND_OP_LINEARLIGHT_EXT";
        case 1000148029:
            return "VK_BLEND_OP_PINLIGHT_EXT";
        case 1000148030:
            return "VK_BLEND_OP_HARDMIX_EXT";
        case 1000148031:
            return "VK_BLEND_OP_HSL_HUE_EXT";
        case 1000148032:
            return "VK_BLEND_OP_HSL_SATURATION_EXT";
        case 1000148033:
            return "VK_BLEND_OP_HSL_COLOR_EXT";
        case 1000148034:
            return "VK_BLEND_OP_HSL_LUMINOSITY_EXT";
        case 1000148035:
            return "VK_BLEND_OP_PLUS_EXT";
        case 1000148036:
            return "VK_BLEND_OP_PLUS_CLAMPED_EXT";
        case 1000148037:
            return "VK_BLEND_OP_PLUS_CLAMPED_ALPHA_EXT";
        case 1000148038:
            return "VK_BLEND_OP_PLUS_DARKER_EXT";
        case 1000148039:
            return "VK_BLEND_OP_MINUS_EXT";
        case 1000148040:
            return "VK_BLEND_OP_MINUS_CLAMPED_EXT";
        case 1000148041:
            return "VK_BLEND_OP_CONTRAST_EXT";
        case 1000148042:
            return "VK_BLEND_OP_INVERT_OVG_EXT";
        case 1000148043:
            return "VK_BLEND_OP_RED_EXT";
        case 1000148044:
            return "VK_BLEND_OP_GREEN_EXT";
        case 1000148045:
            return "VK_BLEND_OP_BLUE_EXT";
        case VK_BLEND_OP_MAX_ENUM: return "VK_BLEND_OP_MAX_ENUM";
        default:
            return "Unknown VkBlendOp value.";
        }
    }


    const char *
    vk_BlendOverlapEXT_to_str(VkBlendOverlapEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BLEND_OVERLAP_UNCORRELATED_EXT";
        case 1:
            return "VK_BLEND_OVERLAP_DISJOINT_EXT";
        case 2:
            return "VK_BLEND_OVERLAP_CONJOINT_EXT";
        case VK_BLEND_OVERLAP_MAX_ENUM_EXT: return "VK_BLEND_OVERLAP_MAX_ENUM_EXT";
        default:
            return "Unknown VkBlendOverlapEXT value.";
        }
    }


    const char *
    vk_BlockMatchWindowCompareModeQCOM_to_str(VkBlockMatchWindowCompareModeQCOM input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BLOCK_MATCH_WINDOW_COMPARE_MODE_MIN_QCOM";
        case 1:
            return "VK_BLOCK_MATCH_WINDOW_COMPARE_MODE_MAX_QCOM";
        case VK_BLOCK_MATCH_WINDOW_COMPARE_MODE_MAX_ENUM_QCOM: return "VK_BLOCK_MATCH_WINDOW_COMPARE_MODE_MAX_ENUM_QCOM";
        default:
            return "Unknown VkBlockMatchWindowCompareModeQCOM value.";
        }
    }


    const char *
    vk_BorderColor_to_str(VkBorderColor input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK";
        case 1:
            return "VK_BORDER_COLOR_INT_TRANSPARENT_BLACK";
        case 2:
            return "VK_BORDER_COLOR_FLOAT_OPAQUE_BLACK";
        case 3:
            return "VK_BORDER_COLOR_INT_OPAQUE_BLACK";
        case 4:
            return "VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE";
        case 5:
            return "VK_BORDER_COLOR_INT_OPAQUE_WHITE";
        case 1000287003:
            return "VK_BORDER_COLOR_FLOAT_CUSTOM_EXT";
        case 1000287004:
            return "VK_BORDER_COLOR_INT_CUSTOM_EXT";
        case VK_BORDER_COLOR_MAX_ENUM: return "VK_BORDER_COLOR_MAX_ENUM";
        default:
            return "Unknown VkBorderColor value.";
        }
    }


    const char *
    vk_BuildAccelerationStructureModeKHR_to_str(VkBuildAccelerationStructureModeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR";
        case 1:
            return "VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR";
        case VK_BUILD_ACCELERATION_STRUCTURE_MODE_MAX_ENUM_KHR: return "VK_BUILD_ACCELERATION_STRUCTURE_MODE_MAX_ENUM_KHR";
        default:
            return "Unknown VkBuildAccelerationStructureModeKHR value.";
        }
    }


    const char *
    vk_BuildMicromapModeEXT_to_str(VkBuildMicromapModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_BUILD_MICROMAP_MODE_BUILD_EXT";
        case VK_BUILD_MICROMAP_MODE_MAX_ENUM_EXT: return "VK_BUILD_MICROMAP_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkBuildMicromapModeEXT value.";
        }
    }


    const char *
    vk_ChromaLocation_to_str(VkChromaLocation input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_CHROMA_LOCATION_COSITED_EVEN";
        case 1:
            return "VK_CHROMA_LOCATION_MIDPOINT";
        case VK_CHROMA_LOCATION_MAX_ENUM: return "VK_CHROMA_LOCATION_MAX_ENUM";
        default:
            return "Unknown VkChromaLocation value.";
        }
    }


    const char *
    vk_CoarseSampleOrderTypeNV_to_str(VkCoarseSampleOrderTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COARSE_SAMPLE_ORDER_TYPE_DEFAULT_NV";
        case 1:
            return "VK_COARSE_SAMPLE_ORDER_TYPE_CUSTOM_NV";
        case 2:
            return "VK_COARSE_SAMPLE_ORDER_TYPE_PIXEL_MAJOR_NV";
        case 3:
            return "VK_COARSE_SAMPLE_ORDER_TYPE_SAMPLE_MAJOR_NV";
        case VK_COARSE_SAMPLE_ORDER_TYPE_MAX_ENUM_NV: return "VK_COARSE_SAMPLE_ORDER_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkCoarseSampleOrderTypeNV value.";
        }
    }


    const char *
    vk_ColorSpaceKHR_to_str(VkColorSpaceKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COLORSPACE_SRGB_NONLINEAR_KHR";
        case 1000104001:
            return "VK_COLOR_SPACE_DISPLAY_P3_NONLINEAR_EXT";
        case 1000104002:
            return "VK_COLOR_SPACE_EXTENDED_SRGB_LINEAR_EXT";
        case 1000104003:
            return "VK_COLOR_SPACE_DCI_P3_LINEAR_EXT";
        case 1000104004:
            return "VK_COLOR_SPACE_DCI_P3_NONLINEAR_EXT";
        case 1000104005:
            return "VK_COLOR_SPACE_BT709_LINEAR_EXT";
        case 1000104006:
            return "VK_COLOR_SPACE_BT709_NONLINEAR_EXT";
        case 1000104007:
            return "VK_COLOR_SPACE_BT2020_LINEAR_EXT";
        case 1000104008:
            return "VK_COLOR_SPACE_HDR10_ST2084_EXT";
        case 1000104009:
            return "VK_COLOR_SPACE_DOLBYVISION_EXT";
        case 1000104010:
            return "VK_COLOR_SPACE_HDR10_HLG_EXT";
        case 1000104011:
            return "VK_COLOR_SPACE_ADOBERGB_LINEAR_EXT";
        case 1000104012:
            return "VK_COLOR_SPACE_ADOBERGB_NONLINEAR_EXT";
        case 1000104013:
            return "VK_COLOR_SPACE_PASS_THROUGH_EXT";
        case 1000104014:
            return "VK_COLOR_SPACE_EXTENDED_SRGB_NONLINEAR_EXT";
        case 1000213000:
            return "VK_COLOR_SPACE_DISPLAY_NATIVE_AMD";
        case VK_COLOR_SPACE_MAX_ENUM_KHR: return "VK_COLOR_SPACE_MAX_ENUM_KHR";
        default:
            return "Unknown VkColorSpaceKHR value.";
        }
    }


    const char *
    vk_CommandBufferLevel_to_str(VkCommandBufferLevel input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COMMAND_BUFFER_LEVEL_PRIMARY";
        case 1:
            return "VK_COMMAND_BUFFER_LEVEL_SECONDARY";
        case VK_COMMAND_BUFFER_LEVEL_MAX_ENUM: return "VK_COMMAND_BUFFER_LEVEL_MAX_ENUM";
        default:
            return "Unknown VkCommandBufferLevel value.";
        }
    }


    const char *
    vk_CompareOp_to_str(VkCompareOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COMPARE_OP_NEVER";
        case 1:
            return "VK_COMPARE_OP_LESS";
        case 2:
            return "VK_COMPARE_OP_EQUAL";
        case 3:
            return "VK_COMPARE_OP_LESS_OR_EQUAL";
        case 4:
            return "VK_COMPARE_OP_GREATER";
        case 5:
            return "VK_COMPARE_OP_NOT_EQUAL";
        case 6:
            return "VK_COMPARE_OP_GREATER_OR_EQUAL";
        case 7:
            return "VK_COMPARE_OP_ALWAYS";
        case VK_COMPARE_OP_MAX_ENUM: return "VK_COMPARE_OP_MAX_ENUM";
        default:
            return "Unknown VkCompareOp value.";
        }
    }


    const char *
    vk_ComponentSwizzle_to_str(VkComponentSwizzle input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COMPONENT_SWIZZLE_IDENTITY";
        case 1:
            return "VK_COMPONENT_SWIZZLE_ZERO";
        case 2:
            return "VK_COMPONENT_SWIZZLE_ONE";
        case 3:
            return "VK_COMPONENT_SWIZZLE_R";
        case 4:
            return "VK_COMPONENT_SWIZZLE_G";
        case 5:
            return "VK_COMPONENT_SWIZZLE_B";
        case 6:
            return "VK_COMPONENT_SWIZZLE_A";
        case VK_COMPONENT_SWIZZLE_MAX_ENUM: return "VK_COMPONENT_SWIZZLE_MAX_ENUM";
        default:
            return "Unknown VkComponentSwizzle value.";
        }
    }


    const char *
    vk_ComponentTypeKHR_to_str(VkComponentTypeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COMPONENT_TYPE_FLOAT16_NV";
        case 1:
            return "VK_COMPONENT_TYPE_FLOAT32_NV";
        case 2:
            return "VK_COMPONENT_TYPE_FLOAT64_NV";
        case 3:
            return "VK_COMPONENT_TYPE_SINT8_NV";
        case 4:
            return "VK_COMPONENT_TYPE_SINT16_NV";
        case 5:
            return "VK_COMPONENT_TYPE_SINT32_NV";
        case 6:
            return "VK_COMPONENT_TYPE_SINT64_NV";
        case 7:
            return "VK_COMPONENT_TYPE_UINT8_NV";
        case 8:
            return "VK_COMPONENT_TYPE_UINT16_NV";
        case 9:
            return "VK_COMPONENT_TYPE_UINT32_NV";
        case 10:
            return "VK_COMPONENT_TYPE_UINT64_NV";
        case VK_COMPONENT_TYPE_MAX_ENUM_KHR: return "VK_COMPONENT_TYPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkComponentTypeKHR value.";
        }
    }


    const char *
    vk_ConservativeRasterizationModeEXT_to_str(VkConservativeRasterizationModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT";
        case 1:
            return "VK_CONSERVATIVE_RASTERIZATION_MODE_OVERESTIMATE_EXT";
        case 2:
            return "VK_CONSERVATIVE_RASTERIZATION_MODE_UNDERESTIMATE_EXT";
        case VK_CONSERVATIVE_RASTERIZATION_MODE_MAX_ENUM_EXT: return "VK_CONSERVATIVE_RASTERIZATION_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkConservativeRasterizationModeEXT value.";
        }
    }


    const char *
    vk_CopyAccelerationStructureModeKHR_to_str(VkCopyAccelerationStructureModeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COPY_ACCELERATION_STRUCTURE_MODE_CLONE_NV";
        case 1:
            return "VK_COPY_ACCELERATION_STRUCTURE_MODE_COMPACT_NV";
        case 2:
            return "VK_COPY_ACCELERATION_STRUCTURE_MODE_SERIALIZE_KHR";
        case 3:
            return "VK_COPY_ACCELERATION_STRUCTURE_MODE_DESERIALIZE_KHR";
        case VK_COPY_ACCELERATION_STRUCTURE_MODE_MAX_ENUM_KHR: return "VK_COPY_ACCELERATION_STRUCTURE_MODE_MAX_ENUM_KHR";
        default:
            return "Unknown VkCopyAccelerationStructureModeKHR value.";
        }
    }


    const char *
    vk_CopyMicromapModeEXT_to_str(VkCopyMicromapModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COPY_MICROMAP_MODE_CLONE_EXT";
        case 1:
            return "VK_COPY_MICROMAP_MODE_SERIALIZE_EXT";
        case 2:
            return "VK_COPY_MICROMAP_MODE_DESERIALIZE_EXT";
        case 3:
            return "VK_COPY_MICROMAP_MODE_COMPACT_EXT";
        case VK_COPY_MICROMAP_MODE_MAX_ENUM_EXT: return "VK_COPY_MICROMAP_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkCopyMicromapModeEXT value.";
        }
    }


    const char *
    vk_CoverageModulationModeNV_to_str(VkCoverageModulationModeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COVERAGE_MODULATION_MODE_NONE_NV";
        case 1:
            return "VK_COVERAGE_MODULATION_MODE_RGB_NV";
        case 2:
            return "VK_COVERAGE_MODULATION_MODE_ALPHA_NV";
        case 3:
            return "VK_COVERAGE_MODULATION_MODE_RGBA_NV";
        case VK_COVERAGE_MODULATION_MODE_MAX_ENUM_NV: return "VK_COVERAGE_MODULATION_MODE_MAX_ENUM_NV";
        default:
            return "Unknown VkCoverageModulationModeNV value.";
        }
    }


    const char *
    vk_CoverageReductionModeNV_to_str(VkCoverageReductionModeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_COVERAGE_REDUCTION_MODE_MERGE_NV";
        case 1:
            return "VK_COVERAGE_REDUCTION_MODE_TRUNCATE_NV";
        case VK_COVERAGE_REDUCTION_MODE_MAX_ENUM_NV: return "VK_COVERAGE_REDUCTION_MODE_MAX_ENUM_NV";
        default:
            return "Unknown VkCoverageReductionModeNV value.";
        }
    }


    const char *
    vk_CubicFilterWeightsQCOM_to_str(VkCubicFilterWeightsQCOM input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_CUBIC_FILTER_WEIGHTS_CATMULL_ROM_QCOM";
        case 1:
            return "VK_CUBIC_FILTER_WEIGHTS_ZERO_TANGENT_CARDINAL_QCOM";
        case 2:
            return "VK_CUBIC_FILTER_WEIGHTS_B_SPLINE_QCOM";
        case 3:
            return "VK_CUBIC_FILTER_WEIGHTS_MITCHELL_NETRAVALI_QCOM";
        case VK_CUBIC_FILTER_WEIGHTS_MAX_ENUM_QCOM: return "VK_CUBIC_FILTER_WEIGHTS_MAX_ENUM_QCOM";
        default:
            return "Unknown VkCubicFilterWeightsQCOM value.";
        }
    }


    const char *
    vk_DebugReportObjectTypeEXT_to_str(VkDebugReportObjectTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_UNKNOWN_EXT";
        case 1:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_INSTANCE_EXT";
        case 2:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_PHYSICAL_DEVICE_EXT";
        case 3:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DEVICE_EXT";
        case 4:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_QUEUE_EXT";
        case 5:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SEMAPHORE_EXT";
        case 6:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_COMMAND_BUFFER_EXT";
        case 7:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_FENCE_EXT";
        case 8:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DEVICE_MEMORY_EXT";
        case 9:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_BUFFER_EXT";
        case 10:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_IMAGE_EXT";
        case 11:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_EVENT_EXT";
        case 12:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_QUERY_POOL_EXT";
        case 13:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_BUFFER_VIEW_EXT";
        case 14:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_IMAGE_VIEW_EXT";
        case 15:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SHADER_MODULE_EXT";
        case 16:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_PIPELINE_CACHE_EXT";
        case 17:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_PIPELINE_LAYOUT_EXT";
        case 18:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_RENDER_PASS_EXT";
        case 19:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_PIPELINE_EXT";
        case 20:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DESCRIPTOR_SET_LAYOUT_EXT";
        case 21:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SAMPLER_EXT";
        case 22:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DESCRIPTOR_POOL_EXT";
        case 23:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DESCRIPTOR_SET_EXT";
        case 24:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_FRAMEBUFFER_EXT";
        case 25:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_COMMAND_POOL_EXT";
        case 26:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SURFACE_KHR_EXT";
        case 27:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SWAPCHAIN_KHR_EXT";
        case 28:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DEBUG_REPORT_EXT";
        case 29:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DISPLAY_KHR_EXT";
        case 30:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DISPLAY_MODE_KHR_EXT";
        case 33:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_VALIDATION_CACHE_EXT";
        case 1000029000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_CU_MODULE_NVX_EXT";
        case 1000029001:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_CU_FUNCTION_NVX_EXT";
        case 1000085000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_DESCRIPTOR_UPDATE_TEMPLATE_EXT";
        case 1000150000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_ACCELERATION_STRUCTURE_KHR_EXT";
        case 1000156000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_SAMPLER_YCBCR_CONVERSION_EXT";
        case 1000165000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_ACCELERATION_STRUCTURE_NV_EXT";
        case 1000307000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_CUDA_MODULE_NV_EXT";
        case 1000307001:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_CUDA_FUNCTION_NV_EXT";
        case 1000366000:
            return "VK_DEBUG_REPORT_OBJECT_TYPE_BUFFER_COLLECTION_FUCHSIA_EXT";
        case VK_DEBUG_REPORT_OBJECT_TYPE_MAX_ENUM_EXT: return "VK_DEBUG_REPORT_OBJECT_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDebugReportObjectTypeEXT value.";
        }
    }


    const char *
    vk_DepthBiasRepresentationEXT_to_str(VkDepthBiasRepresentationEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEPTH_BIAS_REPRESENTATION_LEAST_REPRESENTABLE_VALUE_FORMAT_EXT";
        case 1:
            return "VK_DEPTH_BIAS_REPRESENTATION_LEAST_REPRESENTABLE_VALUE_FORCE_UNORM_EXT";
        case 2:
            return "VK_DEPTH_BIAS_REPRESENTATION_FLOAT_EXT";
        case VK_DEPTH_BIAS_REPRESENTATION_MAX_ENUM_EXT: return "VK_DEPTH_BIAS_REPRESENTATION_MAX_ENUM_EXT";
        default:
            return "Unknown VkDepthBiasRepresentationEXT value.";
        }
    }


    const char *
    vk_DescriptorType_to_str(VkDescriptorType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DESCRIPTOR_TYPE_SAMPLER";
        case 1:
            return "VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER";
        case 2:
            return "VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE";
        case 3:
            return "VK_DESCRIPTOR_TYPE_STORAGE_IMAGE";
        case 4:
            return "VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER";
        case 5:
            return "VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER";
        case 6:
            return "VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER";
        case 7:
            return "VK_DESCRIPTOR_TYPE_STORAGE_BUFFER";
        case 8:
            return "VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC";
        case 9:
            return "VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC";
        case 10:
            return "VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT";
        case 1000138000:
            return "VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK";
        case 1000150000:
            return "VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR";
        case 1000165000:
            return "VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_NV";
        case 1000351000:
            return "VK_DESCRIPTOR_TYPE_MUTABLE_EXT";
        case 1000440000:
            return "VK_DESCRIPTOR_TYPE_SAMPLE_WEIGHT_IMAGE_QCOM";
        case 1000440001:
            return "VK_DESCRIPTOR_TYPE_BLOCK_MATCH_IMAGE_QCOM";
        case VK_DESCRIPTOR_TYPE_MAX_ENUM: return "VK_DESCRIPTOR_TYPE_MAX_ENUM";
        default:
            return "Unknown VkDescriptorType value.";
        }
    }


    const char *
    vk_DescriptorUpdateTemplateType_to_str(VkDescriptorUpdateTemplateType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_DESCRIPTOR_SET";
        case 1:
            return "VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_PUSH_DESCRIPTORS_KHR";
        case VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_MAX_ENUM: return "VK_DESCRIPTOR_UPDATE_TEMPLATE_TYPE_MAX_ENUM";
        default:
            return "Unknown VkDescriptorUpdateTemplateType value.";
        }
    }


    const char *
    vk_DeviceAddressBindingTypeEXT_to_str(VkDeviceAddressBindingTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEVICE_ADDRESS_BINDING_TYPE_BIND_EXT";
        case 1:
            return "VK_DEVICE_ADDRESS_BINDING_TYPE_UNBIND_EXT";
        case VK_DEVICE_ADDRESS_BINDING_TYPE_MAX_ENUM_EXT: return "VK_DEVICE_ADDRESS_BINDING_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDeviceAddressBindingTypeEXT value.";
        }
    }


    const char *
    vk_DeviceEventTypeEXT_to_str(VkDeviceEventTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEVICE_EVENT_TYPE_DISPLAY_HOTPLUG_EXT";
        case VK_DEVICE_EVENT_TYPE_MAX_ENUM_EXT: return "VK_DEVICE_EVENT_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDeviceEventTypeEXT value.";
        }
    }


    const char *
    vk_DeviceFaultAddressTypeEXT_to_str(VkDeviceFaultAddressTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_NONE_EXT";
        case 1:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_READ_INVALID_EXT";
        case 2:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_WRITE_INVALID_EXT";
        case 3:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_EXECUTE_INVALID_EXT";
        case 4:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_INSTRUCTION_POINTER_UNKNOWN_EXT";
        case 5:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_INSTRUCTION_POINTER_INVALID_EXT";
        case 6:
            return "VK_DEVICE_FAULT_ADDRESS_TYPE_INSTRUCTION_POINTER_FAULT_EXT";
        case VK_DEVICE_FAULT_ADDRESS_TYPE_MAX_ENUM_EXT: return "VK_DEVICE_FAULT_ADDRESS_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDeviceFaultAddressTypeEXT value.";
        }
    }


    const char *
    vk_DeviceFaultVendorBinaryHeaderVersionEXT_to_str(VkDeviceFaultVendorBinaryHeaderVersionEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEVICE_FAULT_VENDOR_BINARY_HEADER_VERSION_ONE_EXT";
        case VK_DEVICE_FAULT_VENDOR_BINARY_HEADER_VERSION_MAX_ENUM_EXT: return "VK_DEVICE_FAULT_VENDOR_BINARY_HEADER_VERSION_MAX_ENUM_EXT";
        default:
            return "Unknown VkDeviceFaultVendorBinaryHeaderVersionEXT value.";
        }
    }


    const char *
    vk_DeviceMemoryReportEventTypeEXT_to_str(VkDeviceMemoryReportEventTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_ALLOCATE_EXT";
        case 1:
            return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_FREE_EXT";
        case 2:
            return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_IMPORT_EXT";
        case 3:
            return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_UNIMPORT_EXT";
        case 4:
            return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_ALLOCATION_FAILED_EXT";
        case VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_MAX_ENUM_EXT: return "VK_DEVICE_MEMORY_REPORT_EVENT_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDeviceMemoryReportEventTypeEXT value.";
        }
    }


    const char *
    vk_DirectDriverLoadingModeLUNARG_to_str(VkDirectDriverLoadingModeLUNARG input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DIRECT_DRIVER_LOADING_MODE_EXCLUSIVE_LUNARG";
        case 1:
            return "VK_DIRECT_DRIVER_LOADING_MODE_INCLUSIVE_LUNARG";
        case VK_DIRECT_DRIVER_LOADING_MODE_MAX_ENUM_LUNARG: return "VK_DIRECT_DRIVER_LOADING_MODE_MAX_ENUM_LUNARG";
        default:
            return "Unknown VkDirectDriverLoadingModeLUNARG value.";
        }
    }


    const char *
    vk_DiscardRectangleModeEXT_to_str(VkDiscardRectangleModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DISCARD_RECTANGLE_MODE_INCLUSIVE_EXT";
        case 1:
            return "VK_DISCARD_RECTANGLE_MODE_EXCLUSIVE_EXT";
        case VK_DISCARD_RECTANGLE_MODE_MAX_ENUM_EXT: return "VK_DISCARD_RECTANGLE_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDiscardRectangleModeEXT value.";
        }
    }


    const char *
    vk_DisplayEventTypeEXT_to_str(VkDisplayEventTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DISPLAY_EVENT_TYPE_FIRST_PIXEL_OUT_EXT";
        case VK_DISPLAY_EVENT_TYPE_MAX_ENUM_EXT: return "VK_DISPLAY_EVENT_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDisplayEventTypeEXT value.";
        }
    }


    const char *
    vk_DisplayPowerStateEXT_to_str(VkDisplayPowerStateEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DISPLAY_POWER_STATE_OFF_EXT";
        case 1:
            return "VK_DISPLAY_POWER_STATE_SUSPEND_EXT";
        case 2:
            return "VK_DISPLAY_POWER_STATE_ON_EXT";
        case VK_DISPLAY_POWER_STATE_MAX_ENUM_EXT: return "VK_DISPLAY_POWER_STATE_MAX_ENUM_EXT";
        default:
            return "Unknown VkDisplayPowerStateEXT value.";
        }
    }


    const char *
    vk_DriverId_to_str(VkDriverId input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DRIVER_ID_AMD_PROPRIETARY";
        case 2:
            return "VK_DRIVER_ID_AMD_OPEN_SOURCE";
        case 3:
            return "VK_DRIVER_ID_MESA_RADV";
        case 4:
            return "VK_DRIVER_ID_NVIDIA_PROPRIETARY";
        case 5:
            return "VK_DRIVER_ID_INTEL_PROPRIETARY_WINDOWS";
        case 6:
            return "VK_DRIVER_ID_INTEL_OPEN_SOURCE_MESA";
        case 7:
            return "VK_DRIVER_ID_IMAGINATION_PROPRIETARY";
        case 8:
            return "VK_DRIVER_ID_QUALCOMM_PROPRIETARY";
        case 9:
            return "VK_DRIVER_ID_ARM_PROPRIETARY";
        case 10:
            return "VK_DRIVER_ID_GOOGLE_SWIFTSHADER";
        case 11:
            return "VK_DRIVER_ID_GGP_PROPRIETARY";
        case 12:
            return "VK_DRIVER_ID_BROADCOM_PROPRIETARY";
        case 13:
            return "VK_DRIVER_ID_MESA_LLVMPIPE";
        case 14:
            return "VK_DRIVER_ID_MOLTENVK";
        case 15:
            return "VK_DRIVER_ID_COREAVI_PROPRIETARY";
        case 16:
            return "VK_DRIVER_ID_JUICE_PROPRIETARY";
        case 17:
            return "VK_DRIVER_ID_VERISILICON_PROPRIETARY";
        case 18:
            return "VK_DRIVER_ID_MESA_TURNIP";
        case 19:
            return "VK_DRIVER_ID_MESA_V3DV";
        case 20:
            return "VK_DRIVER_ID_MESA_PANVK";
        case 21:
            return "VK_DRIVER_ID_SAMSUNG_PROPRIETARY";
        case 22:
            return "VK_DRIVER_ID_MESA_VENUS";
        case 23:
            return "VK_DRIVER_ID_MESA_DOZEN";
        case 24:
            return "VK_DRIVER_ID_MESA_NVK";
        case 25:
            return "VK_DRIVER_ID_IMAGINATION_OPEN_SOURCE_MESA";
        case 26:
            return "VK_DRIVER_ID_MESA_AGXV";
        case VK_DRIVER_ID_MAX_ENUM: return "VK_DRIVER_ID_MAX_ENUM";
        default:
            return "Unknown VkDriverId value.";
        }
    }


    const char *
    vk_DynamicState_to_str(VkDynamicState input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_DYNAMIC_STATE_VIEWPORT";
        case 1:
            return "VK_DYNAMIC_STATE_SCISSOR";
        case 2:
            return "VK_DYNAMIC_STATE_LINE_WIDTH";
        case 3:
            return "VK_DYNAMIC_STATE_DEPTH_BIAS";
        case 4:
            return "VK_DYNAMIC_STATE_BLEND_CONSTANTS";
        case 5:
            return "VK_DYNAMIC_STATE_DEPTH_BOUNDS";
        case 6:
            return "VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK";
        case 7:
            return "VK_DYNAMIC_STATE_STENCIL_WRITE_MASK";
        case 8:
            return "VK_DYNAMIC_STATE_STENCIL_REFERENCE";
        case 1000087000:
            return "VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_NV";
        case 1000099000:
            return "VK_DYNAMIC_STATE_DISCARD_RECTANGLE_EXT";
        case 1000099001:
            return "VK_DYNAMIC_STATE_DISCARD_RECTANGLE_ENABLE_EXT";
        case 1000099002:
            return "VK_DYNAMIC_STATE_DISCARD_RECTANGLE_MODE_EXT";
        case 1000143000:
            return "VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_EXT";
        case 1000164004:
            return "VK_DYNAMIC_STATE_VIEWPORT_SHADING_RATE_PALETTE_NV";
        case 1000164006:
            return "VK_DYNAMIC_STATE_VIEWPORT_COARSE_SAMPLE_ORDER_NV";
        case 1000205000:
            return "VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_ENABLE_NV";
        case 1000205001:
            return "VK_DYNAMIC_STATE_EXCLUSIVE_SCISSOR_NV";
        case 1000226000:
            return "VK_DYNAMIC_STATE_FRAGMENT_SHADING_RATE_KHR";
        case 1000259000:
            return "VK_DYNAMIC_STATE_LINE_STIPPLE_EXT";
        case 1000267000:
            return "VK_DYNAMIC_STATE_CULL_MODE";
        case 1000267001:
            return "VK_DYNAMIC_STATE_FRONT_FACE";
        case 1000267002:
            return "VK_DYNAMIC_STATE_PRIMITIVE_TOPOLOGY";
        case 1000267003:
            return "VK_DYNAMIC_STATE_VIEWPORT_WITH_COUNT";
        case 1000267004:
            return "VK_DYNAMIC_STATE_SCISSOR_WITH_COUNT";
        case 1000267005:
            return "VK_DYNAMIC_STATE_VERTEX_INPUT_BINDING_STRIDE";
        case 1000267006:
            return "VK_DYNAMIC_STATE_DEPTH_TEST_ENABLE";
        case 1000267007:
            return "VK_DYNAMIC_STATE_DEPTH_WRITE_ENABLE";
        case 1000267008:
            return "VK_DYNAMIC_STATE_DEPTH_COMPARE_OP";
        case 1000267009:
            return "VK_DYNAMIC_STATE_DEPTH_BOUNDS_TEST_ENABLE";
        case 1000267010:
            return "VK_DYNAMIC_STATE_STENCIL_TEST_ENABLE";
        case 1000267011:
            return "VK_DYNAMIC_STATE_STENCIL_OP";
        case 1000347000:
            return "VK_DYNAMIC_STATE_RAY_TRACING_PIPELINE_STACK_SIZE_KHR";
        case 1000352000:
            return "VK_DYNAMIC_STATE_VERTEX_INPUT_EXT";
        case 1000377000:
            return "VK_DYNAMIC_STATE_PATCH_CONTROL_POINTS_EXT";
        case 1000377001:
            return "VK_DYNAMIC_STATE_RASTERIZER_DISCARD_ENABLE";
        case 1000377002:
            return "VK_DYNAMIC_STATE_DEPTH_BIAS_ENABLE";
        case 1000377003:
            return "VK_DYNAMIC_STATE_LOGIC_OP_EXT";
        case 1000377004:
            return "VK_DYNAMIC_STATE_PRIMITIVE_RESTART_ENABLE";
        case 1000381000:
            return "VK_DYNAMIC_STATE_COLOR_WRITE_ENABLE_EXT";
        case 1000455002:
            return "VK_DYNAMIC_STATE_TESSELLATION_DOMAIN_ORIGIN_EXT";
        case 1000455003:
            return "VK_DYNAMIC_STATE_DEPTH_CLAMP_ENABLE_EXT";
        case 1000455004:
            return "VK_DYNAMIC_STATE_POLYGON_MODE_EXT";
        case 1000455005:
            return "VK_DYNAMIC_STATE_RASTERIZATION_SAMPLES_EXT";
        case 1000455006:
            return "VK_DYNAMIC_STATE_SAMPLE_MASK_EXT";
        case 1000455007:
            return "VK_DYNAMIC_STATE_ALPHA_TO_COVERAGE_ENABLE_EXT";
        case 1000455008:
            return "VK_DYNAMIC_STATE_ALPHA_TO_ONE_ENABLE_EXT";
        case 1000455009:
            return "VK_DYNAMIC_STATE_LOGIC_OP_ENABLE_EXT";
        case 1000455010:
            return "VK_DYNAMIC_STATE_COLOR_BLEND_ENABLE_EXT";
        case 1000455011:
            return "VK_DYNAMIC_STATE_COLOR_BLEND_EQUATION_EXT";
        case 1000455012:
            return "VK_DYNAMIC_STATE_COLOR_WRITE_MASK_EXT";
        case 1000455013:
            return "VK_DYNAMIC_STATE_RASTERIZATION_STREAM_EXT";
        case 1000455014:
            return "VK_DYNAMIC_STATE_CONSERVATIVE_RASTERIZATION_MODE_EXT";
        case 1000455015:
            return "VK_DYNAMIC_STATE_EXTRA_PRIMITIVE_OVERESTIMATION_SIZE_EXT";
        case 1000455016:
            return "VK_DYNAMIC_STATE_DEPTH_CLIP_ENABLE_EXT";
        case 1000455017:
            return "VK_DYNAMIC_STATE_SAMPLE_LOCATIONS_ENABLE_EXT";
        case 1000455018:
            return "VK_DYNAMIC_STATE_COLOR_BLEND_ADVANCED_EXT";
        case 1000455019:
            return "VK_DYNAMIC_STATE_PROVOKING_VERTEX_MODE_EXT";
        case 1000455020:
            return "VK_DYNAMIC_STATE_LINE_RASTERIZATION_MODE_EXT";
        case 1000455021:
            return "VK_DYNAMIC_STATE_LINE_STIPPLE_ENABLE_EXT";
        case 1000455022:
            return "VK_DYNAMIC_STATE_DEPTH_CLIP_NEGATIVE_ONE_TO_ONE_EXT";
        case 1000455023:
            return "VK_DYNAMIC_STATE_VIEWPORT_W_SCALING_ENABLE_NV";
        case 1000455024:
            return "VK_DYNAMIC_STATE_VIEWPORT_SWIZZLE_NV";
        case 1000455025:
            return "VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_ENABLE_NV";
        case 1000455026:
            return "VK_DYNAMIC_STATE_COVERAGE_TO_COLOR_LOCATION_NV";
        case 1000455027:
            return "VK_DYNAMIC_STATE_COVERAGE_MODULATION_MODE_NV";
        case 1000455028:
            return "VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_ENABLE_NV";
        case 1000455029:
            return "VK_DYNAMIC_STATE_COVERAGE_MODULATION_TABLE_NV";
        case 1000455030:
            return "VK_DYNAMIC_STATE_SHADING_RATE_IMAGE_ENABLE_NV";
        case 1000455031:
            return "VK_DYNAMIC_STATE_REPRESENTATIVE_FRAGMENT_TEST_ENABLE_NV";
        case 1000455032:
            return "VK_DYNAMIC_STATE_COVERAGE_REDUCTION_MODE_NV";
        case 1000524000:
            return "VK_DYNAMIC_STATE_ATTACHMENT_FEEDBACK_LOOP_ENABLE_EXT";
        case VK_DYNAMIC_STATE_MAX_ENUM: return "VK_DYNAMIC_STATE_MAX_ENUM";
        default:
            return "Unknown VkDynamicState value.";
        }
    }


    const char *
    vk_Filter_to_str(VkFilter input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FILTER_NEAREST";
        case 1:
            return "VK_FILTER_LINEAR";
        case 1000015000:
            return "VK_FILTER_CUBIC_EXT";
        case VK_FILTER_MAX_ENUM: return "VK_FILTER_MAX_ENUM";
        default:
            return "Unknown VkFilter value.";
        }
    }


    const char *
    vk_Format_to_str(VkFormat input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FORMAT_UNDEFINED";
        case 1:
            return "VK_FORMAT_R4G4_UNORM_PACK8";
        case 2:
            return "VK_FORMAT_R4G4B4A4_UNORM_PACK16";
        case 3:
            return "VK_FORMAT_B4G4R4A4_UNORM_PACK16";
        case 4:
            return "VK_FORMAT_R5G6B5_UNORM_PACK16";
        case 5:
            return "VK_FORMAT_B5G6R5_UNORM_PACK16";
        case 6:
            return "VK_FORMAT_R5G5B5A1_UNORM_PACK16";
        case 7:
            return "VK_FORMAT_B5G5R5A1_UNORM_PACK16";
        case 8:
            return "VK_FORMAT_A1R5G5B5_UNORM_PACK16";
        case 9:
            return "VK_FORMAT_R8_UNORM";
        case 10:
            return "VK_FORMAT_R8_SNORM";
        case 11:
            return "VK_FORMAT_R8_USCALED";
        case 12:
            return "VK_FORMAT_R8_SSCALED";
        case 13:
            return "VK_FORMAT_R8_UINT";
        case 14:
            return "VK_FORMAT_R8_SINT";
        case 15:
            return "VK_FORMAT_R8_SRGB";
        case 16:
            return "VK_FORMAT_R8G8_UNORM";
        case 17:
            return "VK_FORMAT_R8G8_SNORM";
        case 18:
            return "VK_FORMAT_R8G8_USCALED";
        case 19:
            return "VK_FORMAT_R8G8_SSCALED";
        case 20:
            return "VK_FORMAT_R8G8_UINT";
        case 21:
            return "VK_FORMAT_R8G8_SINT";
        case 22:
            return "VK_FORMAT_R8G8_SRGB";
        case 23:
            return "VK_FORMAT_R8G8B8_UNORM";
        case 24:
            return "VK_FORMAT_R8G8B8_SNORM";
        case 25:
            return "VK_FORMAT_R8G8B8_USCALED";
        case 26:
            return "VK_FORMAT_R8G8B8_SSCALED";
        case 27:
            return "VK_FORMAT_R8G8B8_UINT";
        case 28:
            return "VK_FORMAT_R8G8B8_SINT";
        case 29:
            return "VK_FORMAT_R8G8B8_SRGB";
        case 30:
            return "VK_FORMAT_B8G8R8_UNORM";
        case 31:
            return "VK_FORMAT_B8G8R8_SNORM";
        case 32:
            return "VK_FORMAT_B8G8R8_USCALED";
        case 33:
            return "VK_FORMAT_B8G8R8_SSCALED";
        case 34:
            return "VK_FORMAT_B8G8R8_UINT";
        case 35:
            return "VK_FORMAT_B8G8R8_SINT";
        case 36:
            return "VK_FORMAT_B8G8R8_SRGB";
        case 37:
            return "VK_FORMAT_R8G8B8A8_UNORM";
        case 38:
            return "VK_FORMAT_R8G8B8A8_SNORM";
        case 39:
            return "VK_FORMAT_R8G8B8A8_USCALED";
        case 40:
            return "VK_FORMAT_R8G8B8A8_SSCALED";
        case 41:
            return "VK_FORMAT_R8G8B8A8_UINT";
        case 42:
            return "VK_FORMAT_R8G8B8A8_SINT";
        case 43:
            return "VK_FORMAT_R8G8B8A8_SRGB";
        case 44:
            return "VK_FORMAT_B8G8R8A8_UNORM";
        case 45:
            return "VK_FORMAT_B8G8R8A8_SNORM";
        case 46:
            return "VK_FORMAT_B8G8R8A8_USCALED";
        case 47:
            return "VK_FORMAT_B8G8R8A8_SSCALED";
        case 48:
            return "VK_FORMAT_B8G8R8A8_UINT";
        case 49:
            return "VK_FORMAT_B8G8R8A8_SINT";
        case 50:
            return "VK_FORMAT_B8G8R8A8_SRGB";
        case 51:
            return "VK_FORMAT_A8B8G8R8_UNORM_PACK32";
        case 52:
            return "VK_FORMAT_A8B8G8R8_SNORM_PACK32";
        case 53:
            return "VK_FORMAT_A8B8G8R8_USCALED_PACK32";
        case 54:
            return "VK_FORMAT_A8B8G8R8_SSCALED_PACK32";
        case 55:
            return "VK_FORMAT_A8B8G8R8_UINT_PACK32";
        case 56:
            return "VK_FORMAT_A8B8G8R8_SINT_PACK32";
        case 57:
            return "VK_FORMAT_A8B8G8R8_SRGB_PACK32";
        case 58:
            return "VK_FORMAT_A2R10G10B10_UNORM_PACK32";
        case 59:
            return "VK_FORMAT_A2R10G10B10_SNORM_PACK32";
        case 60:
            return "VK_FORMAT_A2R10G10B10_USCALED_PACK32";
        case 61:
            return "VK_FORMAT_A2R10G10B10_SSCALED_PACK32";
        case 62:
            return "VK_FORMAT_A2R10G10B10_UINT_PACK32";
        case 63:
            return "VK_FORMAT_A2R10G10B10_SINT_PACK32";
        case 64:
            return "VK_FORMAT_A2B10G10R10_UNORM_PACK32";
        case 65:
            return "VK_FORMAT_A2B10G10R10_SNORM_PACK32";
        case 66:
            return "VK_FORMAT_A2B10G10R10_USCALED_PACK32";
        case 67:
            return "VK_FORMAT_A2B10G10R10_SSCALED_PACK32";
        case 68:
            return "VK_FORMAT_A2B10G10R10_UINT_PACK32";
        case 69:
            return "VK_FORMAT_A2B10G10R10_SINT_PACK32";
        case 70:
            return "VK_FORMAT_R16_UNORM";
        case 71:
            return "VK_FORMAT_R16_SNORM";
        case 72:
            return "VK_FORMAT_R16_USCALED";
        case 73:
            return "VK_FORMAT_R16_SSCALED";
        case 74:
            return "VK_FORMAT_R16_UINT";
        case 75:
            return "VK_FORMAT_R16_SINT";
        case 76:
            return "VK_FORMAT_R16_SFLOAT";
        case 77:
            return "VK_FORMAT_R16G16_UNORM";
        case 78:
            return "VK_FORMAT_R16G16_SNORM";
        case 79:
            return "VK_FORMAT_R16G16_USCALED";
        case 80:
            return "VK_FORMAT_R16G16_SSCALED";
        case 81:
            return "VK_FORMAT_R16G16_UINT";
        case 82:
            return "VK_FORMAT_R16G16_SINT";
        case 83:
            return "VK_FORMAT_R16G16_SFLOAT";
        case 84:
            return "VK_FORMAT_R16G16B16_UNORM";
        case 85:
            return "VK_FORMAT_R16G16B16_SNORM";
        case 86:
            return "VK_FORMAT_R16G16B16_USCALED";
        case 87:
            return "VK_FORMAT_R16G16B16_SSCALED";
        case 88:
            return "VK_FORMAT_R16G16B16_UINT";
        case 89:
            return "VK_FORMAT_R16G16B16_SINT";
        case 90:
            return "VK_FORMAT_R16G16B16_SFLOAT";
        case 91:
            return "VK_FORMAT_R16G16B16A16_UNORM";
        case 92:
            return "VK_FORMAT_R16G16B16A16_SNORM";
        case 93:
            return "VK_FORMAT_R16G16B16A16_USCALED";
        case 94:
            return "VK_FORMAT_R16G16B16A16_SSCALED";
        case 95:
            return "VK_FORMAT_R16G16B16A16_UINT";
        case 96:
            return "VK_FORMAT_R16G16B16A16_SINT";
        case 97:
            return "VK_FORMAT_R16G16B16A16_SFLOAT";
        case 98:
            return "VK_FORMAT_R32_UINT";
        case 99:
            return "VK_FORMAT_R32_SINT";
        case 100:
            return "VK_FORMAT_R32_SFLOAT";
        case 101:
            return "VK_FORMAT_R32G32_UINT";
        case 102:
            return "VK_FORMAT_R32G32_SINT";
        case 103:
            return "VK_FORMAT_R32G32_SFLOAT";
        case 104:
            return "VK_FORMAT_R32G32B32_UINT";
        case 105:
            return "VK_FORMAT_R32G32B32_SINT";
        case 106:
            return "VK_FORMAT_R32G32B32_SFLOAT";
        case 107:
            return "VK_FORMAT_R32G32B32A32_UINT";
        case 108:
            return "VK_FORMAT_R32G32B32A32_SINT";
        case 109:
            return "VK_FORMAT_R32G32B32A32_SFLOAT";
        case 110:
            return "VK_FORMAT_R64_UINT";
        case 111:
            return "VK_FORMAT_R64_SINT";
        case 112:
            return "VK_FORMAT_R64_SFLOAT";
        case 113:
            return "VK_FORMAT_R64G64_UINT";
        case 114:
            return "VK_FORMAT_R64G64_SINT";
        case 115:
            return "VK_FORMAT_R64G64_SFLOAT";
        case 116:
            return "VK_FORMAT_R64G64B64_UINT";
        case 117:
            return "VK_FORMAT_R64G64B64_SINT";
        case 118:
            return "VK_FORMAT_R64G64B64_SFLOAT";
        case 119:
            return "VK_FORMAT_R64G64B64A64_UINT";
        case 120:
            return "VK_FORMAT_R64G64B64A64_SINT";
        case 121:
            return "VK_FORMAT_R64G64B64A64_SFLOAT";
        case 122:
            return "VK_FORMAT_B10G11R11_UFLOAT_PACK32";
        case 123:
            return "VK_FORMAT_E5B9G9R9_UFLOAT_PACK32";
        case 124:
            return "VK_FORMAT_D16_UNORM";
        case 125:
            return "VK_FORMAT_X8_D24_UNORM_PACK32";
        case 126:
            return "VK_FORMAT_D32_SFLOAT";
        case 127:
            return "VK_FORMAT_S8_UINT";
        case 128:
            return "VK_FORMAT_D16_UNORM_S8_UINT";
        case 129:
            return "VK_FORMAT_D24_UNORM_S8_UINT";
        case 130:
            return "VK_FORMAT_D32_SFLOAT_S8_UINT";
        case 131:
            return "VK_FORMAT_BC1_RGB_UNORM_BLOCK";
        case 132:
            return "VK_FORMAT_BC1_RGB_SRGB_BLOCK";
        case 133:
            return "VK_FORMAT_BC1_RGBA_UNORM_BLOCK";
        case 134:
            return "VK_FORMAT_BC1_RGBA_SRGB_BLOCK";
        case 135:
            return "VK_FORMAT_BC2_UNORM_BLOCK";
        case 136:
            return "VK_FORMAT_BC2_SRGB_BLOCK";
        case 137:
            return "VK_FORMAT_BC3_UNORM_BLOCK";
        case 138:
            return "VK_FORMAT_BC3_SRGB_BLOCK";
        case 139:
            return "VK_FORMAT_BC4_UNORM_BLOCK";
        case 140:
            return "VK_FORMAT_BC4_SNORM_BLOCK";
        case 141:
            return "VK_FORMAT_BC5_UNORM_BLOCK";
        case 142:
            return "VK_FORMAT_BC5_SNORM_BLOCK";
        case 143:
            return "VK_FORMAT_BC6H_UFLOAT_BLOCK";
        case 144:
            return "VK_FORMAT_BC6H_SFLOAT_BLOCK";
        case 145:
            return "VK_FORMAT_BC7_UNORM_BLOCK";
        case 146:
            return "VK_FORMAT_BC7_SRGB_BLOCK";
        case 147:
            return "VK_FORMAT_ETC2_R8G8B8_UNORM_BLOCK";
        case 148:
            return "VK_FORMAT_ETC2_R8G8B8_SRGB_BLOCK";
        case 149:
            return "VK_FORMAT_ETC2_R8G8B8A1_UNORM_BLOCK";
        case 150:
            return "VK_FORMAT_ETC2_R8G8B8A1_SRGB_BLOCK";
        case 151:
            return "VK_FORMAT_ETC2_R8G8B8A8_UNORM_BLOCK";
        case 152:
            return "VK_FORMAT_ETC2_R8G8B8A8_SRGB_BLOCK";
        case 153:
            return "VK_FORMAT_EAC_R11_UNORM_BLOCK";
        case 154:
            return "VK_FORMAT_EAC_R11_SNORM_BLOCK";
        case 155:
            return "VK_FORMAT_EAC_R11G11_UNORM_BLOCK";
        case 156:
            return "VK_FORMAT_EAC_R11G11_SNORM_BLOCK";
        case 157:
            return "VK_FORMAT_ASTC_4x4_UNORM_BLOCK";
        case 158:
            return "VK_FORMAT_ASTC_4x4_SRGB_BLOCK";
        case 159:
            return "VK_FORMAT_ASTC_5x4_UNORM_BLOCK";
        case 160:
            return "VK_FORMAT_ASTC_5x4_SRGB_BLOCK";
        case 161:
            return "VK_FORMAT_ASTC_5x5_UNORM_BLOCK";
        case 162:
            return "VK_FORMAT_ASTC_5x5_SRGB_BLOCK";
        case 163:
            return "VK_FORMAT_ASTC_6x5_UNORM_BLOCK";
        case 164:
            return "VK_FORMAT_ASTC_6x5_SRGB_BLOCK";
        case 165:
            return "VK_FORMAT_ASTC_6x6_UNORM_BLOCK";
        case 166:
            return "VK_FORMAT_ASTC_6x6_SRGB_BLOCK";
        case 167:
            return "VK_FORMAT_ASTC_8x5_UNORM_BLOCK";
        case 168:
            return "VK_FORMAT_ASTC_8x5_SRGB_BLOCK";
        case 169:
            return "VK_FORMAT_ASTC_8x6_UNORM_BLOCK";
        case 170:
            return "VK_FORMAT_ASTC_8x6_SRGB_BLOCK";
        case 171:
            return "VK_FORMAT_ASTC_8x8_UNORM_BLOCK";
        case 172:
            return "VK_FORMAT_ASTC_8x8_SRGB_BLOCK";
        case 173:
            return "VK_FORMAT_ASTC_10x5_UNORM_BLOCK";
        case 174:
            return "VK_FORMAT_ASTC_10x5_SRGB_BLOCK";
        case 175:
            return "VK_FORMAT_ASTC_10x6_UNORM_BLOCK";
        case 176:
            return "VK_FORMAT_ASTC_10x6_SRGB_BLOCK";
        case 177:
            return "VK_FORMAT_ASTC_10x8_UNORM_BLOCK";
        case 178:
            return "VK_FORMAT_ASTC_10x8_SRGB_BLOCK";
        case 179:
            return "VK_FORMAT_ASTC_10x10_UNORM_BLOCK";
        case 180:
            return "VK_FORMAT_ASTC_10x10_SRGB_BLOCK";
        case 181:
            return "VK_FORMAT_ASTC_12x10_UNORM_BLOCK";
        case 182:
            return "VK_FORMAT_ASTC_12x10_SRGB_BLOCK";
        case 183:
            return "VK_FORMAT_ASTC_12x12_UNORM_BLOCK";
        case 184:
            return "VK_FORMAT_ASTC_12x12_SRGB_BLOCK";
        case 1000054000:
            return "VK_FORMAT_PVRTC1_2BPP_UNORM_BLOCK_IMG";
        case 1000054001:
            return "VK_FORMAT_PVRTC1_4BPP_UNORM_BLOCK_IMG";
        case 1000054002:
            return "VK_FORMAT_PVRTC2_2BPP_UNORM_BLOCK_IMG";
        case 1000054003:
            return "VK_FORMAT_PVRTC2_4BPP_UNORM_BLOCK_IMG";
        case 1000054004:
            return "VK_FORMAT_PVRTC1_2BPP_SRGB_BLOCK_IMG";
        case 1000054005:
            return "VK_FORMAT_PVRTC1_4BPP_SRGB_BLOCK_IMG";
        case 1000054006:
            return "VK_FORMAT_PVRTC2_2BPP_SRGB_BLOCK_IMG";
        case 1000054007:
            return "VK_FORMAT_PVRTC2_4BPP_SRGB_BLOCK_IMG";
        case 1000066000:
            return "VK_FORMAT_ASTC_4x4_SFLOAT_BLOCK";
        case 1000066001:
            return "VK_FORMAT_ASTC_5x4_SFLOAT_BLOCK";
        case 1000066002:
            return "VK_FORMAT_ASTC_5x5_SFLOAT_BLOCK";
        case 1000066003:
            return "VK_FORMAT_ASTC_6x5_SFLOAT_BLOCK";
        case 1000066004:
            return "VK_FORMAT_ASTC_6x6_SFLOAT_BLOCK";
        case 1000066005:
            return "VK_FORMAT_ASTC_8x5_SFLOAT_BLOCK";
        case 1000066006:
            return "VK_FORMAT_ASTC_8x6_SFLOAT_BLOCK";
        case 1000066007:
            return "VK_FORMAT_ASTC_8x8_SFLOAT_BLOCK";
        case 1000066008:
            return "VK_FORMAT_ASTC_10x5_SFLOAT_BLOCK";
        case 1000066009:
            return "VK_FORMAT_ASTC_10x6_SFLOAT_BLOCK";
        case 1000066010:
            return "VK_FORMAT_ASTC_10x8_SFLOAT_BLOCK";
        case 1000066011:
            return "VK_FORMAT_ASTC_10x10_SFLOAT_BLOCK";
        case 1000066012:
            return "VK_FORMAT_ASTC_12x10_SFLOAT_BLOCK";
        case 1000066013:
            return "VK_FORMAT_ASTC_12x12_SFLOAT_BLOCK";
        case 1000156000:
            return "VK_FORMAT_G8B8G8R8_422_UNORM";
        case 1000156001:
            return "VK_FORMAT_B8G8R8G8_422_UNORM";
        case 1000156002:
            return "VK_FORMAT_G8_B8_R8_3PLANE_420_UNORM";
        case 1000156003:
            return "VK_FORMAT_G8_B8R8_2PLANE_420_UNORM";
        case 1000156004:
            return "VK_FORMAT_G8_B8_R8_3PLANE_422_UNORM";
        case 1000156005:
            return "VK_FORMAT_G8_B8R8_2PLANE_422_UNORM";
        case 1000156006:
            return "VK_FORMAT_G8_B8_R8_3PLANE_444_UNORM";
        case 1000156007:
            return "VK_FORMAT_R10X6_UNORM_PACK16";
        case 1000156008:
            return "VK_FORMAT_R10X6G10X6_UNORM_2PACK16";
        case 1000156009:
            return "VK_FORMAT_R10X6G10X6B10X6A10X6_UNORM_4PACK16";
        case 1000156010:
            return "VK_FORMAT_G10X6B10X6G10X6R10X6_422_UNORM_4PACK16";
        case 1000156011:
            return "VK_FORMAT_B10X6G10X6R10X6G10X6_422_UNORM_4PACK16";
        case 1000156012:
            return "VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_420_UNORM_3PACK16";
        case 1000156013:
            return "VK_FORMAT_G10X6_B10X6R10X6_2PLANE_420_UNORM_3PACK16";
        case 1000156014:
            return "VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_422_UNORM_3PACK16";
        case 1000156015:
            return "VK_FORMAT_G10X6_B10X6R10X6_2PLANE_422_UNORM_3PACK16";
        case 1000156016:
            return "VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_444_UNORM_3PACK16";
        case 1000156017:
            return "VK_FORMAT_R12X4_UNORM_PACK16";
        case 1000156018:
            return "VK_FORMAT_R12X4G12X4_UNORM_2PACK16";
        case 1000156019:
            return "VK_FORMAT_R12X4G12X4B12X4A12X4_UNORM_4PACK16";
        case 1000156020:
            return "VK_FORMAT_G12X4B12X4G12X4R12X4_422_UNORM_4PACK16";
        case 1000156021:
            return "VK_FORMAT_B12X4G12X4R12X4G12X4_422_UNORM_4PACK16";
        case 1000156022:
            return "VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_420_UNORM_3PACK16";
        case 1000156023:
            return "VK_FORMAT_G12X4_B12X4R12X4_2PLANE_420_UNORM_3PACK16";
        case 1000156024:
            return "VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_422_UNORM_3PACK16";
        case 1000156025:
            return "VK_FORMAT_G12X4_B12X4R12X4_2PLANE_422_UNORM_3PACK16";
        case 1000156026:
            return "VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_444_UNORM_3PACK16";
        case 1000156027:
            return "VK_FORMAT_G16B16G16R16_422_UNORM";
        case 1000156028:
            return "VK_FORMAT_B16G16R16G16_422_UNORM";
        case 1000156029:
            return "VK_FORMAT_G16_B16_R16_3PLANE_420_UNORM";
        case 1000156030:
            return "VK_FORMAT_G16_B16R16_2PLANE_420_UNORM";
        case 1000156031:
            return "VK_FORMAT_G16_B16_R16_3PLANE_422_UNORM";
        case 1000156032:
            return "VK_FORMAT_G16_B16R16_2PLANE_422_UNORM";
        case 1000156033:
            return "VK_FORMAT_G16_B16_R16_3PLANE_444_UNORM";
        case 1000330000:
            return "VK_FORMAT_G8_B8R8_2PLANE_444_UNORM";
        case 1000330001:
            return "VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16";
        case 1000330002:
            return "VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16";
        case 1000330003:
            return "VK_FORMAT_G16_B16R16_2PLANE_444_UNORM";
        case 1000340000:
            return "VK_FORMAT_A4R4G4B4_UNORM_PACK16";
        case 1000340001:
            return "VK_FORMAT_A4B4G4R4_UNORM_PACK16";
        case 1000464000:
            return "VK_FORMAT_R16G16_S10_5_NV";
        case 1000470000:
            return "VK_FORMAT_A1B5G5R5_UNORM_PACK16_KHR";
        case 1000470001:
            return "VK_FORMAT_A8_UNORM_KHR";
        case VK_FORMAT_MAX_ENUM: return "VK_FORMAT_MAX_ENUM";
        default:
            return "Unknown VkFormat value.";
        }
    }


    const char *
    vk_FragmentShadingRateCombinerOpKHR_to_str(VkFragmentShadingRateCombinerOpKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_KEEP_KHR";
        case 1:
            return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_REPLACE_KHR";
        case 2:
            return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_MIN_KHR";
        case 3:
            return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_MAX_KHR";
        case 4:
            return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_MUL_KHR";
        case VK_FRAGMENT_SHADING_RATE_COMBINER_OP_MAX_ENUM_KHR: return "VK_FRAGMENT_SHADING_RATE_COMBINER_OP_MAX_ENUM_KHR";
        default:
            return "Unknown VkFragmentShadingRateCombinerOpKHR value.";
        }
    }


    const char *
    vk_FragmentShadingRateNV_to_str(VkFragmentShadingRateNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_PIXEL_NV";
        case 1:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_1X2_PIXELS_NV";
        case 4:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_2X1_PIXELS_NV";
        case 5:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_2X2_PIXELS_NV";
        case 6:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_2X4_PIXELS_NV";
        case 9:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_4X2_PIXELS_NV";
        case 10:
            return "VK_FRAGMENT_SHADING_RATE_1_INVOCATION_PER_4X4_PIXELS_NV";
        case 11:
            return "VK_FRAGMENT_SHADING_RATE_2_INVOCATIONS_PER_PIXEL_NV";
        case 12:
            return "VK_FRAGMENT_SHADING_RATE_4_INVOCATIONS_PER_PIXEL_NV";
        case 13:
            return "VK_FRAGMENT_SHADING_RATE_8_INVOCATIONS_PER_PIXEL_NV";
        case 14:
            return "VK_FRAGMENT_SHADING_RATE_16_INVOCATIONS_PER_PIXEL_NV";
        case 15:
            return "VK_FRAGMENT_SHADING_RATE_NO_INVOCATIONS_NV";
        case VK_FRAGMENT_SHADING_RATE_MAX_ENUM_NV: return "VK_FRAGMENT_SHADING_RATE_MAX_ENUM_NV";
        default:
            return "Unknown VkFragmentShadingRateNV value.";
        }
    }


    const char *
    vk_FragmentShadingRateTypeNV_to_str(VkFragmentShadingRateTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FRAGMENT_SHADING_RATE_TYPE_FRAGMENT_SIZE_NV";
        case 1:
            return "VK_FRAGMENT_SHADING_RATE_TYPE_ENUMS_NV";
        case VK_FRAGMENT_SHADING_RATE_TYPE_MAX_ENUM_NV: return "VK_FRAGMENT_SHADING_RATE_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkFragmentShadingRateTypeNV value.";
        }
    }


    const char *
    vk_FrontFace_to_str(VkFrontFace input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FRONT_FACE_COUNTER_CLOCKWISE";
        case 1:
            return "VK_FRONT_FACE_CLOCKWISE";
        case VK_FRONT_FACE_MAX_ENUM: return "VK_FRONT_FACE_MAX_ENUM";
        default:
            return "Unknown VkFrontFace value.";
        }
    }


#ifdef VK_USE_PLATFORM_WIN32_KHR
    const char *
    vk_FullScreenExclusiveEXT_to_str(VkFullScreenExclusiveEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_FULL_SCREEN_EXCLUSIVE_DEFAULT_EXT";
        case 1:
            return "VK_FULL_SCREEN_EXCLUSIVE_ALLOWED_EXT";
        case 2:
            return "VK_FULL_SCREEN_EXCLUSIVE_DISALLOWED_EXT";
        case 3:
            return "VK_FULL_SCREEN_EXCLUSIVE_APPLICATION_CONTROLLED_EXT";
        case VK_FULL_SCREEN_EXCLUSIVE_MAX_ENUM_EXT: return "VK_FULL_SCREEN_EXCLUSIVE_MAX_ENUM_EXT";
        default:
            return "Unknown VkFullScreenExclusiveEXT value.";
        }
    }

#endif

    const char *
    vk_GeometryTypeKHR_to_str(VkGeometryTypeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_GEOMETRY_TYPE_TRIANGLES_NV";
        case 1:
            return "VK_GEOMETRY_TYPE_AABBS_NV";
        case 2:
            return "VK_GEOMETRY_TYPE_INSTANCES_KHR";
        case VK_GEOMETRY_TYPE_MAX_ENUM_KHR: return "VK_GEOMETRY_TYPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkGeometryTypeKHR value.";
        }
    }


    const char *
    vk_ImageLayout_to_str(VkImageLayout input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_LAYOUT_UNDEFINED";
        case 1:
            return "VK_IMAGE_LAYOUT_GENERAL";
        case 2:
            return "VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL";
        case 3:
            return "VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL";
        case 4:
            return "VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL";
        case 5:
            return "VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL";
        case 6:
            return "VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL";
        case 7:
            return "VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL";
        case 8:
            return "VK_IMAGE_LAYOUT_PREINITIALIZED";
        case 1000001002:
            return "VK_IMAGE_LAYOUT_PRESENT_SRC_KHR";
        case 1000024000:
            return "VK_IMAGE_LAYOUT_VIDEO_DECODE_DST_KHR";
        case 1000024001:
            return "VK_IMAGE_LAYOUT_VIDEO_DECODE_SRC_KHR";
        case 1000024002:
            return "VK_IMAGE_LAYOUT_VIDEO_DECODE_DPB_KHR";
        case 1000111000:
            return "VK_IMAGE_LAYOUT_SHARED_PRESENT_KHR";
        case 1000117000:
            return "VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_STENCIL_ATTACHMENT_OPTIMAL";
        case 1000117001:
            return "VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_STENCIL_READ_ONLY_OPTIMAL";
        case 1000164003:
            return "VK_IMAGE_LAYOUT_SHADING_RATE_OPTIMAL_NV";
        case 1000218000:
            return "VK_IMAGE_LAYOUT_FRAGMENT_DENSITY_MAP_OPTIMAL_EXT";
        case 1000241000:
            return "VK_IMAGE_LAYOUT_DEPTH_ATTACHMENT_OPTIMAL";
        case 1000241001:
            return "VK_IMAGE_LAYOUT_DEPTH_READ_ONLY_OPTIMAL";
        case 1000241002:
            return "VK_IMAGE_LAYOUT_STENCIL_ATTACHMENT_OPTIMAL";
        case 1000241003:
            return "VK_IMAGE_LAYOUT_STENCIL_READ_ONLY_OPTIMAL";
        case 1000299000:
            return "VK_IMAGE_LAYOUT_VIDEO_ENCODE_DST_KHR";
        case 1000299001:
            return "VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR";
        case 1000299002:
            return "VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR";
        case 1000314000:
            return "VK_IMAGE_LAYOUT_READ_ONLY_OPTIMAL";
        case 1000314001:
            return "VK_IMAGE_LAYOUT_ATTACHMENT_OPTIMAL";
        case 1000339000:
            return "VK_IMAGE_LAYOUT_ATTACHMENT_FEEDBACK_LOOP_OPTIMAL_EXT";
        case VK_IMAGE_LAYOUT_MAX_ENUM: return "VK_IMAGE_LAYOUT_MAX_ENUM";
        default:
            return "Unknown VkImageLayout value.";
        }
    }


    const char *
    vk_ImageTiling_to_str(VkImageTiling input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_TILING_OPTIMAL";
        case 1:
            return "VK_IMAGE_TILING_LINEAR";
        case 1000158000:
            return "VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT";
        case VK_IMAGE_TILING_MAX_ENUM: return "VK_IMAGE_TILING_MAX_ENUM";
        default:
            return "Unknown VkImageTiling value.";
        }
    }


    const char *
    vk_ImageType_to_str(VkImageType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_TYPE_1D";
        case 1:
            return "VK_IMAGE_TYPE_2D";
        case 2:
            return "VK_IMAGE_TYPE_3D";
        case VK_IMAGE_TYPE_MAX_ENUM: return "VK_IMAGE_TYPE_MAX_ENUM";
        default:
            return "Unknown VkImageType value.";
        }
    }


    const char *
    vk_ImageViewType_to_str(VkImageViewType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_VIEW_TYPE_1D";
        case 1:
            return "VK_IMAGE_VIEW_TYPE_2D";
        case 2:
            return "VK_IMAGE_VIEW_TYPE_3D";
        case 3:
            return "VK_IMAGE_VIEW_TYPE_CUBE";
        case 4:
            return "VK_IMAGE_VIEW_TYPE_1D_ARRAY";
        case 5:
            return "VK_IMAGE_VIEW_TYPE_2D_ARRAY";
        case 6:
            return "VK_IMAGE_VIEW_TYPE_CUBE_ARRAY";
        case VK_IMAGE_VIEW_TYPE_MAX_ENUM: return "VK_IMAGE_VIEW_TYPE_MAX_ENUM";
        default:
            return "Unknown VkImageViewType value.";
        }
    }


    const char *
    vk_IndexType_to_str(VkIndexType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_INDEX_TYPE_UINT16";
        case 1:
            return "VK_INDEX_TYPE_UINT32";
        case 1000165000:
            return "VK_INDEX_TYPE_NONE_NV";
        case 1000265000:
            return "VK_INDEX_TYPE_UINT8_EXT";
        case VK_INDEX_TYPE_MAX_ENUM: return "VK_INDEX_TYPE_MAX_ENUM";
        default:
            return "Unknown VkIndexType value.";
        }
    }


    const char *
    vk_IndirectCommandsTokenTypeNV_to_str(VkIndirectCommandsTokenTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_SHADER_GROUP_NV";
        case 1:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_STATE_FLAGS_NV";
        case 2:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_INDEX_BUFFER_NV";
        case 3:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_VERTEX_BUFFER_NV";
        case 4:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_PUSH_CONSTANT_NV";
        case 5:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_INDEXED_NV";
        case 6:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_NV";
        case 7:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_TASKS_NV";
        case 1000328000:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_DRAW_MESH_TASKS_NV";
        case 1000428003:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_PIPELINE_NV";
        case 1000428004:
            return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_DISPATCH_NV";
        case VK_INDIRECT_COMMANDS_TOKEN_TYPE_MAX_ENUM_NV: return "VK_INDIRECT_COMMANDS_TOKEN_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkIndirectCommandsTokenTypeNV value.";
        }
    }


    const char *
    vk_InternalAllocationType_to_str(VkInternalAllocationType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_INTERNAL_ALLOCATION_TYPE_EXECUTABLE";
        case VK_INTERNAL_ALLOCATION_TYPE_MAX_ENUM: return "VK_INTERNAL_ALLOCATION_TYPE_MAX_ENUM";
        default:
            return "Unknown VkInternalAllocationType value.";
        }
    }


    const char *
    vk_LatencyMarkerNV_to_str(VkLatencyMarkerNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_LATENCY_MARKER_SIMULATION_START_NV";
        case 1:
            return "VK_LATENCY_MARKER_SIMULATION_END_NV";
        case 2:
            return "VK_LATENCY_MARKER_RENDERSUBMIT_START_NV";
        case 3:
            return "VK_LATENCY_MARKER_RENDERSUBMIT_END_NV";
        case 4:
            return "VK_LATENCY_MARKER_PRESENT_START_NV";
        case 5:
            return "VK_LATENCY_MARKER_PRESENT_END_NV";
        case 6:
            return "VK_LATENCY_MARKER_INPUT_SAMPLE_NV";
        case 7:
            return "VK_LATENCY_MARKER_TRIGGER_FLASH_NV";
        case 8:
            return "VK_LATENCY_MARKER_OUT_OF_BAND_RENDERSUBMIT_START_NV";
        case 9:
            return "VK_LATENCY_MARKER_OUT_OF_BAND_RENDERSUBMIT_END_NV";
        case 10:
            return "VK_LATENCY_MARKER_OUT_OF_BAND_PRESENT_START_NV";
        case 11:
            return "VK_LATENCY_MARKER_OUT_OF_BAND_PRESENT_END_NV";
        case VK_LATENCY_MARKER_MAX_ENUM_NV: return "VK_LATENCY_MARKER_MAX_ENUM_NV";
        default:
            return "Unknown VkLatencyMarkerNV value.";
        }
    }


    const char *
    vk_LayerSettingTypeEXT_to_str(VkLayerSettingTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_LAYER_SETTING_TYPE_BOOL32_EXT";
        case 1:
            return "VK_LAYER_SETTING_TYPE_INT32_EXT";
        case 2:
            return "VK_LAYER_SETTING_TYPE_INT64_EXT";
        case 3:
            return "VK_LAYER_SETTING_TYPE_UINT32_EXT";
        case 4:
            return "VK_LAYER_SETTING_TYPE_UINT64_EXT";
        case 5:
            return "VK_LAYER_SETTING_TYPE_FLOAT32_EXT";
        case 6:
            return "VK_LAYER_SETTING_TYPE_FLOAT64_EXT";
        case 7:
            return "VK_LAYER_SETTING_TYPE_STRING_EXT";
        case VK_LAYER_SETTING_TYPE_MAX_ENUM_EXT: return "VK_LAYER_SETTING_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkLayerSettingTypeEXT value.";
        }
    }


    const char *
    vk_LayeredDriverUnderlyingApiMSFT_to_str(VkLayeredDriverUnderlyingApiMSFT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_LAYERED_DRIVER_UNDERLYING_API_NONE_MSFT";
        case 1:
            return "VK_LAYERED_DRIVER_UNDERLYING_API_D3D12_MSFT";
        case VK_LAYERED_DRIVER_UNDERLYING_API_MAX_ENUM_MSFT: return "VK_LAYERED_DRIVER_UNDERLYING_API_MAX_ENUM_MSFT";
        default:
            return "Unknown VkLayeredDriverUnderlyingApiMSFT value.";
        }
    }


    const char *
    vk_LineRasterizationModeEXT_to_str(VkLineRasterizationModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_LINE_RASTERIZATION_MODE_DEFAULT_EXT";
        case 1:
            return "VK_LINE_RASTERIZATION_MODE_RECTANGULAR_EXT";
        case 2:
            return "VK_LINE_RASTERIZATION_MODE_BRESENHAM_EXT";
        case 3:
            return "VK_LINE_RASTERIZATION_MODE_RECTANGULAR_SMOOTH_EXT";
        case VK_LINE_RASTERIZATION_MODE_MAX_ENUM_EXT: return "VK_LINE_RASTERIZATION_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkLineRasterizationModeEXT value.";
        }
    }


    const char *
    vk_LogicOp_to_str(VkLogicOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_LOGIC_OP_CLEAR";
        case 1:
            return "VK_LOGIC_OP_AND";
        case 2:
            return "VK_LOGIC_OP_AND_REVERSE";
        case 3:
            return "VK_LOGIC_OP_COPY";
        case 4:
            return "VK_LOGIC_OP_AND_INVERTED";
        case 5:
            return "VK_LOGIC_OP_NO_OP";
        case 6:
            return "VK_LOGIC_OP_XOR";
        case 7:
            return "VK_LOGIC_OP_OR";
        case 8:
            return "VK_LOGIC_OP_NOR";
        case 9:
            return "VK_LOGIC_OP_EQUIVALENT";
        case 10:
            return "VK_LOGIC_OP_INVERT";
        case 11:
            return "VK_LOGIC_OP_OR_REVERSE";
        case 12:
            return "VK_LOGIC_OP_COPY_INVERTED";
        case 13:
            return "VK_LOGIC_OP_OR_INVERTED";
        case 14:
            return "VK_LOGIC_OP_NAND";
        case 15:
            return "VK_LOGIC_OP_SET";
        case VK_LOGIC_OP_MAX_ENUM: return "VK_LOGIC_OP_MAX_ENUM";
        default:
            return "Unknown VkLogicOp value.";
        }
    }


    const char *
    vk_MemoryOverallocationBehaviorAMD_to_str(VkMemoryOverallocationBehaviorAMD input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_MEMORY_OVERALLOCATION_BEHAVIOR_DEFAULT_AMD";
        case 1:
            return "VK_MEMORY_OVERALLOCATION_BEHAVIOR_ALLOWED_AMD";
        case 2:
            return "VK_MEMORY_OVERALLOCATION_BEHAVIOR_DISALLOWED_AMD";
        case VK_MEMORY_OVERALLOCATION_BEHAVIOR_MAX_ENUM_AMD: return "VK_MEMORY_OVERALLOCATION_BEHAVIOR_MAX_ENUM_AMD";
        default:
            return "Unknown VkMemoryOverallocationBehaviorAMD value.";
        }
    }


    const char *
    vk_MicromapTypeEXT_to_str(VkMicromapTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_MICROMAP_TYPE_OPACITY_MICROMAP_EXT";
        case 1000397000:
            return "VK_MICROMAP_TYPE_DISPLACEMENT_MICROMAP_NV";
        case VK_MICROMAP_TYPE_MAX_ENUM_EXT: return "VK_MICROMAP_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkMicromapTypeEXT value.";
        }
    }


    const char *
    vk_ObjectType_to_str(VkObjectType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OBJECT_TYPE_UNKNOWN";
        case 1:
            return "VK_OBJECT_TYPE_INSTANCE";
        case 2:
            return "VK_OBJECT_TYPE_PHYSICAL_DEVICE";
        case 3:
            return "VK_OBJECT_TYPE_DEVICE";
        case 4:
            return "VK_OBJECT_TYPE_QUEUE";
        case 5:
            return "VK_OBJECT_TYPE_SEMAPHORE";
        case 6:
            return "VK_OBJECT_TYPE_COMMAND_BUFFER";
        case 7:
            return "VK_OBJECT_TYPE_FENCE";
        case 8:
            return "VK_OBJECT_TYPE_DEVICE_MEMORY";
        case 9:
            return "VK_OBJECT_TYPE_BUFFER";
        case 10:
            return "VK_OBJECT_TYPE_IMAGE";
        case 11:
            return "VK_OBJECT_TYPE_EVENT";
        case 12:
            return "VK_OBJECT_TYPE_QUERY_POOL";
        case 13:
            return "VK_OBJECT_TYPE_BUFFER_VIEW";
        case 14:
            return "VK_OBJECT_TYPE_IMAGE_VIEW";
        case 15:
            return "VK_OBJECT_TYPE_SHADER_MODULE";
        case 16:
            return "VK_OBJECT_TYPE_PIPELINE_CACHE";
        case 17:
            return "VK_OBJECT_TYPE_PIPELINE_LAYOUT";
        case 18:
            return "VK_OBJECT_TYPE_RENDER_PASS";
        case 19:
            return "VK_OBJECT_TYPE_PIPELINE";
        case 20:
            return "VK_OBJECT_TYPE_DESCRIPTOR_SET_LAYOUT";
        case 21:
            return "VK_OBJECT_TYPE_SAMPLER";
        case 22:
            return "VK_OBJECT_TYPE_DESCRIPTOR_POOL";
        case 23:
            return "VK_OBJECT_TYPE_DESCRIPTOR_SET";
        case 24:
            return "VK_OBJECT_TYPE_FRAMEBUFFER";
        case 25:
            return "VK_OBJECT_TYPE_COMMAND_POOL";
        case 1000000000:
            return "VK_OBJECT_TYPE_SURFACE_KHR";
        case 1000001000:
            return "VK_OBJECT_TYPE_SWAPCHAIN_KHR";
        case 1000002000:
            return "VK_OBJECT_TYPE_DISPLAY_KHR";
        case 1000002001:
            return "VK_OBJECT_TYPE_DISPLAY_MODE_KHR";
        case 1000011000:
            return "VK_OBJECT_TYPE_DEBUG_REPORT_CALLBACK_EXT";
        case 1000023000:
            return "VK_OBJECT_TYPE_VIDEO_SESSION_KHR";
        case 1000023001:
            return "VK_OBJECT_TYPE_VIDEO_SESSION_PARAMETERS_KHR";
        case 1000029000:
            return "VK_OBJECT_TYPE_CU_MODULE_NVX";
        case 1000029001:
            return "VK_OBJECT_TYPE_CU_FUNCTION_NVX";
        case 1000085000:
            return "VK_OBJECT_TYPE_DESCRIPTOR_UPDATE_TEMPLATE";
        case 1000128000:
            return "VK_OBJECT_TYPE_DEBUG_UTILS_MESSENGER_EXT";
        case 1000150000:
            return "VK_OBJECT_TYPE_ACCELERATION_STRUCTURE_KHR";
        case 1000156000:
            return "VK_OBJECT_TYPE_SAMPLER_YCBCR_CONVERSION";
        case 1000160000:
            return "VK_OBJECT_TYPE_VALIDATION_CACHE_EXT";
        case 1000165000:
            return "VK_OBJECT_TYPE_ACCELERATION_STRUCTURE_NV";
        case 1000210000:
            return "VK_OBJECT_TYPE_PERFORMANCE_CONFIGURATION_INTEL";
        case 1000268000:
            return "VK_OBJECT_TYPE_DEFERRED_OPERATION_KHR";
        case 1000277000:
            return "VK_OBJECT_TYPE_INDIRECT_COMMANDS_LAYOUT_NV";
        case 1000295000:
            return "VK_OBJECT_TYPE_PRIVATE_DATA_SLOT";
        case 1000307000:
            return "VK_OBJECT_TYPE_CUDA_MODULE_NV";
        case 1000307001:
            return "VK_OBJECT_TYPE_CUDA_FUNCTION_NV";
        case 1000366000:
            return "VK_OBJECT_TYPE_BUFFER_COLLECTION_FUCHSIA";
        case 1000396000:
            return "VK_OBJECT_TYPE_MICROMAP_EXT";
        case 1000464000:
            return "VK_OBJECT_TYPE_OPTICAL_FLOW_SESSION_NV";
        case 1000482000:
            return "VK_OBJECT_TYPE_SHADER_EXT";
        case VK_OBJECT_TYPE_MAX_ENUM: return "VK_OBJECT_TYPE_MAX_ENUM";
        default:
            return "Unknown VkObjectType value.";
        }
    }


    const char *
    vk_OpacityMicromapFormatEXT_to_str(VkOpacityMicromapFormatEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_OPACITY_MICROMAP_FORMAT_2_STATE_EXT";
        case 2:
            return "VK_OPACITY_MICROMAP_FORMAT_4_STATE_EXT";
        case VK_OPACITY_MICROMAP_FORMAT_MAX_ENUM_EXT: return "VK_OPACITY_MICROMAP_FORMAT_MAX_ENUM_EXT";
        default:
            return "Unknown VkOpacityMicromapFormatEXT value.";
        }
    }


    const char *
    vk_OpacityMicromapSpecialIndexEXT_to_str(VkOpacityMicromapSpecialIndexEXT input)
    {
        switch((int64_t)input) {
        case -4:
            return "VK_OPACITY_MICROMAP_SPECIAL_INDEX_FULLY_UNKNOWN_OPAQUE_EXT";
        case -3:
            return "VK_OPACITY_MICROMAP_SPECIAL_INDEX_FULLY_UNKNOWN_TRANSPARENT_EXT";
        case -2:
            return "VK_OPACITY_MICROMAP_SPECIAL_INDEX_FULLY_OPAQUE_EXT";
        case -1:
            return "VK_OPACITY_MICROMAP_SPECIAL_INDEX_FULLY_TRANSPARENT_EXT";
        case VK_OPACITY_MICROMAP_SPECIAL_INDEX_MAX_ENUM_EXT: return "VK_OPACITY_MICROMAP_SPECIAL_INDEX_MAX_ENUM_EXT";
        default:
            return "Unknown VkOpacityMicromapSpecialIndexEXT value.";
        }
    }


    const char *
    vk_OpticalFlowPerformanceLevelNV_to_str(VkOpticalFlowPerformanceLevelNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_UNKNOWN_NV";
        case 1:
            return "VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_SLOW_NV";
        case 2:
            return "VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_MEDIUM_NV";
        case 3:
            return "VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_FAST_NV";
        case VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_MAX_ENUM_NV: return "VK_OPTICAL_FLOW_PERFORMANCE_LEVEL_MAX_ENUM_NV";
        default:
            return "Unknown VkOpticalFlowPerformanceLevelNV value.";
        }
    }


    const char *
    vk_OpticalFlowSessionBindingPointNV_to_str(VkOpticalFlowSessionBindingPointNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_UNKNOWN_NV";
        case 1:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_INPUT_NV";
        case 2:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_REFERENCE_NV";
        case 3:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_HINT_NV";
        case 4:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_FLOW_VECTOR_NV";
        case 5:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_BACKWARD_FLOW_VECTOR_NV";
        case 6:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_COST_NV";
        case 7:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_BACKWARD_COST_NV";
        case 8:
            return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_GLOBAL_FLOW_NV";
        case VK_OPTICAL_FLOW_SESSION_BINDING_POINT_MAX_ENUM_NV: return "VK_OPTICAL_FLOW_SESSION_BINDING_POINT_MAX_ENUM_NV";
        default:
            return "Unknown VkOpticalFlowSessionBindingPointNV value.";
        }
    }


    const char *
    vk_OutOfBandQueueTypeNV_to_str(VkOutOfBandQueueTypeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OUT_OF_BAND_QUEUE_TYPE_RENDER_NV";
        case 1:
            return "VK_OUT_OF_BAND_QUEUE_TYPE_PRESENT_NV";
        case VK_OUT_OF_BAND_QUEUE_TYPE_MAX_ENUM_NV: return "VK_OUT_OF_BAND_QUEUE_TYPE_MAX_ENUM_NV";
        default:
            return "Unknown VkOutOfBandQueueTypeNV value.";
        }
    }


    const char *
    vk_PerformanceConfigurationTypeINTEL_to_str(VkPerformanceConfigurationTypeINTEL input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_CONFIGURATION_TYPE_COMMAND_QUEUE_METRICS_DISCOVERY_ACTIVATED_INTEL";
        case VK_PERFORMANCE_CONFIGURATION_TYPE_MAX_ENUM_INTEL: return "VK_PERFORMANCE_CONFIGURATION_TYPE_MAX_ENUM_INTEL";
        default:
            return "Unknown VkPerformanceConfigurationTypeINTEL value.";
        }
    }


    const char *
    vk_PerformanceCounterScopeKHR_to_str(VkPerformanceCounterScopeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_QUERY_SCOPE_COMMAND_BUFFER_KHR";
        case 1:
            return "VK_QUERY_SCOPE_RENDER_PASS_KHR";
        case 2:
            return "VK_QUERY_SCOPE_COMMAND_KHR";
        case VK_PERFORMANCE_COUNTER_SCOPE_MAX_ENUM_KHR: return "VK_PERFORMANCE_COUNTER_SCOPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkPerformanceCounterScopeKHR value.";
        }
    }


    const char *
    vk_PerformanceCounterStorageKHR_to_str(VkPerformanceCounterStorageKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_COUNTER_STORAGE_INT32_KHR";
        case 1:
            return "VK_PERFORMANCE_COUNTER_STORAGE_INT64_KHR";
        case 2:
            return "VK_PERFORMANCE_COUNTER_STORAGE_UINT32_KHR";
        case 3:
            return "VK_PERFORMANCE_COUNTER_STORAGE_UINT64_KHR";
        case 4:
            return "VK_PERFORMANCE_COUNTER_STORAGE_FLOAT32_KHR";
        case 5:
            return "VK_PERFORMANCE_COUNTER_STORAGE_FLOAT64_KHR";
        case VK_PERFORMANCE_COUNTER_STORAGE_MAX_ENUM_KHR: return "VK_PERFORMANCE_COUNTER_STORAGE_MAX_ENUM_KHR";
        default:
            return "Unknown VkPerformanceCounterStorageKHR value.";
        }
    }


    const char *
    vk_PerformanceCounterUnitKHR_to_str(VkPerformanceCounterUnitKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_COUNTER_UNIT_GENERIC_KHR";
        case 1:
            return "VK_PERFORMANCE_COUNTER_UNIT_PERCENTAGE_KHR";
        case 2:
            return "VK_PERFORMANCE_COUNTER_UNIT_NANOSECONDS_KHR";
        case 3:
            return "VK_PERFORMANCE_COUNTER_UNIT_BYTES_KHR";
        case 4:
            return "VK_PERFORMANCE_COUNTER_UNIT_BYTES_PER_SECOND_KHR";
        case 5:
            return "VK_PERFORMANCE_COUNTER_UNIT_KELVIN_KHR";
        case 6:
            return "VK_PERFORMANCE_COUNTER_UNIT_WATTS_KHR";
        case 7:
            return "VK_PERFORMANCE_COUNTER_UNIT_VOLTS_KHR";
        case 8:
            return "VK_PERFORMANCE_COUNTER_UNIT_AMPS_KHR";
        case 9:
            return "VK_PERFORMANCE_COUNTER_UNIT_HERTZ_KHR";
        case 10:
            return "VK_PERFORMANCE_COUNTER_UNIT_CYCLES_KHR";
        case VK_PERFORMANCE_COUNTER_UNIT_MAX_ENUM_KHR: return "VK_PERFORMANCE_COUNTER_UNIT_MAX_ENUM_KHR";
        default:
            return "Unknown VkPerformanceCounterUnitKHR value.";
        }
    }


    const char *
    vk_PerformanceOverrideTypeINTEL_to_str(VkPerformanceOverrideTypeINTEL input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_OVERRIDE_TYPE_NULL_HARDWARE_INTEL";
        case 1:
            return "VK_PERFORMANCE_OVERRIDE_TYPE_FLUSH_GPU_CACHES_INTEL";
        case VK_PERFORMANCE_OVERRIDE_TYPE_MAX_ENUM_INTEL: return "VK_PERFORMANCE_OVERRIDE_TYPE_MAX_ENUM_INTEL";
        default:
            return "Unknown VkPerformanceOverrideTypeINTEL value.";
        }
    }


    const char *
    vk_PerformanceParameterTypeINTEL_to_str(VkPerformanceParameterTypeINTEL input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_PARAMETER_TYPE_HW_COUNTERS_SUPPORTED_INTEL";
        case 1:
            return "VK_PERFORMANCE_PARAMETER_TYPE_STREAM_MARKER_VALID_BITS_INTEL";
        case VK_PERFORMANCE_PARAMETER_TYPE_MAX_ENUM_INTEL: return "VK_PERFORMANCE_PARAMETER_TYPE_MAX_ENUM_INTEL";
        default:
            return "Unknown VkPerformanceParameterTypeINTEL value.";
        }
    }


    const char *
    vk_PerformanceValueTypeINTEL_to_str(VkPerformanceValueTypeINTEL input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PERFORMANCE_VALUE_TYPE_UINT32_INTEL";
        case 1:
            return "VK_PERFORMANCE_VALUE_TYPE_UINT64_INTEL";
        case 2:
            return "VK_PERFORMANCE_VALUE_TYPE_FLOAT_INTEL";
        case 3:
            return "VK_PERFORMANCE_VALUE_TYPE_BOOL_INTEL";
        case 4:
            return "VK_PERFORMANCE_VALUE_TYPE_STRING_INTEL";
        case VK_PERFORMANCE_VALUE_TYPE_MAX_ENUM_INTEL: return "VK_PERFORMANCE_VALUE_TYPE_MAX_ENUM_INTEL";
        default:
            return "Unknown VkPerformanceValueTypeINTEL value.";
        }
    }


    const char *
    vk_PhysicalDeviceType_to_str(VkPhysicalDeviceType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PHYSICAL_DEVICE_TYPE_OTHER";
        case 1:
            return "VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU";
        case 2:
            return "VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU";
        case 3:
            return "VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU";
        case 4:
            return "VK_PHYSICAL_DEVICE_TYPE_CPU";
        case VK_PHYSICAL_DEVICE_TYPE_MAX_ENUM: return "VK_PHYSICAL_DEVICE_TYPE_MAX_ENUM";
        default:
            return "Unknown VkPhysicalDeviceType value.";
        }
    }


    const char *
    vk_PipelineBindPoint_to_str(VkPipelineBindPoint input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_BIND_POINT_GRAPHICS";
        case 1:
            return "VK_PIPELINE_BIND_POINT_COMPUTE";
        case 1000134000:
            return "VK_PIPELINE_BIND_POINT_EXECUTION_GRAPH_AMDX";
        case 1000165000:
            return "VK_PIPELINE_BIND_POINT_RAY_TRACING_NV";
        case 1000369003:
            return "VK_PIPELINE_BIND_POINT_SUBPASS_SHADING_HUAWEI";
        case VK_PIPELINE_BIND_POINT_MAX_ENUM: return "VK_PIPELINE_BIND_POINT_MAX_ENUM";
        default:
            return "Unknown VkPipelineBindPoint value.";
        }
    }


    const char *
    vk_PipelineCacheHeaderVersion_to_str(VkPipelineCacheHeaderVersion input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_CACHE_HEADER_VERSION_ONE";
        case VK_PIPELINE_CACHE_HEADER_VERSION_MAX_ENUM: return "VK_PIPELINE_CACHE_HEADER_VERSION_MAX_ENUM";
        default:
            return "Unknown VkPipelineCacheHeaderVersion value.";
        }
    }


    const char *
    vk_PipelineExecutableStatisticFormatKHR_to_str(VkPipelineExecutableStatisticFormatKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_BOOL32_KHR";
        case 1:
            return "VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_INT64_KHR";
        case 2:
            return "VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_UINT64_KHR";
        case 3:
            return "VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_FLOAT64_KHR";
        case VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_MAX_ENUM_KHR: return "VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_MAX_ENUM_KHR";
        default:
            return "Unknown VkPipelineExecutableStatisticFormatKHR value.";
        }
    }


    const char *
    vk_PipelineRobustnessBufferBehaviorEXT_to_str(VkPipelineRobustnessBufferBehaviorEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DEVICE_DEFAULT_EXT";
        case 1:
            return "VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DISABLED_EXT";
        case 2:
            return "VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_EXT";
        case 3:
            return "VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2_EXT";
        case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_MAX_ENUM_EXT: return "VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_MAX_ENUM_EXT";
        default:
            return "Unknown VkPipelineRobustnessBufferBehaviorEXT value.";
        }
    }


    const char *
    vk_PipelineRobustnessImageBehaviorEXT_to_str(VkPipelineRobustnessImageBehaviorEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_DEVICE_DEFAULT_EXT";
        case 1:
            return "VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_DISABLED_EXT";
        case 2:
            return "VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS_EXT";
        case 3:
            return "VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_ROBUST_IMAGE_ACCESS_2_EXT";
        case VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_MAX_ENUM_EXT: return "VK_PIPELINE_ROBUSTNESS_IMAGE_BEHAVIOR_MAX_ENUM_EXT";
        default:
            return "Unknown VkPipelineRobustnessImageBehaviorEXT value.";
        }
    }


    const char *
    vk_PointClippingBehavior_to_str(VkPointClippingBehavior input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_POINT_CLIPPING_BEHAVIOR_ALL_CLIP_PLANES";
        case 1:
            return "VK_POINT_CLIPPING_BEHAVIOR_USER_CLIP_PLANES_ONLY";
        case VK_POINT_CLIPPING_BEHAVIOR_MAX_ENUM: return "VK_POINT_CLIPPING_BEHAVIOR_MAX_ENUM";
        default:
            return "Unknown VkPointClippingBehavior value.";
        }
    }


    const char *
    vk_PolygonMode_to_str(VkPolygonMode input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_POLYGON_MODE_FILL";
        case 1:
            return "VK_POLYGON_MODE_LINE";
        case 2:
            return "VK_POLYGON_MODE_POINT";
        case 1000153000:
            return "VK_POLYGON_MODE_FILL_RECTANGLE_NV";
        case VK_POLYGON_MODE_MAX_ENUM: return "VK_POLYGON_MODE_MAX_ENUM";
        default:
            return "Unknown VkPolygonMode value.";
        }
    }


    const char *
    vk_PresentModeKHR_to_str(VkPresentModeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PRESENT_MODE_IMMEDIATE_KHR";
        case 1:
            return "VK_PRESENT_MODE_MAILBOX_KHR";
        case 2:
            return "VK_PRESENT_MODE_FIFO_KHR";
        case 3:
            return "VK_PRESENT_MODE_FIFO_RELAXED_KHR";
        case 1000111000:
            return "VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR";
        case 1000111001:
            return "VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR";
        case VK_PRESENT_MODE_MAX_ENUM_KHR: return "VK_PRESENT_MODE_MAX_ENUM_KHR";
        default:
            return "Unknown VkPresentModeKHR value.";
        }
    }


    const char *
    vk_PrimitiveTopology_to_str(VkPrimitiveTopology input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PRIMITIVE_TOPOLOGY_POINT_LIST";
        case 1:
            return "VK_PRIMITIVE_TOPOLOGY_LINE_LIST";
        case 2:
            return "VK_PRIMITIVE_TOPOLOGY_LINE_STRIP";
        case 3:
            return "VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST";
        case 4:
            return "VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP";
        case 5:
            return "VK_PRIMITIVE_TOPOLOGY_TRIANGLE_FAN";
        case 6:
            return "VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY";
        case 7:
            return "VK_PRIMITIVE_TOPOLOGY_LINE_STRIP_WITH_ADJACENCY";
        case 8:
            return "VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY";
        case 9:
            return "VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP_WITH_ADJACENCY";
        case 10:
            return "VK_PRIMITIVE_TOPOLOGY_PATCH_LIST";
        case VK_PRIMITIVE_TOPOLOGY_MAX_ENUM: return "VK_PRIMITIVE_TOPOLOGY_MAX_ENUM";
        default:
            return "Unknown VkPrimitiveTopology value.";
        }
    }


    const char *
    vk_ProvokingVertexModeEXT_to_str(VkProvokingVertexModeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PROVOKING_VERTEX_MODE_FIRST_VERTEX_EXT";
        case 1:
            return "VK_PROVOKING_VERTEX_MODE_LAST_VERTEX_EXT";
        case VK_PROVOKING_VERTEX_MODE_MAX_ENUM_EXT: return "VK_PROVOKING_VERTEX_MODE_MAX_ENUM_EXT";
        default:
            return "Unknown VkProvokingVertexModeEXT value.";
        }
    }


    const char *
    vk_QueryPoolSamplingModeINTEL_to_str(VkQueryPoolSamplingModeINTEL input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_QUERY_POOL_SAMPLING_MODE_MANUAL_INTEL";
        case VK_QUERY_POOL_SAMPLING_MODE_MAX_ENUM_INTEL: return "VK_QUERY_POOL_SAMPLING_MODE_MAX_ENUM_INTEL";
        default:
            return "Unknown VkQueryPoolSamplingModeINTEL value.";
        }
    }


    const char *
    vk_QueryResultStatusKHR_to_str(VkQueryResultStatusKHR input)
    {
        switch((int64_t)input) {
        case -1000299000:
            return "VK_QUERY_RESULT_STATUS_INSUFFICIENT_BITSTREAM_BUFFER_RANGE_KHR";
        case -1:
            return "VK_QUERY_RESULT_STATUS_ERROR_KHR";
        case 0:
            return "VK_QUERY_RESULT_STATUS_NOT_READY_KHR";
        case 1:
            return "VK_QUERY_RESULT_STATUS_COMPLETE_KHR";
        case VK_QUERY_RESULT_STATUS_MAX_ENUM_KHR: return "VK_QUERY_RESULT_STATUS_MAX_ENUM_KHR";
        default:
            return "Unknown VkQueryResultStatusKHR value.";
        }
    }


    const char *
    vk_QueryType_to_str(VkQueryType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_QUERY_TYPE_OCCLUSION";
        case 1:
            return "VK_QUERY_TYPE_PIPELINE_STATISTICS";
        case 2:
            return "VK_QUERY_TYPE_TIMESTAMP";
        case 1000023000:
            return "VK_QUERY_TYPE_RESULT_STATUS_ONLY_KHR";
        case 1000028004:
            return "VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT";
        case 1000116000:
            return "VK_QUERY_TYPE_PERFORMANCE_QUERY_KHR";
        case 1000150000:
            return "VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_KHR";
        case 1000150001:
            return "VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_SIZE_KHR";
        case 1000165000:
            return "VK_QUERY_TYPE_ACCELERATION_STRUCTURE_COMPACTED_SIZE_NV";
        case 1000210000:
            return "VK_QUERY_TYPE_PERFORMANCE_QUERY_INTEL";
        case 1000299000:
            return "VK_QUERY_TYPE_VIDEO_ENCODE_FEEDBACK_KHR";
        case 1000328000:
            return "VK_QUERY_TYPE_MESH_PRIMITIVES_GENERATED_EXT";
        case 1000382000:
            return "VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT";
        case 1000386000:
            return "VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SERIALIZATION_BOTTOM_LEVEL_POINTERS_KHR";
        case 1000386001:
            return "VK_QUERY_TYPE_ACCELERATION_STRUCTURE_SIZE_KHR";
        case 1000396000:
            return "VK_QUERY_TYPE_MICROMAP_SERIALIZATION_SIZE_EXT";
        case 1000396001:
            return "VK_QUERY_TYPE_MICROMAP_COMPACTED_SIZE_EXT";
        case VK_QUERY_TYPE_MAX_ENUM: return "VK_QUERY_TYPE_MAX_ENUM";
        default:
            return "Unknown VkQueryType value.";
        }
    }


    const char *
    vk_QueueGlobalPriorityKHR_to_str(VkQueueGlobalPriorityKHR input)
    {
        switch((int64_t)input) {
        case 128:
            return "VK_QUEUE_GLOBAL_PRIORITY_LOW_KHR";
        case 256:
            return "VK_QUEUE_GLOBAL_PRIORITY_MEDIUM_KHR";
        case 512:
            return "VK_QUEUE_GLOBAL_PRIORITY_HIGH_KHR";
        case 1024:
            return "VK_QUEUE_GLOBAL_PRIORITY_REALTIME_KHR";
        case VK_QUEUE_GLOBAL_PRIORITY_MAX_ENUM_KHR: return "VK_QUEUE_GLOBAL_PRIORITY_MAX_ENUM_KHR";
        default:
            return "Unknown VkQueueGlobalPriorityKHR value.";
        }
    }


    const char *
    vk_RasterizationOrderAMD_to_str(VkRasterizationOrderAMD input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_RASTERIZATION_ORDER_STRICT_AMD";
        case 1:
            return "VK_RASTERIZATION_ORDER_RELAXED_AMD";
        case VK_RASTERIZATION_ORDER_MAX_ENUM_AMD: return "VK_RASTERIZATION_ORDER_MAX_ENUM_AMD";
        default:
            return "Unknown VkRasterizationOrderAMD value.";
        }
    }


    const char *
    vk_RayTracingInvocationReorderModeNV_to_str(VkRayTracingInvocationReorderModeNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_RAY_TRACING_INVOCATION_REORDER_MODE_NONE_NV";
        case 1:
            return "VK_RAY_TRACING_INVOCATION_REORDER_MODE_REORDER_NV";
        case VK_RAY_TRACING_INVOCATION_REORDER_MODE_MAX_ENUM_NV: return "VK_RAY_TRACING_INVOCATION_REORDER_MODE_MAX_ENUM_NV";
        default:
            return "Unknown VkRayTracingInvocationReorderModeNV value.";
        }
    }


    const char *
    vk_RayTracingShaderGroupTypeKHR_to_str(VkRayTracingShaderGroupTypeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_RAY_TRACING_SHADER_GROUP_TYPE_GENERAL_NV";
        case 1:
            return "VK_RAY_TRACING_SHADER_GROUP_TYPE_TRIANGLES_HIT_GROUP_NV";
        case 2:
            return "VK_RAY_TRACING_SHADER_GROUP_TYPE_PROCEDURAL_HIT_GROUP_NV";
        case VK_RAY_TRACING_SHADER_GROUP_TYPE_MAX_ENUM_KHR: return "VK_RAY_TRACING_SHADER_GROUP_TYPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkRayTracingShaderGroupTypeKHR value.";
        }
    }


    const char *
    vk_Result_to_str(VkResult input)
    {
        switch((int64_t)input) {
        case -1000338000:
            return "VK_ERROR_COMPRESSION_EXHAUSTED_EXT";
        case -1000299000:
            return "VK_ERROR_INVALID_VIDEO_STD_PARAMETERS_KHR";
        case -1000257000:
            return "VK_ERROR_INVALID_DEVICE_ADDRESS_EXT";
        case -1000255000:
            return "VK_ERROR_FULL_SCREEN_EXCLUSIVE_MODE_LOST_EXT";
        case -1000174001:
            return "VK_ERROR_NOT_PERMITTED_KHR";
        case -1000161000:
            return "VK_ERROR_FRAGMENTATION";
        case -1000158000:
            return "VK_ERROR_INVALID_DRM_FORMAT_MODIFIER_PLANE_LAYOUT_EXT";
        case -1000072003:
            return "VK_ERROR_INVALID_EXTERNAL_HANDLE";
        case -1000069000:
            return "VK_ERROR_OUT_OF_POOL_MEMORY";
        case -1000023005:
            return "VK_ERROR_VIDEO_STD_VERSION_NOT_SUPPORTED_KHR";
        case -1000023004:
            return "VK_ERROR_VIDEO_PROFILE_CODEC_NOT_SUPPORTED_KHR";
        case -1000023003:
            return "VK_ERROR_VIDEO_PROFILE_FORMAT_NOT_SUPPORTED_KHR";
        case -1000023002:
            return "VK_ERROR_VIDEO_PROFILE_OPERATION_NOT_SUPPORTED_KHR";
        case -1000023001:
            return "VK_ERROR_VIDEO_PICTURE_LAYOUT_NOT_SUPPORTED_KHR";
        case -1000023000:
            return "VK_ERROR_IMAGE_USAGE_NOT_SUPPORTED_KHR";
        case -1000012000:
            return "VK_ERROR_INVALID_SHADER_NV";
        case -1000011001:
            return "VK_ERROR_VALIDATION_FAILED_EXT";
        case -1000003001:
            return "VK_ERROR_INCOMPATIBLE_DISPLAY_KHR";
        case -1000001004:
            return "VK_ERROR_OUT_OF_DATE_KHR";
        case -1000000001:
            return "VK_ERROR_NATIVE_WINDOW_IN_USE_KHR";
        case -1000000000:
            return "VK_ERROR_SURFACE_LOST_KHR";
        case -13:
            return "VK_ERROR_UNKNOWN";
        case -12:
            return "VK_ERROR_FRAGMENTED_POOL";
        case -11:
            return "VK_ERROR_FORMAT_NOT_SUPPORTED";
        case -10:
            return "VK_ERROR_TOO_MANY_OBJECTS";
        case -9:
            return "VK_ERROR_INCOMPATIBLE_DRIVER";
        case -8:
            return "VK_ERROR_FEATURE_NOT_PRESENT";
        case -7:
            return "VK_ERROR_EXTENSION_NOT_PRESENT";
        case -6:
            return "VK_ERROR_LAYER_NOT_PRESENT";
        case -5:
            return "VK_ERROR_MEMORY_MAP_FAILED";
        case -4:
            return "VK_ERROR_DEVICE_LOST";
        case -3:
            return "VK_ERROR_INITIALIZATION_FAILED";
        case -2:
            return "VK_ERROR_OUT_OF_DEVICE_MEMORY";
        case -1:
            return "VK_ERROR_OUT_OF_HOST_MEMORY";
        case 0:
            return "VK_SUCCESS";
        case 1:
            return "VK_NOT_READY";
        case 2:
            return "VK_TIMEOUT";
        case 3:
            return "VK_EVENT_SET";
        case 4:
            return "VK_EVENT_RESET";
        case 5:
            return "VK_INCOMPLETE";
        case 1000001003:
            return "VK_SUBOPTIMAL_KHR";
        case 1000268000:
            return "VK_THREAD_IDLE_KHR";
        case 1000268001:
            return "VK_THREAD_DONE_KHR";
        case 1000268002:
            return "VK_OPERATION_DEFERRED_KHR";
        case 1000268003:
            return "VK_OPERATION_NOT_DEFERRED_KHR";
        case 1000297000:
            return "VK_PIPELINE_COMPILE_REQUIRED";
        case 1000482000:
            return "VK_ERROR_INCOMPATIBLE_SHADER_BINARY_EXT";
        case VK_RESULT_MAX_ENUM: return "VK_RESULT_MAX_ENUM";
        default:
            return "Unknown VkResult value.";
        }
    }


    const char *
    vk_SamplerAddressMode_to_str(VkSamplerAddressMode input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SAMPLER_ADDRESS_MODE_REPEAT";
        case 1:
            return "VK_SAMPLER_ADDRESS_MODE_MIRRORED_REPEAT";
        case 2:
            return "VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE";
        case 3:
            return "VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER";
        case 4:
            return "VK_SAMPLER_ADDRESS_MODE_MIRROR_CLAMP_TO_EDGE";
        case VK_SAMPLER_ADDRESS_MODE_MAX_ENUM: return "VK_SAMPLER_ADDRESS_MODE_MAX_ENUM";
        default:
            return "Unknown VkSamplerAddressMode value.";
        }
    }


    const char *
    vk_SamplerMipmapMode_to_str(VkSamplerMipmapMode input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SAMPLER_MIPMAP_MODE_NEAREST";
        case 1:
            return "VK_SAMPLER_MIPMAP_MODE_LINEAR";
        case VK_SAMPLER_MIPMAP_MODE_MAX_ENUM: return "VK_SAMPLER_MIPMAP_MODE_MAX_ENUM";
        default:
            return "Unknown VkSamplerMipmapMode value.";
        }
    }


    const char *
    vk_SamplerReductionMode_to_str(VkSamplerReductionMode input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE";
        case 1:
            return "VK_SAMPLER_REDUCTION_MODE_MIN";
        case 2:
            return "VK_SAMPLER_REDUCTION_MODE_MAX";
        case 1000521000:
            return "VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE_RANGECLAMP_QCOM";
        case VK_SAMPLER_REDUCTION_MODE_MAX_ENUM: return "VK_SAMPLER_REDUCTION_MODE_MAX_ENUM";
        default:
            return "Unknown VkSamplerReductionMode value.";
        }
    }


    const char *
    vk_SamplerYcbcrModelConversion_to_str(VkSamplerYcbcrModelConversion input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_RGB_IDENTITY";
        case 1:
            return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_YCBCR_IDENTITY";
        case 2:
            return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_YCBCR_709";
        case 3:
            return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_YCBCR_601";
        case 4:
            return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_YCBCR_2020";
        case VK_SAMPLER_YCBCR_MODEL_CONVERSION_MAX_ENUM: return "VK_SAMPLER_YCBCR_MODEL_CONVERSION_MAX_ENUM";
        default:
            return "Unknown VkSamplerYcbcrModelConversion value.";
        }
    }


    const char *
    vk_SamplerYcbcrRange_to_str(VkSamplerYcbcrRange input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SAMPLER_YCBCR_RANGE_ITU_FULL";
        case 1:
            return "VK_SAMPLER_YCBCR_RANGE_ITU_NARROW";
        case VK_SAMPLER_YCBCR_RANGE_MAX_ENUM: return "VK_SAMPLER_YCBCR_RANGE_MAX_ENUM";
        default:
            return "Unknown VkSamplerYcbcrRange value.";
        }
    }


    const char *
    vk_ScopeKHR_to_str(VkScopeKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SCOPE_DEVICE_NV";
        case 2:
            return "VK_SCOPE_WORKGROUP_NV";
        case 3:
            return "VK_SCOPE_SUBGROUP_NV";
        case 5:
            return "VK_SCOPE_QUEUE_FAMILY_NV";
        case VK_SCOPE_MAX_ENUM_KHR: return "VK_SCOPE_MAX_ENUM_KHR";
        default:
            return "Unknown VkScopeKHR value.";
        }
    }


    const char *
    vk_SemaphoreType_to_str(VkSemaphoreType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SEMAPHORE_TYPE_BINARY";
        case 1:
            return "VK_SEMAPHORE_TYPE_TIMELINE";
        case VK_SEMAPHORE_TYPE_MAX_ENUM: return "VK_SEMAPHORE_TYPE_MAX_ENUM";
        default:
            return "Unknown VkSemaphoreType value.";
        }
    }


    const char *
    vk_ShaderCodeTypeEXT_to_str(VkShaderCodeTypeEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHADER_CODE_TYPE_BINARY_EXT";
        case 1:
            return "VK_SHADER_CODE_TYPE_SPIRV_EXT";
        case VK_SHADER_CODE_TYPE_MAX_ENUM_EXT: return "VK_SHADER_CODE_TYPE_MAX_ENUM_EXT";
        default:
            return "Unknown VkShaderCodeTypeEXT value.";
        }
    }


    const char *
    vk_ShaderFloatControlsIndependence_to_str(VkShaderFloatControlsIndependence input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_32_BIT_ONLY";
        case 1:
            return "VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_ALL";
        case 2:
            return "VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_NONE";
        case VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_MAX_ENUM: return "VK_SHADER_FLOAT_CONTROLS_INDEPENDENCE_MAX_ENUM";
        default:
            return "Unknown VkShaderFloatControlsIndependence value.";
        }
    }


    const char *
    vk_ShaderGroupShaderKHR_to_str(VkShaderGroupShaderKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHADER_GROUP_SHADER_GENERAL_KHR";
        case 1:
            return "VK_SHADER_GROUP_SHADER_CLOSEST_HIT_KHR";
        case 2:
            return "VK_SHADER_GROUP_SHADER_ANY_HIT_KHR";
        case 3:
            return "VK_SHADER_GROUP_SHADER_INTERSECTION_KHR";
        case VK_SHADER_GROUP_SHADER_MAX_ENUM_KHR: return "VK_SHADER_GROUP_SHADER_MAX_ENUM_KHR";
        default:
            return "Unknown VkShaderGroupShaderKHR value.";
        }
    }


    const char *
    vk_ShaderInfoTypeAMD_to_str(VkShaderInfoTypeAMD input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHADER_INFO_TYPE_STATISTICS_AMD";
        case 1:
            return "VK_SHADER_INFO_TYPE_BINARY_AMD";
        case 2:
            return "VK_SHADER_INFO_TYPE_DISASSEMBLY_AMD";
        case VK_SHADER_INFO_TYPE_MAX_ENUM_AMD: return "VK_SHADER_INFO_TYPE_MAX_ENUM_AMD";
        default:
            return "Unknown VkShaderInfoTypeAMD value.";
        }
    }


    const char *
    vk_ShadingRatePaletteEntryNV_to_str(VkShadingRatePaletteEntryNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHADING_RATE_PALETTE_ENTRY_NO_INVOCATIONS_NV";
        case 1:
            return "VK_SHADING_RATE_PALETTE_ENTRY_16_INVOCATIONS_PER_PIXEL_NV";
        case 2:
            return "VK_SHADING_RATE_PALETTE_ENTRY_8_INVOCATIONS_PER_PIXEL_NV";
        case 3:
            return "VK_SHADING_RATE_PALETTE_ENTRY_4_INVOCATIONS_PER_PIXEL_NV";
        case 4:
            return "VK_SHADING_RATE_PALETTE_ENTRY_2_INVOCATIONS_PER_PIXEL_NV";
        case 5:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_PIXEL_NV";
        case 6:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_2X1_PIXELS_NV";
        case 7:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_1X2_PIXELS_NV";
        case 8:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_2X2_PIXELS_NV";
        case 9:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_4X2_PIXELS_NV";
        case 10:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_2X4_PIXELS_NV";
        case 11:
            return "VK_SHADING_RATE_PALETTE_ENTRY_1_INVOCATION_PER_4X4_PIXELS_NV";
        case VK_SHADING_RATE_PALETTE_ENTRY_MAX_ENUM_NV: return "VK_SHADING_RATE_PALETTE_ENTRY_MAX_ENUM_NV";
        default:
            return "Unknown VkShadingRatePaletteEntryNV value.";
        }
    }


    const char *
    vk_SharingMode_to_str(VkSharingMode input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SHARING_MODE_EXCLUSIVE";
        case 1:
            return "VK_SHARING_MODE_CONCURRENT";
        case VK_SHARING_MODE_MAX_ENUM: return "VK_SHARING_MODE_MAX_ENUM";
        default:
            return "Unknown VkSharingMode value.";
        }
    }


    const char *
    vk_StencilOp_to_str(VkStencilOp input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_STENCIL_OP_KEEP";
        case 1:
            return "VK_STENCIL_OP_ZERO";
        case 2:
            return "VK_STENCIL_OP_REPLACE";
        case 3:
            return "VK_STENCIL_OP_INCREMENT_AND_CLAMP";
        case 4:
            return "VK_STENCIL_OP_DECREMENT_AND_CLAMP";
        case 5:
            return "VK_STENCIL_OP_INVERT";
        case 6:
            return "VK_STENCIL_OP_INCREMENT_AND_WRAP";
        case 7:
            return "VK_STENCIL_OP_DECREMENT_AND_WRAP";
        case VK_STENCIL_OP_MAX_ENUM: return "VK_STENCIL_OP_MAX_ENUM";
        default:
            return "Unknown VkStencilOp value.";
        }
    }


    const char *
    vk_StructureType_to_str(VkStructureType input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_STRUCTURE_TYPE_APPLICATION_INFO";
        case 1:
            return "VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO";
        case 2:
            return "VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO";
        case 3:
            return "VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO";
        case 4:
            return "VK_STRUCTURE_TYPE_SUBMIT_INFO";
        case 5:
            return "VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO";
        case 6:
            return "VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE";
        case 7:
            return "VK_STRUCTURE_TYPE_BIND_SPARSE_INFO";
        case 8:
            return "VK_STRUCTURE_TYPE_FENCE_CREATE_INFO";
        case 9:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO";
        case 10:
            return "VK_STRUCTURE_TYPE_EVENT_CREATE_INFO";
        case 11:
            return "VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO";
        case 12:
            return "VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO";
        case 13:
            return "VK_STRUCTURE_TYPE_BUFFER_VIEW_CREATE_INFO";
        case 14:
            return "VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO";
        case 15:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO";
        case 16:
            return "VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO";
        case 17:
            return "VK_STRUCTURE_TYPE_PIPELINE_CACHE_CREATE_INFO";
        case 18:
            return "VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO";
        case 19:
            return "VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO";
        case 20:
            return "VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO";
        case 21:
            return "VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO";
        case 22:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO";
        case 23:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO";
        case 24:
            return "VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO";
        case 25:
            return "VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO";
        case 26:
            return "VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO";
        case 27:
            return "VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO";
        case 28:
            return "VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO";
        case 29:
            return "VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO";
        case 30:
            return "VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO";
        case 31:
            return "VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO";
        case 32:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO";
        case 33:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO";
        case 34:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO";
        case 35:
            return "VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET";
        case 36:
            return "VK_STRUCTURE_TYPE_COPY_DESCRIPTOR_SET";
        case 37:
            return "VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO";
        case 38:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO";
        case 39:
            return "VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO";
        case 40:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO";
        case 41:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_INFO";
        case 42:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO";
        case 43:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO";
        case 44:
            return "VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER";
        case 45:
            return "VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER";
        case 46:
            return "VK_STRUCTURE_TYPE_MEMORY_BARRIER";
        case 47:
            return "VK_STRUCTURE_TYPE_LOADER_INSTANCE_CREATE_INFO";
        case 48:
            return "VK_STRUCTURE_TYPE_LOADER_DEVICE_CREATE_INFO";
        case 49:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES";
        case 50:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_PROPERTIES";
        case 51:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES";
        case 52:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_PROPERTIES";
        case 53:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES";
        case 54:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_PROPERTIES";
        case 1000001000:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR";
        case 1000001001:
            return "VK_STRUCTURE_TYPE_PRESENT_INFO_KHR";
        case 1000002000:
            return "VK_STRUCTURE_TYPE_DISPLAY_MODE_CREATE_INFO_KHR";
        case 1000002001:
            return "VK_STRUCTURE_TYPE_DISPLAY_SURFACE_CREATE_INFO_KHR";
        case 1000003000:
            return "VK_STRUCTURE_TYPE_DISPLAY_PRESENT_INFO_KHR";
        case 1000004000:
            return "VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR";
        case 1000005000:
            return "VK_STRUCTURE_TYPE_XCB_SURFACE_CREATE_INFO_KHR";
        case 1000006000:
            return "VK_STRUCTURE_TYPE_WAYLAND_SURFACE_CREATE_INFO_KHR";
        case 1000008000:
            return "VK_STRUCTURE_TYPE_ANDROID_SURFACE_CREATE_INFO_KHR";
        case 1000009000:
            return "VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR";
        case 1000010000:
            return "VK_STRUCTURE_TYPE_NATIVE_BUFFER_ANDROID";
        case 1000010001:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_IMAGE_CREATE_INFO_ANDROID";
        case 1000010002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENTATION_PROPERTIES_ANDROID";
        case 1000011000:
            return "VK_STRUCTURE_TYPE_DEBUG_REPORT_CREATE_INFO_EXT";
        case 1000018000:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_RASTERIZATION_ORDER_AMD";
        case 1000022000:
            return "VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_NAME_INFO_EXT";
        case 1000022001:
            return "VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_TAG_INFO_EXT";
        case 1000022002:
            return "VK_STRUCTURE_TYPE_DEBUG_MARKER_MARKER_INFO_EXT";
        case 1000023000:
            return "VK_STRUCTURE_TYPE_VIDEO_PROFILE_INFO_KHR";
        case 1000023001:
            return "VK_STRUCTURE_TYPE_VIDEO_CAPABILITIES_KHR";
        case 1000023002:
            return "VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_INFO_KHR";
        case 1000023003:
            return "VK_STRUCTURE_TYPE_VIDEO_SESSION_MEMORY_REQUIREMENTS_KHR";
        case 1000023004:
            return "VK_STRUCTURE_TYPE_BIND_VIDEO_SESSION_MEMORY_INFO_KHR";
        case 1000023005:
            return "VK_STRUCTURE_TYPE_VIDEO_SESSION_CREATE_INFO_KHR";
        case 1000023006:
            return "VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_CREATE_INFO_KHR";
        case 1000023007:
            return "VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_UPDATE_INFO_KHR";
        case 1000023008:
            return "VK_STRUCTURE_TYPE_VIDEO_BEGIN_CODING_INFO_KHR";
        case 1000023009:
            return "VK_STRUCTURE_TYPE_VIDEO_END_CODING_INFO_KHR";
        case 1000023010:
            return "VK_STRUCTURE_TYPE_VIDEO_CODING_CONTROL_INFO_KHR";
        case 1000023011:
            return "VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_INFO_KHR";
        case 1000023012:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_VIDEO_PROPERTIES_KHR";
        case 1000023013:
            return "VK_STRUCTURE_TYPE_VIDEO_PROFILE_LIST_INFO_KHR";
        case 1000023014:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_FORMAT_INFO_KHR";
        case 1000023015:
            return "VK_STRUCTURE_TYPE_VIDEO_FORMAT_PROPERTIES_KHR";
        case 1000023016:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_QUERY_RESULT_STATUS_PROPERTIES_KHR";
        case 1000024000:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_INFO_KHR";
        case 1000024001:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_CAPABILITIES_KHR";
        case 1000024002:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_USAGE_INFO_KHR";
        case 1000026000:
            return "VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_IMAGE_CREATE_INFO_NV";
        case 1000026001:
            return "VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_BUFFER_CREATE_INFO_NV";
        case 1000026002:
            return "VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_MEMORY_ALLOCATE_INFO_NV";
        case 1000028000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_FEATURES_EXT";
        case 1000028001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_PROPERTIES_EXT";
        case 1000028002:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_STREAM_CREATE_INFO_EXT";
        case 1000029000:
            return "VK_STRUCTURE_TYPE_CU_MODULE_CREATE_INFO_NVX";
        case 1000029001:
            return "VK_STRUCTURE_TYPE_CU_FUNCTION_CREATE_INFO_NVX";
        case 1000029002:
            return "VK_STRUCTURE_TYPE_CU_LAUNCH_INFO_NVX";
        case 1000030000:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_HANDLE_INFO_NVX";
        case 1000030001:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_ADDRESS_PROPERTIES_NVX";
        case 1000038000:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_CAPABILITIES_KHR";
        case 1000038001:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR";
        case 1000038002:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR";
        case 1000038003:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PICTURE_INFO_KHR";
        case 1000038004:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_DPB_SLOT_INFO_KHR";
        case 1000038005:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_NALU_SLICE_INFO_KHR";
        case 1000038006:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_GOP_REMAINING_FRAME_INFO_KHR";
        case 1000038007:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PROFILE_INFO_KHR";
        case 1000038008:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_INFO_KHR";
        case 1000038009:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_LAYER_INFO_KHR";
        case 1000038010:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_CREATE_INFO_KHR";
        case 1000038011:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUALITY_LEVEL_PROPERTIES_KHR";
        case 1000038012:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_GET_INFO_KHR";
        case 1000038013:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_FEEDBACK_INFO_KHR";
        case 1000039000:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_CAPABILITIES_KHR";
        case 1000039001:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR";
        case 1000039002:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR";
        case 1000039003:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PICTURE_INFO_KHR";
        case 1000039004:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_DPB_SLOT_INFO_KHR";
        case 1000039005:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_NALU_SLICE_SEGMENT_INFO_KHR";
        case 1000039006:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_GOP_REMAINING_FRAME_INFO_KHR";
        case 1000039007:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PROFILE_INFO_KHR";
        case 1000039009:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_INFO_KHR";
        case 1000039010:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_LAYER_INFO_KHR";
        case 1000039011:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_CREATE_INFO_KHR";
        case 1000039012:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUALITY_LEVEL_PROPERTIES_KHR";
        case 1000039013:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_GET_INFO_KHR";
        case 1000039014:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_FEEDBACK_INFO_KHR";
        case 1000040000:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_CAPABILITIES_KHR";
        case 1000040001:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PICTURE_INFO_KHR";
        case 1000040003:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PROFILE_INFO_KHR";
        case 1000040004:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR";
        case 1000040005:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR";
        case 1000040006:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_DPB_SLOT_INFO_KHR";
        case 1000041000:
            return "VK_STRUCTURE_TYPE_TEXTURE_LOD_GATHER_FORMAT_PROPERTIES_AMD";
        case 1000044000:
            return "VK_STRUCTURE_TYPE_RENDERING_INFO";
        case 1000044001:
            return "VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO";
        case 1000044002:
            return "VK_STRUCTURE_TYPE_PIPELINE_RENDERING_CREATE_INFO";
        case 1000044003:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_FEATURES";
        case 1000044004:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDERING_INFO";
        case 1000044006:
            return "VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR";
        case 1000044007:
            return "VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_INFO_EXT";
        case 1000044008:
            return "VK_STRUCTURE_TYPE_ATTACHMENT_SAMPLE_COUNT_INFO_NV";
        case 1000044009:
            return "VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_ATTRIBUTES_INFO_NVX";
        case 1000049000:
            return "VK_STRUCTURE_TYPE_STREAM_DESCRIPTOR_SURFACE_CREATE_INFO_GGP";
        case 1000050000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CORNER_SAMPLED_IMAGE_FEATURES_NV";
        case 1000053000:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_MULTIVIEW_CREATE_INFO";
        case 1000053001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_FEATURES";
        case 1000053002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PROPERTIES";
        case 1000056000:
            return "VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO_NV";
        case 1000056001:
            return "VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO_NV";
        case 1000057000:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_NV";
        case 1000057001:
            return "VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_NV";
        case 1000058000:
            return "VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_NV";
        case 1000059000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2";
        case 1000059001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2";
        case 1000059002:
            return "VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2";
        case 1000059003:
            return "VK_STRUCTURE_TYPE_IMAGE_FORMAT_PROPERTIES_2";
        case 1000059004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2";
        case 1000059005:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_PROPERTIES_2";
        case 1000059006:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PROPERTIES_2";
        case 1000059007:
            return "VK_STRUCTURE_TYPE_SPARSE_IMAGE_FORMAT_PROPERTIES_2";
        case 1000059008:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SPARSE_IMAGE_FORMAT_INFO_2";
        case 1000060000:
            return "VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_FLAGS_INFO";
        case 1000060003:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_RENDER_PASS_BEGIN_INFO";
        case 1000060004:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_COMMAND_BUFFER_BEGIN_INFO";
        case 1000060005:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_SUBMIT_INFO";
        case 1000060006:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_BIND_SPARSE_INFO";
        case 1000060007:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_CAPABILITIES_KHR";
        case 1000060008:
            return "VK_STRUCTURE_TYPE_IMAGE_SWAPCHAIN_CREATE_INFO_KHR";
        case 1000060009:
            return "VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_SWAPCHAIN_INFO_KHR";
        case 1000060010:
            return "VK_STRUCTURE_TYPE_ACQUIRE_NEXT_IMAGE_INFO_KHR";
        case 1000060011:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_INFO_KHR";
        case 1000060012:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_SWAPCHAIN_CREATE_INFO_KHR";
        case 1000060013:
            return "VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_DEVICE_GROUP_INFO";
        case 1000060014:
            return "VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_DEVICE_GROUP_INFO";
        case 1000061000:
            return "VK_STRUCTURE_TYPE_VALIDATION_FLAGS_EXT";
        case 1000062000:
            return "VK_STRUCTURE_TYPE_VI_SURFACE_CREATE_INFO_NN";
        case 1000063000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETER_FEATURES";
        case 1000066000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_HDR_FEATURES";
        case 1000067000:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_ASTC_DECODE_MODE_EXT";
        case 1000067001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ASTC_DECODE_FEATURES_EXT";
        case 1000068000:
            return "VK_STRUCTURE_TYPE_PIPELINE_ROBUSTNESS_CREATE_INFO_EXT";
        case 1000068001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_FEATURES_EXT";
        case 1000068002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_PROPERTIES_EXT";
        case 1000070000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GROUP_PROPERTIES";
        case 1000070001:
            return "VK_STRUCTURE_TYPE_DEVICE_GROUP_DEVICE_CREATE_INFO";
        case 1000071000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_IMAGE_FORMAT_INFO";
        case 1000071001:
            return "VK_STRUCTURE_TYPE_EXTERNAL_IMAGE_FORMAT_PROPERTIES";
        case 1000071002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_BUFFER_INFO";
        case 1000071003:
            return "VK_STRUCTURE_TYPE_EXTERNAL_BUFFER_PROPERTIES";
        case 1000071004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ID_PROPERTIES";
        case 1000072000:
            return "VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_BUFFER_CREATE_INFO";
        case 1000072001:
            return "VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO";
        case 1000072002:
            return "VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO";
        case 1000073000:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_KHR";
        case 1000073001:
            return "VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_KHR";
        case 1000073002:
            return "VK_STRUCTURE_TYPE_MEMORY_WIN32_HANDLE_PROPERTIES_KHR";
        case 1000073003:
            return "VK_STRUCTURE_TYPE_MEMORY_GET_WIN32_HANDLE_INFO_KHR";
        case 1000074000:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_FD_INFO_KHR";
        case 1000074001:
            return "VK_STRUCTURE_TYPE_MEMORY_FD_PROPERTIES_KHR";
        case 1000074002:
            return "VK_STRUCTURE_TYPE_MEMORY_GET_FD_INFO_KHR";
        case 1000075000:
            return "VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_KHR";
        case 1000076000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_SEMAPHORE_INFO";
        case 1000076001:
            return "VK_STRUCTURE_TYPE_EXTERNAL_SEMAPHORE_PROPERTIES";
        case 1000077000:
            return "VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO";
        case 1000078000:
            return "VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR";
        case 1000078001:
            return "VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR";
        case 1000078002:
            return "VK_STRUCTURE_TYPE_D3D12_FENCE_SUBMIT_INFO_KHR";
        case 1000078003:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR";
        case 1000079000:
            return "VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_FD_INFO_KHR";
        case 1000079001:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR";
        case 1000080000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_DESCRIPTOR_PROPERTIES_KHR";
        case 1000081000:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_CONDITIONAL_RENDERING_INFO_EXT";
        case 1000081001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT";
        case 1000081002:
            return "VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_EXT";
        case 1000082000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT16_INT8_FEATURES_KHR";
        case 1000083000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_16BIT_STORAGE_FEATURES";
        case 1000084000:
            return "VK_STRUCTURE_TYPE_PRESENT_REGIONS_KHR";
        case 1000085000:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_UPDATE_TEMPLATE_CREATE_INFO";
        case 1000087000:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_W_SCALING_STATE_CREATE_INFO_NV";
        case 1000090000:
            return "VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES2_EXT";
        case 1000091000:
            return "VK_STRUCTURE_TYPE_DISPLAY_POWER_INFO_EXT";
        case 1000091001:
            return "VK_STRUCTURE_TYPE_DEVICE_EVENT_INFO_EXT";
        case 1000091002:
            return "VK_STRUCTURE_TYPE_DISPLAY_EVENT_INFO_EXT";
        case 1000091003:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_COUNTER_CREATE_INFO_EXT";
        case 1000092000:
            return "VK_STRUCTURE_TYPE_PRESENT_TIMES_INFO_GOOGLE";
        case 1000094000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES";
        case 1000097000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_ATTRIBUTES_PROPERTIES_NVX";
        case 1000098000:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SWIZZLE_STATE_CREATE_INFO_NV";
        case 1000099000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISCARD_RECTANGLE_PROPERTIES_EXT";
        case 1000099001:
            return "VK_STRUCTURE_TYPE_PIPELINE_DISCARD_RECTANGLE_STATE_CREATE_INFO_EXT";
        case 1000101000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONSERVATIVE_RASTERIZATION_PROPERTIES_EXT";
        case 1000101001:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_CONSERVATIVE_STATE_CREATE_INFO_EXT";
        case 1000102000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_ENABLE_FEATURES_EXT";
        case 1000102001:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_DEPTH_CLIP_STATE_CREATE_INFO_EXT";
        case 1000105000:
            return "VK_STRUCTURE_TYPE_HDR_METADATA_EXT";
        case 1000108000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGELESS_FRAMEBUFFER_FEATURES";
        case 1000108001:
            return "VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENTS_CREATE_INFO";
        case 1000108002:
            return "VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENT_IMAGE_INFO";
        case 1000108003:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_ATTACHMENT_BEGIN_INFO";
        case 1000109000:
            return "VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_2";
        case 1000109001:
            return "VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_2";
        case 1000109002:
            return "VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_2";
        case 1000109003:
            return "VK_STRUCTURE_TYPE_SUBPASS_DEPENDENCY_2";
        case 1000109004:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO_2";
        case 1000109005:
            return "VK_STRUCTURE_TYPE_SUBPASS_BEGIN_INFO";
        case 1000109006:
            return "VK_STRUCTURE_TYPE_SUBPASS_END_INFO";
        case 1000110000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RELAXED_LINE_RASTERIZATION_FEATURES_IMG";
        case 1000111000:
            return "VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_KHR";
        case 1000112000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FENCE_INFO";
        case 1000112001:
            return "VK_STRUCTURE_TYPE_EXTERNAL_FENCE_PROPERTIES";
        case 1000113000:
            return "VK_STRUCTURE_TYPE_EXPORT_FENCE_CREATE_INFO";
        case 1000114000:
            return "VK_STRUCTURE_TYPE_IMPORT_FENCE_WIN32_HANDLE_INFO_KHR";
        case 1000114001:
            return "VK_STRUCTURE_TYPE_EXPORT_FENCE_WIN32_HANDLE_INFO_KHR";
        case 1000114002:
            return "VK_STRUCTURE_TYPE_FENCE_GET_WIN32_HANDLE_INFO_KHR";
        case 1000115000:
            return "VK_STRUCTURE_TYPE_IMPORT_FENCE_FD_INFO_KHR";
        case 1000115001:
            return "VK_STRUCTURE_TYPE_FENCE_GET_FD_INFO_KHR";
        case 1000116000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_FEATURES_KHR";
        case 1000116001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_PROPERTIES_KHR";
        case 1000116002:
            return "VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_CREATE_INFO_KHR";
        case 1000116003:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_QUERY_SUBMIT_INFO_KHR";
        case 1000116004:
            return "VK_STRUCTURE_TYPE_ACQUIRE_PROFILING_LOCK_INFO_KHR";
        case 1000116005:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_KHR";
        case 1000116006:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_KHR";
        case 1000117000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_POINT_CLIPPING_PROPERTIES";
        case 1000117001:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_INPUT_ATTACHMENT_ASPECT_CREATE_INFO";
        case 1000117002:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_CREATE_INFO";
        case 1000117003:
            return "VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_DOMAIN_ORIGIN_STATE_CREATE_INFO";
        case 1000119000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SURFACE_INFO_2_KHR";
        case 1000119001:
            return "VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_KHR";
        case 1000119002:
            return "VK_STRUCTURE_TYPE_SURFACE_FORMAT_2_KHR";
        case 1000120000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VARIABLE_POINTER_FEATURES";
        case 1000121000:
            return "VK_STRUCTURE_TYPE_DISPLAY_PROPERTIES_2_KHR";
        case 1000121001:
            return "VK_STRUCTURE_TYPE_DISPLAY_PLANE_PROPERTIES_2_KHR";
        case 1000121002:
            return "VK_STRUCTURE_TYPE_DISPLAY_MODE_PROPERTIES_2_KHR";
        case 1000121003:
            return "VK_STRUCTURE_TYPE_DISPLAY_PLANE_INFO_2_KHR";
        case 1000121004:
            return "VK_STRUCTURE_TYPE_DISPLAY_PLANE_CAPABILITIES_2_KHR";
        case 1000122000:
            return "VK_STRUCTURE_TYPE_IOS_SURFACE_CREATE_INFO_MVK";
        case 1000123000:
            return "VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK";
        case 1000127000:
            return "VK_STRUCTURE_TYPE_MEMORY_DEDICATED_REQUIREMENTS";
        case 1000127001:
            return "VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO";
        case 1000128000:
            return "VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT";
        case 1000128001:
            return "VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_TAG_INFO_EXT";
        case 1000128002:
            return "VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT";
        case 1000128003:
            return "VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CALLBACK_DATA_EXT";
        case 1000128004:
            return "VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT";
        case 1000129000:
            return "VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_USAGE_ANDROID";
        case 1000129001:
            return "VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_PROPERTIES_ANDROID";
        case 1000129002:
            return "VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_ANDROID";
        case 1000129003:
            return "VK_STRUCTURE_TYPE_IMPORT_ANDROID_HARDWARE_BUFFER_INFO_ANDROID";
        case 1000129004:
            return "VK_STRUCTURE_TYPE_MEMORY_GET_ANDROID_HARDWARE_BUFFER_INFO_ANDROID";
        case 1000129005:
            return "VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_ANDROID";
        case 1000129006:
            return "VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_2_ANDROID";
        case 1000130000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_FILTER_MINMAX_PROPERTIES";
        case 1000130001:
            return "VK_STRUCTURE_TYPE_SAMPLER_REDUCTION_MODE_CREATE_INFO";
        case 1000134000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_FEATURES_AMDX";
        case 1000134001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_PROPERTIES_AMDX";
        case 1000134002:
            return "VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_SCRATCH_SIZE_AMDX";
        case 1000134003:
            return "VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_CREATE_INFO_AMDX";
        case 1000134004:
            return "VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_NODE_CREATE_INFO_AMDX";
        case 1000138000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_FEATURES";
        case 1000138001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_PROPERTIES";
        case 1000138002:
            return "VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_INLINE_UNIFORM_BLOCK";
        case 1000138003:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_INLINE_UNIFORM_BLOCK_CREATE_INFO";
        case 1000143000:
            return "VK_STRUCTURE_TYPE_SAMPLE_LOCATIONS_INFO_EXT";
        case 1000143001:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_SAMPLE_LOCATIONS_BEGIN_INFO_EXT";
        case 1000143002:
            return "VK_STRUCTURE_TYPE_PIPELINE_SAMPLE_LOCATIONS_STATE_CREATE_INFO_EXT";
        case 1000143003:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLE_LOCATIONS_PROPERTIES_EXT";
        case 1000143004:
            return "VK_STRUCTURE_TYPE_MULTISAMPLE_PROPERTIES_EXT";
        case 1000145000:
            return "VK_STRUCTURE_TYPE_PROTECTED_SUBMIT_INFO";
        case 1000145001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_FEATURES";
        case 1000145002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_PROPERTIES";
        case 1000145003:
            return "VK_STRUCTURE_TYPE_DEVICE_QUEUE_INFO_2";
        case 1000146000:
            return "VK_STRUCTURE_TYPE_BUFFER_MEMORY_REQUIREMENTS_INFO_2";
        case 1000146001:
            return "VK_STRUCTURE_TYPE_IMAGE_MEMORY_REQUIREMENTS_INFO_2";
        case 1000146002:
            return "VK_STRUCTURE_TYPE_IMAGE_SPARSE_MEMORY_REQUIREMENTS_INFO_2";
        case 1000146003:
            return "VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2";
        case 1000146004:
            return "VK_STRUCTURE_TYPE_SPARSE_IMAGE_MEMORY_REQUIREMENTS_2";
        case 1000147000:
            return "VK_STRUCTURE_TYPE_IMAGE_FORMAT_LIST_CREATE_INFO";
        case 1000148000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_FEATURES_EXT";
        case 1000148001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_PROPERTIES_EXT";
        case 1000148002:
            return "VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_ADVANCED_STATE_CREATE_INFO_EXT";
        case 1000149000:
            return "VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_TO_COLOR_STATE_CREATE_INFO_NV";
        case 1000150000:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_GEOMETRY_INFO_KHR";
        case 1000150002:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DEVICE_ADDRESS_INFO_KHR";
        case 1000150003:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_AABBS_DATA_KHR";
        case 1000150004:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_INSTANCES_DATA_KHR";
        case 1000150005:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_TRIANGLES_DATA_KHR";
        case 1000150006:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_KHR";
        case 1000150007:
            return "VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_KHR";
        case 1000150009:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_VERSION_INFO_KHR";
        case 1000150010:
            return "VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_INFO_KHR";
        case 1000150011:
            return "VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_TO_MEMORY_INFO_KHR";
        case 1000150012:
            return "VK_STRUCTURE_TYPE_COPY_MEMORY_TO_ACCELERATION_STRUCTURE_INFO_KHR";
        case 1000150013:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_FEATURES_KHR";
        case 1000150014:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_PROPERTIES_KHR";
        case 1000150015:
            return "VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_KHR";
        case 1000150016:
            return "VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_KHR";
        case 1000150017:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_KHR";
        case 1000150018:
            return "VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_INTERFACE_CREATE_INFO_KHR";
        case 1000150020:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_SIZES_INFO_KHR";
        case 1000152000:
            return "VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_MODULATION_STATE_CREATE_INFO_NV";
        case 1000154000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_FEATURES_NV";
        case 1000154001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_PROPERTIES_NV";
        case 1000156000:
            return "VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_CREATE_INFO";
        case 1000156001:
            return "VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_INFO";
        case 1000156002:
            return "VK_STRUCTURE_TYPE_BIND_IMAGE_PLANE_MEMORY_INFO";
        case 1000156003:
            return "VK_STRUCTURE_TYPE_IMAGE_PLANE_MEMORY_REQUIREMENTS_INFO";
        case 1000156004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_YCBCR_CONVERSION_FEATURES";
        case 1000156005:
            return "VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_IMAGE_FORMAT_PROPERTIES";
        case 1000157000:
            return "VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_INFO";
        case 1000157001:
            return "VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO";
        case 1000158000:
            return "VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_EXT";
        case 1000158002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT";
        case 1000158003:
            return "VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_LIST_CREATE_INFO_EXT";
        case 1000158004:
            return "VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_EXPLICIT_CREATE_INFO_EXT";
        case 1000158005:
            return "VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_PROPERTIES_EXT";
        case 1000158006:
            return "VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_2_EXT";
        case 1000160000:
            return "VK_STRUCTURE_TYPE_VALIDATION_CACHE_CREATE_INFO_EXT";
        case 1000160001:
            return "VK_STRUCTURE_TYPE_SHADER_MODULE_VALIDATION_CACHE_CREATE_INFO_EXT";
        case 1000161000:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_BINDING_FLAGS_CREATE_INFO";
        case 1000161001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_FEATURES";
        case 1000161002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_PROPERTIES";
        case 1000161003:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_ALLOCATE_INFO";
        case 1000161004:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_LAYOUT_SUPPORT";
        case 1000163000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_FEATURES_KHR";
        case 1000163001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_PROPERTIES_KHR";
        case 1000164000:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SHADING_RATE_IMAGE_STATE_CREATE_INFO_NV";
        case 1000164001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_FEATURES_NV";
        case 1000164002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_PROPERTIES_NV";
        case 1000164005:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_COARSE_SAMPLE_ORDER_STATE_CREATE_INFO_NV";
        case 1000165000:
            return "VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_NV";
        case 1000165001:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_NV";
        case 1000165003:
            return "VK_STRUCTURE_TYPE_GEOMETRY_NV";
        case 1000165004:
            return "VK_STRUCTURE_TYPE_GEOMETRY_TRIANGLES_NV";
        case 1000165005:
            return "VK_STRUCTURE_TYPE_GEOMETRY_AABB_NV";
        case 1000165006:
            return "VK_STRUCTURE_TYPE_BIND_ACCELERATION_STRUCTURE_MEMORY_INFO_NV";
        case 1000165007:
            return "VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_NV";
        case 1000165008:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_INFO_NV";
        case 1000165009:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PROPERTIES_NV";
        case 1000165011:
            return "VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_NV";
        case 1000165012:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_INFO_NV";
        case 1000166000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_REPRESENTATIVE_FRAGMENT_TEST_FEATURES_NV";
        case 1000166001:
            return "VK_STRUCTURE_TYPE_PIPELINE_REPRESENTATIVE_FRAGMENT_TEST_STATE_CREATE_INFO_NV";
        case 1000168000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_3_PROPERTIES";
        case 1000168001:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_SUPPORT";
        case 1000170000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_IMAGE_FORMAT_INFO_EXT";
        case 1000170001:
            return "VK_STRUCTURE_TYPE_FILTER_CUBIC_IMAGE_VIEW_IMAGE_FORMAT_PROPERTIES_EXT";
        case 1000174000:
            return "VK_STRUCTURE_TYPE_DEVICE_QUEUE_GLOBAL_PRIORITY_CREATE_INFO_KHR";
        case 1000175000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_EXTENDED_TYPES_FEATURES";
        case 1000177000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_8BIT_STORAGE_FEATURES";
        case 1000178000:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_HOST_POINTER_INFO_EXT";
        case 1000178001:
            return "VK_STRUCTURE_TYPE_MEMORY_HOST_POINTER_PROPERTIES_EXT";
        case 1000178002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_HOST_PROPERTIES_EXT";
        case 1000180000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_INT64_FEATURES";
        case 1000181000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CLOCK_FEATURES_KHR";
        case 1000183000:
            return "VK_STRUCTURE_TYPE_PIPELINE_COMPILER_CONTROL_CREATE_INFO_AMD";
        case 1000184000:
            return "VK_STRUCTURE_TYPE_CALIBRATED_TIMESTAMP_INFO_KHR";
        case 1000185000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_AMD";
        case 1000187000:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_CAPABILITIES_KHR";
        case 1000187001:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR";
        case 1000187002:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR";
        case 1000187003:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PROFILE_INFO_KHR";
        case 1000187004:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PICTURE_INFO_KHR";
        case 1000187005:
            return "VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_DPB_SLOT_INFO_KHR";
        case 1000189000:
            return "VK_STRUCTURE_TYPE_DEVICE_MEMORY_OVERALLOCATION_CREATE_INFO_AMD";
        case 1000190000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_EXT";
        case 1000190001:
            return "VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_DIVISOR_STATE_CREATE_INFO_KHR";
        case 1000190002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_FEATURES_KHR";
        case 1000191000:
            return "VK_STRUCTURE_TYPE_PRESENT_FRAME_TOKEN_GGP";
        case 1000192000:
            return "VK_STRUCTURE_TYPE_PIPELINE_CREATION_FEEDBACK_CREATE_INFO";
        case 1000196000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRIVER_PROPERTIES";
        case 1000197000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT_CONTROLS_PROPERTIES";
        case 1000199000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_STENCIL_RESOLVE_PROPERTIES";
        case 1000199001:
            return "VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_DEPTH_STENCIL_RESOLVE";
        case 1000201000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_FEATURES_NV";
        case 1000202000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_NV";
        case 1000202001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_NV";
        case 1000203000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_FEATURES_NV";
        case 1000204000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_FOOTPRINT_FEATURES_NV";
        case 1000205000:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_EXCLUSIVE_SCISSOR_STATE_CREATE_INFO_NV";
        case 1000205002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXCLUSIVE_SCISSOR_FEATURES_NV";
        case 1000206000:
            return "VK_STRUCTURE_TYPE_CHECKPOINT_DATA_NV";
        case 1000206001:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_NV";
        case 1000207000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES";
        case 1000207001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_PROPERTIES";
        case 1000207002:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO";
        case 1000207003:
            return "VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO";
        case 1000207004:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_WAIT_INFO";
        case 1000207005:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_SIGNAL_INFO";
        case 1000209000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_FUNCTIONS_2_FEATURES_INTEL";
        case 1000210000:
            return "VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO_INTEL";
        case 1000210001:
            return "VK_STRUCTURE_TYPE_INITIALIZE_PERFORMANCE_API_INFO_INTEL";
        case 1000210002:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_MARKER_INFO_INTEL";
        case 1000210003:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_STREAM_MARKER_INFO_INTEL";
        case 1000210004:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_OVERRIDE_INFO_INTEL";
        case 1000210005:
            return "VK_STRUCTURE_TYPE_PERFORMANCE_CONFIGURATION_ACQUIRE_INFO_INTEL";
        case 1000211000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_MEMORY_MODEL_FEATURES";
        case 1000212000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PCI_BUS_INFO_PROPERTIES_EXT";
        case 1000213000:
            return "VK_STRUCTURE_TYPE_DISPLAY_NATIVE_HDR_SURFACE_CAPABILITIES_AMD";
        case 1000213001:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_DISPLAY_NATIVE_HDR_CREATE_INFO_AMD";
        case 1000214000:
            return "VK_STRUCTURE_TYPE_IMAGEPIPE_SURFACE_CREATE_INFO_FUCHSIA";
        case 1000215000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TERMINATE_INVOCATION_FEATURES";
        case 1000217000:
            return "VK_STRUCTURE_TYPE_METAL_SURFACE_CREATE_INFO_EXT";
        case 1000218000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_FEATURES_EXT";
        case 1000218001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_PROPERTIES_EXT";
        case 1000218002:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_CREATE_INFO_EXT";
        case 1000221000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCALAR_BLOCK_LAYOUT_FEATURES";
        case 1000225000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_PROPERTIES";
        case 1000225001:
            return "VK_STRUCTURE_TYPE_SHADER_REQUIRED_SUBGROUP_SIZE_CREATE_INFO_EXT";
        case 1000225002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_FEATURES";
        case 1000226000:
            return "VK_STRUCTURE_TYPE_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR";
        case 1000226001:
            return "VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_STATE_CREATE_INFO_KHR";
        case 1000226002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_PROPERTIES_KHR";
        case 1000226003:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_FEATURES_KHR";
        case 1000226004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_KHR";
        case 1000227000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_2_AMD";
        case 1000229000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COHERENT_MEMORY_FEATURES_AMD";
        case 1000234000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_ATOMIC_INT64_FEATURES_EXT";
        case 1000237000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_BUDGET_PROPERTIES_EXT";
        case 1000238000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PRIORITY_FEATURES_EXT";
        case 1000238001:
            return "VK_STRUCTURE_TYPE_MEMORY_PRIORITY_ALLOCATE_INFO_EXT";
        case 1000239000:
            return "VK_STRUCTURE_TYPE_SURFACE_PROTECTED_CAPABILITIES_KHR";
        case 1000240000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEDICATED_ALLOCATION_IMAGE_ALIASING_FEATURES_NV";
        case 1000241000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SEPARATE_DEPTH_STENCIL_LAYOUTS_FEATURES";
        case 1000241001:
            return "VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_STENCIL_LAYOUT";
        case 1000241002:
            return "VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_STENCIL_LAYOUT";
        case 1000244000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_ADDRESS_FEATURES_EXT";
        case 1000244001:
            return "VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_INFO";
        case 1000244002:
            return "VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_CREATE_INFO_EXT";
        case 1000245000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TOOL_PROPERTIES";
        case 1000246000:
            return "VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_CREATE_INFO";
        case 1000247000:
            return "VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT";
        case 1000248000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_FEATURES_KHR";
        case 1000249000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_NV";
        case 1000249001:
            return "VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_NV";
        case 1000249002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_NV";
        case 1000250000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COVERAGE_REDUCTION_MODE_FEATURES_NV";
        case 1000250001:
            return "VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_REDUCTION_STATE_CREATE_INFO_NV";
        case 1000250002:
            return "VK_STRUCTURE_TYPE_FRAMEBUFFER_MIXED_SAMPLES_COMBINATION_NV";
        case 1000251000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_INTERLOCK_FEATURES_EXT";
        case 1000252000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_IMAGE_ARRAYS_FEATURES_EXT";
        case 1000253000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFORM_BUFFER_STANDARD_LAYOUT_FEATURES";
        case 1000254000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_FEATURES_EXT";
        case 1000254001:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_PROVOKING_VERTEX_STATE_CREATE_INFO_EXT";
        case 1000254002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_PROPERTIES_EXT";
        case 1000255000:
            return "VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_INFO_EXT";
        case 1000255001:
            return "VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_WIN32_INFO_EXT";
        case 1000255002:
            return "VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_FULL_SCREEN_EXCLUSIVE_EXT";
        case 1000256000:
            return "VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT";
        case 1000257000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES";
        case 1000257002:
            return "VK_STRUCTURE_TYPE_BUFFER_OPAQUE_CAPTURE_ADDRESS_CREATE_INFO";
        case 1000257003:
            return "VK_STRUCTURE_TYPE_MEMORY_OPAQUE_CAPTURE_ADDRESS_ALLOCATE_INFO";
        case 1000257004:
            return "VK_STRUCTURE_TYPE_DEVICE_MEMORY_OPAQUE_CAPTURE_ADDRESS_INFO";
        case 1000259000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_FEATURES_EXT";
        case 1000259001:
            return "VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_LINE_STATE_CREATE_INFO_EXT";
        case 1000259002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_PROPERTIES_EXT";
        case 1000260000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_FEATURES_EXT";
        case 1000261000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES";
        case 1000265000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INDEX_TYPE_UINT8_FEATURES_EXT";
        case 1000267000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT";
        case 1000269000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_EXECUTABLE_PROPERTIES_FEATURES_KHR";
        case 1000269001:
            return "VK_STRUCTURE_TYPE_PIPELINE_INFO_KHR";
        case 1000269002:
            return "VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_PROPERTIES_KHR";
        case 1000269003:
            return "VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INFO_KHR";
        case 1000269004:
            return "VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_STATISTIC_KHR";
        case 1000269005:
            return "VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INTERNAL_REPRESENTATION_KHR";
        case 1000270000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_FEATURES_EXT";
        case 1000270001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_PROPERTIES_EXT";
        case 1000270002:
            return "VK_STRUCTURE_TYPE_MEMORY_TO_IMAGE_COPY_EXT";
        case 1000270003:
            return "VK_STRUCTURE_TYPE_IMAGE_TO_MEMORY_COPY_EXT";
        case 1000270004:
            return "VK_STRUCTURE_TYPE_COPY_IMAGE_TO_MEMORY_INFO_EXT";
        case 1000270005:
            return "VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INFO_EXT";
        case 1000270006:
            return "VK_STRUCTURE_TYPE_HOST_IMAGE_LAYOUT_TRANSITION_INFO_EXT";
        case 1000270007:
            return "VK_STRUCTURE_TYPE_COPY_IMAGE_TO_IMAGE_INFO_EXT";
        case 1000270008:
            return "VK_STRUCTURE_TYPE_SUBRESOURCE_HOST_MEMCPY_SIZE_EXT";
        case 1000270009:
            return "VK_STRUCTURE_TYPE_HOST_IMAGE_COPY_DEVICE_PERFORMANCE_QUERY_EXT";
        case 1000271000:
            return "VK_STRUCTURE_TYPE_MEMORY_MAP_INFO_KHR";
        case 1000271001:
            return "VK_STRUCTURE_TYPE_MEMORY_UNMAP_INFO_KHR";
        case 1000273000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_2_FEATURES_EXT";
        case 1000274000:
            return "VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_EXT";
        case 1000274001:
            return "VK_STRUCTURE_TYPE_SURFACE_PRESENT_SCALING_CAPABILITIES_EXT";
        case 1000274002:
            return "VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_COMPATIBILITY_EXT";
        case 1000275000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SWAPCHAIN_MAINTENANCE_1_FEATURES_EXT";
        case 1000275001:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_FENCE_INFO_EXT";
        case 1000275002:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODES_CREATE_INFO_EXT";
        case 1000275003:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODE_INFO_EXT";
        case 1000275004:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_SCALING_CREATE_INFO_EXT";
        case 1000275005:
            return "VK_STRUCTURE_TYPE_RELEASE_SWAPCHAIN_IMAGES_INFO_EXT";
        case 1000276000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DEMOTE_TO_HELPER_INVOCATION_FEATURES";
        case 1000277000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_NV";
        case 1000277001:
            return "VK_STRUCTURE_TYPE_GRAPHICS_SHADER_GROUP_CREATE_INFO_NV";
        case 1000277002:
            return "VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_SHADER_GROUPS_CREATE_INFO_NV";
        case 1000277003:
            return "VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_NV";
        case 1000277004:
            return "VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_NV";
        case 1000277005:
            return "VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_NV";
        case 1000277006:
            return "VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_NV";
        case 1000277007:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_NV";
        case 1000278000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INHERITED_VIEWPORT_SCISSOR_FEATURES_NV";
        case 1000278001:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_VIEWPORT_SCISSOR_INFO_NV";
        case 1000280000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_FEATURES";
        case 1000280001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_PROPERTIES";
        case 1000281000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_FEATURES_EXT";
        case 1000281001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_PROPERTIES";
        case 1000282000:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDER_PASS_TRANSFORM_INFO_QCOM";
        case 1000282001:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_TRANSFORM_BEGIN_INFO_QCOM";
        case 1000283000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_BIAS_CONTROL_FEATURES_EXT";
        case 1000283001:
            return "VK_STRUCTURE_TYPE_DEPTH_BIAS_INFO_EXT";
        case 1000283002:
            return "VK_STRUCTURE_TYPE_DEPTH_BIAS_REPRESENTATION_INFO_EXT";
        case 1000284000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_MEMORY_REPORT_FEATURES_EXT";
        case 1000284001:
            return "VK_STRUCTURE_TYPE_DEVICE_DEVICE_MEMORY_REPORT_CREATE_INFO_EXT";
        case 1000284002:
            return "VK_STRUCTURE_TYPE_DEVICE_MEMORY_REPORT_CALLBACK_DATA_EXT";
        case 1000286000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_FEATURES_EXT";
        case 1000286001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_PROPERTIES_EXT";
        case 1000287000:
            return "VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_CREATE_INFO_EXT";
        case 1000287001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_PROPERTIES_EXT";
        case 1000287002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_FEATURES_EXT";
        case 1000290000:
            return "VK_STRUCTURE_TYPE_PIPELINE_LIBRARY_CREATE_INFO_KHR";
        case 1000292000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_BARRIER_FEATURES_NV";
        case 1000292001:
            return "VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_BARRIER_NV";
        case 1000292002:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_BARRIER_CREATE_INFO_NV";
        case 1000294000:
            return "VK_STRUCTURE_TYPE_PRESENT_ID_KHR";
        case 1000294001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_FEATURES_KHR";
        case 1000295000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIVATE_DATA_FEATURES";
        case 1000295001:
            return "VK_STRUCTURE_TYPE_DEVICE_PRIVATE_DATA_CREATE_INFO";
        case 1000295002:
            return "VK_STRUCTURE_TYPE_PRIVATE_DATA_SLOT_CREATE_INFO";
        case 1000297000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CREATION_CACHE_CONTROL_FEATURES";
        case 1000299000:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_INFO_KHR";
        case 1000299001:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_INFO_KHR";
        case 1000299002:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_LAYER_INFO_KHR";
        case 1000299003:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_CAPABILITIES_KHR";
        case 1000299004:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_USAGE_INFO_KHR";
        case 1000299005:
            return "VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_FEEDBACK_CREATE_INFO_KHR";
        case 1000299006:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR";
        case 1000299007:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_PROPERTIES_KHR";
        case 1000299008:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR";
        case 1000299009:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_GET_INFO_KHR";
        case 1000299010:
            return "VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_FEEDBACK_INFO_KHR";
        case 1000300000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DIAGNOSTICS_CONFIG_FEATURES_NV";
        case 1000300001:
            return "VK_STRUCTURE_TYPE_DEVICE_DIAGNOSTICS_CONFIG_CREATE_INFO_NV";
        case 1000307000:
            return "VK_STRUCTURE_TYPE_CUDA_MODULE_CREATE_INFO_NV";
        case 1000307001:
            return "VK_STRUCTURE_TYPE_CUDA_FUNCTION_CREATE_INFO_NV";
        case 1000307002:
            return "VK_STRUCTURE_TYPE_CUDA_LAUNCH_INFO_NV";
        case 1000307003:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_FEATURES_NV";
        case 1000307004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_PROPERTIES_NV";
        case 1000310000:
            return "VK_STRUCTURE_TYPE_QUERY_LOW_LATENCY_SUPPORT_NV";
        case 1000311000:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECT_CREATE_INFO_EXT";
        case 1000311001:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECTS_INFO_EXT";
        case 1000311002:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_DEVICE_INFO_EXT";
        case 1000311003:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_COMMAND_QUEUE_INFO_EXT";
        case 1000311004:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_BUFFER_INFO_EXT";
        case 1000311005:
            return "VK_STRUCTURE_TYPE_IMPORT_METAL_BUFFER_INFO_EXT";
        case 1000311006:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_TEXTURE_INFO_EXT";
        case 1000311007:
            return "VK_STRUCTURE_TYPE_IMPORT_METAL_TEXTURE_INFO_EXT";
        case 1000311008:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_IO_SURFACE_INFO_EXT";
        case 1000311009:
            return "VK_STRUCTURE_TYPE_IMPORT_METAL_IO_SURFACE_INFO_EXT";
        case 1000311010:
            return "VK_STRUCTURE_TYPE_EXPORT_METAL_SHARED_EVENT_INFO_EXT";
        case 1000311011:
            return "VK_STRUCTURE_TYPE_IMPORT_METAL_SHARED_EVENT_INFO_EXT";
        case 1000314000:
            return "VK_STRUCTURE_TYPE_MEMORY_BARRIER_2";
        case 1000314001:
            return "VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2";
        case 1000314002:
            return "VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2";
        case 1000314003:
            return "VK_STRUCTURE_TYPE_DEPENDENCY_INFO";
        case 1000314004:
            return "VK_STRUCTURE_TYPE_SUBMIT_INFO_2";
        case 1000314005:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO";
        case 1000314006:
            return "VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO";
        case 1000314007:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SYNCHRONIZATION_2_FEATURES";
        case 1000314008:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_2_NV";
        case 1000314009:
            return "VK_STRUCTURE_TYPE_CHECKPOINT_DATA_2_NV";
        case 1000316000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_PROPERTIES_EXT";
        case 1000316001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_DENSITY_MAP_PROPERTIES_EXT";
        case 1000316002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_FEATURES_EXT";
        case 1000316003:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_ADDRESS_INFO_EXT";
        case 1000316004:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_GET_INFO_EXT";
        case 1000316005:
            return "VK_STRUCTURE_TYPE_BUFFER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT";
        case 1000316006:
            return "VK_STRUCTURE_TYPE_IMAGE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT";
        case 1000316007:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_EXT";
        case 1000316008:
            return "VK_STRUCTURE_TYPE_SAMPLER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT";
        case 1000316009:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT";
        case 1000316010:
            return "VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DESCRIPTOR_DATA_CREATE_INFO_EXT";
        case 1000316011:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_INFO_EXT";
        case 1000316012:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_PUSH_DESCRIPTOR_BUFFER_HANDLE_EXT";
        case 1000320000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_FEATURES_EXT";
        case 1000320001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_PROPERTIES_EXT";
        case 1000320002:
            return "VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_LIBRARY_CREATE_INFO_EXT";
        case 1000321000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EARLY_AND_LATE_FRAGMENT_TESTS_FEATURES_AMD";
        case 1000322000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_PROPERTIES_KHR";
        case 1000323000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_UNIFORM_CONTROL_FLOW_FEATURES_KHR";
        case 1000325000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_WORKGROUP_MEMORY_FEATURES";
        case 1000326000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_PROPERTIES_NV";
        case 1000326001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_FEATURES_NV";
        case 1000326002:
            return "VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_ENUM_STATE_CREATE_INFO_NV";
        case 1000327000:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MOTION_TRIANGLES_DATA_NV";
        case 1000327001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MOTION_BLUR_FEATURES_NV";
        case 1000327002:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MOTION_INFO_NV";
        case 1000328000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_EXT";
        case 1000328001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_EXT";
        case 1000330000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_2_PLANE_444_FORMATS_FEATURES_EXT";
        case 1000332000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_FEATURES_EXT";
        case 1000332001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_PROPERTIES_EXT";
        case 1000333000:
            return "VK_STRUCTURE_TYPE_COPY_COMMAND_TRANSFORM_INFO_QCOM";
        case 1000335000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ROBUSTNESS_FEATURES";
        case 1000336000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_WORKGROUP_MEMORY_EXPLICIT_LAYOUT_FEATURES_KHR";
        case 1000337000:
            return "VK_STRUCTURE_TYPE_COPY_BUFFER_INFO_2";
        case 1000337001:
            return "VK_STRUCTURE_TYPE_COPY_IMAGE_INFO_2";
        case 1000337002:
            return "VK_STRUCTURE_TYPE_COPY_BUFFER_TO_IMAGE_INFO_2";
        case 1000337003:
            return "VK_STRUCTURE_TYPE_COPY_IMAGE_TO_BUFFER_INFO_2";
        case 1000337004:
            return "VK_STRUCTURE_TYPE_BLIT_IMAGE_INFO_2";
        case 1000337005:
            return "VK_STRUCTURE_TYPE_RESOLVE_IMAGE_INFO_2";
        case 1000337006:
            return "VK_STRUCTURE_TYPE_BUFFER_COPY_2";
        case 1000337007:
            return "VK_STRUCTURE_TYPE_IMAGE_COPY_2";
        case 1000337008:
            return "VK_STRUCTURE_TYPE_IMAGE_BLIT_2";
        case 1000337009:
            return "VK_STRUCTURE_TYPE_BUFFER_IMAGE_COPY_2";
        case 1000337010:
            return "VK_STRUCTURE_TYPE_IMAGE_RESOLVE_2";
        case 1000338000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_FEATURES_EXT";
        case 1000338001:
            return "VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_CONTROL_EXT";
        case 1000338002:
            return "VK_STRUCTURE_TYPE_SUBRESOURCE_LAYOUT_2_KHR";
        case 1000338003:
            return "VK_STRUCTURE_TYPE_IMAGE_SUBRESOURCE_2_KHR";
        case 1000338004:
            return "VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_PROPERTIES_EXT";
        case 1000339000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_LAYOUT_FEATURES_EXT";
        case 1000340000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT";
        case 1000341000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_EXT";
        case 1000341001:
            return "VK_STRUCTURE_TYPE_DEVICE_FAULT_COUNTS_EXT";
        case 1000341002:
            return "VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_EXT";
        case 1000342000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_FEATURES_EXT";
        case 1000344000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RGBA10X6_FORMATS_FEATURES_EXT";
        case 1000346000:
            return "VK_STRUCTURE_TYPE_DIRECTFB_SURFACE_CREATE_INFO_EXT";
        case 1000347000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_FEATURES_KHR";
        case 1000347001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_PROPERTIES_KHR";
        case 1000348013:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_QUERY_FEATURES_KHR";
        case 1000351000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MUTABLE_DESCRIPTOR_TYPE_FEATURES_EXT";
        case 1000351002:
            return "VK_STRUCTURE_TYPE_MUTABLE_DESCRIPTOR_TYPE_CREATE_INFO_EXT";
        case 1000352000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_INPUT_DYNAMIC_STATE_FEATURES_EXT";
        case 1000352001:
            return "VK_STRUCTURE_TYPE_VERTEX_INPUT_BINDING_DESCRIPTION_2_EXT";
        case 1000352002:
            return "VK_STRUCTURE_TYPE_VERTEX_INPUT_ATTRIBUTE_DESCRIPTION_2_EXT";
        case 1000353000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRM_PROPERTIES_EXT";
        case 1000354000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ADDRESS_BINDING_REPORT_FEATURES_EXT";
        case 1000354001:
            return "VK_STRUCTURE_TYPE_DEVICE_ADDRESS_BINDING_CALLBACK_DATA_EXT";
        case 1000355000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_CONTROL_FEATURES_EXT";
        case 1000355001:
            return "VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLIP_CONTROL_CREATE_INFO_EXT";
        case 1000356000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_TOPOLOGY_LIST_RESTART_FEATURES_EXT";
        case 1000360000:
            return "VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_3";
        case 1000364000:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_ZIRCON_HANDLE_INFO_FUCHSIA";
        case 1000364001:
            return "VK_STRUCTURE_TYPE_MEMORY_ZIRCON_HANDLE_PROPERTIES_FUCHSIA";
        case 1000364002:
            return "VK_STRUCTURE_TYPE_MEMORY_GET_ZIRCON_HANDLE_INFO_FUCHSIA";
        case 1000365000:
            return "VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_ZIRCON_HANDLE_INFO_FUCHSIA";
        case 1000365001:
            return "VK_STRUCTURE_TYPE_SEMAPHORE_GET_ZIRCON_HANDLE_INFO_FUCHSIA";
        case 1000366000:
            return "VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CREATE_INFO_FUCHSIA";
        case 1000366001:
            return "VK_STRUCTURE_TYPE_IMPORT_MEMORY_BUFFER_COLLECTION_FUCHSIA";
        case 1000366002:
            return "VK_STRUCTURE_TYPE_BUFFER_COLLECTION_IMAGE_CREATE_INFO_FUCHSIA";
        case 1000366003:
            return "VK_STRUCTURE_TYPE_BUFFER_COLLECTION_PROPERTIES_FUCHSIA";
        case 1000366004:
            return "VK_STRUCTURE_TYPE_BUFFER_CONSTRAINTS_INFO_FUCHSIA";
        case 1000366005:
            return "VK_STRUCTURE_TYPE_BUFFER_COLLECTION_BUFFER_CREATE_INFO_FUCHSIA";
        case 1000366006:
            return "VK_STRUCTURE_TYPE_IMAGE_CONSTRAINTS_INFO_FUCHSIA";
        case 1000366007:
            return "VK_STRUCTURE_TYPE_IMAGE_FORMAT_CONSTRAINTS_INFO_FUCHSIA";
        case 1000366008:
            return "VK_STRUCTURE_TYPE_SYSMEM_COLOR_SPACE_FUCHSIA";
        case 1000366009:
            return "VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CONSTRAINTS_INFO_FUCHSIA";
        case 1000369000:
            return "VK_STRUCTURE_TYPE_SUBPASS_SHADING_PIPELINE_CREATE_INFO_HUAWEI";
        case 1000369001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_FEATURES_HUAWEI";
        case 1000369002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_PROPERTIES_HUAWEI";
        case 1000370000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INVOCATION_MASK_FEATURES_HUAWEI";
        case 1000371000:
            return "VK_STRUCTURE_TYPE_MEMORY_GET_REMOTE_ADDRESS_INFO_NV";
        case 1000371001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_RDMA_FEATURES_NV";
        case 1000372000:
            return "VK_STRUCTURE_TYPE_PIPELINE_PROPERTIES_IDENTIFIER_EXT";
        case 1000372001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROPERTIES_FEATURES_EXT";
        case 1000375000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAME_BOUNDARY_FEATURES_EXT";
        case 1000375001:
            return "VK_STRUCTURE_TYPE_FRAME_BOUNDARY_EXT";
        case 1000376000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_FEATURES_EXT";
        case 1000376001:
            return "VK_STRUCTURE_TYPE_SUBPASS_RESOLVE_PERFORMANCE_QUERY_EXT";
        case 1000376002:
            return "VK_STRUCTURE_TYPE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_INFO_EXT";
        case 1000377000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_2_FEATURES_EXT";
        case 1000378000:
            return "VK_STRUCTURE_TYPE_SCREEN_SURFACE_CREATE_INFO_QNX";
        case 1000381000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COLOR_WRITE_ENABLE_FEATURES_EXT";
        case 1000381001:
            return "VK_STRUCTURE_TYPE_PIPELINE_COLOR_WRITE_CREATE_INFO_EXT";
        case 1000382000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVES_GENERATED_QUERY_FEATURES_EXT";
        case 1000386000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MAINTENANCE_1_FEATURES_KHR";
        case 1000388000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GLOBAL_PRIORITY_QUERY_FEATURES_KHR";
        case 1000388001:
            return "VK_STRUCTURE_TYPE_QUEUE_FAMILY_GLOBAL_PRIORITY_PROPERTIES_KHR";
        case 1000391000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_MIN_LOD_FEATURES_EXT";
        case 1000391001:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_MIN_LOD_CREATE_INFO_EXT";
        case 1000392000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_FEATURES_EXT";
        case 1000392001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_PROPERTIES_EXT";
        case 1000393000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_2D_VIEW_OF_3D_FEATURES_EXT";
        case 1000395000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_FEATURES_EXT";
        case 1000395001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_PROPERTIES_EXT";
        case 1000396000:
            return "VK_STRUCTURE_TYPE_MICROMAP_BUILD_INFO_EXT";
        case 1000396001:
            return "VK_STRUCTURE_TYPE_MICROMAP_VERSION_INFO_EXT";
        case 1000396002:
            return "VK_STRUCTURE_TYPE_COPY_MICROMAP_INFO_EXT";
        case 1000396003:
            return "VK_STRUCTURE_TYPE_COPY_MICROMAP_TO_MEMORY_INFO_EXT";
        case 1000396004:
            return "VK_STRUCTURE_TYPE_COPY_MEMORY_TO_MICROMAP_INFO_EXT";
        case 1000396005:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_EXT";
        case 1000396006:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_EXT";
        case 1000396007:
            return "VK_STRUCTURE_TYPE_MICROMAP_CREATE_INFO_EXT";
        case 1000396008:
            return "VK_STRUCTURE_TYPE_MICROMAP_BUILD_SIZES_INFO_EXT";
        case 1000396009:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_EXT";
        case 1000397000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_FEATURES_NV";
        case 1000397001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_PROPERTIES_NV";
        case 1000397002:
            return "VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_DISPLACEMENT_MICROMAP_NV";
        case 1000404000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_FEATURES_HUAWEI";
        case 1000404001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_PROPERTIES_HUAWEI";
        case 1000404002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_VRS_FEATURES_HUAWEI";
        case 1000411000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BORDER_COLOR_SWIZZLE_FEATURES_EXT";
        case 1000411001:
            return "VK_STRUCTURE_TYPE_SAMPLER_BORDER_COLOR_COMPONENT_MAPPING_CREATE_INFO_EXT";
        case 1000412000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PAGEABLE_DEVICE_LOCAL_MEMORY_FEATURES_EXT";
        case 1000413000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES";
        case 1000413001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_PROPERTIES";
        case 1000413002:
            return "VK_STRUCTURE_TYPE_DEVICE_BUFFER_MEMORY_REQUIREMENTS";
        case 1000413003:
            return "VK_STRUCTURE_TYPE_DEVICE_IMAGE_MEMORY_REQUIREMENTS";
        case 1000415000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_ARM";
        case 1000417000:
            return "VK_STRUCTURE_TYPE_DEVICE_QUEUE_SHADER_CORE_CONTROL_CREATE_INFO_ARM";
        case 1000417001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_FEATURES_ARM";
        case 1000417002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_PROPERTIES_ARM";
        case 1000418000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_SLICED_VIEW_OF_3D_FEATURES_EXT";
        case 1000418001:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_SLICED_CREATE_INFO_EXT";
        case 1000420000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_SET_HOST_MAPPING_FEATURES_VALVE";
        case 1000420001:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_BINDING_REFERENCE_VALVE";
        case 1000420002:
            return "VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_HOST_MAPPING_INFO_VALVE";
        case 1000421000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_ZERO_ONE_FEATURES_EXT";
        case 1000422000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT";
        case 1000424000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_FEATURES_ARM";
        case 1000424001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_PROPERTIES_ARM";
        case 1000424002:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_BEGIN_INFO_ARM";
        case 1000424003:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_INFO_ARM";
        case 1000424004:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_SUBMIT_INFO_ARM";
        case 1000425000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_FEATURES_QCOM";
        case 1000425001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_PROPERTIES_QCOM";
        case 1000425002:
            return "VK_STRUCTURE_TYPE_SUBPASS_FRAGMENT_DENSITY_MAP_OFFSET_END_INFO_QCOM";
        case 1000426000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_NV";
        case 1000426001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_PROPERTIES_NV";
        case 1000427000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_FEATURES_NV";
        case 1000427001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_PROPERTIES_NV";
        case 1000428000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_COMPUTE_FEATURES_NV";
        case 1000428001:
            return "VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_INDIRECT_BUFFER_INFO_NV";
        case 1000428002:
            return "VK_STRUCTURE_TYPE_PIPELINE_INDIRECT_DEVICE_ADDRESS_INFO_NV";
        case 1000430000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINEAR_COLOR_ATTACHMENT_FEATURES_NV";
        case 1000437000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_SWAPCHAIN_FEATURES_EXT";
        case 1000440000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_FEATURES_QCOM";
        case 1000440001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_PROPERTIES_QCOM";
        case 1000440002:
            return "VK_STRUCTURE_TYPE_IMAGE_VIEW_SAMPLE_WEIGHT_CREATE_INFO_QCOM";
        case 1000451000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_FEATURES_EXT";
        case 1000451001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_PROPERTIES_EXT";
        case 1000453000:
            return "VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_ACQUIRE_UNMODIFIED_EXT";
        case 1000455000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_FEATURES_EXT";
        case 1000455001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_PROPERTIES_EXT";
        case 1000458000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_MERGE_FEEDBACK_FEATURES_EXT";
        case 1000458001:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_CONTROL_EXT";
        case 1000458002:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_FEEDBACK_CREATE_INFO_EXT";
        case 1000458003:
            return "VK_STRUCTURE_TYPE_RENDER_PASS_SUBPASS_FEEDBACK_CREATE_INFO_EXT";
        case 1000459000:
            return "VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_INFO_LUNARG";
        case 1000459001:
            return "VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_LIST_LUNARG";
        case 1000462000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_FEATURES_EXT";
        case 1000462001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_PROPERTIES_EXT";
        case 1000462002:
            return "VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_MODULE_IDENTIFIER_CREATE_INFO_EXT";
        case 1000462003:
            return "VK_STRUCTURE_TYPE_SHADER_MODULE_IDENTIFIER_EXT";
        case 1000464000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_FEATURES_NV";
        case 1000464001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_PROPERTIES_NV";
        case 1000464002:
            return "VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_INFO_NV";
        case 1000464003:
            return "VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_NV";
        case 1000464004:
            return "VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_INFO_NV";
        case 1000464005:
            return "VK_STRUCTURE_TYPE_OPTICAL_FLOW_EXECUTE_INFO_NV";
        case 1000464010:
            return "VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_PRIVATE_DATA_INFO_NV";
        case 1000465000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_DITHERING_FEATURES_EXT";
        case 1000466000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROTECTED_ACCESS_FEATURES_EXT";
        case 1000468000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_FEATURES_ANDROID";
        case 1000468001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_PROPERTIES_ANDROID";
        case 1000468002:
            return "VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_RESOLVE_PROPERTIES_ANDROID";
        case 1000470000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_FEATURES_KHR";
        case 1000470001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_PROPERTIES_KHR";
        case 1000470003:
            return "VK_STRUCTURE_TYPE_RENDERING_AREA_INFO_KHR";
        case 1000470004:
            return "VK_STRUCTURE_TYPE_DEVICE_IMAGE_SUBRESOURCE_INFO_KHR";
        case 1000470005:
            return "VK_STRUCTURE_TYPE_PIPELINE_CREATE_FLAGS_2_CREATE_INFO_KHR";
        case 1000470006:
            return "VK_STRUCTURE_TYPE_BUFFER_USAGE_FLAGS_2_CREATE_INFO_KHR";
        case 1000481000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_POSITION_FETCH_FEATURES_KHR";
        case 1000482000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_FEATURES_EXT";
        case 1000482001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_PROPERTIES_EXT";
        case 1000482002:
            return "VK_STRUCTURE_TYPE_SHADER_CREATE_INFO_EXT";
        case 1000484000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_PROPERTIES_FEATURES_QCOM";
        case 1000484001:
            return "VK_STRUCTURE_TYPE_TILE_PROPERTIES_QCOM";
        case 1000485000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_AMIGO_PROFILING_FEATURES_SEC";
        case 1000485001:
            return "VK_STRUCTURE_TYPE_AMIGO_PROFILING_SUBMIT_INFO_SEC";
        case 1000488000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_VIEWPORTS_FEATURES_QCOM";
        case 1000490000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_NV";
        case 1000490001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_NV";
        case 1000492000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_FEATURES_NV";
        case 1000492001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_PROPERTIES_NV";
        case 1000496000:
            return "VK_STRUCTURE_TYPE_LAYER_SETTINGS_CREATE_INFO_EXT";
        case 1000497000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_FEATURES_ARM";
        case 1000497001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_PROPERTIES_ARM";
        case 1000498000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_LIBRARY_GROUP_HANDLES_FEATURES_EXT";
        case 1000499000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_UNUSED_ATTACHMENTS_FEATURES_EXT";
        case 1000505000:
            return "VK_STRUCTURE_TYPE_LATENCY_SLEEP_MODE_INFO_NV";
        case 1000505001:
            return "VK_STRUCTURE_TYPE_LATENCY_SLEEP_INFO_NV";
        case 1000505002:
            return "VK_STRUCTURE_TYPE_SET_LATENCY_MARKER_INFO_NV";
        case 1000505003:
            return "VK_STRUCTURE_TYPE_GET_LATENCY_MARKER_INFO_NV";
        case 1000505004:
            return "VK_STRUCTURE_TYPE_LATENCY_TIMINGS_FRAME_REPORT_NV";
        case 1000505005:
            return "VK_STRUCTURE_TYPE_LATENCY_SUBMISSION_PRESENT_ID_NV";
        case 1000505006:
            return "VK_STRUCTURE_TYPE_OUT_OF_BAND_QUEUE_TYPE_INFO_NV";
        case 1000505007:
            return "VK_STRUCTURE_TYPE_SWAPCHAIN_LATENCY_CREATE_INFO_NV";
        case 1000505008:
            return "VK_STRUCTURE_TYPE_LATENCY_SURFACE_CAPABILITIES_NV";
        case 1000506000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_KHR";
        case 1000506001:
            return "VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_KHR";
        case 1000506002:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_KHR";
        case 1000510000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_RENDER_AREAS_FEATURES_QCOM";
        case 1000510001:
            return "VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_RENDER_AREAS_RENDER_PASS_BEGIN_INFO_QCOM";
        case 1000515000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_1_FEATURES_KHR";
        case 1000515001:
            return "VK_STRUCTURE_TYPE_VIDEO_INLINE_QUERY_INFO_KHR";
        case 1000516000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PER_STAGE_DESCRIPTOR_SET_FEATURES_NV";
        case 1000518000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_FEATURES_QCOM";
        case 1000518001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_PROPERTIES_QCOM";
        case 1000518002:
            return "VK_STRUCTURE_TYPE_SAMPLER_BLOCK_MATCH_WINDOW_CREATE_INFO_QCOM";
        case 1000519000:
            return "VK_STRUCTURE_TYPE_SAMPLER_CUBIC_WEIGHTS_CREATE_INFO_QCOM";
        case 1000519001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_WEIGHTS_FEATURES_QCOM";
        case 1000519002:
            return "VK_STRUCTURE_TYPE_BLIT_IMAGE_CUBIC_WEIGHTS_INFO_QCOM";
        case 1000520000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_DEGAMMA_FEATURES_QCOM";
        case 1000520001:
            return "VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_YCBCR_DEGAMMA_CREATE_INFO_QCOM";
        case 1000521000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_CLAMP_FEATURES_QCOM";
        case 1000524000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_DYNAMIC_STATE_FEATURES_EXT";
        case 1000525000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_KHR";
        case 1000529000:
            return "VK_STRUCTURE_TYPE_SCREEN_BUFFER_PROPERTIES_QNX";
        case 1000529001:
            return "VK_STRUCTURE_TYPE_SCREEN_BUFFER_FORMAT_PROPERTIES_QNX";
        case 1000529002:
            return "VK_STRUCTURE_TYPE_IMPORT_SCREEN_BUFFER_INFO_QNX";
        case 1000529003:
            return "VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_QNX";
        case 1000529004:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_SCREEN_BUFFER_FEATURES_QNX";
        case 1000530000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_DRIVER_PROPERTIES_MSFT";
        case 1000545000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_FEATURES_KHR";
        case 1000545001:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_PROPERTIES_KHR";
        case 1000545002:
            return "VK_STRUCTURE_TYPE_BIND_MEMORY_STATUS_KHR";
        case 1000545003:
            return "VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_SETS_INFO_KHR";
        case 1000545004:
            return "VK_STRUCTURE_TYPE_PUSH_CONSTANTS_INFO_KHR";
        case 1000545005:
            return "VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_INFO_KHR";
        case 1000545006:
            return "VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_WITH_TEMPLATE_INFO_KHR";
        case 1000545007:
            return "VK_STRUCTURE_TYPE_SET_DESCRIPTOR_BUFFER_OFFSETS_INFO_EXT";
        case 1000545008:
            return "VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_BUFFER_EMBEDDED_SAMPLERS_INFO_EXT";
        case 1000546000:
            return "VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV";
        case VK_STRUCTURE_TYPE_MAX_ENUM: return "VK_STRUCTURE_TYPE_MAX_ENUM";
        default:
            return "Unknown VkStructureType value.";
        }
    }


    const char *
    vk_SubpassContents_to_str(VkSubpassContents input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SUBPASS_CONTENTS_INLINE";
        case 1:
            return "VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS";
        case 1000451000:
            return "VK_SUBPASS_CONTENTS_INLINE_AND_SECONDARY_COMMAND_BUFFERS_EXT";
        case VK_SUBPASS_CONTENTS_MAX_ENUM: return "VK_SUBPASS_CONTENTS_MAX_ENUM";
        default:
            return "Unknown VkSubpassContents value.";
        }
    }


    const char *
    vk_SubpassMergeStatusEXT_to_str(VkSubpassMergeStatusEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SUBPASS_MERGE_STATUS_MERGED_EXT";
        case 1:
            return "VK_SUBPASS_MERGE_STATUS_DISALLOWED_EXT";
        case 2:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_SIDE_EFFECTS_EXT";
        case 3:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_SAMPLES_MISMATCH_EXT";
        case 4:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_VIEWS_MISMATCH_EXT";
        case 5:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_ALIASING_EXT";
        case 6:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_DEPENDENCIES_EXT";
        case 7:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_INCOMPATIBLE_INPUT_ATTACHMENT_EXT";
        case 8:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_TOO_MANY_ATTACHMENTS_EXT";
        case 9:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_INSUFFICIENT_STORAGE_EXT";
        case 10:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_DEPTH_STENCIL_COUNT_EXT";
        case 11:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_RESOLVE_ATTACHMENT_REUSE_EXT";
        case 12:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_SINGLE_SUBPASS_EXT";
        case 13:
            return "VK_SUBPASS_MERGE_STATUS_NOT_MERGED_UNSPECIFIED_EXT";
        case VK_SUBPASS_MERGE_STATUS_MAX_ENUM_EXT: return "VK_SUBPASS_MERGE_STATUS_MAX_ENUM_EXT";
        default:
            return "Unknown VkSubpassMergeStatusEXT value.";
        }
    }


    const char *
    vk_SystemAllocationScope_to_str(VkSystemAllocationScope input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_SYSTEM_ALLOCATION_SCOPE_COMMAND";
        case 1:
            return "VK_SYSTEM_ALLOCATION_SCOPE_OBJECT";
        case 2:
            return "VK_SYSTEM_ALLOCATION_SCOPE_CACHE";
        case 3:
            return "VK_SYSTEM_ALLOCATION_SCOPE_DEVICE";
        case 4:
            return "VK_SYSTEM_ALLOCATION_SCOPE_INSTANCE";
        case VK_SYSTEM_ALLOCATION_SCOPE_MAX_ENUM: return "VK_SYSTEM_ALLOCATION_SCOPE_MAX_ENUM";
        default:
            return "Unknown VkSystemAllocationScope value.";
        }
    }


    const char *
    vk_TessellationDomainOrigin_to_str(VkTessellationDomainOrigin input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_TESSELLATION_DOMAIN_ORIGIN_UPPER_LEFT";
        case 1:
            return "VK_TESSELLATION_DOMAIN_ORIGIN_LOWER_LEFT";
        case VK_TESSELLATION_DOMAIN_ORIGIN_MAX_ENUM: return "VK_TESSELLATION_DOMAIN_ORIGIN_MAX_ENUM";
        default:
            return "Unknown VkTessellationDomainOrigin value.";
        }
    }


    const char *
    vk_TimeDomainKHR_to_str(VkTimeDomainKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_TIME_DOMAIN_DEVICE_KHR";
        case 1:
            return "VK_TIME_DOMAIN_CLOCK_MONOTONIC_KHR";
        case 2:
            return "VK_TIME_DOMAIN_CLOCK_MONOTONIC_RAW_KHR";
        case 3:
            return "VK_TIME_DOMAIN_QUERY_PERFORMANCE_COUNTER_KHR";
        case VK_TIME_DOMAIN_MAX_ENUM_KHR: return "VK_TIME_DOMAIN_MAX_ENUM_KHR";
        default:
            return "Unknown VkTimeDomainKHR value.";
        }
    }


    const char *
    vk_ValidationCacheHeaderVersionEXT_to_str(VkValidationCacheHeaderVersionEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VALIDATION_CACHE_HEADER_VERSION_ONE_EXT";
        case VK_VALIDATION_CACHE_HEADER_VERSION_MAX_ENUM_EXT: return "VK_VALIDATION_CACHE_HEADER_VERSION_MAX_ENUM_EXT";
        default:
            return "Unknown VkValidationCacheHeaderVersionEXT value.";
        }
    }


    const char *
    vk_ValidationCheckEXT_to_str(VkValidationCheckEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VALIDATION_CHECK_ALL_EXT";
        case 1:
            return "VK_VALIDATION_CHECK_SHADERS_EXT";
        case VK_VALIDATION_CHECK_MAX_ENUM_EXT: return "VK_VALIDATION_CHECK_MAX_ENUM_EXT";
        default:
            return "Unknown VkValidationCheckEXT value.";
        }
    }


    const char *
    vk_ValidationFeatureDisableEXT_to_str(VkValidationFeatureDisableEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VALIDATION_FEATURE_DISABLE_ALL_EXT";
        case 1:
            return "VK_VALIDATION_FEATURE_DISABLE_SHADERS_EXT";
        case 2:
            return "VK_VALIDATION_FEATURE_DISABLE_THREAD_SAFETY_EXT";
        case 3:
            return "VK_VALIDATION_FEATURE_DISABLE_API_PARAMETERS_EXT";
        case 4:
            return "VK_VALIDATION_FEATURE_DISABLE_OBJECT_LIFETIMES_EXT";
        case 5:
            return "VK_VALIDATION_FEATURE_DISABLE_CORE_CHECKS_EXT";
        case 6:
            return "VK_VALIDATION_FEATURE_DISABLE_UNIQUE_HANDLES_EXT";
        case 7:
            return "VK_VALIDATION_FEATURE_DISABLE_SHADER_VALIDATION_CACHE_EXT";
        case VK_VALIDATION_FEATURE_DISABLE_MAX_ENUM_EXT: return "VK_VALIDATION_FEATURE_DISABLE_MAX_ENUM_EXT";
        default:
            return "Unknown VkValidationFeatureDisableEXT value.";
        }
    }


    const char *
    vk_ValidationFeatureEnableEXT_to_str(VkValidationFeatureEnableEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_EXT";
        case 1:
            return "VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_RESERVE_BINDING_SLOT_EXT";
        case 2:
            return "VK_VALIDATION_FEATURE_ENABLE_BEST_PRACTICES_EXT";
        case 3:
            return "VK_VALIDATION_FEATURE_ENABLE_DEBUG_PRINTF_EXT";
        case 4:
            return "VK_VALIDATION_FEATURE_ENABLE_SYNCHRONIZATION_VALIDATION_EXT";
        case VK_VALIDATION_FEATURE_ENABLE_MAX_ENUM_EXT: return "VK_VALIDATION_FEATURE_ENABLE_MAX_ENUM_EXT";
        default:
            return "Unknown VkValidationFeatureEnableEXT value.";
        }
    }


    const char *
    vk_VendorId_to_str(VkVendorId input)
    {
        switch((int64_t)input) {
        case 65537:
            return "VK_VENDOR_ID_VIV";
        case 65538:
            return "VK_VENDOR_ID_VSI";
        case 65539:
            return "VK_VENDOR_ID_KAZAN";
        case 65540:
            return "VK_VENDOR_ID_CODEPLAY";
        case 65541:
            return "VK_VENDOR_ID_MESA";
        case 65542:
            return "VK_VENDOR_ID_POCL";
        case 65543:
            return "VK_VENDOR_ID_MOBILEYE";
        case VK_VENDOR_ID_MAX_ENUM: return "VK_VENDOR_ID_MAX_ENUM";
        default:
            return "Unknown VkVendorId value.";
        }
    }


    const char *
    vk_VertexInputRate_to_str(VkVertexInputRate input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VERTEX_INPUT_RATE_VERTEX";
        case 1:
            return "VK_VERTEX_INPUT_RATE_INSTANCE";
        case VK_VERTEX_INPUT_RATE_MAX_ENUM: return "VK_VERTEX_INPUT_RATE_MAX_ENUM";
        default:
            return "Unknown VkVertexInputRate value.";
        }
    }


    const char *
    vk_VideoEncodeTuningModeKHR_to_str(VkVideoEncodeTuningModeKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_ENCODE_TUNING_MODE_DEFAULT_KHR";
        case 1:
            return "VK_VIDEO_ENCODE_TUNING_MODE_HIGH_QUALITY_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_TUNING_MODE_LOW_LATENCY_KHR";
        case 3:
            return "VK_VIDEO_ENCODE_TUNING_MODE_ULTRA_LOW_LATENCY_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_TUNING_MODE_LOSSLESS_KHR";
        case VK_VIDEO_ENCODE_TUNING_MODE_MAX_ENUM_KHR: return "VK_VIDEO_ENCODE_TUNING_MODE_MAX_ENUM_KHR";
        default:
            return "Unknown VkVideoEncodeTuningModeKHR value.";
        }
    }


    const char *
    vk_ViewportCoordinateSwizzleNV_to_str(VkViewportCoordinateSwizzleNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_POSITIVE_X_NV";
        case 1:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_NEGATIVE_X_NV";
        case 2:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_POSITIVE_Y_NV";
        case 3:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_NEGATIVE_Y_NV";
        case 4:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_POSITIVE_Z_NV";
        case 5:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_NEGATIVE_Z_NV";
        case 6:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_POSITIVE_W_NV";
        case 7:
            return "VK_VIEWPORT_COORDINATE_SWIZZLE_NEGATIVE_W_NV";
        case VK_VIEWPORT_COORDINATE_SWIZZLE_MAX_ENUM_NV: return "VK_VIEWPORT_COORDINATE_SWIZZLE_MAX_ENUM_NV";
        default:
            return "Unknown VkViewportCoordinateSwizzleNV value.";
        }
    }



    const char *
    vk_AccelerationStructureCreateFlagBitsKHR_to_str(VkAccelerationStructureCreateFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_ACCELERATION_STRUCTURE_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_KHR";
        case 4:
            return "VK_ACCELERATION_STRUCTURE_CREATE_MOTION_BIT_NV";
        case 8:
            return "VK_ACCELERATION_STRUCTURE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT";
        default:
            return "Unknown VkAccelerationStructureCreateFlagBitsKHR value.";
        }
    }


    const char *
    vk_AccessFlagBits_to_str(VkAccessFlagBits input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCESS_NONE";
        case 1:
            return "VK_ACCESS_INDIRECT_COMMAND_READ_BIT";
        case 2:
            return "VK_ACCESS_INDEX_READ_BIT";
        case 4:
            return "VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT";
        case 8:
            return "VK_ACCESS_UNIFORM_READ_BIT";
        case 16:
            return "VK_ACCESS_INPUT_ATTACHMENT_READ_BIT";
        case 32:
            return "VK_ACCESS_SHADER_READ_BIT";
        case 64:
            return "VK_ACCESS_SHADER_WRITE_BIT";
        case 128:
            return "VK_ACCESS_COLOR_ATTACHMENT_READ_BIT";
        case 256:
            return "VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT";
        case 512:
            return "VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT";
        case 1024:
            return "VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT";
        case 2048:
            return "VK_ACCESS_TRANSFER_READ_BIT";
        case 4096:
            return "VK_ACCESS_TRANSFER_WRITE_BIT";
        case 8192:
            return "VK_ACCESS_HOST_READ_BIT";
        case 16384:
            return "VK_ACCESS_HOST_WRITE_BIT";
        case 32768:
            return "VK_ACCESS_MEMORY_READ_BIT";
        case 65536:
            return "VK_ACCESS_MEMORY_WRITE_BIT";
        case 131072:
            return "VK_ACCESS_COMMAND_PREPROCESS_READ_BIT_NV";
        case 262144:
            return "VK_ACCESS_COMMAND_PREPROCESS_WRITE_BIT_NV";
        case 524288:
            return "VK_ACCESS_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT";
        case 1048576:
            return "VK_ACCESS_CONDITIONAL_RENDERING_READ_BIT_EXT";
        case 2097152:
            return "VK_ACCESS_ACCELERATION_STRUCTURE_READ_BIT_NV";
        case 4194304:
            return "VK_ACCESS_ACCELERATION_STRUCTURE_WRITE_BIT_NV";
        case 8388608:
            return "VK_ACCESS_SHADING_RATE_IMAGE_READ_BIT_NV";
        case 16777216:
            return "VK_ACCESS_FRAGMENT_DENSITY_MAP_READ_BIT_EXT";
        case 33554432:
            return "VK_ACCESS_TRANSFORM_FEEDBACK_WRITE_BIT_EXT";
        case 67108864:
            return "VK_ACCESS_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT";
        case 134217728:
            return "VK_ACCESS_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT";
        default:
            return "Unknown VkAccessFlagBits value.";
        }
    }


    const char *
    vk_AccessFlagBits2_to_str(VkAccessFlagBits2 input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_ACCESS_2_NONE";
        case 1:
            return "VK_ACCESS_2_INDIRECT_COMMAND_READ_BIT";
        case 2:
            return "VK_ACCESS_2_INDEX_READ_BIT";
        case 4:
            return "VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT";
        case 8:
            return "VK_ACCESS_2_UNIFORM_READ_BIT";
        case 16:
            return "VK_ACCESS_2_INPUT_ATTACHMENT_READ_BIT";
        case 32:
            return "VK_ACCESS_2_SHADER_READ_BIT";
        case 64:
            return "VK_ACCESS_2_SHADER_WRITE_BIT";
        case 128:
            return "VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT";
        case 256:
            return "VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT";
        case 512:
            return "VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_READ_BIT";
        case 1024:
            return "VK_ACCESS_2_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT";
        case 2048:
            return "VK_ACCESS_2_TRANSFER_READ_BIT";
        case 4096:
            return "VK_ACCESS_2_TRANSFER_WRITE_BIT";
        case 8192:
            return "VK_ACCESS_2_HOST_READ_BIT";
        case 16384:
            return "VK_ACCESS_2_HOST_WRITE_BIT";
        case 32768:
            return "VK_ACCESS_2_MEMORY_READ_BIT";
        case 65536:
            return "VK_ACCESS_2_MEMORY_WRITE_BIT";
        case 131072:
            return "VK_ACCESS_2_COMMAND_PREPROCESS_READ_BIT_NV";
        case 262144:
            return "VK_ACCESS_2_COMMAND_PREPROCESS_WRITE_BIT_NV";
        case 524288:
            return "VK_ACCESS_2_COLOR_ATTACHMENT_READ_NONCOHERENT_BIT_EXT";
        case 1048576:
            return "VK_ACCESS_2_CONDITIONAL_RENDERING_READ_BIT_EXT";
        case 2097152:
            return "VK_ACCESS_2_ACCELERATION_STRUCTURE_READ_BIT_NV";
        case 4194304:
            return "VK_ACCESS_2_ACCELERATION_STRUCTURE_WRITE_BIT_NV";
        case 8388608:
            return "VK_ACCESS_2_SHADING_RATE_IMAGE_READ_BIT_NV";
        case 16777216:
            return "VK_ACCESS_2_FRAGMENT_DENSITY_MAP_READ_BIT_EXT";
        case 33554432:
            return "VK_ACCESS_2_TRANSFORM_FEEDBACK_WRITE_BIT_EXT";
        case 67108864:
            return "VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_READ_BIT_EXT";
        case 134217728:
            return "VK_ACCESS_2_TRANSFORM_FEEDBACK_COUNTER_WRITE_BIT_EXT";
        case 4294967296:
            return "VK_ACCESS_2_SHADER_SAMPLED_READ_BIT";
        case 8589934592:
            return "VK_ACCESS_2_SHADER_STORAGE_READ_BIT";
        case 17179869184:
            return "VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT";
        case 34359738368:
            return "VK_ACCESS_2_VIDEO_DECODE_READ_BIT_KHR";
        case 68719476736:
            return "VK_ACCESS_2_VIDEO_DECODE_WRITE_BIT_KHR";
        case 137438953472:
            return "VK_ACCESS_2_VIDEO_ENCODE_READ_BIT_KHR";
        case 274877906944:
            return "VK_ACCESS_2_VIDEO_ENCODE_WRITE_BIT_KHR";
        case 549755813888:
            return "VK_ACCESS_2_INVOCATION_MASK_READ_BIT_HUAWEI";
        case 1099511627776:
            return "VK_ACCESS_2_SHADER_BINDING_TABLE_READ_BIT_KHR";
        case 2199023255552:
            return "VK_ACCESS_2_DESCRIPTOR_BUFFER_READ_BIT_EXT";
        case 4398046511104:
            return "VK_ACCESS_2_OPTICAL_FLOW_READ_BIT_NV";
        case 8796093022208:
            return "VK_ACCESS_2_OPTICAL_FLOW_WRITE_BIT_NV";
        case 17592186044416:
            return "VK_ACCESS_2_MICROMAP_READ_BIT_EXT";
        case 35184372088832:
            return "VK_ACCESS_2_MICROMAP_WRITE_BIT_EXT";
        default:
            return "Unknown VkAccessFlagBits2 value.";
        }
    }


    const char *
    vk_AcquireProfilingLockFlagBitsKHR_to_str(VkAcquireProfilingLockFlagBitsKHR input)
    {
        switch((int64_t)input) {
        default:
            return "Unknown VkAcquireProfilingLockFlagBitsKHR value.";
        }
    }


    const char *
    vk_AttachmentDescriptionFlagBits_to_str(VkAttachmentDescriptionFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_ATTACHMENT_DESCRIPTION_MAY_ALIAS_BIT";
        default:
            return "Unknown VkAttachmentDescriptionFlagBits value.";
        }
    }


    const char *
    vk_BufferCreateFlagBits_to_str(VkBufferCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_BUFFER_CREATE_SPARSE_BINDING_BIT";
        case 2:
            return "VK_BUFFER_CREATE_SPARSE_RESIDENCY_BIT";
        case 4:
            return "VK_BUFFER_CREATE_SPARSE_ALIASED_BIT";
        case 8:
            return "VK_BUFFER_CREATE_PROTECTED_BIT";
        case 16:
            return "VK_BUFFER_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT";
        case 32:
            return "VK_BUFFER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT";
        case 64:
            return "VK_BUFFER_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR";
        default:
            return "Unknown VkBufferCreateFlagBits value.";
        }
    }


    const char *
    vk_BufferUsageFlagBits_to_str(VkBufferUsageFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_BUFFER_USAGE_TRANSFER_SRC_BIT";
        case 2:
            return "VK_BUFFER_USAGE_TRANSFER_DST_BIT";
        case 4:
            return "VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT";
        case 8:
            return "VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT";
        case 16:
            return "VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT";
        case 32:
            return "VK_BUFFER_USAGE_STORAGE_BUFFER_BIT";
        case 64:
            return "VK_BUFFER_USAGE_INDEX_BUFFER_BIT";
        case 128:
            return "VK_BUFFER_USAGE_VERTEX_BUFFER_BIT";
        case 256:
            return "VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT";
        case 512:
            return "VK_BUFFER_USAGE_CONDITIONAL_RENDERING_BIT_EXT";
        case 1024:
            return "VK_BUFFER_USAGE_RAY_TRACING_BIT_NV";
        case 2048:
            return "VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT";
        case 4096:
            return "VK_BUFFER_USAGE_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT";
        case 8192:
            return "VK_BUFFER_USAGE_VIDEO_DECODE_SRC_BIT_KHR";
        case 16384:
            return "VK_BUFFER_USAGE_VIDEO_DECODE_DST_BIT_KHR";
        case 32768:
            return "VK_BUFFER_USAGE_VIDEO_ENCODE_DST_BIT_KHR";
        case 65536:
            return "VK_BUFFER_USAGE_VIDEO_ENCODE_SRC_BIT_KHR";
        case 131072:
            return "VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT";
        case 524288:
            return "VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR";
        case 1048576:
            return "VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR";
        case 2097152:
            return "VK_BUFFER_USAGE_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT";
        case 4194304:
            return "VK_BUFFER_USAGE_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT";
        case 8388608:
            return "VK_BUFFER_USAGE_MICROMAP_BUILD_INPUT_READ_ONLY_BIT_EXT";
        case 16777216:
            return "VK_BUFFER_USAGE_MICROMAP_STORAGE_BIT_EXT";
        case 33554432:
            return "VK_BUFFER_USAGE_EXECUTION_GRAPH_SCRATCH_BIT_AMDX";
        case 67108864:
            return "VK_BUFFER_USAGE_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT";
        default:
            return "Unknown VkBufferUsageFlagBits value.";
        }
    }


    const char *
    vk_BufferUsageFlagBits2KHR_to_str(VkBufferUsageFlagBits2KHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_BUFFER_USAGE_2_TRANSFER_SRC_BIT_KHR";
        case 2:
            return "VK_BUFFER_USAGE_2_TRANSFER_DST_BIT_KHR";
        case 4:
            return "VK_BUFFER_USAGE_2_UNIFORM_TEXEL_BUFFER_BIT_KHR";
        case 8:
            return "VK_BUFFER_USAGE_2_STORAGE_TEXEL_BUFFER_BIT_KHR";
        case 16:
            return "VK_BUFFER_USAGE_2_UNIFORM_BUFFER_BIT_KHR";
        case 32:
            return "VK_BUFFER_USAGE_2_STORAGE_BUFFER_BIT_KHR";
        case 64:
            return "VK_BUFFER_USAGE_2_INDEX_BUFFER_BIT_KHR";
        case 128:
            return "VK_BUFFER_USAGE_2_VERTEX_BUFFER_BIT_KHR";
        case 256:
            return "VK_BUFFER_USAGE_2_INDIRECT_BUFFER_BIT_KHR";
        case 512:
            return "VK_BUFFER_USAGE_2_CONDITIONAL_RENDERING_BIT_EXT";
        case 1024:
            return "VK_BUFFER_USAGE_2_RAY_TRACING_BIT_NV";
        case 2048:
            return "VK_BUFFER_USAGE_2_TRANSFORM_FEEDBACK_BUFFER_BIT_EXT";
        case 4096:
            return "VK_BUFFER_USAGE_2_TRANSFORM_FEEDBACK_COUNTER_BUFFER_BIT_EXT";
        case 8192:
            return "VK_BUFFER_USAGE_2_VIDEO_DECODE_SRC_BIT_KHR";
        case 16384:
            return "VK_BUFFER_USAGE_2_VIDEO_DECODE_DST_BIT_KHR";
        case 32768:
            return "VK_BUFFER_USAGE_2_VIDEO_ENCODE_DST_BIT_KHR";
        case 65536:
            return "VK_BUFFER_USAGE_2_VIDEO_ENCODE_SRC_BIT_KHR";
        case 131072:
            return "VK_BUFFER_USAGE_2_SHADER_DEVICE_ADDRESS_BIT_KHR";
        case 524288:
            return "VK_BUFFER_USAGE_2_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR";
        case 1048576:
            return "VK_BUFFER_USAGE_2_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR";
        case 2097152:
            return "VK_BUFFER_USAGE_2_SAMPLER_DESCRIPTOR_BUFFER_BIT_EXT";
        case 4194304:
            return "VK_BUFFER_USAGE_2_RESOURCE_DESCRIPTOR_BUFFER_BIT_EXT";
        case 8388608:
            return "VK_BUFFER_USAGE_2_MICROMAP_BUILD_INPUT_READ_ONLY_BIT_EXT";
        case 16777216:
            return "VK_BUFFER_USAGE_2_MICROMAP_STORAGE_BIT_EXT";
        case 33554432:
            return "VK_BUFFER_USAGE_2_EXECUTION_GRAPH_SCRATCH_BIT_AMDX";
        case 67108864:
            return "VK_BUFFER_USAGE_2_PUSH_DESCRIPTORS_DESCRIPTOR_BUFFER_BIT_EXT";
        default:
            return "Unknown VkBufferUsageFlagBits2KHR value.";
        }
    }


    const char *
    vk_BuildAccelerationStructureFlagBitsKHR_to_str(VkBuildAccelerationStructureFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_NV";
        case 2:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_COMPACTION_BIT_NV";
        case 4:
            return "VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_TRACE_BIT_NV";
        case 8:
            return "VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_BUILD_BIT_NV";
        case 16:
            return "VK_BUILD_ACCELERATION_STRUCTURE_LOW_MEMORY_BIT_NV";
        case 32:
            return "VK_BUILD_ACCELERATION_STRUCTURE_MOTION_BIT_NV";
        case 64:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_UPDATE_EXT";
        case 128:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISABLE_OPACITY_MICROMAPS_EXT";
        case 256:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_OPACITY_MICROMAP_DATA_UPDATE_EXT";
        case 512:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DISPLACEMENT_MICROMAP_UPDATE_NV";
        case 2048:
            return "VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_DATA_ACCESS_KHR";
        default:
            return "Unknown VkBuildAccelerationStructureFlagBitsKHR value.";
        }
    }


    const char *
    vk_BuildMicromapFlagBitsEXT_to_str(VkBuildMicromapFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_BUILD_MICROMAP_PREFER_FAST_TRACE_BIT_EXT";
        case 2:
            return "VK_BUILD_MICROMAP_PREFER_FAST_BUILD_BIT_EXT";
        case 4:
            return "VK_BUILD_MICROMAP_ALLOW_COMPACTION_BIT_EXT";
        default:
            return "Unknown VkBuildMicromapFlagBitsEXT value.";
        }
    }


    const char *
    vk_ColorComponentFlagBits_to_str(VkColorComponentFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COLOR_COMPONENT_R_BIT";
        case 2:
            return "VK_COLOR_COMPONENT_G_BIT";
        case 4:
            return "VK_COLOR_COMPONENT_B_BIT";
        case 8:
            return "VK_COLOR_COMPONENT_A_BIT";
        default:
            return "Unknown VkColorComponentFlagBits value.";
        }
    }


    const char *
    vk_CommandBufferResetFlagBits_to_str(VkCommandBufferResetFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COMMAND_BUFFER_RESET_RELEASE_RESOURCES_BIT";
        default:
            return "Unknown VkCommandBufferResetFlagBits value.";
        }
    }


    const char *
    vk_CommandBufferUsageFlagBits_to_str(VkCommandBufferUsageFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT";
        case 2:
            return "VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT";
        case 4:
            return "VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT";
        default:
            return "Unknown VkCommandBufferUsageFlagBits value.";
        }
    }


    const char *
    vk_CommandPoolCreateFlagBits_to_str(VkCommandPoolCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COMMAND_POOL_CREATE_TRANSIENT_BIT";
        case 2:
            return "VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT";
        case 4:
            return "VK_COMMAND_POOL_CREATE_PROTECTED_BIT";
        default:
            return "Unknown VkCommandPoolCreateFlagBits value.";
        }
    }


    const char *
    vk_CommandPoolResetFlagBits_to_str(VkCommandPoolResetFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COMMAND_POOL_RESET_RELEASE_RESOURCES_BIT";
        default:
            return "Unknown VkCommandPoolResetFlagBits value.";
        }
    }


    const char *
    vk_CompositeAlphaFlagBitsKHR_to_str(VkCompositeAlphaFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR";
        case 2:
            return "VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR";
        case 4:
            return "VK_COMPOSITE_ALPHA_POST_MULTIPLIED_BIT_KHR";
        case 8:
            return "VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR";
        default:
            return "Unknown VkCompositeAlphaFlagBitsKHR value.";
        }
    }


    const char *
    vk_ConditionalRenderingFlagBitsEXT_to_str(VkConditionalRenderingFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_CONDITIONAL_RENDERING_INVERTED_BIT_EXT";
        default:
            return "Unknown VkConditionalRenderingFlagBitsEXT value.";
        }
    }


    const char *
    vk_CullModeFlagBits_to_str(VkCullModeFlagBits input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_CULL_MODE_NONE";
        case 1:
            return "VK_CULL_MODE_FRONT_BIT";
        case 2:
            return "VK_CULL_MODE_BACK_BIT";
        case 3:
            return "VK_CULL_MODE_FRONT_AND_BACK";
        default:
            return "Unknown VkCullModeFlagBits value.";
        }
    }


    const char *
    vk_DebugReportFlagBitsEXT_to_str(VkDebugReportFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEBUG_REPORT_INFORMATION_BIT_EXT";
        case 2:
            return "VK_DEBUG_REPORT_WARNING_BIT_EXT";
        case 4:
            return "VK_DEBUG_REPORT_PERFORMANCE_WARNING_BIT_EXT";
        case 8:
            return "VK_DEBUG_REPORT_ERROR_BIT_EXT";
        case 16:
            return "VK_DEBUG_REPORT_DEBUG_BIT_EXT";
        default:
            return "Unknown VkDebugReportFlagBitsEXT value.";
        }
    }


    const char *
    vk_DebugUtilsMessageSeverityFlagBitsEXT_to_str(VkDebugUtilsMessageSeverityFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT";
        case 16:
            return "VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT";
        case 256:
            return "VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT";
        case 4096:
            return "VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT";
        default:
            return "Unknown VkDebugUtilsMessageSeverityFlagBitsEXT value.";
        }
    }


    const char *
    vk_DebugUtilsMessageTypeFlagBitsEXT_to_str(VkDebugUtilsMessageTypeFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT";
        case 2:
            return "VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT";
        case 4:
            return "VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT";
        case 8:
            return "VK_DEBUG_UTILS_MESSAGE_TYPE_DEVICE_ADDRESS_BINDING_BIT_EXT";
        default:
            return "Unknown VkDebugUtilsMessageTypeFlagBitsEXT value.";
        }
    }


    const char *
    vk_DependencyFlagBits_to_str(VkDependencyFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEPENDENCY_BY_REGION_BIT";
        case 2:
            return "VK_DEPENDENCY_VIEW_LOCAL_BIT";
        case 4:
            return "VK_DEPENDENCY_DEVICE_GROUP_BIT";
        case 8:
            return "VK_DEPENDENCY_FEEDBACK_LOOP_BIT_EXT";
        default:
            return "Unknown VkDependencyFlagBits value.";
        }
    }


    const char *
    vk_DescriptorBindingFlagBits_to_str(VkDescriptorBindingFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DESCRIPTOR_BINDING_UPDATE_AFTER_BIND_BIT";
        case 2:
            return "VK_DESCRIPTOR_BINDING_UPDATE_UNUSED_WHILE_PENDING_BIT";
        case 4:
            return "VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT";
        case 8:
            return "VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT";
        default:
            return "Unknown VkDescriptorBindingFlagBits value.";
        }
    }


    const char *
    vk_DescriptorPoolCreateFlagBits_to_str(VkDescriptorPoolCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT";
        case 2:
            return "VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT";
        case 4:
            return "VK_DESCRIPTOR_POOL_CREATE_HOST_ONLY_BIT_EXT";
        case 8:
            return "VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_SETS_BIT_NV";
        case 16:
            return "VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_POOLS_BIT_NV";
        default:
            return "Unknown VkDescriptorPoolCreateFlagBits value.";
        }
    }


    const char *
    vk_DescriptorSetLayoutCreateFlagBits_to_str(VkDescriptorSetLayoutCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT_KHR";
        case 2:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_UPDATE_AFTER_BIND_POOL_BIT";
        case 4:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_HOST_ONLY_POOL_BIT_EXT";
        case 16:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_DESCRIPTOR_BUFFER_BIT_EXT";
        case 32:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_EMBEDDED_IMMUTABLE_SAMPLERS_BIT_EXT";
        case 64:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_PER_STAGE_BIT_NV";
        case 128:
            return "VK_DESCRIPTOR_SET_LAYOUT_CREATE_INDIRECT_BINDABLE_BIT_NV";
        default:
            return "Unknown VkDescriptorSetLayoutCreateFlagBits value.";
        }
    }


    const char *
    vk_DeviceAddressBindingFlagBitsEXT_to_str(VkDeviceAddressBindingFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEVICE_ADDRESS_BINDING_INTERNAL_OBJECT_BIT_EXT";
        default:
            return "Unknown VkDeviceAddressBindingFlagBitsEXT value.";
        }
    }


    const char *
    vk_DeviceDiagnosticsConfigFlagBitsNV_to_str(VkDeviceDiagnosticsConfigFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEVICE_DIAGNOSTICS_CONFIG_ENABLE_SHADER_DEBUG_INFO_BIT_NV";
        case 2:
            return "VK_DEVICE_DIAGNOSTICS_CONFIG_ENABLE_RESOURCE_TRACKING_BIT_NV";
        case 4:
            return "VK_DEVICE_DIAGNOSTICS_CONFIG_ENABLE_AUTOMATIC_CHECKPOINTS_BIT_NV";
        case 8:
            return "VK_DEVICE_DIAGNOSTICS_CONFIG_ENABLE_SHADER_ERROR_REPORTING_BIT_NV";
        default:
            return "Unknown VkDeviceDiagnosticsConfigFlagBitsNV value.";
        }
    }


    const char *
    vk_DeviceGroupPresentModeFlagBitsKHR_to_str(VkDeviceGroupPresentModeFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEVICE_GROUP_PRESENT_MODE_LOCAL_BIT_KHR";
        case 2:
            return "VK_DEVICE_GROUP_PRESENT_MODE_REMOTE_BIT_KHR";
        case 4:
            return "VK_DEVICE_GROUP_PRESENT_MODE_SUM_BIT_KHR";
        case 8:
            return "VK_DEVICE_GROUP_PRESENT_MODE_LOCAL_MULTI_DEVICE_BIT_KHR";
        default:
            return "Unknown VkDeviceGroupPresentModeFlagBitsKHR value.";
        }
    }


    const char *
    vk_DeviceQueueCreateFlagBits_to_str(VkDeviceQueueCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DEVICE_QUEUE_CREATE_PROTECTED_BIT";
        default:
            return "Unknown VkDeviceQueueCreateFlagBits value.";
        }
    }


    const char *
    vk_DisplayPlaneAlphaFlagBitsKHR_to_str(VkDisplayPlaneAlphaFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_DISPLAY_PLANE_ALPHA_OPAQUE_BIT_KHR";
        case 2:
            return "VK_DISPLAY_PLANE_ALPHA_GLOBAL_BIT_KHR";
        case 4:
            return "VK_DISPLAY_PLANE_ALPHA_PER_PIXEL_BIT_KHR";
        case 8:
            return "VK_DISPLAY_PLANE_ALPHA_PER_PIXEL_PREMULTIPLIED_BIT_KHR";
        default:
            return "Unknown VkDisplayPlaneAlphaFlagBitsKHR value.";
        }
    }


    const char *
    vk_EventCreateFlagBits_to_str(VkEventCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EVENT_CREATE_DEVICE_ONLY_BIT";
        default:
            return "Unknown VkEventCreateFlagBits value.";
        }
    }


#ifdef VK_USE_PLATFORM_METAL_EXT
    const char *
    vk_ExportMetalObjectTypeFlagBitsEXT_to_str(VkExportMetalObjectTypeFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_DEVICE_BIT_EXT";
        case 2:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_COMMAND_QUEUE_BIT_EXT";
        case 4:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_BUFFER_BIT_EXT";
        case 8:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_TEXTURE_BIT_EXT";
        case 16:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_IOSURFACE_BIT_EXT";
        case 32:
            return "VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT";
        default:
            return "Unknown VkExportMetalObjectTypeFlagBitsEXT value.";
        }
    }

#endif

    const char *
    vk_ExternalFenceFeatureFlagBits_to_str(VkExternalFenceFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_FENCE_FEATURE_EXPORTABLE_BIT";
        case 2:
            return "VK_EXTERNAL_FENCE_FEATURE_IMPORTABLE_BIT";
        default:
            return "Unknown VkExternalFenceFeatureFlagBits value.";
        }
    }


    const char *
    vk_ExternalFenceHandleTypeFlagBits_to_str(VkExternalFenceHandleTypeFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_FENCE_HANDLE_TYPE_OPAQUE_FD_BIT";
        case 2:
            return "VK_EXTERNAL_FENCE_HANDLE_TYPE_OPAQUE_WIN32_BIT";
        case 4:
            return "VK_EXTERNAL_FENCE_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT";
        case 8:
            return "VK_EXTERNAL_FENCE_HANDLE_TYPE_SYNC_FD_BIT";
        default:
            return "Unknown VkExternalFenceHandleTypeFlagBits value.";
        }
    }


    const char *
    vk_ExternalMemoryFeatureFlagBits_to_str(VkExternalMemoryFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_MEMORY_FEATURE_DEDICATED_ONLY_BIT";
        case 2:
            return "VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT";
        case 4:
            return "VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT";
        default:
            return "Unknown VkExternalMemoryFeatureFlagBits value.";
        }
    }


    const char *
    vk_ExternalMemoryFeatureFlagBitsNV_to_str(VkExternalMemoryFeatureFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_MEMORY_FEATURE_DEDICATED_ONLY_BIT_NV";
        case 2:
            return "VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT_NV";
        case 4:
            return "VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT_NV";
        default:
            return "Unknown VkExternalMemoryFeatureFlagBitsNV value.";
        }
    }


    const char *
    vk_ExternalMemoryHandleTypeFlagBits_to_str(VkExternalMemoryHandleTypeFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT";
        case 2:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT";
        case 4:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT";
        case 8:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_BIT";
        case 16:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_TEXTURE_KMT_BIT";
        case 32:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_HEAP_BIT";
        case 64:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D12_RESOURCE_BIT";
        case 128:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT";
        case 256:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_MAPPED_FOREIGN_MEMORY_BIT_EXT";
        case 512:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT";
        case 1024:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_ANDROID_HARDWARE_BUFFER_BIT_ANDROID";
        case 2048:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_ZIRCON_VMO_BIT_FUCHSIA";
        case 4096:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_RDMA_ADDRESS_BIT_NV";
        case 16384:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_SCREEN_BUFFER_BIT_QNX";
        default:
            return "Unknown VkExternalMemoryHandleTypeFlagBits value.";
        }
    }


    const char *
    vk_ExternalMemoryHandleTypeFlagBitsNV_to_str(VkExternalMemoryHandleTypeFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_BIT_NV";
        case 2:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT_NV";
        case 4:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_IMAGE_BIT_NV";
        case 8:
            return "VK_EXTERNAL_MEMORY_HANDLE_TYPE_D3D11_IMAGE_KMT_BIT_NV";
        default:
            return "Unknown VkExternalMemoryHandleTypeFlagBitsNV value.";
        }
    }


    const char *
    vk_ExternalSemaphoreFeatureFlagBits_to_str(VkExternalSemaphoreFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_SEMAPHORE_FEATURE_EXPORTABLE_BIT";
        case 2:
            return "VK_EXTERNAL_SEMAPHORE_FEATURE_IMPORTABLE_BIT";
        default:
            return "Unknown VkExternalSemaphoreFeatureFlagBits value.";
        }
    }


    const char *
    vk_ExternalSemaphoreHandleTypeFlagBits_to_str(VkExternalSemaphoreHandleTypeFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT";
        case 2:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_BIT";
        case 4:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_WIN32_KMT_BIT";
        case 8:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_D3D12_FENCE_BIT";
        case 16:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_SYNC_FD_BIT";
        case 128:
            return "VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_ZIRCON_EVENT_BIT_FUCHSIA";
        default:
            return "Unknown VkExternalSemaphoreHandleTypeFlagBits value.";
        }
    }


    const char *
    vk_FenceCreateFlagBits_to_str(VkFenceCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FENCE_CREATE_SIGNALED_BIT";
        default:
            return "Unknown VkFenceCreateFlagBits value.";
        }
    }


    const char *
    vk_FenceImportFlagBits_to_str(VkFenceImportFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FENCE_IMPORT_TEMPORARY_BIT";
        default:
            return "Unknown VkFenceImportFlagBits value.";
        }
    }


    const char *
    vk_FormatFeatureFlagBits_to_str(VkFormatFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT";
        case 2:
            return "VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT";
        case 4:
            return "VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT";
        case 8:
            return "VK_FORMAT_FEATURE_UNIFORM_TEXEL_BUFFER_BIT";
        case 16:
            return "VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_BIT";
        case 32:
            return "VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT";
        case 64:
            return "VK_FORMAT_FEATURE_VERTEX_BUFFER_BIT";
        case 128:
            return "VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT";
        case 256:
            return "VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT";
        case 512:
            return "VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT";
        case 1024:
            return "VK_FORMAT_FEATURE_BLIT_SRC_BIT";
        case 2048:
            return "VK_FORMAT_FEATURE_BLIT_DST_BIT";
        case 4096:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT";
        case 8192:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_CUBIC_BIT_EXT";
        case 16384:
            return "VK_FORMAT_FEATURE_TRANSFER_SRC_BIT";
        case 32768:
            return "VK_FORMAT_FEATURE_TRANSFER_DST_BIT";
        case 65536:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_MINMAX_BIT";
        case 131072:
            return "VK_FORMAT_FEATURE_MIDPOINT_CHROMA_SAMPLES_BIT";
        case 262144:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_LINEAR_FILTER_BIT";
        case 524288:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_SEPARATE_RECONSTRUCTION_FILTER_BIT";
        case 1048576:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_CHROMA_RECONSTRUCTION_EXPLICIT_BIT";
        case 2097152:
            return "VK_FORMAT_FEATURE_SAMPLED_IMAGE_YCBCR_CONVERSION_CHROMA_RECONSTRUCTION_EXPLICIT_FORCEABLE_BIT";
        case 4194304:
            return "VK_FORMAT_FEATURE_DISJOINT_BIT";
        case 8388608:
            return "VK_FORMAT_FEATURE_COSITED_CHROMA_SAMPLES_BIT";
        case 16777216:
            return "VK_FORMAT_FEATURE_FRAGMENT_DENSITY_MAP_BIT_EXT";
        case 33554432:
            return "VK_FORMAT_FEATURE_VIDEO_DECODE_OUTPUT_BIT_KHR";
        case 67108864:
            return "VK_FORMAT_FEATURE_VIDEO_DECODE_DPB_BIT_KHR";
        case 134217728:
            return "VK_FORMAT_FEATURE_VIDEO_ENCODE_INPUT_BIT_KHR";
        case 268435456:
            return "VK_FORMAT_FEATURE_VIDEO_ENCODE_DPB_BIT_KHR";
        case 536870912:
            return "VK_FORMAT_FEATURE_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR";
        case 1073741824:
            return "VK_FORMAT_FEATURE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR";
        default:
            return "Unknown VkFormatFeatureFlagBits value.";
        }
    }


    const char *
    vk_FormatFeatureFlagBits2_to_str(VkFormatFeatureFlagBits2 input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_BIT";
        case 2:
            return "VK_FORMAT_FEATURE_2_STORAGE_IMAGE_BIT";
        case 4:
            return "VK_FORMAT_FEATURE_2_STORAGE_IMAGE_ATOMIC_BIT";
        case 8:
            return "VK_FORMAT_FEATURE_2_UNIFORM_TEXEL_BUFFER_BIT";
        case 16:
            return "VK_FORMAT_FEATURE_2_STORAGE_TEXEL_BUFFER_BIT";
        case 32:
            return "VK_FORMAT_FEATURE_2_STORAGE_TEXEL_BUFFER_ATOMIC_BIT";
        case 64:
            return "VK_FORMAT_FEATURE_2_VERTEX_BUFFER_BIT";
        case 128:
            return "VK_FORMAT_FEATURE_2_COLOR_ATTACHMENT_BIT";
        case 256:
            return "VK_FORMAT_FEATURE_2_COLOR_ATTACHMENT_BLEND_BIT";
        case 512:
            return "VK_FORMAT_FEATURE_2_DEPTH_STENCIL_ATTACHMENT_BIT";
        case 1024:
            return "VK_FORMAT_FEATURE_2_BLIT_SRC_BIT";
        case 2048:
            return "VK_FORMAT_FEATURE_2_BLIT_DST_BIT";
        case 4096:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_FILTER_LINEAR_BIT";
        case 8192:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_FILTER_CUBIC_BIT";
        case 16384:
            return "VK_FORMAT_FEATURE_2_TRANSFER_SRC_BIT";
        case 32768:
            return "VK_FORMAT_FEATURE_2_TRANSFER_DST_BIT";
        case 65536:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_FILTER_MINMAX_BIT";
        case 131072:
            return "VK_FORMAT_FEATURE_2_MIDPOINT_CHROMA_SAMPLES_BIT";
        case 262144:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_YCBCR_CONVERSION_LINEAR_FILTER_BIT";
        case 524288:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_YCBCR_CONVERSION_SEPARATE_RECONSTRUCTION_FILTER_BIT";
        case 1048576:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_YCBCR_CONVERSION_CHROMA_RECONSTRUCTION_EXPLICIT_BIT";
        case 2097152:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_YCBCR_CONVERSION_CHROMA_RECONSTRUCTION_EXPLICIT_FORCEABLE_BIT";
        case 4194304:
            return "VK_FORMAT_FEATURE_2_DISJOINT_BIT";
        case 8388608:
            return "VK_FORMAT_FEATURE_2_COSITED_CHROMA_SAMPLES_BIT";
        case 16777216:
            return "VK_FORMAT_FEATURE_2_FRAGMENT_DENSITY_MAP_BIT_EXT";
        case 33554432:
            return "VK_FORMAT_FEATURE_2_VIDEO_DECODE_OUTPUT_BIT_KHR";
        case 67108864:
            return "VK_FORMAT_FEATURE_2_VIDEO_DECODE_DPB_BIT_KHR";
        case 134217728:
            return "VK_FORMAT_FEATURE_2_VIDEO_ENCODE_INPUT_BIT_KHR";
        case 268435456:
            return "VK_FORMAT_FEATURE_2_VIDEO_ENCODE_DPB_BIT_KHR";
        case 536870912:
            return "VK_FORMAT_FEATURE_2_ACCELERATION_STRUCTURE_VERTEX_BUFFER_BIT_KHR";
        case 1073741824:
            return "VK_FORMAT_FEATURE_2_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR";
        case 2147483648:
            return "VK_FORMAT_FEATURE_2_STORAGE_READ_WITHOUT_FORMAT_BIT";
        case 4294967296:
            return "VK_FORMAT_FEATURE_2_STORAGE_WRITE_WITHOUT_FORMAT_BIT";
        case 8589934592:
            return "VK_FORMAT_FEATURE_2_SAMPLED_IMAGE_DEPTH_COMPARISON_BIT";
        case 17179869184:
            return "VK_FORMAT_FEATURE_2_WEIGHT_IMAGE_BIT_QCOM";
        case 34359738368:
            return "VK_FORMAT_FEATURE_2_WEIGHT_SAMPLED_IMAGE_BIT_QCOM";
        case 68719476736:
            return "VK_FORMAT_FEATURE_2_BLOCK_MATCHING_BIT_QCOM";
        case 137438953472:
            return "VK_FORMAT_FEATURE_2_BOX_FILTER_SAMPLED_BIT_QCOM";
        case 274877906944:
            return "VK_FORMAT_FEATURE_2_LINEAR_COLOR_ATTACHMENT_BIT_NV";
        case 1099511627776:
            return "VK_FORMAT_FEATURE_2_OPTICAL_FLOW_IMAGE_BIT_NV";
        case 2199023255552:
            return "VK_FORMAT_FEATURE_2_OPTICAL_FLOW_VECTOR_BIT_NV";
        case 4398046511104:
            return "VK_FORMAT_FEATURE_2_OPTICAL_FLOW_COST_BIT_NV";
        case 70368744177664:
            return "VK_FORMAT_FEATURE_2_HOST_IMAGE_TRANSFER_BIT_EXT";
        default:
            return "Unknown VkFormatFeatureFlagBits2 value.";
        }
    }


    const char *
    vk_FrameBoundaryFlagBitsEXT_to_str(VkFrameBoundaryFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FRAME_BOUNDARY_FRAME_END_BIT_EXT";
        default:
            return "Unknown VkFrameBoundaryFlagBitsEXT value.";
        }
    }


    const char *
    vk_FramebufferCreateFlagBits_to_str(VkFramebufferCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_FRAMEBUFFER_CREATE_IMAGELESS_BIT";
        default:
            return "Unknown VkFramebufferCreateFlagBits value.";
        }
    }


    const char *
    vk_GeometryFlagBitsKHR_to_str(VkGeometryFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_GEOMETRY_OPAQUE_BIT_NV";
        case 2:
            return "VK_GEOMETRY_NO_DUPLICATE_ANY_HIT_INVOCATION_BIT_NV";
        default:
            return "Unknown VkGeometryFlagBitsKHR value.";
        }
    }


    const char *
    vk_GeometryInstanceFlagBitsKHR_to_str(VkGeometryInstanceFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_GEOMETRY_INSTANCE_TRIANGLE_CULL_DISABLE_BIT_NV";
        case 2:
            return "VK_GEOMETRY_INSTANCE_TRIANGLE_FLIP_FACING_BIT_KHR";
        case 4:
            return "VK_GEOMETRY_INSTANCE_FORCE_OPAQUE_BIT_NV";
        case 8:
            return "VK_GEOMETRY_INSTANCE_FORCE_NO_OPAQUE_BIT_NV";
        case 16:
            return "VK_GEOMETRY_INSTANCE_FORCE_OPACITY_MICROMAP_2_STATE_EXT";
        case 32:
            return "VK_GEOMETRY_INSTANCE_DISABLE_OPACITY_MICROMAPS_EXT";
        default:
            return "Unknown VkGeometryInstanceFlagBitsKHR value.";
        }
    }


    const char *
    vk_GraphicsPipelineLibraryFlagBitsEXT_to_str(VkGraphicsPipelineLibraryFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_GRAPHICS_PIPELINE_LIBRARY_VERTEX_INPUT_INTERFACE_BIT_EXT";
        case 2:
            return "VK_GRAPHICS_PIPELINE_LIBRARY_PRE_RASTERIZATION_SHADERS_BIT_EXT";
        case 4:
            return "VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT";
        case 8:
            return "VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_OUTPUT_INTERFACE_BIT_EXT";
        default:
            return "Unknown VkGraphicsPipelineLibraryFlagBitsEXT value.";
        }
    }


    const char *
    vk_HostImageCopyFlagBitsEXT_to_str(VkHostImageCopyFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_HOST_IMAGE_COPY_MEMCPY_EXT";
        default:
            return "Unknown VkHostImageCopyFlagBitsEXT value.";
        }
    }


    const char *
    vk_ImageAspectFlagBits_to_str(VkImageAspectFlagBits input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_ASPECT_NONE";
        case 1:
            return "VK_IMAGE_ASPECT_COLOR_BIT";
        case 2:
            return "VK_IMAGE_ASPECT_DEPTH_BIT";
        case 4:
            return "VK_IMAGE_ASPECT_STENCIL_BIT";
        case 8:
            return "VK_IMAGE_ASPECT_METADATA_BIT";
        case 16:
            return "VK_IMAGE_ASPECT_PLANE_0_BIT";
        case 32:
            return "VK_IMAGE_ASPECT_PLANE_1_BIT";
        case 64:
            return "VK_IMAGE_ASPECT_PLANE_2_BIT";
        case 128:
            return "VK_IMAGE_ASPECT_MEMORY_PLANE_0_BIT_EXT";
        case 256:
            return "VK_IMAGE_ASPECT_MEMORY_PLANE_1_BIT_EXT";
        case 512:
            return "VK_IMAGE_ASPECT_MEMORY_PLANE_2_BIT_EXT";
        case 1024:
            return "VK_IMAGE_ASPECT_MEMORY_PLANE_3_BIT_EXT";
        default:
            return "Unknown VkImageAspectFlagBits value.";
        }
    }


    const char *
    vk_ImageCompressionFixedRateFlagBitsEXT_to_str(VkImageCompressionFixedRateFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_NONE_EXT";
        case 1:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_1BPC_BIT_EXT";
        case 2:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_2BPC_BIT_EXT";
        case 4:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_3BPC_BIT_EXT";
        case 8:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_4BPC_BIT_EXT";
        case 16:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_5BPC_BIT_EXT";
        case 32:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_6BPC_BIT_EXT";
        case 64:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_7BPC_BIT_EXT";
        case 128:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_8BPC_BIT_EXT";
        case 256:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_9BPC_BIT_EXT";
        case 512:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_10BPC_BIT_EXT";
        case 1024:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_11BPC_BIT_EXT";
        case 2048:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_12BPC_BIT_EXT";
        case 4096:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_13BPC_BIT_EXT";
        case 8192:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_14BPC_BIT_EXT";
        case 16384:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_15BPC_BIT_EXT";
        case 32768:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_16BPC_BIT_EXT";
        case 65536:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_17BPC_BIT_EXT";
        case 131072:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_18BPC_BIT_EXT";
        case 262144:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_19BPC_BIT_EXT";
        case 524288:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_20BPC_BIT_EXT";
        case 1048576:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_21BPC_BIT_EXT";
        case 2097152:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_22BPC_BIT_EXT";
        case 4194304:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_23BPC_BIT_EXT";
        case 8388608:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_24BPC_BIT_EXT";
        default:
            return "Unknown VkImageCompressionFixedRateFlagBitsEXT value.";
        }
    }


    const char *
    vk_ImageCompressionFlagBitsEXT_to_str(VkImageCompressionFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_IMAGE_COMPRESSION_DEFAULT_EXT";
        case 1:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_DEFAULT_EXT";
        case 2:
            return "VK_IMAGE_COMPRESSION_FIXED_RATE_EXPLICIT_EXT";
        case 4:
            return "VK_IMAGE_COMPRESSION_DISABLED_EXT";
        default:
            return "Unknown VkImageCompressionFlagBitsEXT value.";
        }
    }


#ifdef VK_USE_PLATFORM_FUCHSIA
    const char *
    vk_ImageConstraintsInfoFlagBitsFUCHSIA_to_str(VkImageConstraintsInfoFlagBitsFUCHSIA input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_IMAGE_CONSTRAINTS_INFO_CPU_READ_RARELY_FUCHSIA";
        case 2:
            return "VK_IMAGE_CONSTRAINTS_INFO_CPU_READ_OFTEN_FUCHSIA";
        case 4:
            return "VK_IMAGE_CONSTRAINTS_INFO_CPU_WRITE_RARELY_FUCHSIA";
        case 8:
            return "VK_IMAGE_CONSTRAINTS_INFO_CPU_WRITE_OFTEN_FUCHSIA";
        case 16:
            return "VK_IMAGE_CONSTRAINTS_INFO_PROTECTED_OPTIONAL_FUCHSIA";
        default:
            return "Unknown VkImageConstraintsInfoFlagBitsFUCHSIA value.";
        }
    }

#endif

    const char *
    vk_ImageCreateFlagBits_to_str(VkImageCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_IMAGE_CREATE_SPARSE_BINDING_BIT";
        case 2:
            return "VK_IMAGE_CREATE_SPARSE_RESIDENCY_BIT";
        case 4:
            return "VK_IMAGE_CREATE_SPARSE_ALIASED_BIT";
        case 8:
            return "VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT";
        case 16:
            return "VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT";
        case 32:
            return "VK_IMAGE_CREATE_2D_ARRAY_COMPATIBLE_BIT";
        case 64:
            return "VK_IMAGE_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT";
        case 128:
            return "VK_IMAGE_CREATE_BLOCK_TEXEL_VIEW_COMPATIBLE_BIT";
        case 256:
            return "VK_IMAGE_CREATE_EXTENDED_USAGE_BIT";
        case 512:
            return "VK_IMAGE_CREATE_DISJOINT_BIT";
        case 1024:
            return "VK_IMAGE_CREATE_ALIAS_BIT";
        case 2048:
            return "VK_IMAGE_CREATE_PROTECTED_BIT";
        case 4096:
            return "VK_IMAGE_CREATE_SAMPLE_LOCATIONS_COMPATIBLE_DEPTH_BIT_EXT";
        case 8192:
            return "VK_IMAGE_CREATE_CORNER_SAMPLED_BIT_NV";
        case 16384:
            return "VK_IMAGE_CREATE_SUBSAMPLED_BIT_EXT";
        case 32768:
            return "VK_IMAGE_CREATE_FRAGMENT_DENSITY_MAP_OFFSET_BIT_QCOM";
        case 65536:
            return "VK_IMAGE_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT";
        case 131072:
            return "VK_IMAGE_CREATE_2D_VIEW_COMPATIBLE_BIT_EXT";
        case 262144:
            return "VK_IMAGE_CREATE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_BIT_EXT";
        case 1048576:
            return "VK_IMAGE_CREATE_VIDEO_PROFILE_INDEPENDENT_BIT_KHR";
        default:
            return "Unknown VkImageCreateFlagBits value.";
        }
    }


    const char *
    vk_ImageUsageFlagBits_to_str(VkImageUsageFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_IMAGE_USAGE_TRANSFER_SRC_BIT";
        case 2:
            return "VK_IMAGE_USAGE_TRANSFER_DST_BIT";
        case 4:
            return "VK_IMAGE_USAGE_SAMPLED_BIT";
        case 8:
            return "VK_IMAGE_USAGE_STORAGE_BIT";
        case 16:
            return "VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT";
        case 32:
            return "VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT";
        case 64:
            return "VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT";
        case 128:
            return "VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT";
        case 256:
            return "VK_IMAGE_USAGE_SHADING_RATE_IMAGE_BIT_NV";
        case 512:
            return "VK_IMAGE_USAGE_FRAGMENT_DENSITY_MAP_BIT_EXT";
        case 1024:
            return "VK_IMAGE_USAGE_VIDEO_DECODE_DST_BIT_KHR";
        case 2048:
            return "VK_IMAGE_USAGE_VIDEO_DECODE_SRC_BIT_KHR";
        case 4096:
            return "VK_IMAGE_USAGE_VIDEO_DECODE_DPB_BIT_KHR";
        case 8192:
            return "VK_IMAGE_USAGE_VIDEO_ENCODE_DST_BIT_KHR";
        case 16384:
            return "VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR";
        case 32768:
            return "VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR";
        case 262144:
            return "VK_IMAGE_USAGE_INVOCATION_MASK_BIT_HUAWEI";
        case 524288:
            return "VK_IMAGE_USAGE_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT";
        case 1048576:
            return "VK_IMAGE_USAGE_SAMPLE_WEIGHT_BIT_QCOM";
        case 2097152:
            return "VK_IMAGE_USAGE_SAMPLE_BLOCK_MATCH_BIT_QCOM";
        case 4194304:
            return "VK_IMAGE_USAGE_HOST_TRANSFER_BIT_EXT";
        default:
            return "Unknown VkImageUsageFlagBits value.";
        }
    }


    const char *
    vk_ImageViewCreateFlagBits_to_str(VkImageViewCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DYNAMIC_BIT_EXT";
        case 2:
            return "VK_IMAGE_VIEW_CREATE_FRAGMENT_DENSITY_MAP_DEFERRED_BIT_EXT";
        case 4:
            return "VK_IMAGE_VIEW_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT";
        default:
            return "Unknown VkImageViewCreateFlagBits value.";
        }
    }


    const char *
    vk_IndirectCommandsLayoutUsageFlagBitsNV_to_str(VkIndirectCommandsLayoutUsageFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_INDIRECT_COMMANDS_LAYOUT_USAGE_EXPLICIT_PREPROCESS_BIT_NV";
        case 2:
            return "VK_INDIRECT_COMMANDS_LAYOUT_USAGE_INDEXED_SEQUENCES_BIT_NV";
        case 4:
            return "VK_INDIRECT_COMMANDS_LAYOUT_USAGE_UNORDERED_SEQUENCES_BIT_NV";
        default:
            return "Unknown VkIndirectCommandsLayoutUsageFlagBitsNV value.";
        }
    }


    const char *
    vk_IndirectStateFlagBitsNV_to_str(VkIndirectStateFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_INDIRECT_STATE_FLAG_FRONTFACE_BIT_NV";
        default:
            return "Unknown VkIndirectStateFlagBitsNV value.";
        }
    }


    const char *
    vk_InstanceCreateFlagBits_to_str(VkInstanceCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_INSTANCE_CREATE_ENUMERATE_PORTABILITY_BIT_KHR";
        default:
            return "Unknown VkInstanceCreateFlagBits value.";
        }
    }


    const char *
    vk_MemoryAllocateFlagBits_to_str(VkMemoryAllocateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_MEMORY_ALLOCATE_DEVICE_MASK_BIT";
        case 2:
            return "VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_BIT";
        case 4:
            return "VK_MEMORY_ALLOCATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT";
        default:
            return "Unknown VkMemoryAllocateFlagBits value.";
        }
    }


    const char *
    vk_MemoryDecompressionMethodFlagBitsNV_to_str(VkMemoryDecompressionMethodFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_MEMORY_DECOMPRESSION_METHOD_GDEFLATE_1_0_BIT_NV";
        default:
            return "Unknown VkMemoryDecompressionMethodFlagBitsNV value.";
        }
    }


    const char *
    vk_MemoryHeapFlagBits_to_str(VkMemoryHeapFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_MEMORY_HEAP_DEVICE_LOCAL_BIT";
        case 2:
            return "VK_MEMORY_HEAP_MULTI_INSTANCE_BIT";
        default:
            return "Unknown VkMemoryHeapFlagBits value.";
        }
    }


    const char *
    vk_MemoryPropertyFlagBits_to_str(VkMemoryPropertyFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT";
        case 2:
            return "VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT";
        case 4:
            return "VK_MEMORY_PROPERTY_HOST_COHERENT_BIT";
        case 8:
            return "VK_MEMORY_PROPERTY_HOST_CACHED_BIT";
        case 16:
            return "VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT";
        case 32:
            return "VK_MEMORY_PROPERTY_PROTECTED_BIT";
        case 64:
            return "VK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD";
        case 128:
            return "VK_MEMORY_PROPERTY_DEVICE_UNCACHED_BIT_AMD";
        case 256:
            return "VK_MEMORY_PROPERTY_RDMA_CAPABLE_BIT_NV";
        default:
            return "Unknown VkMemoryPropertyFlagBits value.";
        }
    }


    const char *
    vk_MicromapCreateFlagBitsEXT_to_str(VkMicromapCreateFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_MICROMAP_CREATE_DEVICE_ADDRESS_CAPTURE_REPLAY_BIT_EXT";
        default:
            return "Unknown VkMicromapCreateFlagBitsEXT value.";
        }
    }


    const char *
    vk_OpticalFlowExecuteFlagBitsNV_to_str(VkOpticalFlowExecuteFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_OPTICAL_FLOW_EXECUTE_DISABLE_TEMPORAL_HINTS_BIT_NV";
        default:
            return "Unknown VkOpticalFlowExecuteFlagBitsNV value.";
        }
    }


    const char *
    vk_OpticalFlowGridSizeFlagBitsNV_to_str(VkOpticalFlowGridSizeFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OPTICAL_FLOW_GRID_SIZE_UNKNOWN_NV";
        case 1:
            return "VK_OPTICAL_FLOW_GRID_SIZE_1X1_BIT_NV";
        case 2:
            return "VK_OPTICAL_FLOW_GRID_SIZE_2X2_BIT_NV";
        case 4:
            return "VK_OPTICAL_FLOW_GRID_SIZE_4X4_BIT_NV";
        case 8:
            return "VK_OPTICAL_FLOW_GRID_SIZE_8X8_BIT_NV";
        default:
            return "Unknown VkOpticalFlowGridSizeFlagBitsNV value.";
        }
    }


    const char *
    vk_OpticalFlowSessionCreateFlagBitsNV_to_str(VkOpticalFlowSessionCreateFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_HINT_BIT_NV";
        case 2:
            return "VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_COST_BIT_NV";
        case 4:
            return "VK_OPTICAL_FLOW_SESSION_CREATE_ENABLE_GLOBAL_FLOW_BIT_NV";
        case 8:
            return "VK_OPTICAL_FLOW_SESSION_CREATE_ALLOW_REGIONS_BIT_NV";
        case 16:
            return "VK_OPTICAL_FLOW_SESSION_CREATE_BOTH_DIRECTIONS_BIT_NV";
        default:
            return "Unknown VkOpticalFlowSessionCreateFlagBitsNV value.";
        }
    }


    const char *
    vk_OpticalFlowUsageFlagBitsNV_to_str(VkOpticalFlowUsageFlagBitsNV input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_OPTICAL_FLOW_USAGE_UNKNOWN_NV";
        case 1:
            return "VK_OPTICAL_FLOW_USAGE_INPUT_BIT_NV";
        case 2:
            return "VK_OPTICAL_FLOW_USAGE_OUTPUT_BIT_NV";
        case 4:
            return "VK_OPTICAL_FLOW_USAGE_HINT_BIT_NV";
        case 8:
            return "VK_OPTICAL_FLOW_USAGE_COST_BIT_NV";
        case 16:
            return "VK_OPTICAL_FLOW_USAGE_GLOBAL_FLOW_BIT_NV";
        default:
            return "Unknown VkOpticalFlowUsageFlagBitsNV value.";
        }
    }


    const char *
    vk_PeerMemoryFeatureFlagBits_to_str(VkPeerMemoryFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PEER_MEMORY_FEATURE_COPY_SRC_BIT";
        case 2:
            return "VK_PEER_MEMORY_FEATURE_COPY_DST_BIT";
        case 4:
            return "VK_PEER_MEMORY_FEATURE_GENERIC_SRC_BIT";
        case 8:
            return "VK_PEER_MEMORY_FEATURE_GENERIC_DST_BIT";
        default:
            return "Unknown VkPeerMemoryFeatureFlagBits value.";
        }
    }


    const char *
    vk_PerformanceCounterDescriptionFlagBitsKHR_to_str(VkPerformanceCounterDescriptionFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PERFORMANCE_COUNTER_DESCRIPTION_PERFORMANCE_IMPACTING_KHR";
        case 2:
            return "VK_PERFORMANCE_COUNTER_DESCRIPTION_CONCURRENTLY_IMPACTED_KHR";
        default:
            return "Unknown VkPerformanceCounterDescriptionFlagBitsKHR value.";
        }
    }


    const char *
    vk_PhysicalDeviceSchedulingControlsFlagBitsARM_to_str(VkPhysicalDeviceSchedulingControlsFlagBitsARM input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_SHADER_CORE_COUNT_ARM";
        default:
            return "Unknown VkPhysicalDeviceSchedulingControlsFlagBitsARM value.";
        }
    }


    const char *
    vk_PipelineCacheCreateFlagBits_to_str(VkPipelineCacheCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_CACHE_CREATE_EXTERNALLY_SYNCHRONIZED_BIT";
        default:
            return "Unknown VkPipelineCacheCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineColorBlendStateCreateFlagBits_to_str(VkPipelineColorBlendStateCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_COLOR_BLEND_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_BIT_EXT";
        default:
            return "Unknown VkPipelineColorBlendStateCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineCompilerControlFlagBitsAMD_to_str(VkPipelineCompilerControlFlagBitsAMD input)
    {
        switch((int64_t)input) {
        default:
            return "Unknown VkPipelineCompilerControlFlagBitsAMD value.";
        }
    }


    const char *
    vk_PipelineCreateFlagBits_to_str(VkPipelineCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_CREATE_DISABLE_OPTIMIZATION_BIT";
        case 2:
            return "VK_PIPELINE_CREATE_ALLOW_DERIVATIVES_BIT";
        case 4:
            return "VK_PIPELINE_CREATE_DERIVATIVE_BIT";
        case 8:
            return "VK_PIPELINE_CREATE_VIEW_INDEX_FROM_DEVICE_INDEX_BIT";
        case 16:
            return "VK_PIPELINE_CREATE_DISPATCH_BASE";
        case 32:
            return "VK_PIPELINE_CREATE_DEFER_COMPILE_BIT_NV";
        case 64:
            return "VK_PIPELINE_CREATE_CAPTURE_STATISTICS_BIT_KHR";
        case 128:
            return "VK_PIPELINE_CREATE_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR";
        case 256:
            return "VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT";
        case 512:
            return "VK_PIPELINE_CREATE_EARLY_RETURN_ON_FAILURE_BIT";
        case 1024:
            return "VK_PIPELINE_CREATE_LINK_TIME_OPTIMIZATION_BIT_EXT";
        case 2048:
            return "VK_PIPELINE_CREATE_LIBRARY_BIT_KHR";
        case 4096:
            return "VK_PIPELINE_CREATE_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR";
        case 8192:
            return "VK_PIPELINE_CREATE_RAY_TRACING_SKIP_AABBS_BIT_KHR";
        case 16384:
            return "VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR";
        case 32768:
            return "VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR";
        case 65536:
            return "VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR";
        case 131072:
            return "VK_PIPELINE_CREATE_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR";
        case 262144:
            return "VK_PIPELINE_CREATE_INDIRECT_BINDABLE_BIT_NV";
        case 524288:
            return "VK_PIPELINE_CREATE_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR";
        case 1048576:
            return "VK_PIPELINE_CREATE_RAY_TRACING_ALLOW_MOTION_BIT_NV";
        case 2097152:
            return "VK_PIPELINE_CREATE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR";
        case 4194304:
            return "VK_PIPELINE_CREATE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT";
        case 8388608:
            return "VK_PIPELINE_CREATE_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT";
        case 16777216:
            return "VK_PIPELINE_CREATE_RAY_TRACING_OPACITY_MICROMAP_BIT_EXT";
        case 33554432:
            return "VK_PIPELINE_CREATE_COLOR_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT";
        case 67108864:
            return "VK_PIPELINE_CREATE_DEPTH_STENCIL_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT";
        case 134217728:
            return "VK_PIPELINE_CREATE_NO_PROTECTED_ACCESS_BIT_EXT";
        case 268435456:
            return "VK_PIPELINE_CREATE_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV";
        case 536870912:
            return "VK_PIPELINE_CREATE_DESCRIPTOR_BUFFER_BIT_EXT";
        case 1073741824:
            return "VK_PIPELINE_CREATE_PROTECTED_ACCESS_ONLY_BIT_EXT";
        default:
            return "Unknown VkPipelineCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineCreateFlagBits2KHR_to_str(VkPipelineCreateFlagBits2KHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_CREATE_2_DISABLE_OPTIMIZATION_BIT_KHR";
        case 2:
            return "VK_PIPELINE_CREATE_2_ALLOW_DERIVATIVES_BIT_KHR";
        case 4:
            return "VK_PIPELINE_CREATE_2_DERIVATIVE_BIT_KHR";
        case 8:
            return "VK_PIPELINE_CREATE_2_VIEW_INDEX_FROM_DEVICE_INDEX_BIT_KHR";
        case 16:
            return "VK_PIPELINE_CREATE_2_DISPATCH_BASE_BIT_KHR";
        case 32:
            return "VK_PIPELINE_CREATE_2_DEFER_COMPILE_BIT_NV";
        case 64:
            return "VK_PIPELINE_CREATE_2_CAPTURE_STATISTICS_BIT_KHR";
        case 128:
            return "VK_PIPELINE_CREATE_2_CAPTURE_INTERNAL_REPRESENTATIONS_BIT_KHR";
        case 256:
            return "VK_PIPELINE_CREATE_2_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT_KHR";
        case 512:
            return "VK_PIPELINE_CREATE_2_EARLY_RETURN_ON_FAILURE_BIT_KHR";
        case 1024:
            return "VK_PIPELINE_CREATE_2_LINK_TIME_OPTIMIZATION_BIT_EXT";
        case 2048:
            return "VK_PIPELINE_CREATE_2_LIBRARY_BIT_KHR";
        case 4096:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_SKIP_TRIANGLES_BIT_KHR";
        case 8192:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_SKIP_AABBS_BIT_KHR";
        case 16384:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_NO_NULL_ANY_HIT_SHADERS_BIT_KHR";
        case 32768:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_NO_NULL_CLOSEST_HIT_SHADERS_BIT_KHR";
        case 65536:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_NO_NULL_MISS_SHADERS_BIT_KHR";
        case 131072:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_NO_NULL_INTERSECTION_SHADERS_BIT_KHR";
        case 262144:
            return "VK_PIPELINE_CREATE_2_INDIRECT_BINDABLE_BIT_NV";
        case 524288:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_SHADER_GROUP_HANDLE_CAPTURE_REPLAY_BIT_KHR";
        case 1048576:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_ALLOW_MOTION_BIT_NV";
        case 2097152:
            return "VK_PIPELINE_CREATE_2_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_KHR";
        case 4194304:
            return "VK_PIPELINE_CREATE_2_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT";
        case 8388608:
            return "VK_PIPELINE_CREATE_2_RETAIN_LINK_TIME_OPTIMIZATION_INFO_BIT_EXT";
        case 16777216:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_OPACITY_MICROMAP_BIT_EXT";
        case 33554432:
            return "VK_PIPELINE_CREATE_2_COLOR_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT";
        case 67108864:
            return "VK_PIPELINE_CREATE_2_DEPTH_STENCIL_ATTACHMENT_FEEDBACK_LOOP_BIT_EXT";
        case 134217728:
            return "VK_PIPELINE_CREATE_2_NO_PROTECTED_ACCESS_BIT_EXT";
        case 268435456:
            return "VK_PIPELINE_CREATE_2_RAY_TRACING_DISPLACEMENT_MICROMAP_BIT_NV";
        case 536870912:
            return "VK_PIPELINE_CREATE_2_DESCRIPTOR_BUFFER_BIT_EXT";
        case 1073741824:
            return "VK_PIPELINE_CREATE_2_PROTECTED_ACCESS_ONLY_BIT_EXT";
        default:
            return "Unknown VkPipelineCreateFlagBits2KHR value.";
        }
    }


    const char *
    vk_PipelineCreationFeedbackFlagBits_to_str(VkPipelineCreationFeedbackFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_CREATION_FEEDBACK_VALID_BIT";
        case 2:
            return "VK_PIPELINE_CREATION_FEEDBACK_APPLICATION_PIPELINE_CACHE_HIT_BIT";
        case 4:
            return "VK_PIPELINE_CREATION_FEEDBACK_BASE_PIPELINE_ACCELERATION_BIT";
        default:
            return "Unknown VkPipelineCreationFeedbackFlagBits value.";
        }
    }


    const char *
    vk_PipelineDepthStencilStateCreateFlagBits_to_str(VkPipelineDepthStencilStateCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT";
        case 2:
            return "VK_PIPELINE_DEPTH_STENCIL_STATE_CREATE_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT";
        default:
            return "Unknown VkPipelineDepthStencilStateCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineLayoutCreateFlagBits_to_str(VkPipelineLayoutCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 2:
            return "VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT";
        default:
            return "Unknown VkPipelineLayoutCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineShaderStageCreateFlagBits_to_str(VkPipelineShaderStageCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PIPELINE_SHADER_STAGE_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT";
        case 2:
            return "VK_PIPELINE_SHADER_STAGE_CREATE_REQUIRE_FULL_SUBGROUPS_BIT";
        default:
            return "Unknown VkPipelineShaderStageCreateFlagBits value.";
        }
    }


    const char *
    vk_PipelineStageFlagBits_to_str(VkPipelineStageFlagBits input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_STAGE_NONE";
        case 1:
            return "VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT";
        case 2:
            return "VK_PIPELINE_STAGE_DRAW_INDIRECT_BIT";
        case 4:
            return "VK_PIPELINE_STAGE_VERTEX_INPUT_BIT";
        case 8:
            return "VK_PIPELINE_STAGE_VERTEX_SHADER_BIT";
        case 16:
            return "VK_PIPELINE_STAGE_TESSELLATION_CONTROL_SHADER_BIT";
        case 32:
            return "VK_PIPELINE_STAGE_TESSELLATION_EVALUATION_SHADER_BIT";
        case 64:
            return "VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT";
        case 128:
            return "VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT";
        case 256:
            return "VK_PIPELINE_STAGE_EARLY_FRAGMENT_TESTS_BIT";
        case 512:
            return "VK_PIPELINE_STAGE_LATE_FRAGMENT_TESTS_BIT";
        case 1024:
            return "VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT";
        case 2048:
            return "VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT";
        case 4096:
            return "VK_PIPELINE_STAGE_TRANSFER_BIT";
        case 8192:
            return "VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT";
        case 16384:
            return "VK_PIPELINE_STAGE_HOST_BIT";
        case 32768:
            return "VK_PIPELINE_STAGE_ALL_GRAPHICS_BIT";
        case 65536:
            return "VK_PIPELINE_STAGE_ALL_COMMANDS_BIT";
        case 131072:
            return "VK_PIPELINE_STAGE_COMMAND_PREPROCESS_BIT_NV";
        case 262144:
            return "VK_PIPELINE_STAGE_CONDITIONAL_RENDERING_BIT_EXT";
        case 524288:
            return "VK_PIPELINE_STAGE_TASK_SHADER_BIT_NV";
        case 1048576:
            return "VK_PIPELINE_STAGE_MESH_SHADER_BIT_NV";
        case 2097152:
            return "VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_NV";
        case 4194304:
            return "VK_PIPELINE_STAGE_SHADING_RATE_IMAGE_BIT_NV";
        case 8388608:
            return "VK_PIPELINE_STAGE_FRAGMENT_DENSITY_PROCESS_BIT_EXT";
        case 16777216:
            return "VK_PIPELINE_STAGE_TRANSFORM_FEEDBACK_BIT_EXT";
        case 33554432:
            return "VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_NV";
        default:
            return "Unknown VkPipelineStageFlagBits value.";
        }
    }


    const char *
    vk_PipelineStageFlagBits2_to_str(VkPipelineStageFlagBits2 input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_PIPELINE_STAGE_2_NONE";
        case 1:
            return "VK_PIPELINE_STAGE_2_TOP_OF_PIPE_BIT";
        case 2:
            return "VK_PIPELINE_STAGE_2_DRAW_INDIRECT_BIT";
        case 4:
            return "VK_PIPELINE_STAGE_2_VERTEX_INPUT_BIT";
        case 8:
            return "VK_PIPELINE_STAGE_2_VERTEX_SHADER_BIT";
        case 16:
            return "VK_PIPELINE_STAGE_2_TESSELLATION_CONTROL_SHADER_BIT";
        case 32:
            return "VK_PIPELINE_STAGE_2_TESSELLATION_EVALUATION_SHADER_BIT";
        case 64:
            return "VK_PIPELINE_STAGE_2_GEOMETRY_SHADER_BIT";
        case 128:
            return "VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT";
        case 256:
            return "VK_PIPELINE_STAGE_2_EARLY_FRAGMENT_TESTS_BIT";
        case 512:
            return "VK_PIPELINE_STAGE_2_LATE_FRAGMENT_TESTS_BIT";
        case 1024:
            return "VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT";
        case 2048:
            return "VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT";
        case 4096:
            return "VK_PIPELINE_STAGE_2_TRANSFER_BIT";
        case 8192:
            return "VK_PIPELINE_STAGE_2_BOTTOM_OF_PIPE_BIT";
        case 16384:
            return "VK_PIPELINE_STAGE_2_HOST_BIT";
        case 32768:
            return "VK_PIPELINE_STAGE_2_ALL_GRAPHICS_BIT";
        case 65536:
            return "VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT";
        case 131072:
            return "VK_PIPELINE_STAGE_2_COMMAND_PREPROCESS_BIT_NV";
        case 262144:
            return "VK_PIPELINE_STAGE_2_CONDITIONAL_RENDERING_BIT_EXT";
        case 524288:
            return "VK_PIPELINE_STAGE_2_TASK_SHADER_BIT_NV";
        case 1048576:
            return "VK_PIPELINE_STAGE_2_MESH_SHADER_BIT_NV";
        case 2097152:
            return "VK_PIPELINE_STAGE_2_RAY_TRACING_SHADER_BIT_NV";
        case 4194304:
            return "VK_PIPELINE_STAGE_2_SHADING_RATE_IMAGE_BIT_NV";
        case 8388608:
            return "VK_PIPELINE_STAGE_2_FRAGMENT_DENSITY_PROCESS_BIT_EXT";
        case 16777216:
            return "VK_PIPELINE_STAGE_2_TRANSFORM_FEEDBACK_BIT_EXT";
        case 33554432:
            return "VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_BUILD_BIT_NV";
        case 67108864:
            return "VK_PIPELINE_STAGE_2_VIDEO_DECODE_BIT_KHR";
        case 134217728:
            return "VK_PIPELINE_STAGE_2_VIDEO_ENCODE_BIT_KHR";
        case 268435456:
            return "VK_PIPELINE_STAGE_2_ACCELERATION_STRUCTURE_COPY_BIT_KHR";
        case 536870912:
            return "VK_PIPELINE_STAGE_2_OPTICAL_FLOW_BIT_NV";
        case 1073741824:
            return "VK_PIPELINE_STAGE_2_MICROMAP_BUILD_BIT_EXT";
        case 4294967296:
            return "VK_PIPELINE_STAGE_2_COPY_BIT";
        case 8589934592:
            return "VK_PIPELINE_STAGE_2_RESOLVE_BIT";
        case 17179869184:
            return "VK_PIPELINE_STAGE_2_BLIT_BIT";
        case 34359738368:
            return "VK_PIPELINE_STAGE_2_CLEAR_BIT";
        case 68719476736:
            return "VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT";
        case 137438953472:
            return "VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT";
        case 274877906944:
            return "VK_PIPELINE_STAGE_2_PRE_RASTERIZATION_SHADERS_BIT";
        case 549755813888:
            return "VK_PIPELINE_STAGE_2_SUBPASS_SHADER_BIT_HUAWEI";
        case 1099511627776:
            return "VK_PIPELINE_STAGE_2_INVOCATION_MASK_BIT_HUAWEI";
        case 2199023255552:
            return "VK_PIPELINE_STAGE_2_CLUSTER_CULLING_SHADER_BIT_HUAWEI";
        default:
            return "Unknown VkPipelineStageFlagBits2 value.";
        }
    }


    const char *
    vk_PresentGravityFlagBitsEXT_to_str(VkPresentGravityFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PRESENT_GRAVITY_MIN_BIT_EXT";
        case 2:
            return "VK_PRESENT_GRAVITY_MAX_BIT_EXT";
        case 4:
            return "VK_PRESENT_GRAVITY_CENTERED_BIT_EXT";
        default:
            return "Unknown VkPresentGravityFlagBitsEXT value.";
        }
    }


    const char *
    vk_PresentScalingFlagBitsEXT_to_str(VkPresentScalingFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_PRESENT_SCALING_ONE_TO_ONE_BIT_EXT";
        case 2:
            return "VK_PRESENT_SCALING_ASPECT_RATIO_STRETCH_BIT_EXT";
        case 4:
            return "VK_PRESENT_SCALING_STRETCH_BIT_EXT";
        default:
            return "Unknown VkPresentScalingFlagBitsEXT value.";
        }
    }


    const char *
    vk_QueryControlFlagBits_to_str(VkQueryControlFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_QUERY_CONTROL_PRECISE_BIT";
        default:
            return "Unknown VkQueryControlFlagBits value.";
        }
    }


    const char *
    vk_QueryPipelineStatisticFlagBits_to_str(VkQueryPipelineStatisticFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_VERTICES_BIT";
        case 2:
            return "VK_QUERY_PIPELINE_STATISTIC_INPUT_ASSEMBLY_PRIMITIVES_BIT";
        case 4:
            return "VK_QUERY_PIPELINE_STATISTIC_VERTEX_SHADER_INVOCATIONS_BIT";
        case 8:
            return "VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_INVOCATIONS_BIT";
        case 16:
            return "VK_QUERY_PIPELINE_STATISTIC_GEOMETRY_SHADER_PRIMITIVES_BIT";
        case 32:
            return "VK_QUERY_PIPELINE_STATISTIC_CLIPPING_INVOCATIONS_BIT";
        case 64:
            return "VK_QUERY_PIPELINE_STATISTIC_CLIPPING_PRIMITIVES_BIT";
        case 128:
            return "VK_QUERY_PIPELINE_STATISTIC_FRAGMENT_SHADER_INVOCATIONS_BIT";
        case 256:
            return "VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_CONTROL_SHADER_PATCHES_BIT";
        case 512:
            return "VK_QUERY_PIPELINE_STATISTIC_TESSELLATION_EVALUATION_SHADER_INVOCATIONS_BIT";
        case 1024:
            return "VK_QUERY_PIPELINE_STATISTIC_COMPUTE_SHADER_INVOCATIONS_BIT";
        case 2048:
            return "VK_QUERY_PIPELINE_STATISTIC_TASK_SHADER_INVOCATIONS_BIT_EXT";
        case 4096:
            return "VK_QUERY_PIPELINE_STATISTIC_MESH_SHADER_INVOCATIONS_BIT_EXT";
        case 8192:
            return "VK_QUERY_PIPELINE_STATISTIC_CLUSTER_CULLING_SHADER_INVOCATIONS_BIT_HUAWEI";
        default:
            return "Unknown VkQueryPipelineStatisticFlagBits value.";
        }
    }


    const char *
    vk_QueryResultFlagBits_to_str(VkQueryResultFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_QUERY_RESULT_64_BIT";
        case 2:
            return "VK_QUERY_RESULT_WAIT_BIT";
        case 4:
            return "VK_QUERY_RESULT_WITH_AVAILABILITY_BIT";
        case 8:
            return "VK_QUERY_RESULT_PARTIAL_BIT";
        case 16:
            return "VK_QUERY_RESULT_WITH_STATUS_BIT_KHR";
        default:
            return "Unknown VkQueryResultFlagBits value.";
        }
    }


    const char *
    vk_QueueFlagBits_to_str(VkQueueFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_QUEUE_GRAPHICS_BIT";
        case 2:
            return "VK_QUEUE_COMPUTE_BIT";
        case 4:
            return "VK_QUEUE_TRANSFER_BIT";
        case 8:
            return "VK_QUEUE_SPARSE_BINDING_BIT";
        case 16:
            return "VK_QUEUE_PROTECTED_BIT";
        case 32:
            return "VK_QUEUE_VIDEO_DECODE_BIT_KHR";
        case 64:
            return "VK_QUEUE_VIDEO_ENCODE_BIT_KHR";
        case 256:
            return "VK_QUEUE_OPTICAL_FLOW_BIT_NV";
        default:
            return "Unknown VkQueueFlagBits value.";
        }
    }


    const char *
    vk_RenderPassCreateFlagBits_to_str(VkRenderPassCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 2:
            return "VK_RENDER_PASS_CREATE_TRANSFORM_BIT_QCOM";
        default:
            return "Unknown VkRenderPassCreateFlagBits value.";
        }
    }


    const char *
    vk_RenderingFlagBits_to_str(VkRenderingFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_RENDERING_CONTENTS_SECONDARY_COMMAND_BUFFERS_BIT";
        case 2:
            return "VK_RENDERING_SUSPENDING_BIT";
        case 4:
            return "VK_RENDERING_RESUMING_BIT";
        case 8:
            return "VK_RENDERING_ENABLE_LEGACY_DITHERING_BIT_EXT";
        case 16:
            return "VK_RENDERING_CONTENTS_INLINE_BIT_EXT";
        default:
            return "Unknown VkRenderingFlagBits value.";
        }
    }


    const char *
    vk_ResolveModeFlagBits_to_str(VkResolveModeFlagBits input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_RESOLVE_MODE_NONE";
        case 1:
            return "VK_RESOLVE_MODE_SAMPLE_ZERO_BIT";
        case 2:
            return "VK_RESOLVE_MODE_AVERAGE_BIT";
        case 4:
            return "VK_RESOLVE_MODE_MIN_BIT";
        case 8:
            return "VK_RESOLVE_MODE_MAX_BIT";
        case 16:
            return "VK_RESOLVE_MODE_EXTERNAL_FORMAT_DOWNSAMPLE_ANDROID";
        default:
            return "Unknown VkResolveModeFlagBits value.";
        }
    }


    const char *
    vk_SampleCountFlagBits_to_str(VkSampleCountFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SAMPLE_COUNT_1_BIT";
        case 2:
            return "VK_SAMPLE_COUNT_2_BIT";
        case 4:
            return "VK_SAMPLE_COUNT_4_BIT";
        case 8:
            return "VK_SAMPLE_COUNT_8_BIT";
        case 16:
            return "VK_SAMPLE_COUNT_16_BIT";
        case 32:
            return "VK_SAMPLE_COUNT_32_BIT";
        case 64:
            return "VK_SAMPLE_COUNT_64_BIT";
        default:
            return "Unknown VkSampleCountFlagBits value.";
        }
    }


    const char *
    vk_SamplerCreateFlagBits_to_str(VkSamplerCreateFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SAMPLER_CREATE_SUBSAMPLED_BIT_EXT";
        case 2:
            return "VK_SAMPLER_CREATE_SUBSAMPLED_COARSE_RECONSTRUCTION_BIT_EXT";
        case 4:
            return "VK_SAMPLER_CREATE_NON_SEAMLESS_CUBE_MAP_BIT_EXT";
        case 8:
            return "VK_SAMPLER_CREATE_DESCRIPTOR_BUFFER_CAPTURE_REPLAY_BIT_EXT";
        case 16:
            return "VK_SAMPLER_CREATE_IMAGE_PROCESSING_BIT_QCOM";
        default:
            return "Unknown VkSamplerCreateFlagBits value.";
        }
    }


    const char *
    vk_SemaphoreImportFlagBits_to_str(VkSemaphoreImportFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SEMAPHORE_IMPORT_TEMPORARY_BIT";
        default:
            return "Unknown VkSemaphoreImportFlagBits value.";
        }
    }


    const char *
    vk_SemaphoreWaitFlagBits_to_str(VkSemaphoreWaitFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SEMAPHORE_WAIT_ANY_BIT";
        default:
            return "Unknown VkSemaphoreWaitFlagBits value.";
        }
    }


    const char *
    vk_ShaderCorePropertiesFlagBitsAMD_to_str(VkShaderCorePropertiesFlagBitsAMD input)
    {
        switch((int64_t)input) {
        default:
            return "Unknown VkShaderCorePropertiesFlagBitsAMD value.";
        }
    }


    const char *
    vk_ShaderCreateFlagBitsEXT_to_str(VkShaderCreateFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SHADER_CREATE_LINK_STAGE_BIT_EXT";
        case 2:
            return "VK_SHADER_CREATE_ALLOW_VARYING_SUBGROUP_SIZE_BIT_EXT";
        case 4:
            return "VK_SHADER_CREATE_REQUIRE_FULL_SUBGROUPS_BIT_EXT";
        case 8:
            return "VK_SHADER_CREATE_NO_TASK_SHADER_BIT_EXT";
        case 16:
            return "VK_SHADER_CREATE_DISPATCH_BASE_BIT_EXT";
        case 32:
            return "VK_SHADER_CREATE_FRAGMENT_SHADING_RATE_ATTACHMENT_BIT_EXT";
        case 64:
            return "VK_SHADER_CREATE_FRAGMENT_DENSITY_MAP_ATTACHMENT_BIT_EXT";
        default:
            return "Unknown VkShaderCreateFlagBitsEXT value.";
        }
    }


    const char *
    vk_ShaderStageFlagBits_to_str(VkShaderStageFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SHADER_STAGE_VERTEX_BIT";
        case 2:
            return "VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT";
        case 4:
            return "VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT";
        case 8:
            return "VK_SHADER_STAGE_GEOMETRY_BIT";
        case 16:
            return "VK_SHADER_STAGE_FRAGMENT_BIT";
        case 31:
            return "VK_SHADER_STAGE_ALL_GRAPHICS";
        case 32:
            return "VK_SHADER_STAGE_COMPUTE_BIT";
        case 64:
            return "VK_SHADER_STAGE_TASK_BIT_NV";
        case 128:
            return "VK_SHADER_STAGE_MESH_BIT_NV";
        case 256:
            return "VK_SHADER_STAGE_RAYGEN_BIT_NV";
        case 512:
            return "VK_SHADER_STAGE_ANY_HIT_BIT_NV";
        case 1024:
            return "VK_SHADER_STAGE_CLOSEST_HIT_BIT_NV";
        case 2048:
            return "VK_SHADER_STAGE_MISS_BIT_NV";
        case 4096:
            return "VK_SHADER_STAGE_INTERSECTION_BIT_NV";
        case 8192:
            return "VK_SHADER_STAGE_CALLABLE_BIT_NV";
        case 16384:
            return "VK_SHADER_STAGE_SUBPASS_SHADING_BIT_HUAWEI";
        case 524288:
            return "VK_SHADER_STAGE_CLUSTER_CULLING_BIT_HUAWEI";
        case 2147483647:
            return "VK_SHADER_STAGE_ALL";
        default:
            return "Unknown VkShaderStageFlagBits value.";
        }
    }


    const char *
    vk_SparseImageFormatFlagBits_to_str(VkSparseImageFormatFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SPARSE_IMAGE_FORMAT_SINGLE_MIPTAIL_BIT";
        case 2:
            return "VK_SPARSE_IMAGE_FORMAT_ALIGNED_MIP_SIZE_BIT";
        case 4:
            return "VK_SPARSE_IMAGE_FORMAT_NONSTANDARD_BLOCK_SIZE_BIT";
        default:
            return "Unknown VkSparseImageFormatFlagBits value.";
        }
    }


    const char *
    vk_SparseMemoryBindFlagBits_to_str(VkSparseMemoryBindFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SPARSE_MEMORY_BIND_METADATA_BIT";
        default:
            return "Unknown VkSparseMemoryBindFlagBits value.";
        }
    }


    const char *
    vk_StencilFaceFlagBits_to_str(VkStencilFaceFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_STENCIL_FACE_FRONT_BIT";
        case 2:
            return "VK_STENCIL_FACE_BACK_BIT";
        case 3:
            return "VK_STENCIL_FRONT_AND_BACK";
        default:
            return "Unknown VkStencilFaceFlagBits value.";
        }
    }


    const char *
    vk_SubgroupFeatureFlagBits_to_str(VkSubgroupFeatureFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SUBGROUP_FEATURE_BASIC_BIT";
        case 2:
            return "VK_SUBGROUP_FEATURE_VOTE_BIT";
        case 4:
            return "VK_SUBGROUP_FEATURE_ARITHMETIC_BIT";
        case 8:
            return "VK_SUBGROUP_FEATURE_BALLOT_BIT";
        case 16:
            return "VK_SUBGROUP_FEATURE_SHUFFLE_BIT";
        case 32:
            return "VK_SUBGROUP_FEATURE_SHUFFLE_RELATIVE_BIT";
        case 64:
            return "VK_SUBGROUP_FEATURE_CLUSTERED_BIT";
        case 128:
            return "VK_SUBGROUP_FEATURE_QUAD_BIT";
        case 256:
            return "VK_SUBGROUP_FEATURE_PARTITIONED_BIT_NV";
        default:
            return "Unknown VkSubgroupFeatureFlagBits value.";
        }
    }


    const char *
    vk_SubmitFlagBits_to_str(VkSubmitFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SUBMIT_PROTECTED_BIT";
        default:
            return "Unknown VkSubmitFlagBits value.";
        }
    }


    const char *
    vk_SubpassDescriptionFlagBits_to_str(VkSubpassDescriptionFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SUBPASS_DESCRIPTION_PER_VIEW_ATTRIBUTES_BIT_NVX";
        case 2:
            return "VK_SUBPASS_DESCRIPTION_PER_VIEW_POSITION_X_ONLY_BIT_NVX";
        case 4:
            return "VK_SUBPASS_DESCRIPTION_FRAGMENT_REGION_BIT_QCOM";
        case 8:
            return "VK_SUBPASS_DESCRIPTION_SHADER_RESOLVE_BIT_QCOM";
        case 16:
            return "VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_COLOR_ACCESS_BIT_EXT";
        case 32:
            return "VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_DEPTH_ACCESS_BIT_EXT";
        case 64:
            return "VK_SUBPASS_DESCRIPTION_RASTERIZATION_ORDER_ATTACHMENT_STENCIL_ACCESS_BIT_EXT";
        case 128:
            return "VK_SUBPASS_DESCRIPTION_ENABLE_LEGACY_DITHERING_BIT_EXT";
        default:
            return "Unknown VkSubpassDescriptionFlagBits value.";
        }
    }


    const char *
    vk_SurfaceCounterFlagBitsEXT_to_str(VkSurfaceCounterFlagBitsEXT input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SURFACE_COUNTER_VBLANK_EXT";
        default:
            return "Unknown VkSurfaceCounterFlagBitsEXT value.";
        }
    }


    const char *
    vk_SurfaceTransformFlagBitsKHR_to_str(VkSurfaceTransformFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR";
        case 2:
            return "VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR";
        case 4:
            return "VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR";
        case 8:
            return "VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR";
        case 16:
            return "VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_BIT_KHR";
        case 32:
            return "VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_90_BIT_KHR";
        case 64:
            return "VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_180_BIT_KHR";
        case 128:
            return "VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_270_BIT_KHR";
        case 256:
            return "VK_SURFACE_TRANSFORM_INHERIT_BIT_KHR";
        default:
            return "Unknown VkSurfaceTransformFlagBitsKHR value.";
        }
    }


    const char *
    vk_SwapchainCreateFlagBitsKHR_to_str(VkSwapchainCreateFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SWAPCHAIN_CREATE_SPLIT_INSTANCE_BIND_REGIONS_BIT_KHR";
        case 2:
            return "VK_SWAPCHAIN_CREATE_PROTECTED_BIT_KHR";
        case 4:
            return "VK_SWAPCHAIN_CREATE_MUTABLE_FORMAT_BIT_KHR";
        case 8:
            return "VK_SWAPCHAIN_CREATE_DEFERRED_MEMORY_ALLOCATION_BIT_EXT";
        default:
            return "Unknown VkSwapchainCreateFlagBitsKHR value.";
        }
    }


#ifdef VK_USE_PLATFORM_ANDROID_KHR
    const char *
    vk_SwapchainImageUsageFlagBitsANDROID_to_str(VkSwapchainImageUsageFlagBitsANDROID input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_SWAPCHAIN_IMAGE_USAGE_SHARED_BIT_ANDROID";
        default:
            return "Unknown VkSwapchainImageUsageFlagBitsANDROID value.";
        }
    }

#endif

    const char *
    vk_ToolPurposeFlagBits_to_str(VkToolPurposeFlagBits input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_TOOL_PURPOSE_VALIDATION_BIT";
        case 2:
            return "VK_TOOL_PURPOSE_PROFILING_BIT";
        case 4:
            return "VK_TOOL_PURPOSE_TRACING_BIT";
        case 8:
            return "VK_TOOL_PURPOSE_ADDITIONAL_FEATURES_BIT";
        case 16:
            return "VK_TOOL_PURPOSE_MODIFYING_FEATURES_BIT";
        case 32:
            return "VK_TOOL_PURPOSE_DEBUG_REPORTING_BIT_EXT";
        case 64:
            return "VK_TOOL_PURPOSE_DEBUG_MARKERS_BIT_EXT";
        default:
            return "Unknown VkToolPurposeFlagBits value.";
        }
    }


    const char *
    vk_VideoCapabilityFlagBitsKHR_to_str(VkVideoCapabilityFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_CAPABILITY_PROTECTED_CONTENT_BIT_KHR";
        case 2:
            return "VK_VIDEO_CAPABILITY_SEPARATE_REFERENCE_IMAGES_BIT_KHR";
        default:
            return "Unknown VkVideoCapabilityFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoChromaSubsamplingFlagBitsKHR_to_str(VkVideoChromaSubsamplingFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_CHROMA_SUBSAMPLING_INVALID_KHR";
        case 1:
            return "VK_VIDEO_CHROMA_SUBSAMPLING_MONOCHROME_BIT_KHR";
        case 2:
            return "VK_VIDEO_CHROMA_SUBSAMPLING_420_BIT_KHR";
        case 4:
            return "VK_VIDEO_CHROMA_SUBSAMPLING_422_BIT_KHR";
        case 8:
            return "VK_VIDEO_CHROMA_SUBSAMPLING_444_BIT_KHR";
        default:
            return "Unknown VkVideoChromaSubsamplingFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoCodecOperationFlagBitsKHR_to_str(VkVideoCodecOperationFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_CODEC_OPERATION_NONE_KHR";
        case 1:
            return "VK_VIDEO_CODEC_OPERATION_DECODE_H264_BIT_KHR";
        case 2:
            return "VK_VIDEO_CODEC_OPERATION_DECODE_H265_BIT_KHR";
        case 65536:
            return "VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_KHR";
        case 131072:
            return "VK_VIDEO_CODEC_OPERATION_ENCODE_H265_BIT_KHR";
        default:
            return "Unknown VkVideoCodecOperationFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoCodingControlFlagBitsKHR_to_str(VkVideoCodingControlFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_CODING_CONTROL_RESET_BIT_KHR";
        case 2:
            return "VK_VIDEO_CODING_CONTROL_ENCODE_RATE_CONTROL_BIT_KHR";
        case 4:
            return "VK_VIDEO_CODING_CONTROL_ENCODE_QUALITY_LEVEL_BIT_KHR";
        default:
            return "Unknown VkVideoCodingControlFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoComponentBitDepthFlagBitsKHR_to_str(VkVideoComponentBitDepthFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_COMPONENT_BIT_DEPTH_INVALID_KHR";
        case 1:
            return "VK_VIDEO_COMPONENT_BIT_DEPTH_8_BIT_KHR";
        case 4:
            return "VK_VIDEO_COMPONENT_BIT_DEPTH_10_BIT_KHR";
        case 16:
            return "VK_VIDEO_COMPONENT_BIT_DEPTH_12_BIT_KHR";
        default:
            return "Unknown VkVideoComponentBitDepthFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoDecodeCapabilityFlagBitsKHR_to_str(VkVideoDecodeCapabilityFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_DECODE_CAPABILITY_DPB_AND_OUTPUT_COINCIDE_BIT_KHR";
        case 2:
            return "VK_VIDEO_DECODE_CAPABILITY_DPB_AND_OUTPUT_DISTINCT_BIT_KHR";
        default:
            return "Unknown VkVideoDecodeCapabilityFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoDecodeH264PictureLayoutFlagBitsKHR_to_str(VkVideoDecodeH264PictureLayoutFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_DECODE_H264_PICTURE_LAYOUT_PROGRESSIVE_KHR";
        case 1:
            return "VK_VIDEO_DECODE_H264_PICTURE_LAYOUT_INTERLACED_INTERLEAVED_LINES_BIT_KHR";
        case 2:
            return "VK_VIDEO_DECODE_H264_PICTURE_LAYOUT_INTERLACED_SEPARATE_PLANES_BIT_KHR";
        default:
            return "Unknown VkVideoDecodeH264PictureLayoutFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoDecodeUsageFlagBitsKHR_to_str(VkVideoDecodeUsageFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_DECODE_USAGE_DEFAULT_KHR";
        case 1:
            return "VK_VIDEO_DECODE_USAGE_TRANSCODING_BIT_KHR";
        case 2:
            return "VK_VIDEO_DECODE_USAGE_OFFLINE_BIT_KHR";
        case 4:
            return "VK_VIDEO_DECODE_USAGE_STREAMING_BIT_KHR";
        default:
            return "Unknown VkVideoDecodeUsageFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeCapabilityFlagBitsKHR_to_str(VkVideoEncodeCapabilityFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_CAPABILITY_PRECEDING_EXTERNALLY_ENCODED_BYTES_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_CAPABILITY_INSUFFICIENT_BITSTREAM_BUFFER_RANGE_DETECTION_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeCapabilityFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeContentFlagBitsKHR_to_str(VkVideoEncodeContentFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_ENCODE_CONTENT_DEFAULT_KHR";
        case 1:
            return "VK_VIDEO_ENCODE_CONTENT_CAMERA_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_CONTENT_DESKTOP_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_CONTENT_RENDERED_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeContentFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeFeedbackFlagBitsKHR_to_str(VkVideoEncodeFeedbackFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_FEEDBACK_BITSTREAM_BUFFER_OFFSET_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_FEEDBACK_BITSTREAM_BYTES_WRITTEN_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_FEEDBACK_BITSTREAM_HAS_OVERRIDES_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeFeedbackFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH264CapabilityFlagBitsKHR_to_str(VkVideoEncodeH264CapabilityFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_HRD_COMPLIANCE_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_PREDICTION_WEIGHT_TABLE_GENERATED_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_ROW_UNALIGNED_SLICE_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_DIFFERENT_SLICE_TYPE_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_B_FRAME_IN_L0_LIST_BIT_KHR";
        case 32:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_B_FRAME_IN_L1_LIST_BIT_KHR";
        case 64:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR";
        case 128:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_PER_SLICE_CONSTANT_QP_BIT_KHR";
        case 256:
            return "VK_VIDEO_ENCODE_H264_CAPABILITY_GENERATE_PREFIX_NALU_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH264CapabilityFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH264RateControlFlagBitsKHR_to_str(VkVideoEncodeH264RateControlFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H264_RATE_CONTROL_ATTEMPT_HRD_COMPLIANCE_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H264_RATE_CONTROL_REGULAR_GOP_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H264_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H264_RATE_CONTROL_TEMPORAL_LAYER_PATTERN_DYADIC_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH264RateControlFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH264StdFlagBitsKHR_to_str(VkVideoEncodeH264StdFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H264_STD_SEPARATE_COLOR_PLANE_FLAG_SET_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H264_STD_QPPRIME_Y_ZERO_TRANSFORM_BYPASS_FLAG_SET_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H264_STD_SCALING_MATRIX_PRESENT_FLAG_SET_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H264_STD_CHROMA_QP_INDEX_OFFSET_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H264_STD_SECOND_CHROMA_QP_INDEX_OFFSET_BIT_KHR";
        case 32:
            return "VK_VIDEO_ENCODE_H264_STD_PIC_INIT_QP_MINUS26_BIT_KHR";
        case 64:
            return "VK_VIDEO_ENCODE_H264_STD_WEIGHTED_PRED_FLAG_SET_BIT_KHR";
        case 128:
            return "VK_VIDEO_ENCODE_H264_STD_WEIGHTED_BIPRED_IDC_EXPLICIT_BIT_KHR";
        case 256:
            return "VK_VIDEO_ENCODE_H264_STD_WEIGHTED_BIPRED_IDC_IMPLICIT_BIT_KHR";
        case 512:
            return "VK_VIDEO_ENCODE_H264_STD_TRANSFORM_8X8_MODE_FLAG_SET_BIT_KHR";
        case 1024:
            return "VK_VIDEO_ENCODE_H264_STD_DIRECT_SPATIAL_MV_PRED_FLAG_UNSET_BIT_KHR";
        case 2048:
            return "VK_VIDEO_ENCODE_H264_STD_ENTROPY_CODING_MODE_FLAG_UNSET_BIT_KHR";
        case 4096:
            return "VK_VIDEO_ENCODE_H264_STD_ENTROPY_CODING_MODE_FLAG_SET_BIT_KHR";
        case 8192:
            return "VK_VIDEO_ENCODE_H264_STD_DIRECT_8X8_INFERENCE_FLAG_UNSET_BIT_KHR";
        case 16384:
            return "VK_VIDEO_ENCODE_H264_STD_CONSTRAINED_INTRA_PRED_FLAG_SET_BIT_KHR";
        case 32768:
            return "VK_VIDEO_ENCODE_H264_STD_DEBLOCKING_FILTER_DISABLED_BIT_KHR";
        case 65536:
            return "VK_VIDEO_ENCODE_H264_STD_DEBLOCKING_FILTER_ENABLED_BIT_KHR";
        case 131072:
            return "VK_VIDEO_ENCODE_H264_STD_DEBLOCKING_FILTER_PARTIAL_BIT_KHR";
        case 524288:
            return "VK_VIDEO_ENCODE_H264_STD_SLICE_QP_DELTA_BIT_KHR";
        case 1048576:
            return "VK_VIDEO_ENCODE_H264_STD_DIFFERENT_SLICE_QP_DELTA_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH264StdFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH265CapabilityFlagBitsKHR_to_str(VkVideoEncodeH265CapabilityFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_HRD_COMPLIANCE_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_PREDICTION_WEIGHT_TABLE_GENERATED_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_ROW_UNALIGNED_SLICE_SEGMENT_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_DIFFERENT_SLICE_SEGMENT_TYPE_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_B_FRAME_IN_L0_LIST_BIT_KHR";
        case 32:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_B_FRAME_IN_L1_LIST_BIT_KHR";
        case 64:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_PER_PICTURE_TYPE_MIN_MAX_QP_BIT_KHR";
        case 128:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_PER_SLICE_SEGMENT_CONSTANT_QP_BIT_KHR";
        case 256:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_MULTIPLE_TILES_PER_SLICE_SEGMENT_BIT_KHR";
        case 512:
            return "VK_VIDEO_ENCODE_H265_CAPABILITY_MULTIPLE_SLICE_SEGMENTS_PER_TILE_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH265CapabilityFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH265CtbSizeFlagBitsKHR_to_str(VkVideoEncodeH265CtbSizeFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H265_CTB_SIZE_16_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H265_CTB_SIZE_32_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H265_CTB_SIZE_64_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH265CtbSizeFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH265RateControlFlagBitsKHR_to_str(VkVideoEncodeH265RateControlFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H265_RATE_CONTROL_ATTEMPT_HRD_COMPLIANCE_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H265_RATE_CONTROL_REGULAR_GOP_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_FLAT_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H265_RATE_CONTROL_REFERENCE_PATTERN_DYADIC_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H265_RATE_CONTROL_TEMPORAL_SUB_LAYER_PATTERN_DYADIC_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH265RateControlFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH265StdFlagBitsKHR_to_str(VkVideoEncodeH265StdFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H265_STD_SEPARATE_COLOR_PLANE_FLAG_SET_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H265_STD_SAMPLE_ADAPTIVE_OFFSET_ENABLED_FLAG_SET_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H265_STD_SCALING_LIST_DATA_PRESENT_FLAG_SET_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H265_STD_PCM_ENABLED_FLAG_SET_BIT_KHR";
        case 16:
            return "VK_VIDEO_ENCODE_H265_STD_SPS_TEMPORAL_MVP_ENABLED_FLAG_SET_BIT_KHR";
        case 32:
            return "VK_VIDEO_ENCODE_H265_STD_INIT_QP_MINUS26_BIT_KHR";
        case 64:
            return "VK_VIDEO_ENCODE_H265_STD_WEIGHTED_PRED_FLAG_SET_BIT_KHR";
        case 128:
            return "VK_VIDEO_ENCODE_H265_STD_WEIGHTED_BIPRED_FLAG_SET_BIT_KHR";
        case 256:
            return "VK_VIDEO_ENCODE_H265_STD_LOG2_PARALLEL_MERGE_LEVEL_MINUS2_BIT_KHR";
        case 512:
            return "VK_VIDEO_ENCODE_H265_STD_SIGN_DATA_HIDING_ENABLED_FLAG_SET_BIT_KHR";
        case 1024:
            return "VK_VIDEO_ENCODE_H265_STD_TRANSFORM_SKIP_ENABLED_FLAG_SET_BIT_KHR";
        case 2048:
            return "VK_VIDEO_ENCODE_H265_STD_TRANSFORM_SKIP_ENABLED_FLAG_UNSET_BIT_KHR";
        case 4096:
            return "VK_VIDEO_ENCODE_H265_STD_PPS_SLICE_CHROMA_QP_OFFSETS_PRESENT_FLAG_SET_BIT_KHR";
        case 8192:
            return "VK_VIDEO_ENCODE_H265_STD_TRANSQUANT_BYPASS_ENABLED_FLAG_SET_BIT_KHR";
        case 16384:
            return "VK_VIDEO_ENCODE_H265_STD_CONSTRAINED_INTRA_PRED_FLAG_SET_BIT_KHR";
        case 32768:
            return "VK_VIDEO_ENCODE_H265_STD_ENTROPY_CODING_SYNC_ENABLED_FLAG_SET_BIT_KHR";
        case 65536:
            return "VK_VIDEO_ENCODE_H265_STD_DEBLOCKING_FILTER_OVERRIDE_ENABLED_FLAG_SET_BIT_KHR";
        case 131072:
            return "VK_VIDEO_ENCODE_H265_STD_DEPENDENT_SLICE_SEGMENTS_ENABLED_FLAG_SET_BIT_KHR";
        case 262144:
            return "VK_VIDEO_ENCODE_H265_STD_DEPENDENT_SLICE_SEGMENT_FLAG_SET_BIT_KHR";
        case 524288:
            return "VK_VIDEO_ENCODE_H265_STD_SLICE_QP_DELTA_BIT_KHR";
        case 1048576:
            return "VK_VIDEO_ENCODE_H265_STD_DIFFERENT_SLICE_QP_DELTA_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH265StdFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeH265TransformBlockSizeFlagBitsKHR_to_str(VkVideoEncodeH265TransformBlockSizeFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_ENCODE_H265_TRANSFORM_BLOCK_SIZE_4_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_H265_TRANSFORM_BLOCK_SIZE_8_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_H265_TRANSFORM_BLOCK_SIZE_16_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_H265_TRANSFORM_BLOCK_SIZE_32_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeH265TransformBlockSizeFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeRateControlModeFlagBitsKHR_to_str(VkVideoEncodeRateControlModeFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DEFAULT_KHR";
        case 1:
            return "VK_VIDEO_ENCODE_RATE_CONTROL_MODE_DISABLED_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_RATE_CONTROL_MODE_CBR_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_RATE_CONTROL_MODE_VBR_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeRateControlModeFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoEncodeUsageFlagBitsKHR_to_str(VkVideoEncodeUsageFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 0:
            return "VK_VIDEO_ENCODE_USAGE_DEFAULT_KHR";
        case 1:
            return "VK_VIDEO_ENCODE_USAGE_TRANSCODING_BIT_KHR";
        case 2:
            return "VK_VIDEO_ENCODE_USAGE_STREAMING_BIT_KHR";
        case 4:
            return "VK_VIDEO_ENCODE_USAGE_RECORDING_BIT_KHR";
        case 8:
            return "VK_VIDEO_ENCODE_USAGE_CONFERENCING_BIT_KHR";
        default:
            return "Unknown VkVideoEncodeUsageFlagBitsKHR value.";
        }
    }


    const char *
    vk_VideoSessionCreateFlagBitsKHR_to_str(VkVideoSessionCreateFlagBitsKHR input)
    {
        switch((int64_t)input) {
        case 1:
            return "VK_VIDEO_SESSION_CREATE_PROTECTED_CONTENT_BIT_KHR";
        case 2:
            return "VK_VIDEO_SESSION_CREATE_ALLOW_ENCODE_PARAMETER_OPTIMIZATIONS_BIT_KHR";
        case 4:
            return "VK_VIDEO_SESSION_CREATE_INLINE_QUERIES_BIT_KHR";
        default:
            return "Unknown VkVideoSessionCreateFlagBitsKHR value.";
        }
    }


    size_t vk_structure_type_size(const struct VkBaseInStructure *item)
    {
        switch((int)item->sType) {
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_GEOMETRY_INFO_KHR: return sizeof(VkAccelerationStructureBuildGeometryInfoKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_SIZES_INFO_KHR: return sizeof(VkAccelerationStructureBuildSizesInfoKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT: return sizeof(VkAccelerationStructureCaptureDescriptorDataInfoEXT);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_KHR: return sizeof(VkAccelerationStructureCreateInfoKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_NV: return sizeof(VkAccelerationStructureCreateInfoNV);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DEVICE_ADDRESS_INFO_KHR: return sizeof(VkAccelerationStructureDeviceAddressInfoKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_AABBS_DATA_KHR: return sizeof(VkAccelerationStructureGeometryAabbsDataKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_INSTANCES_DATA_KHR: return sizeof(VkAccelerationStructureGeometryInstancesDataKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_KHR: return sizeof(VkAccelerationStructureGeometryKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MOTION_TRIANGLES_DATA_NV: return sizeof(VkAccelerationStructureGeometryMotionTrianglesDataNV);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_TRIANGLES_DATA_KHR: return sizeof(VkAccelerationStructureGeometryTrianglesDataKHR);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_INFO_NV: return sizeof(VkAccelerationStructureInfoNV);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_INFO_NV: return sizeof(VkAccelerationStructureMemoryRequirementsInfoNV);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MOTION_INFO_NV: return sizeof(VkAccelerationStructureMotionInfoNV);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_EXT: return sizeof(VkAccelerationStructureTrianglesOpacityMicromapEXT);
        case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_VERSION_INFO_KHR: return sizeof(VkAccelerationStructureVersionInfoKHR);
        case VK_STRUCTURE_TYPE_ACQUIRE_NEXT_IMAGE_INFO_KHR: return sizeof(VkAcquireNextImageInfoKHR);
        case VK_STRUCTURE_TYPE_ACQUIRE_PROFILING_LOCK_INFO_KHR: return sizeof(VkAcquireProfilingLockInfoKHR);
        case VK_STRUCTURE_TYPE_AMIGO_PROFILING_SUBMIT_INFO_SEC: return sizeof(VkAmigoProfilingSubmitInfoSEC);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_2_ANDROID: return sizeof(VkAndroidHardwareBufferFormatProperties2ANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_ANDROID: return sizeof(VkAndroidHardwareBufferFormatPropertiesANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_RESOLVE_PROPERTIES_ANDROID: return sizeof(VkAndroidHardwareBufferFormatResolvePropertiesANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_PROPERTIES_ANDROID: return sizeof(VkAndroidHardwareBufferPropertiesANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_USAGE_ANDROID: return sizeof(VkAndroidHardwareBufferUsageANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_ANDROID_SURFACE_CREATE_INFO_KHR: return sizeof(VkAndroidSurfaceCreateInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_APPLICATION_INFO: return sizeof(VkApplicationInfo);
        case VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_2: return sizeof(VkAttachmentDescription2);
        case VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_STENCIL_LAYOUT: return sizeof(VkAttachmentDescriptionStencilLayout);
        case VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_2: return sizeof(VkAttachmentReference2);
        case VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_STENCIL_LAYOUT: return sizeof(VkAttachmentReferenceStencilLayout);
        case VK_STRUCTURE_TYPE_ATTACHMENT_SAMPLE_COUNT_INFO_AMD: return sizeof(VkAttachmentSampleCountInfoAMD);
        case VK_STRUCTURE_TYPE_BIND_ACCELERATION_STRUCTURE_MEMORY_INFO_NV: return sizeof(VkBindAccelerationStructureMemoryInfoNV);
        case VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_DEVICE_GROUP_INFO: return sizeof(VkBindBufferMemoryDeviceGroupInfo);
        case VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_INFO: return sizeof(VkBindBufferMemoryInfo);
        case VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_BUFFER_EMBEDDED_SAMPLERS_INFO_EXT: return sizeof(VkBindDescriptorBufferEmbeddedSamplersInfoEXT);
        case VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_SETS_INFO_KHR: return sizeof(VkBindDescriptorSetsInfoKHR);
        case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_DEVICE_GROUP_INFO: return sizeof(VkBindImageMemoryDeviceGroupInfo);
        case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO: return sizeof(VkBindImageMemoryInfo);
        case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_SWAPCHAIN_INFO_KHR: return sizeof(VkBindImageMemorySwapchainInfoKHR);
        case VK_STRUCTURE_TYPE_BIND_IMAGE_PLANE_MEMORY_INFO: return sizeof(VkBindImagePlaneMemoryInfo);
        case VK_STRUCTURE_TYPE_BIND_MEMORY_STATUS_KHR: return sizeof(VkBindMemoryStatusKHR);
        case VK_STRUCTURE_TYPE_BIND_SPARSE_INFO: return sizeof(VkBindSparseInfo);
        case VK_STRUCTURE_TYPE_BIND_VIDEO_SESSION_MEMORY_INFO_KHR: return sizeof(VkBindVideoSessionMemoryInfoKHR);
        case VK_STRUCTURE_TYPE_BLIT_IMAGE_CUBIC_WEIGHTS_INFO_QCOM: return sizeof(VkBlitImageCubicWeightsInfoQCOM);
        case VK_STRUCTURE_TYPE_BLIT_IMAGE_INFO_2: return sizeof(VkBlitImageInfo2);
        case VK_STRUCTURE_TYPE_BUFFER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT: return sizeof(VkBufferCaptureDescriptorDataInfoEXT);
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_BUFFER_CREATE_INFO_FUCHSIA: return sizeof(VkBufferCollectionBufferCreateInfoFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CONSTRAINTS_INFO_FUCHSIA: return sizeof(VkBufferCollectionConstraintsInfoFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CREATE_INFO_FUCHSIA: return sizeof(VkBufferCollectionCreateInfoFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_IMAGE_CREATE_INFO_FUCHSIA: return sizeof(VkBufferCollectionImageCreateInfoFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_PROPERTIES_FUCHSIA: return sizeof(VkBufferCollectionPropertiesFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_BUFFER_CONSTRAINTS_INFO_FUCHSIA: return sizeof(VkBufferConstraintsInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_BUFFER_COPY_2: return sizeof(VkBufferCopy2);
        case VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO: return sizeof(VkBufferCreateInfo);
        case VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_CREATE_INFO_EXT: return sizeof(VkBufferDeviceAddressCreateInfoEXT);
        case VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_INFO: return sizeof(VkBufferDeviceAddressInfo);
        case VK_STRUCTURE_TYPE_BUFFER_IMAGE_COPY_2: return sizeof(VkBufferImageCopy2);
        case VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER: return sizeof(VkBufferMemoryBarrier);
        case VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2: return sizeof(VkBufferMemoryBarrier2);
        case VK_STRUCTURE_TYPE_BUFFER_MEMORY_REQUIREMENTS_INFO_2: return sizeof(VkBufferMemoryRequirementsInfo2);
        case VK_STRUCTURE_TYPE_BUFFER_OPAQUE_CAPTURE_ADDRESS_CREATE_INFO: return sizeof(VkBufferOpaqueCaptureAddressCreateInfo);
        case VK_STRUCTURE_TYPE_BUFFER_USAGE_FLAGS_2_CREATE_INFO_KHR: return sizeof(VkBufferUsageFlags2CreateInfoKHR);
        case VK_STRUCTURE_TYPE_BUFFER_VIEW_CREATE_INFO: return sizeof(VkBufferViewCreateInfo);
        case VK_STRUCTURE_TYPE_CALIBRATED_TIMESTAMP_INFO_KHR: return sizeof(VkCalibratedTimestampInfoKHR);
        case VK_STRUCTURE_TYPE_CHECKPOINT_DATA_2_NV: return sizeof(VkCheckpointData2NV);
        case VK_STRUCTURE_TYPE_CHECKPOINT_DATA_NV: return sizeof(VkCheckpointDataNV);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO: return sizeof(VkCommandBufferAllocateInfo);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO: return sizeof(VkCommandBufferBeginInfo);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_CONDITIONAL_RENDERING_INFO_EXT: return sizeof(VkCommandBufferInheritanceConditionalRenderingInfoEXT);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_INFO: return sizeof(VkCommandBufferInheritanceInfo);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDER_PASS_TRANSFORM_INFO_QCOM: return sizeof(VkCommandBufferInheritanceRenderPassTransformInfoQCOM);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDERING_INFO: return sizeof(VkCommandBufferInheritanceRenderingInfo);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_VIEWPORT_SCISSOR_INFO_NV: return sizeof(VkCommandBufferInheritanceViewportScissorInfoNV);
        case VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO: return sizeof(VkCommandBufferSubmitInfo);
        case VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO: return sizeof(VkCommandPoolCreateInfo);
        case VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO: return sizeof(VkComputePipelineCreateInfo);
        case VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_INDIRECT_BUFFER_INFO_NV: return sizeof(VkComputePipelineIndirectBufferInfoNV);
        case VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_EXT: return sizeof(VkConditionalRenderingBeginInfoEXT);
        case VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_KHR: return sizeof(VkCooperativeMatrixPropertiesKHR);
        case VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_NV: return sizeof(VkCooperativeMatrixPropertiesNV);
        case VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_INFO_KHR: return sizeof(VkCopyAccelerationStructureInfoKHR);
        case VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_TO_MEMORY_INFO_KHR: return sizeof(VkCopyAccelerationStructureToMemoryInfoKHR);
        case VK_STRUCTURE_TYPE_COPY_BUFFER_INFO_2: return sizeof(VkCopyBufferInfo2);
        case VK_STRUCTURE_TYPE_COPY_BUFFER_TO_IMAGE_INFO_2: return sizeof(VkCopyBufferToImageInfo2);
        case VK_STRUCTURE_TYPE_COPY_COMMAND_TRANSFORM_INFO_QCOM: return sizeof(VkCopyCommandTransformInfoQCOM);
        case VK_STRUCTURE_TYPE_COPY_DESCRIPTOR_SET: return sizeof(VkCopyDescriptorSet);
        case VK_STRUCTURE_TYPE_COPY_IMAGE_INFO_2: return sizeof(VkCopyImageInfo2);
        case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_BUFFER_INFO_2: return sizeof(VkCopyImageToBufferInfo2);
        case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_IMAGE_INFO_EXT: return sizeof(VkCopyImageToImageInfoEXT);
        case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_MEMORY_INFO_EXT: return sizeof(VkCopyImageToMemoryInfoEXT);
        case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_ACCELERATION_STRUCTURE_INFO_KHR: return sizeof(VkCopyMemoryToAccelerationStructureInfoKHR);
        case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INFO_EXT: return sizeof(VkCopyMemoryToImageInfoEXT);
        case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_MICROMAP_INFO_EXT: return sizeof(VkCopyMemoryToMicromapInfoEXT);
        case VK_STRUCTURE_TYPE_COPY_MICROMAP_INFO_EXT: return sizeof(VkCopyMicromapInfoEXT);
        case VK_STRUCTURE_TYPE_COPY_MICROMAP_TO_MEMORY_INFO_EXT: return sizeof(VkCopyMicromapToMemoryInfoEXT);
        case VK_STRUCTURE_TYPE_CU_FUNCTION_CREATE_INFO_NVX: return sizeof(VkCuFunctionCreateInfoNVX);
        case VK_STRUCTURE_TYPE_CU_LAUNCH_INFO_NVX: return sizeof(VkCuLaunchInfoNVX);
        case VK_STRUCTURE_TYPE_CU_MODULE_CREATE_INFO_NVX: return sizeof(VkCuModuleCreateInfoNVX);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_D3D12_FENCE_SUBMIT_INFO_KHR: return sizeof(VkD3D12FenceSubmitInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_DEBUG_MARKER_MARKER_INFO_EXT: return sizeof(VkDebugMarkerMarkerInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_NAME_INFO_EXT: return sizeof(VkDebugMarkerObjectNameInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_TAG_INFO_EXT: return sizeof(VkDebugMarkerObjectTagInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_REPORT_CALLBACK_CREATE_INFO_EXT: return sizeof(VkDebugReportCallbackCreateInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT: return sizeof(VkDebugUtilsLabelEXT);
        case VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CALLBACK_DATA_EXT: return sizeof(VkDebugUtilsMessengerCallbackDataEXT);
        case VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT: return sizeof(VkDebugUtilsMessengerCreateInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT: return sizeof(VkDebugUtilsObjectNameInfoEXT);
        case VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_TAG_INFO_EXT: return sizeof(VkDebugUtilsObjectTagInfoEXT);
        case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_BUFFER_CREATE_INFO_NV: return sizeof(VkDedicatedAllocationBufferCreateInfoNV);
        case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_IMAGE_CREATE_INFO_NV: return sizeof(VkDedicatedAllocationImageCreateInfoNV);
        case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_MEMORY_ALLOCATE_INFO_NV: return sizeof(VkDedicatedAllocationMemoryAllocateInfoNV);
        case VK_STRUCTURE_TYPE_DEPENDENCY_INFO: return sizeof(VkDependencyInfo);
        case VK_STRUCTURE_TYPE_DEPTH_BIAS_INFO_EXT: return sizeof(VkDepthBiasInfoEXT);
        case VK_STRUCTURE_TYPE_DEPTH_BIAS_REPRESENTATION_INFO_EXT: return sizeof(VkDepthBiasRepresentationInfoEXT);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_ADDRESS_INFO_EXT: return sizeof(VkDescriptorAddressInfoEXT);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_INFO_EXT: return sizeof(VkDescriptorBufferBindingInfoEXT);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_PUSH_DESCRIPTOR_BUFFER_HANDLE_EXT: return sizeof(VkDescriptorBufferBindingPushDescriptorBufferHandleEXT);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_GET_INFO_EXT: return sizeof(VkDescriptorGetInfoEXT);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO: return sizeof(VkDescriptorPoolCreateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_INLINE_UNIFORM_BLOCK_CREATE_INFO: return sizeof(VkDescriptorPoolInlineUniformBlockCreateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO: return sizeof(VkDescriptorSetAllocateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_BINDING_REFERENCE_VALVE: return sizeof(VkDescriptorSetBindingReferenceVALVE);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_BINDING_FLAGS_CREATE_INFO: return sizeof(VkDescriptorSetLayoutBindingFlagsCreateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO: return sizeof(VkDescriptorSetLayoutCreateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_HOST_MAPPING_INFO_VALVE: return sizeof(VkDescriptorSetLayoutHostMappingInfoVALVE);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_SUPPORT: return sizeof(VkDescriptorSetLayoutSupport);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_ALLOCATE_INFO: return sizeof(VkDescriptorSetVariableDescriptorCountAllocateInfo);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_LAYOUT_SUPPORT: return sizeof(VkDescriptorSetVariableDescriptorCountLayoutSupport);
        case VK_STRUCTURE_TYPE_DESCRIPTOR_UPDATE_TEMPLATE_CREATE_INFO: return sizeof(VkDescriptorUpdateTemplateCreateInfo);
        case VK_STRUCTURE_TYPE_DEVICE_ADDRESS_BINDING_CALLBACK_DATA_EXT: return sizeof(VkDeviceAddressBindingCallbackDataEXT);
        case VK_STRUCTURE_TYPE_DEVICE_BUFFER_MEMORY_REQUIREMENTS: return sizeof(VkDeviceBufferMemoryRequirements);
        case VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO: return sizeof(VkDeviceCreateInfo);
        case VK_STRUCTURE_TYPE_DEVICE_DEVICE_MEMORY_REPORT_CREATE_INFO_EXT: return sizeof(VkDeviceDeviceMemoryReportCreateInfoEXT);
        case VK_STRUCTURE_TYPE_DEVICE_DIAGNOSTICS_CONFIG_CREATE_INFO_NV: return sizeof(VkDeviceDiagnosticsConfigCreateInfoNV);
        case VK_STRUCTURE_TYPE_DEVICE_EVENT_INFO_EXT: return sizeof(VkDeviceEventInfoEXT);
        case VK_STRUCTURE_TYPE_DEVICE_FAULT_COUNTS_EXT: return sizeof(VkDeviceFaultCountsEXT);
        case VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_EXT: return sizeof(VkDeviceFaultInfoEXT);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_BIND_SPARSE_INFO: return sizeof(VkDeviceGroupBindSparseInfo);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_COMMAND_BUFFER_BEGIN_INFO: return sizeof(VkDeviceGroupCommandBufferBeginInfo);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_DEVICE_CREATE_INFO: return sizeof(VkDeviceGroupDeviceCreateInfo);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_CAPABILITIES_KHR: return sizeof(VkDeviceGroupPresentCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_INFO_KHR: return sizeof(VkDeviceGroupPresentInfoKHR);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_RENDER_PASS_BEGIN_INFO: return sizeof(VkDeviceGroupRenderPassBeginInfo);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_SUBMIT_INFO: return sizeof(VkDeviceGroupSubmitInfo);
        case VK_STRUCTURE_TYPE_DEVICE_GROUP_SWAPCHAIN_CREATE_INFO_KHR: return sizeof(VkDeviceGroupSwapchainCreateInfoKHR);
        case VK_STRUCTURE_TYPE_DEVICE_IMAGE_MEMORY_REQUIREMENTS: return sizeof(VkDeviceImageMemoryRequirements);
        case VK_STRUCTURE_TYPE_DEVICE_IMAGE_SUBRESOURCE_INFO_KHR: return sizeof(VkDeviceImageSubresourceInfoKHR);
        case VK_STRUCTURE_TYPE_DEVICE_MEMORY_OPAQUE_CAPTURE_ADDRESS_INFO: return sizeof(VkDeviceMemoryOpaqueCaptureAddressInfo);
        case VK_STRUCTURE_TYPE_DEVICE_MEMORY_OVERALLOCATION_CREATE_INFO_AMD: return sizeof(VkDeviceMemoryOverallocationCreateInfoAMD);
        case VK_STRUCTURE_TYPE_DEVICE_MEMORY_REPORT_CALLBACK_DATA_EXT: return sizeof(VkDeviceMemoryReportCallbackDataEXT);
        case VK_STRUCTURE_TYPE_DEVICE_PRIVATE_DATA_CREATE_INFO: return sizeof(VkDevicePrivateDataCreateInfo);
        case VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO: return sizeof(VkDeviceQueueCreateInfo);
        case VK_STRUCTURE_TYPE_DEVICE_QUEUE_GLOBAL_PRIORITY_CREATE_INFO_KHR: return sizeof(VkDeviceQueueGlobalPriorityCreateInfoKHR);
        case VK_STRUCTURE_TYPE_DEVICE_QUEUE_INFO_2: return sizeof(VkDeviceQueueInfo2);
        case VK_STRUCTURE_TYPE_DEVICE_QUEUE_SHADER_CORE_CONTROL_CREATE_INFO_ARM: return sizeof(VkDeviceQueueShaderCoreControlCreateInfoARM);
        case VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_INFO_LUNARG: return sizeof(VkDirectDriverLoadingInfoLUNARG);
        case VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_LIST_LUNARG: return sizeof(VkDirectDriverLoadingListLUNARG);
    #ifdef VK_USE_PLATFORM_DIRECTFB_EXT
        case VK_STRUCTURE_TYPE_DIRECTFB_SURFACE_CREATE_INFO_EXT: return sizeof(VkDirectFBSurfaceCreateInfoEXT);
    #endif
        case VK_STRUCTURE_TYPE_DISPLAY_EVENT_INFO_EXT: return sizeof(VkDisplayEventInfoEXT);
        case VK_STRUCTURE_TYPE_DISPLAY_MODE_CREATE_INFO_KHR: return sizeof(VkDisplayModeCreateInfoKHR);
        case VK_STRUCTURE_TYPE_DISPLAY_MODE_PROPERTIES_2_KHR: return sizeof(VkDisplayModeProperties2KHR);
        case VK_STRUCTURE_TYPE_DISPLAY_NATIVE_HDR_SURFACE_CAPABILITIES_AMD: return sizeof(VkDisplayNativeHdrSurfaceCapabilitiesAMD);
        case VK_STRUCTURE_TYPE_DISPLAY_PLANE_CAPABILITIES_2_KHR: return sizeof(VkDisplayPlaneCapabilities2KHR);
        case VK_STRUCTURE_TYPE_DISPLAY_PLANE_INFO_2_KHR: return sizeof(VkDisplayPlaneInfo2KHR);
        case VK_STRUCTURE_TYPE_DISPLAY_PLANE_PROPERTIES_2_KHR: return sizeof(VkDisplayPlaneProperties2KHR);
        case VK_STRUCTURE_TYPE_DISPLAY_POWER_INFO_EXT: return sizeof(VkDisplayPowerInfoEXT);
        case VK_STRUCTURE_TYPE_DISPLAY_PRESENT_INFO_KHR: return sizeof(VkDisplayPresentInfoKHR);
        case VK_STRUCTURE_TYPE_DISPLAY_PROPERTIES_2_KHR: return sizeof(VkDisplayProperties2KHR);
        case VK_STRUCTURE_TYPE_DISPLAY_SURFACE_CREATE_INFO_KHR: return sizeof(VkDisplaySurfaceCreateInfoKHR);
        case VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_2_EXT: return sizeof(VkDrmFormatModifierPropertiesList2EXT);
        case VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_EXT: return sizeof(VkDrmFormatModifierPropertiesListEXT);
        case VK_STRUCTURE_TYPE_EVENT_CREATE_INFO: return sizeof(VkEventCreateInfo);
        case VK_STRUCTURE_TYPE_EXPORT_FENCE_CREATE_INFO: return sizeof(VkExportFenceCreateInfo);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_EXPORT_FENCE_WIN32_HANDLE_INFO_KHR: return sizeof(VkExportFenceWin32HandleInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO: return sizeof(VkExportMemoryAllocateInfo);
        case VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO_NV: return sizeof(VkExportMemoryAllocateInfoNV);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_KHR: return sizeof(VkExportMemoryWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_NV: return sizeof(VkExportMemoryWin32HandleInfoNV);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_BUFFER_INFO_EXT: return sizeof(VkExportMetalBufferInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_COMMAND_QUEUE_INFO_EXT: return sizeof(VkExportMetalCommandQueueInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_DEVICE_INFO_EXT: return sizeof(VkExportMetalDeviceInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_IO_SURFACE_INFO_EXT: return sizeof(VkExportMetalIOSurfaceInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECT_CREATE_INFO_EXT: return sizeof(VkExportMetalObjectCreateInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECTS_INFO_EXT: return sizeof(VkExportMetalObjectsInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_SHARED_EVENT_INFO_EXT: return sizeof(VkExportMetalSharedEventInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_EXPORT_METAL_TEXTURE_INFO_EXT: return sizeof(VkExportMetalTextureInfoEXT);
    #endif
        case VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO: return sizeof(VkExportSemaphoreCreateInfo);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR: return sizeof(VkExportSemaphoreWin32HandleInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_EXTERNAL_BUFFER_PROPERTIES: return sizeof(VkExternalBufferProperties);
        case VK_STRUCTURE_TYPE_EXTERNAL_FENCE_PROPERTIES: return sizeof(VkExternalFenceProperties);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_ANDROID: return sizeof(VkExternalFormatANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_QNX: return sizeof(VkExternalFormatQNX);
    #endif
        case VK_STRUCTURE_TYPE_EXTERNAL_IMAGE_FORMAT_PROPERTIES: return sizeof(VkExternalImageFormatProperties);
        case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_ACQUIRE_UNMODIFIED_EXT: return sizeof(VkExternalMemoryAcquireUnmodifiedEXT);
        case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_BUFFER_CREATE_INFO: return sizeof(VkExternalMemoryBufferCreateInfo);
        case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO: return sizeof(VkExternalMemoryImageCreateInfo);
        case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO_NV: return sizeof(VkExternalMemoryImageCreateInfoNV);
        case VK_STRUCTURE_TYPE_EXTERNAL_SEMAPHORE_PROPERTIES: return sizeof(VkExternalSemaphoreProperties);
        case VK_STRUCTURE_TYPE_FENCE_CREATE_INFO: return sizeof(VkFenceCreateInfo);
        case VK_STRUCTURE_TYPE_FENCE_GET_FD_INFO_KHR: return sizeof(VkFenceGetFdInfoKHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_FENCE_GET_WIN32_HANDLE_INFO_KHR: return sizeof(VkFenceGetWin32HandleInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_FILTER_CUBIC_IMAGE_VIEW_IMAGE_FORMAT_PROPERTIES_EXT: return sizeof(VkFilterCubicImageViewImageFormatPropertiesEXT);
        case VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2: return sizeof(VkFormatProperties2);
        case VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_3: return sizeof(VkFormatProperties3);
        case VK_STRUCTURE_TYPE_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR: return sizeof(VkFragmentShadingRateAttachmentInfoKHR);
        case VK_STRUCTURE_TYPE_FRAME_BOUNDARY_EXT: return sizeof(VkFrameBoundaryEXT);
        case VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENT_IMAGE_INFO: return sizeof(VkFramebufferAttachmentImageInfo);
        case VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENTS_CREATE_INFO: return sizeof(VkFramebufferAttachmentsCreateInfo);
        case VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO: return sizeof(VkFramebufferCreateInfo);
        case VK_STRUCTURE_TYPE_FRAMEBUFFER_MIXED_SAMPLES_COMBINATION_NV: return sizeof(VkFramebufferMixedSamplesCombinationNV);
        case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_NV: return sizeof(VkGeneratedCommandsInfoNV);
        case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_NV: return sizeof(VkGeneratedCommandsMemoryRequirementsInfoNV);
        case VK_STRUCTURE_TYPE_GEOMETRY_AABB_NV: return sizeof(VkGeometryAABBNV);
        case VK_STRUCTURE_TYPE_GEOMETRY_NV: return sizeof(VkGeometryNV);
        case VK_STRUCTURE_TYPE_GEOMETRY_TRIANGLES_NV: return sizeof(VkGeometryTrianglesNV);
        case VK_STRUCTURE_TYPE_GET_LATENCY_MARKER_INFO_NV: return sizeof(VkGetLatencyMarkerInfoNV);
        case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO: return sizeof(VkGraphicsPipelineCreateInfo);
        case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_LIBRARY_CREATE_INFO_EXT: return sizeof(VkGraphicsPipelineLibraryCreateInfoEXT);
        case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_SHADER_GROUPS_CREATE_INFO_NV: return sizeof(VkGraphicsPipelineShaderGroupsCreateInfoNV);
        case VK_STRUCTURE_TYPE_GRAPHICS_SHADER_GROUP_CREATE_INFO_NV: return sizeof(VkGraphicsShaderGroupCreateInfoNV);
        case VK_STRUCTURE_TYPE_HDR_METADATA_EXT: return sizeof(VkHdrMetadataEXT);
        case VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT: return sizeof(VkHeadlessSurfaceCreateInfoEXT);
        case VK_STRUCTURE_TYPE_HOST_IMAGE_COPY_DEVICE_PERFORMANCE_QUERY_EXT: return sizeof(VkHostImageCopyDevicePerformanceQueryEXT);
        case VK_STRUCTURE_TYPE_HOST_IMAGE_LAYOUT_TRANSITION_INFO_EXT: return sizeof(VkHostImageLayoutTransitionInfoEXT);
    #ifdef VK_USE_PLATFORM_IOS_MVK
        case VK_STRUCTURE_TYPE_IOS_SURFACE_CREATE_INFO_MVK: return sizeof(VkIOSSurfaceCreateInfoMVK);
    #endif
        case VK_STRUCTURE_TYPE_IMAGE_BLIT_2: return sizeof(VkImageBlit2);
        case VK_STRUCTURE_TYPE_IMAGE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT: return sizeof(VkImageCaptureDescriptorDataInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_CONTROL_EXT: return sizeof(VkImageCompressionControlEXT);
        case VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_PROPERTIES_EXT: return sizeof(VkImageCompressionPropertiesEXT);
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMAGE_CONSTRAINTS_INFO_FUCHSIA: return sizeof(VkImageConstraintsInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_IMAGE_COPY_2: return sizeof(VkImageCopy2);
        case VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO: return sizeof(VkImageCreateInfo);
        case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_EXPLICIT_CREATE_INFO_EXT: return sizeof(VkImageDrmFormatModifierExplicitCreateInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_LIST_CREATE_INFO_EXT: return sizeof(VkImageDrmFormatModifierListCreateInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_PROPERTIES_EXT: return sizeof(VkImageDrmFormatModifierPropertiesEXT);
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMAGE_FORMAT_CONSTRAINTS_INFO_FUCHSIA: return sizeof(VkImageFormatConstraintsInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_IMAGE_FORMAT_LIST_CREATE_INFO: return sizeof(VkImageFormatListCreateInfo);
        case VK_STRUCTURE_TYPE_IMAGE_FORMAT_PROPERTIES_2: return sizeof(VkImageFormatProperties2);
        case VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER: return sizeof(VkImageMemoryBarrier);
        case VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2: return sizeof(VkImageMemoryBarrier2);
        case VK_STRUCTURE_TYPE_IMAGE_MEMORY_REQUIREMENTS_INFO_2: return sizeof(VkImageMemoryRequirementsInfo2);
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMAGEPIPE_SURFACE_CREATE_INFO_FUCHSIA: return sizeof(VkImagePipeSurfaceCreateInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_IMAGE_PLANE_MEMORY_REQUIREMENTS_INFO: return sizeof(VkImagePlaneMemoryRequirementsInfo);
        case VK_STRUCTURE_TYPE_IMAGE_RESOLVE_2: return sizeof(VkImageResolve2);
        case VK_STRUCTURE_TYPE_IMAGE_SPARSE_MEMORY_REQUIREMENTS_INFO_2: return sizeof(VkImageSparseMemoryRequirementsInfo2);
        case VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_CREATE_INFO: return sizeof(VkImageStencilUsageCreateInfo);
        case VK_STRUCTURE_TYPE_IMAGE_SUBRESOURCE_2_KHR: return sizeof(VkImageSubresource2KHR);
        case VK_STRUCTURE_TYPE_IMAGE_SWAPCHAIN_CREATE_INFO_KHR: return sizeof(VkImageSwapchainCreateInfoKHR);
        case VK_STRUCTURE_TYPE_IMAGE_TO_MEMORY_COPY_EXT: return sizeof(VkImageToMemoryCopyEXT);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_ASTC_DECODE_MODE_EXT: return sizeof(VkImageViewASTCDecodeModeEXT);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_ADDRESS_PROPERTIES_NVX: return sizeof(VkImageViewAddressPropertiesNVX);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_EXT: return sizeof(VkImageViewCaptureDescriptorDataInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO: return sizeof(VkImageViewCreateInfo);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_HANDLE_INFO_NVX: return sizeof(VkImageViewHandleInfoNVX);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_MIN_LOD_CREATE_INFO_EXT: return sizeof(VkImageViewMinLodCreateInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_SAMPLE_WEIGHT_CREATE_INFO_QCOM: return sizeof(VkImageViewSampleWeightCreateInfoQCOM);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_SLICED_CREATE_INFO_EXT: return sizeof(VkImageViewSlicedCreateInfoEXT);
        case VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_CREATE_INFO: return sizeof(VkImageViewUsageCreateInfo);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_IMPORT_ANDROID_HARDWARE_BUFFER_INFO_ANDROID: return sizeof(VkImportAndroidHardwareBufferInfoANDROID);
    #endif
        case VK_STRUCTURE_TYPE_IMPORT_FENCE_FD_INFO_KHR: return sizeof(VkImportFenceFdInfoKHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_IMPORT_FENCE_WIN32_HANDLE_INFO_KHR: return sizeof(VkImportFenceWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_BUFFER_COLLECTION_FUCHSIA: return sizeof(VkImportMemoryBufferCollectionFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_FD_INFO_KHR: return sizeof(VkImportMemoryFdInfoKHR);
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_HOST_POINTER_INFO_EXT: return sizeof(VkImportMemoryHostPointerInfoEXT);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_KHR: return sizeof(VkImportMemoryWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_NV: return sizeof(VkImportMemoryWin32HandleInfoNV);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMPORT_MEMORY_ZIRCON_HANDLE_INFO_FUCHSIA: return sizeof(VkImportMemoryZirconHandleInfoFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_IMPORT_METAL_BUFFER_INFO_EXT: return sizeof(VkImportMetalBufferInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_IMPORT_METAL_IO_SURFACE_INFO_EXT: return sizeof(VkImportMetalIOSurfaceInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_IMPORT_METAL_SHARED_EVENT_INFO_EXT: return sizeof(VkImportMetalSharedEventInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_IMPORT_METAL_TEXTURE_INFO_EXT: return sizeof(VkImportMetalTextureInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_IMPORT_SCREEN_BUFFER_INFO_QNX: return sizeof(VkImportScreenBufferInfoQNX);
    #endif
        case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_FD_INFO_KHR: return sizeof(VkImportSemaphoreFdInfoKHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR: return sizeof(VkImportSemaphoreWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_ZIRCON_HANDLE_INFO_FUCHSIA: return sizeof(VkImportSemaphoreZirconHandleInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_NV: return sizeof(VkIndirectCommandsLayoutCreateInfoNV);
        case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_NV: return sizeof(VkIndirectCommandsLayoutTokenNV);
        case VK_STRUCTURE_TYPE_INITIALIZE_PERFORMANCE_API_INFO_INTEL: return sizeof(VkInitializePerformanceApiInfoINTEL);
        case VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO: return sizeof(VkInstanceCreateInfo);
        case VK_STRUCTURE_TYPE_LATENCY_SLEEP_INFO_NV: return sizeof(VkLatencySleepInfoNV);
        case VK_STRUCTURE_TYPE_LATENCY_SLEEP_MODE_INFO_NV: return sizeof(VkLatencySleepModeInfoNV);
        case VK_STRUCTURE_TYPE_LATENCY_SUBMISSION_PRESENT_ID_NV: return sizeof(VkLatencySubmissionPresentIdNV);
        case VK_STRUCTURE_TYPE_LATENCY_SURFACE_CAPABILITIES_NV: return sizeof(VkLatencySurfaceCapabilitiesNV);
        case VK_STRUCTURE_TYPE_LATENCY_TIMINGS_FRAME_REPORT_NV: return sizeof(VkLatencyTimingsFrameReportNV);
        case VK_STRUCTURE_TYPE_LAYER_SETTINGS_CREATE_INFO_EXT: return sizeof(VkLayerSettingsCreateInfoEXT);
    #ifdef VK_USE_PLATFORM_MACOS_MVK
        case VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK: return sizeof(VkMacOSSurfaceCreateInfoMVK);
    #endif
        case VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE: return sizeof(VkMappedMemoryRange);
        case VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_FLAGS_INFO: return sizeof(VkMemoryAllocateFlagsInfo);
        case VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO: return sizeof(VkMemoryAllocateInfo);
        case VK_STRUCTURE_TYPE_MEMORY_BARRIER: return sizeof(VkMemoryBarrier);
        case VK_STRUCTURE_TYPE_MEMORY_BARRIER_2: return sizeof(VkMemoryBarrier2);
        case VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO: return sizeof(VkMemoryDedicatedAllocateInfo);
        case VK_STRUCTURE_TYPE_MEMORY_DEDICATED_REQUIREMENTS: return sizeof(VkMemoryDedicatedRequirements);
        case VK_STRUCTURE_TYPE_MEMORY_FD_PROPERTIES_KHR: return sizeof(VkMemoryFdPropertiesKHR);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_MEMORY_GET_ANDROID_HARDWARE_BUFFER_INFO_ANDROID: return sizeof(VkMemoryGetAndroidHardwareBufferInfoANDROID);
    #endif
        case VK_STRUCTURE_TYPE_MEMORY_GET_FD_INFO_KHR: return sizeof(VkMemoryGetFdInfoKHR);
        case VK_STRUCTURE_TYPE_MEMORY_GET_REMOTE_ADDRESS_INFO_NV: return sizeof(VkMemoryGetRemoteAddressInfoNV);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_MEMORY_GET_WIN32_HANDLE_INFO_KHR: return sizeof(VkMemoryGetWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_MEMORY_GET_ZIRCON_HANDLE_INFO_FUCHSIA: return sizeof(VkMemoryGetZirconHandleInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_MEMORY_HOST_POINTER_PROPERTIES_EXT: return sizeof(VkMemoryHostPointerPropertiesEXT);
        case VK_STRUCTURE_TYPE_MEMORY_MAP_INFO_KHR: return sizeof(VkMemoryMapInfoKHR);
        case VK_STRUCTURE_TYPE_MEMORY_OPAQUE_CAPTURE_ADDRESS_ALLOCATE_INFO: return sizeof(VkMemoryOpaqueCaptureAddressAllocateInfo);
        case VK_STRUCTURE_TYPE_MEMORY_PRIORITY_ALLOCATE_INFO_EXT: return sizeof(VkMemoryPriorityAllocateInfoEXT);
        case VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2: return sizeof(VkMemoryRequirements2);
        case VK_STRUCTURE_TYPE_MEMORY_TO_IMAGE_COPY_EXT: return sizeof(VkMemoryToImageCopyEXT);
        case VK_STRUCTURE_TYPE_MEMORY_UNMAP_INFO_KHR: return sizeof(VkMemoryUnmapInfoKHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_MEMORY_WIN32_HANDLE_PROPERTIES_KHR: return sizeof(VkMemoryWin32HandlePropertiesKHR);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_MEMORY_ZIRCON_HANDLE_PROPERTIES_FUCHSIA: return sizeof(VkMemoryZirconHandlePropertiesFUCHSIA);
    #endif
    #ifdef VK_USE_PLATFORM_METAL_EXT
        case VK_STRUCTURE_TYPE_METAL_SURFACE_CREATE_INFO_EXT: return sizeof(VkMetalSurfaceCreateInfoEXT);
    #endif
        case VK_STRUCTURE_TYPE_MICROMAP_BUILD_INFO_EXT: return sizeof(VkMicromapBuildInfoEXT);
        case VK_STRUCTURE_TYPE_MICROMAP_BUILD_SIZES_INFO_EXT: return sizeof(VkMicromapBuildSizesInfoEXT);
        case VK_STRUCTURE_TYPE_MICROMAP_CREATE_INFO_EXT: return sizeof(VkMicromapCreateInfoEXT);
        case VK_STRUCTURE_TYPE_MICROMAP_VERSION_INFO_EXT: return sizeof(VkMicromapVersionInfoEXT);
        case VK_STRUCTURE_TYPE_MULTISAMPLE_PROPERTIES_EXT: return sizeof(VkMultisamplePropertiesEXT);
        case VK_STRUCTURE_TYPE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_INFO_EXT: return sizeof(VkMultisampledRenderToSingleSampledInfoEXT);
        case VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_ATTRIBUTES_INFO_NVX: return sizeof(VkMultiviewPerViewAttributesInfoNVX);
        case VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_RENDER_AREAS_RENDER_PASS_BEGIN_INFO_QCOM: return sizeof(VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM);
        case VK_STRUCTURE_TYPE_MUTABLE_DESCRIPTOR_TYPE_CREATE_INFO_EXT: return sizeof(VkMutableDescriptorTypeCreateInfoEXT);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_NATIVE_BUFFER_ANDROID: return sizeof(VkNativeBufferANDROID);
    #endif
        case VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DESCRIPTOR_DATA_CREATE_INFO_EXT: return sizeof(VkOpaqueCaptureDescriptorDataCreateInfoEXT);
        case VK_STRUCTURE_TYPE_OPTICAL_FLOW_EXECUTE_INFO_NV: return sizeof(VkOpticalFlowExecuteInfoNV);
        case VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_INFO_NV: return sizeof(VkOpticalFlowImageFormatInfoNV);
        case VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_NV: return sizeof(VkOpticalFlowImageFormatPropertiesNV);
        case VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_INFO_NV: return sizeof(VkOpticalFlowSessionCreateInfoNV);
        case VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_PRIVATE_DATA_INFO_NV: return sizeof(VkOpticalFlowSessionCreatePrivateDataInfoNV);
        case VK_STRUCTURE_TYPE_OUT_OF_BAND_QUEUE_TYPE_INFO_NV: return sizeof(VkOutOfBandQueueTypeInfoNV);
        case VK_STRUCTURE_TYPE_PERFORMANCE_CONFIGURATION_ACQUIRE_INFO_INTEL: return sizeof(VkPerformanceConfigurationAcquireInfoINTEL);
        case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_KHR: return sizeof(VkPerformanceCounterDescriptionKHR);
        case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_KHR: return sizeof(VkPerformanceCounterKHR);
        case VK_STRUCTURE_TYPE_PERFORMANCE_MARKER_INFO_INTEL: return sizeof(VkPerformanceMarkerInfoINTEL);
        case VK_STRUCTURE_TYPE_PERFORMANCE_OVERRIDE_INFO_INTEL: return sizeof(VkPerformanceOverrideInfoINTEL);
        case VK_STRUCTURE_TYPE_PERFORMANCE_QUERY_SUBMIT_INFO_KHR: return sizeof(VkPerformanceQuerySubmitInfoKHR);
        case VK_STRUCTURE_TYPE_PERFORMANCE_STREAM_MARKER_INFO_INTEL: return sizeof(VkPerformanceStreamMarkerInfoINTEL);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_16BIT_STORAGE_FEATURES: return sizeof(VkPhysicalDevice16BitStorageFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT: return sizeof(VkPhysicalDevice4444FormatsFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_8BIT_STORAGE_FEATURES: return sizeof(VkPhysicalDevice8BitStorageFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ASTC_DECODE_FEATURES_EXT: return sizeof(VkPhysicalDeviceASTCDecodeFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_FEATURES_KHR: return sizeof(VkPhysicalDeviceAccelerationStructureFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceAccelerationStructurePropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ADDRESS_BINDING_REPORT_FEATURES_EXT: return sizeof(VkPhysicalDeviceAddressBindingReportFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_AMIGO_PROFILING_FEATURES_SEC: return sizeof(VkPhysicalDeviceAmigoProfilingFeaturesSEC);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_DYNAMIC_STATE_FEATURES_EXT: return sizeof(VkPhysicalDeviceAttachmentFeedbackLoopDynamicStateFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_LAYOUT_FEATURES_EXT: return sizeof(VkPhysicalDeviceAttachmentFeedbackLoopLayoutFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_FEATURES_EXT: return sizeof(VkPhysicalDeviceBlendOperationAdvancedFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BORDER_COLOR_SWIZZLE_FEATURES_EXT: return sizeof(VkPhysicalDeviceBorderColorSwizzleFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES: return sizeof(VkPhysicalDeviceBufferDeviceAddressFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_EXT: return sizeof(VkPhysicalDeviceBufferDeviceAddressFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_FEATURES_HUAWEI: return sizeof(VkPhysicalDeviceClusterCullingShaderFeaturesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_PROPERTIES_HUAWEI: return sizeof(VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_VRS_FEATURES_HUAWEI: return sizeof(VkPhysicalDeviceClusterCullingShaderVrsFeaturesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COHERENT_MEMORY_FEATURES_AMD: return sizeof(VkPhysicalDeviceCoherentMemoryFeaturesAMD);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COLOR_WRITE_ENABLE_FEATURES_EXT: return sizeof(VkPhysicalDeviceColorWriteEnableFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_FEATURES_NV: return sizeof(VkPhysicalDeviceComputeShaderDerivativesFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT: return sizeof(VkPhysicalDeviceConditionalRenderingFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONSERVATIVE_RASTERIZATION_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceConservativeRasterizationPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_KHR: return sizeof(VkPhysicalDeviceCooperativeMatrixFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_NV: return sizeof(VkPhysicalDeviceCooperativeMatrixFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceCooperativeMatrixPropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_NV: return sizeof(VkPhysicalDeviceCooperativeMatrixPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_NV: return sizeof(VkPhysicalDeviceCopyMemoryIndirectFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_PROPERTIES_NV: return sizeof(VkPhysicalDeviceCopyMemoryIndirectPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CORNER_SAMPLED_IMAGE_FEATURES_NV: return sizeof(VkPhysicalDeviceCornerSampledImageFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COVERAGE_REDUCTION_MODE_FEATURES_NV: return sizeof(VkPhysicalDeviceCoverageReductionModeFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_CLAMP_FEATURES_QCOM: return sizeof(VkPhysicalDeviceCubicClampFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_WEIGHTS_FEATURES_QCOM: return sizeof(VkPhysicalDeviceCubicWeightsFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_FEATURES_EXT: return sizeof(VkPhysicalDeviceCustomBorderColorFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceCustomBorderColorPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEDICATED_ALLOCATION_IMAGE_ALIASING_FEATURES_NV: return sizeof(VkPhysicalDeviceDedicatedAllocationImageAliasingFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_BIAS_CONTROL_FEATURES_EXT: return sizeof(VkPhysicalDeviceDepthBiasControlFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_ZERO_ONE_FEATURES_EXT: return sizeof(VkPhysicalDeviceDepthClampZeroOneFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_CONTROL_FEATURES_EXT: return sizeof(VkPhysicalDeviceDepthClipControlFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_ENABLE_FEATURES_EXT: return sizeof(VkPhysicalDeviceDepthClipEnableFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_STENCIL_RESOLVE_PROPERTIES: return sizeof(VkPhysicalDeviceDepthStencilResolveProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_DENSITY_MAP_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceDescriptorBufferDensityMapPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_FEATURES_EXT: return sizeof(VkPhysicalDeviceDescriptorBufferFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceDescriptorBufferPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_FEATURES: return sizeof(VkPhysicalDeviceDescriptorIndexingFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_PROPERTIES: return sizeof(VkPhysicalDeviceDescriptorIndexingProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV: return sizeof(VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_SET_HOST_MAPPING_FEATURES_VALVE: return sizeof(VkPhysicalDeviceDescriptorSetHostMappingFeaturesVALVE);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_COMPUTE_FEATURES_NV: return sizeof(VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_NV: return sizeof(VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_NV: return sizeof(VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_MEMORY_REPORT_FEATURES_EXT: return sizeof(VkPhysicalDeviceDeviceMemoryReportFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DIAGNOSTICS_CONFIG_FEATURES_NV: return sizeof(VkPhysicalDeviceDiagnosticsConfigFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISCARD_RECTANGLE_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceDiscardRectanglePropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRIVER_PROPERTIES: return sizeof(VkPhysicalDeviceDriverProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRM_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceDrmPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_FEATURES: return sizeof(VkPhysicalDeviceDynamicRenderingFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_UNUSED_ATTACHMENTS_FEATURES_EXT: return sizeof(VkPhysicalDeviceDynamicRenderingUnusedAttachmentsFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXCLUSIVE_SCISSOR_FEATURES_NV: return sizeof(VkPhysicalDeviceExclusiveScissorFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_2_FEATURES_EXT: return sizeof(VkPhysicalDeviceExtendedDynamicState2FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_FEATURES_EXT: return sizeof(VkPhysicalDeviceExtendedDynamicState3FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceExtendedDynamicState3PropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT: return sizeof(VkPhysicalDeviceExtendedDynamicStateFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_FEATURES_NV: return sizeof(VkPhysicalDeviceExtendedSparseAddressSpaceFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_PROPERTIES_NV: return sizeof(VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_BUFFER_INFO: return sizeof(VkPhysicalDeviceExternalBufferInfo);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FENCE_INFO: return sizeof(VkPhysicalDeviceExternalFenceInfo);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_FEATURES_ANDROID: return sizeof(VkPhysicalDeviceExternalFormatResolveFeaturesANDROID);
    #endif
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_PROPERTIES_ANDROID: return sizeof(VkPhysicalDeviceExternalFormatResolvePropertiesANDROID);
    #endif
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_IMAGE_FORMAT_INFO: return sizeof(VkPhysicalDeviceExternalImageFormatInfo);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_HOST_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceExternalMemoryHostPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_RDMA_FEATURES_NV: return sizeof(VkPhysicalDeviceExternalMemoryRDMAFeaturesNV);
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_SCREEN_BUFFER_FEATURES_QNX: return sizeof(VkPhysicalDeviceExternalMemoryScreenBufferFeaturesQNX);
    #endif
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_SEMAPHORE_INFO: return sizeof(VkPhysicalDeviceExternalSemaphoreInfo);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_EXT: return sizeof(VkPhysicalDeviceFaultFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2: return sizeof(VkPhysicalDeviceFeatures2);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT_CONTROLS_PROPERTIES: return sizeof(VkPhysicalDeviceFloatControlsProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_FEATURES_EXT: return sizeof(VkPhysicalDeviceFragmentDensityMap2FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceFragmentDensityMap2PropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_FEATURES_EXT: return sizeof(VkPhysicalDeviceFragmentDensityMapFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_FEATURES_QCOM: return sizeof(VkPhysicalDeviceFragmentDensityMapOffsetFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_PROPERTIES_QCOM: return sizeof(VkPhysicalDeviceFragmentDensityMapOffsetPropertiesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceFragmentDensityMapPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_FEATURES_KHR: return sizeof(VkPhysicalDeviceFragmentShaderBarycentricFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceFragmentShaderBarycentricPropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_INTERLOCK_FEATURES_EXT: return sizeof(VkPhysicalDeviceFragmentShaderInterlockFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_FEATURES_NV: return sizeof(VkPhysicalDeviceFragmentShadingRateEnumsFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_PROPERTIES_NV: return sizeof(VkPhysicalDeviceFragmentShadingRateEnumsPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_FEATURES_KHR: return sizeof(VkPhysicalDeviceFragmentShadingRateFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_KHR: return sizeof(VkPhysicalDeviceFragmentShadingRateKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceFragmentShadingRatePropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAME_BOUNDARY_FEATURES_EXT: return sizeof(VkPhysicalDeviceFrameBoundaryFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GLOBAL_PRIORITY_QUERY_FEATURES_KHR: return sizeof(VkPhysicalDeviceGlobalPriorityQueryFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_FEATURES_EXT: return sizeof(VkPhysicalDeviceGraphicsPipelineLibraryFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceGraphicsPipelineLibraryPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GROUP_PROPERTIES: return sizeof(VkPhysicalDeviceGroupProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_FEATURES_EXT: return sizeof(VkPhysicalDeviceHostImageCopyFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceHostImageCopyPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES: return sizeof(VkPhysicalDeviceHostQueryResetFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ID_PROPERTIES: return sizeof(VkPhysicalDeviceIDProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_2D_VIEW_OF_3D_FEATURES_EXT: return sizeof(VkPhysicalDeviceImage2DViewOf3DFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_FEATURES_EXT: return sizeof(VkPhysicalDeviceImageCompressionControlFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_SWAPCHAIN_FEATURES_EXT: return sizeof(VkPhysicalDeviceImageCompressionControlSwapchainFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT: return sizeof(VkPhysicalDeviceImageDrmFormatModifierInfoEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2: return sizeof(VkPhysicalDeviceImageFormatInfo2);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_FEATURES_QCOM: return sizeof(VkPhysicalDeviceImageProcessing2FeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_PROPERTIES_QCOM: return sizeof(VkPhysicalDeviceImageProcessing2PropertiesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_FEATURES_QCOM: return sizeof(VkPhysicalDeviceImageProcessingFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_PROPERTIES_QCOM: return sizeof(VkPhysicalDeviceImageProcessingPropertiesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ROBUSTNESS_FEATURES: return sizeof(VkPhysicalDeviceImageRobustnessFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_SLICED_VIEW_OF_3D_FEATURES_EXT: return sizeof(VkPhysicalDeviceImageSlicedViewOf3DFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_IMAGE_FORMAT_INFO_EXT: return sizeof(VkPhysicalDeviceImageViewImageFormatInfoEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_MIN_LOD_FEATURES_EXT: return sizeof(VkPhysicalDeviceImageViewMinLodFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGELESS_FRAMEBUFFER_FEATURES: return sizeof(VkPhysicalDeviceImagelessFramebufferFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INDEX_TYPE_UINT8_FEATURES_EXT: return sizeof(VkPhysicalDeviceIndexTypeUint8FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INHERITED_VIEWPORT_SCISSOR_FEATURES_NV: return sizeof(VkPhysicalDeviceInheritedViewportScissorFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_FEATURES: return sizeof(VkPhysicalDeviceInlineUniformBlockFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_PROPERTIES: return sizeof(VkPhysicalDeviceInlineUniformBlockProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INVOCATION_MASK_FEATURES_HUAWEI: return sizeof(VkPhysicalDeviceInvocationMaskFeaturesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_DRIVER_PROPERTIES_MSFT: return sizeof(VkPhysicalDeviceLayeredDriverPropertiesMSFT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_DITHERING_FEATURES_EXT: return sizeof(VkPhysicalDeviceLegacyDitheringFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_FEATURES_EXT: return sizeof(VkPhysicalDeviceLineRasterizationFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceLineRasterizationPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINEAR_COLOR_ATTACHMENT_FEATURES_NV: return sizeof(VkPhysicalDeviceLinearColorAttachmentFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_3_PROPERTIES: return sizeof(VkPhysicalDeviceMaintenance3Properties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES: return sizeof(VkPhysicalDeviceMaintenance4Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_PROPERTIES: return sizeof(VkPhysicalDeviceMaintenance4Properties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_FEATURES_KHR: return sizeof(VkPhysicalDeviceMaintenance5FeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceMaintenance5PropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_FEATURES_KHR: return sizeof(VkPhysicalDeviceMaintenance6FeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceMaintenance6PropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_BUDGET_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceMemoryBudgetPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_FEATURES_NV: return sizeof(VkPhysicalDeviceMemoryDecompressionFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_PROPERTIES_NV: return sizeof(VkPhysicalDeviceMemoryDecompressionPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PRIORITY_FEATURES_EXT: return sizeof(VkPhysicalDeviceMemoryPriorityFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PROPERTIES_2: return sizeof(VkPhysicalDeviceMemoryProperties2);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_EXT: return sizeof(VkPhysicalDeviceMeshShaderFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_NV: return sizeof(VkPhysicalDeviceMeshShaderFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceMeshShaderPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_NV: return sizeof(VkPhysicalDeviceMeshShaderPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_FEATURES_EXT: return sizeof(VkPhysicalDeviceMultiDrawFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceMultiDrawPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_FEATURES_EXT: return sizeof(VkPhysicalDeviceMultisampledRenderToSingleSampledFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_FEATURES: return sizeof(VkPhysicalDeviceMultiviewFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_ATTRIBUTES_PROPERTIES_NVX: return sizeof(VkPhysicalDeviceMultiviewPerViewAttributesPropertiesNVX);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_RENDER_AREAS_FEATURES_QCOM: return sizeof(VkPhysicalDeviceMultiviewPerViewRenderAreasFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_VIEWPORTS_FEATURES_QCOM: return sizeof(VkPhysicalDeviceMultiviewPerViewViewportsFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PROPERTIES: return sizeof(VkPhysicalDeviceMultiviewProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MUTABLE_DESCRIPTOR_TYPE_FEATURES_EXT: return sizeof(VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_FEATURES_EXT: return sizeof(VkPhysicalDeviceNestedCommandBufferFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceNestedCommandBufferPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT: return sizeof(VkPhysicalDeviceNonSeamlessCubeMapFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_EXT: return sizeof(VkPhysicalDeviceOpacityMicromapFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceOpacityMicromapPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_FEATURES_NV: return sizeof(VkPhysicalDeviceOpticalFlowFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_PROPERTIES_NV: return sizeof(VkPhysicalDeviceOpticalFlowPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PCI_BUS_INFO_PROPERTIES_EXT: return sizeof(VkPhysicalDevicePCIBusInfoPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PAGEABLE_DEVICE_LOCAL_MEMORY_FEATURES_EXT: return sizeof(VkPhysicalDevicePageableDeviceLocalMemoryFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PER_STAGE_DESCRIPTOR_SET_FEATURES_NV: return sizeof(VkPhysicalDevicePerStageDescriptorSetFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_FEATURES_KHR: return sizeof(VkPhysicalDevicePerformanceQueryFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_PROPERTIES_KHR: return sizeof(VkPhysicalDevicePerformanceQueryPropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CREATION_CACHE_CONTROL_FEATURES: return sizeof(VkPhysicalDevicePipelineCreationCacheControlFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_EXECUTABLE_PROPERTIES_FEATURES_KHR: return sizeof(VkPhysicalDevicePipelineExecutablePropertiesFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_LIBRARY_GROUP_HANDLES_FEATURES_EXT: return sizeof(VkPhysicalDevicePipelineLibraryGroupHandlesFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROPERTIES_FEATURES_EXT: return sizeof(VkPhysicalDevicePipelinePropertiesFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROTECTED_ACCESS_FEATURES_EXT: return sizeof(VkPhysicalDevicePipelineProtectedAccessFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_FEATURES_EXT: return sizeof(VkPhysicalDevicePipelineRobustnessFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_PROPERTIES_EXT: return sizeof(VkPhysicalDevicePipelineRobustnessPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_POINT_CLIPPING_PROPERTIES: return sizeof(VkPhysicalDevicePointClippingProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_BARRIER_FEATURES_NV: return sizeof(VkPhysicalDevicePresentBarrierFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_FEATURES_KHR: return sizeof(VkPhysicalDevicePresentIdFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_FEATURES_KHR: return sizeof(VkPhysicalDevicePresentWaitFeaturesKHR);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENTATION_PROPERTIES_ANDROID: return sizeof(VkPhysicalDevicePresentationPropertiesANDROID);
    #endif
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_TOPOLOGY_LIST_RESTART_FEATURES_EXT: return sizeof(VkPhysicalDevicePrimitiveTopologyListRestartFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVES_GENERATED_QUERY_FEATURES_EXT: return sizeof(VkPhysicalDevicePrimitivesGeneratedQueryFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIVATE_DATA_FEATURES: return sizeof(VkPhysicalDevicePrivateDataFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2: return sizeof(VkPhysicalDeviceProperties2);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_FEATURES: return sizeof(VkPhysicalDeviceProtectedMemoryFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_PROPERTIES: return sizeof(VkPhysicalDeviceProtectedMemoryProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_FEATURES_EXT: return sizeof(VkPhysicalDeviceProvokingVertexFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceProvokingVertexPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_DESCRIPTOR_PROPERTIES_KHR: return sizeof(VkPhysicalDevicePushDescriptorPropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RGBA10X6_FORMATS_FEATURES_EXT: return sizeof(VkPhysicalDeviceRGBA10X6FormatsFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_FEATURES_EXT: return sizeof(VkPhysicalDeviceRasterizationOrderAttachmentAccessFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_QUERY_FEATURES_KHR: return sizeof(VkPhysicalDeviceRayQueryFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_NV: return sizeof(VkPhysicalDeviceRayTracingInvocationReorderFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_NV: return sizeof(VkPhysicalDeviceRayTracingInvocationReorderPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MAINTENANCE_1_FEATURES_KHR: return sizeof(VkPhysicalDeviceRayTracingMaintenance1FeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MOTION_BLUR_FEATURES_NV: return sizeof(VkPhysicalDeviceRayTracingMotionBlurFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_FEATURES_KHR: return sizeof(VkPhysicalDeviceRayTracingPipelineFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceRayTracingPipelinePropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_POSITION_FETCH_FEATURES_KHR: return sizeof(VkPhysicalDeviceRayTracingPositionFetchFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PROPERTIES_NV: return sizeof(VkPhysicalDeviceRayTracingPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RELAXED_LINE_RASTERIZATION_FEATURES_IMG: return sizeof(VkPhysicalDeviceRelaxedLineRasterizationFeaturesIMG);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_FEATURES_ARM: return sizeof(VkPhysicalDeviceRenderPassStripedFeaturesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_PROPERTIES_ARM: return sizeof(VkPhysicalDeviceRenderPassStripedPropertiesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_REPRESENTATIVE_FRAGMENT_TEST_FEATURES_NV: return sizeof(VkPhysicalDeviceRepresentativeFragmentTestFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_FEATURES_EXT: return sizeof(VkPhysicalDeviceRobustness2FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceRobustness2PropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLE_LOCATIONS_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceSampleLocationsPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_FILTER_MINMAX_PROPERTIES: return sizeof(VkPhysicalDeviceSamplerFilterMinmaxProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_YCBCR_CONVERSION_FEATURES: return sizeof(VkPhysicalDeviceSamplerYcbcrConversionFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCALAR_BLOCK_LAYOUT_FEATURES: return sizeof(VkPhysicalDeviceScalarBlockLayoutFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_FEATURES_ARM: return sizeof(VkPhysicalDeviceSchedulingControlsFeaturesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_PROPERTIES_ARM: return sizeof(VkPhysicalDeviceSchedulingControlsPropertiesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SEPARATE_DEPTH_STENCIL_LAYOUTS_FEATURES: return sizeof(VkPhysicalDeviceSeparateDepthStencilLayoutsFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_2_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderAtomicFloat2FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderAtomicFloatFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_INT64_FEATURES: return sizeof(VkPhysicalDeviceShaderAtomicInt64Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CLOCK_FEATURES_KHR: return sizeof(VkPhysicalDeviceShaderClockFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_FEATURES_ARM: return sizeof(VkPhysicalDeviceShaderCoreBuiltinsFeaturesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_PROPERTIES_ARM: return sizeof(VkPhysicalDeviceShaderCoreBuiltinsPropertiesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_2_AMD: return sizeof(VkPhysicalDeviceShaderCoreProperties2AMD);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_AMD: return sizeof(VkPhysicalDeviceShaderCorePropertiesAMD);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_ARM: return sizeof(VkPhysicalDeviceShaderCorePropertiesARM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DEMOTE_TO_HELPER_INVOCATION_FEATURES: return sizeof(VkPhysicalDeviceShaderDemoteToHelperInvocationFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETERS_FEATURES: return sizeof(VkPhysicalDeviceShaderDrawParametersFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EARLY_AND_LATE_FRAGMENT_TESTS_FEATURES_AMD: return sizeof(VkPhysicalDeviceShaderEarlyAndLateFragmentTestsFeaturesAMD);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT16_INT8_FEATURES: return sizeof(VkPhysicalDeviceShaderFloat16Int8Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_ATOMIC_INT64_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderImageAtomicInt64FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_FOOTPRINT_FEATURES_NV: return sizeof(VkPhysicalDeviceShaderImageFootprintFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_FEATURES: return sizeof(VkPhysicalDeviceShaderIntegerDotProductFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_PROPERTIES: return sizeof(VkPhysicalDeviceShaderIntegerDotProductProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_FUNCTIONS_2_FEATURES_INTEL: return sizeof(VkPhysicalDeviceShaderIntegerFunctions2FeaturesINTEL);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderModuleIdentifierFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceShaderModuleIdentifierPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderObjectFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceShaderObjectPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_FEATURES_NV: return sizeof(VkPhysicalDeviceShaderSMBuiltinsFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_PROPERTIES_NV: return sizeof(VkPhysicalDeviceShaderSMBuiltinsPropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_EXTENDED_TYPES_FEATURES: return sizeof(VkPhysicalDeviceShaderSubgroupExtendedTypesFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_UNIFORM_CONTROL_FLOW_FEATURES_KHR: return sizeof(VkPhysicalDeviceShaderSubgroupUniformControlFlowFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TERMINATE_INVOCATION_FEATURES: return sizeof(VkPhysicalDeviceShaderTerminateInvocationFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_FEATURES_EXT: return sizeof(VkPhysicalDeviceShaderTileImageFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceShaderTileImagePropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_FEATURES_NV: return sizeof(VkPhysicalDeviceShadingRateImageFeaturesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_PROPERTIES_NV: return sizeof(VkPhysicalDeviceShadingRateImagePropertiesNV);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SPARSE_IMAGE_FORMAT_INFO_2: return sizeof(VkPhysicalDeviceSparseImageFormatInfo2);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES: return sizeof(VkPhysicalDeviceSubgroupProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_FEATURES: return sizeof(VkPhysicalDeviceSubgroupSizeControlFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_PROPERTIES: return sizeof(VkPhysicalDeviceSubgroupSizeControlProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_MERGE_FEEDBACK_FEATURES_EXT: return sizeof(VkPhysicalDeviceSubpassMergeFeedbackFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_FEATURES_HUAWEI: return sizeof(VkPhysicalDeviceSubpassShadingFeaturesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_PROPERTIES_HUAWEI: return sizeof(VkPhysicalDeviceSubpassShadingPropertiesHUAWEI);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SURFACE_INFO_2_KHR: return sizeof(VkPhysicalDeviceSurfaceInfo2KHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SWAPCHAIN_MAINTENANCE_1_FEATURES_EXT: return sizeof(VkPhysicalDeviceSwapchainMaintenance1FeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SYNCHRONIZATION_2_FEATURES: return sizeof(VkPhysicalDeviceSynchronization2Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_FEATURES_EXT: return sizeof(VkPhysicalDeviceTexelBufferAlignmentFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_PROPERTIES: return sizeof(VkPhysicalDeviceTexelBufferAlignmentProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_HDR_FEATURES: return sizeof(VkPhysicalDeviceTextureCompressionASTCHDRFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_PROPERTIES_FEATURES_QCOM: return sizeof(VkPhysicalDeviceTilePropertiesFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES: return sizeof(VkPhysicalDeviceTimelineSemaphoreFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_PROPERTIES: return sizeof(VkPhysicalDeviceTimelineSemaphoreProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TOOL_PROPERTIES: return sizeof(VkPhysicalDeviceToolProperties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_FEATURES_EXT: return sizeof(VkPhysicalDeviceTransformFeedbackFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceTransformFeedbackPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFORM_BUFFER_STANDARD_LAYOUT_FEATURES: return sizeof(VkPhysicalDeviceUniformBufferStandardLayoutFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VARIABLE_POINTERS_FEATURES: return sizeof(VkPhysicalDeviceVariablePointersFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_FEATURES_KHR: return sizeof(VkPhysicalDeviceVertexAttributeDivisorFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_EXT: return sizeof(VkPhysicalDeviceVertexAttributeDivisorPropertiesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_KHR: return sizeof(VkPhysicalDeviceVertexAttributeDivisorPropertiesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_INPUT_DYNAMIC_STATE_FEATURES_EXT: return sizeof(VkPhysicalDeviceVertexInputDynamicStateFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR: return sizeof(VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_FORMAT_INFO_KHR: return sizeof(VkPhysicalDeviceVideoFormatInfoKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_1_FEATURES_KHR: return sizeof(VkPhysicalDeviceVideoMaintenance1FeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES: return sizeof(VkPhysicalDeviceVulkan11Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_PROPERTIES: return sizeof(VkPhysicalDeviceVulkan11Properties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES: return sizeof(VkPhysicalDeviceVulkan12Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_PROPERTIES: return sizeof(VkPhysicalDeviceVulkan12Properties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES: return sizeof(VkPhysicalDeviceVulkan13Features);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_PROPERTIES: return sizeof(VkPhysicalDeviceVulkan13Properties);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_MEMORY_MODEL_FEATURES: return sizeof(VkPhysicalDeviceVulkanMemoryModelFeatures);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_WORKGROUP_MEMORY_EXPLICIT_LAYOUT_FEATURES_KHR: return sizeof(VkPhysicalDeviceWorkgroupMemoryExplicitLayoutFeaturesKHR);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_2_PLANE_444_FORMATS_FEATURES_EXT: return sizeof(VkPhysicalDeviceYcbcr2Plane444FormatsFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_DEGAMMA_FEATURES_QCOM: return sizeof(VkPhysicalDeviceYcbcrDegammaFeaturesQCOM);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_IMAGE_ARRAYS_FEATURES_EXT: return sizeof(VkPhysicalDeviceYcbcrImageArraysFeaturesEXT);
        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_WORKGROUP_MEMORY_FEATURES: return sizeof(VkPhysicalDeviceZeroInitializeWorkgroupMemoryFeatures);
        case VK_STRUCTURE_TYPE_PIPELINE_CACHE_CREATE_INFO: return sizeof(VkPipelineCacheCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_ADVANCED_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineColorBlendAdvancedStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO: return sizeof(VkPipelineColorBlendStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_COLOR_WRITE_CREATE_INFO_EXT: return sizeof(VkPipelineColorWriteCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_COMPILER_CONTROL_CREATE_INFO_AMD: return sizeof(VkPipelineCompilerControlCreateInfoAMD);
        case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_MODULATION_STATE_CREATE_INFO_NV: return sizeof(VkPipelineCoverageModulationStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_REDUCTION_STATE_CREATE_INFO_NV: return sizeof(VkPipelineCoverageReductionStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_TO_COLOR_STATE_CREATE_INFO_NV: return sizeof(VkPipelineCoverageToColorStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_CREATE_FLAGS_2_CREATE_INFO_KHR: return sizeof(VkPipelineCreateFlags2CreateInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_CREATION_FEEDBACK_CREATE_INFO: return sizeof(VkPipelineCreationFeedbackCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO: return sizeof(VkPipelineDepthStencilStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_DISCARD_RECTANGLE_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineDiscardRectangleStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO: return sizeof(VkPipelineDynamicStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INFO_KHR: return sizeof(VkPipelineExecutableInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INTERNAL_REPRESENTATION_KHR: return sizeof(VkPipelineExecutableInternalRepresentationKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_PROPERTIES_KHR: return sizeof(VkPipelineExecutablePropertiesKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_STATISTIC_KHR: return sizeof(VkPipelineExecutableStatisticKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_ENUM_STATE_CREATE_INFO_NV: return sizeof(VkPipelineFragmentShadingRateEnumStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_STATE_CREATE_INFO_KHR: return sizeof(VkPipelineFragmentShadingRateStateCreateInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_INDIRECT_DEVICE_ADDRESS_INFO_NV: return sizeof(VkPipelineIndirectDeviceAddressInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_INFO_KHR: return sizeof(VkPipelineInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO: return sizeof(VkPipelineInputAssemblyStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO: return sizeof(VkPipelineLayoutCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_LIBRARY_CREATE_INFO_KHR: return sizeof(VkPipelineLibraryCreateInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO: return sizeof(VkPipelineMultisampleStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_PROPERTIES_IDENTIFIER_EXT: return sizeof(VkPipelinePropertiesIdentifierEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_CONSERVATIVE_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineRasterizationConservativeStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_DEPTH_CLIP_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineRasterizationDepthClipStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_LINE_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineRasterizationLineStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_PROVOKING_VERTEX_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineRasterizationProvokingVertexStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO: return sizeof(VkPipelineRasterizationStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_RASTERIZATION_ORDER_AMD: return sizeof(VkPipelineRasterizationStateRasterizationOrderAMD);
        case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_STREAM_CREATE_INFO_EXT: return sizeof(VkPipelineRasterizationStateStreamCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_RENDERING_CREATE_INFO: return sizeof(VkPipelineRenderingCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_REPRESENTATIVE_FRAGMENT_TEST_STATE_CREATE_INFO_NV: return sizeof(VkPipelineRepresentativeFragmentTestStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_ROBUSTNESS_CREATE_INFO_EXT: return sizeof(VkPipelineRobustnessCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_SAMPLE_LOCATIONS_STATE_CREATE_INFO_EXT: return sizeof(VkPipelineSampleLocationsStateCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO: return sizeof(VkPipelineShaderStageCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_MODULE_IDENTIFIER_CREATE_INFO_EXT: return sizeof(VkPipelineShaderStageModuleIdentifierCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_REQUIRED_SUBGROUP_SIZE_CREATE_INFO: return sizeof(VkPipelineShaderStageRequiredSubgroupSizeCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_DOMAIN_ORIGIN_STATE_CREATE_INFO: return sizeof(VkPipelineTessellationDomainOriginStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO: return sizeof(VkPipelineTessellationStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_DIVISOR_STATE_CREATE_INFO_KHR: return sizeof(VkPipelineVertexInputDivisorStateCreateInfoKHR);
        case VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO: return sizeof(VkPipelineVertexInputStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_COARSE_SAMPLE_ORDER_STATE_CREATE_INFO_NV: return sizeof(VkPipelineViewportCoarseSampleOrderStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLIP_CONTROL_CREATE_INFO_EXT: return sizeof(VkPipelineViewportDepthClipControlCreateInfoEXT);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_EXCLUSIVE_SCISSOR_STATE_CREATE_INFO_NV: return sizeof(VkPipelineViewportExclusiveScissorStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SHADING_RATE_IMAGE_STATE_CREATE_INFO_NV: return sizeof(VkPipelineViewportShadingRateImageStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO: return sizeof(VkPipelineViewportStateCreateInfo);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SWIZZLE_STATE_CREATE_INFO_NV: return sizeof(VkPipelineViewportSwizzleStateCreateInfoNV);
        case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_W_SCALING_STATE_CREATE_INFO_NV: return sizeof(VkPipelineViewportWScalingStateCreateInfoNV);
    #ifdef VK_USE_PLATFORM_GGP
        case VK_STRUCTURE_TYPE_PRESENT_FRAME_TOKEN_GGP: return sizeof(VkPresentFrameTokenGGP);
    #endif
        case VK_STRUCTURE_TYPE_PRESENT_ID_KHR: return sizeof(VkPresentIdKHR);
        case VK_STRUCTURE_TYPE_PRESENT_INFO_KHR: return sizeof(VkPresentInfoKHR);
        case VK_STRUCTURE_TYPE_PRESENT_REGIONS_KHR: return sizeof(VkPresentRegionsKHR);
        case VK_STRUCTURE_TYPE_PRESENT_TIMES_INFO_GOOGLE: return sizeof(VkPresentTimesInfoGOOGLE);
        case VK_STRUCTURE_TYPE_PRIVATE_DATA_SLOT_CREATE_INFO: return sizeof(VkPrivateDataSlotCreateInfo);
        case VK_STRUCTURE_TYPE_PROTECTED_SUBMIT_INFO: return sizeof(VkProtectedSubmitInfo);
        case VK_STRUCTURE_TYPE_PUSH_CONSTANTS_INFO_KHR: return sizeof(VkPushConstantsInfoKHR);
        case VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_INFO_KHR: return sizeof(VkPushDescriptorSetInfoKHR);
        case VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_WITH_TEMPLATE_INFO_KHR: return sizeof(VkPushDescriptorSetWithTemplateInfoKHR);
        case VK_STRUCTURE_TYPE_QUERY_LOW_LATENCY_SUPPORT_NV: return sizeof(VkQueryLowLatencySupportNV);
        case VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO: return sizeof(VkQueryPoolCreateInfo);
        case VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_CREATE_INFO_KHR: return sizeof(VkQueryPoolPerformanceCreateInfoKHR);
        case VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_QUERY_CREATE_INFO_INTEL: return sizeof(VkQueryPoolPerformanceQueryCreateInfoINTEL);
        case VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_FEEDBACK_CREATE_INFO_KHR: return sizeof(VkQueryPoolVideoEncodeFeedbackCreateInfoKHR);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_2_NV: return sizeof(VkQueueFamilyCheckpointProperties2NV);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_NV: return sizeof(VkQueueFamilyCheckpointPropertiesNV);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_GLOBAL_PRIORITY_PROPERTIES_KHR: return sizeof(VkQueueFamilyGlobalPriorityPropertiesKHR);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_PROPERTIES_2: return sizeof(VkQueueFamilyProperties2);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_QUERY_RESULT_STATUS_PROPERTIES_KHR: return sizeof(VkQueueFamilyQueryResultStatusPropertiesKHR);
        case VK_STRUCTURE_TYPE_QUEUE_FAMILY_VIDEO_PROPERTIES_KHR: return sizeof(VkQueueFamilyVideoPropertiesKHR);
        case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_KHR: return sizeof(VkRayTracingPipelineCreateInfoKHR);
        case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_NV: return sizeof(VkRayTracingPipelineCreateInfoNV);
        case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_INTERFACE_CREATE_INFO_KHR: return sizeof(VkRayTracingPipelineInterfaceCreateInfoKHR);
        case VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_KHR: return sizeof(VkRayTracingShaderGroupCreateInfoKHR);
        case VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_NV: return sizeof(VkRayTracingShaderGroupCreateInfoNV);
        case VK_STRUCTURE_TYPE_RELEASE_SWAPCHAIN_IMAGES_INFO_EXT: return sizeof(VkReleaseSwapchainImagesInfoEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_ATTACHMENT_BEGIN_INFO: return sizeof(VkRenderPassAttachmentBeginInfo);
        case VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO: return sizeof(VkRenderPassBeginInfo);
        case VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO: return sizeof(VkRenderPassCreateInfo);
        case VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO_2: return sizeof(VkRenderPassCreateInfo2);
        case VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_CONTROL_EXT: return sizeof(VkRenderPassCreationControlEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_FEEDBACK_CREATE_INFO_EXT: return sizeof(VkRenderPassCreationFeedbackCreateInfoEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_CREATE_INFO_EXT: return sizeof(VkRenderPassFragmentDensityMapCreateInfoEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_INPUT_ATTACHMENT_ASPECT_CREATE_INFO: return sizeof(VkRenderPassInputAttachmentAspectCreateInfo);
        case VK_STRUCTURE_TYPE_RENDER_PASS_MULTIVIEW_CREATE_INFO: return sizeof(VkRenderPassMultiviewCreateInfo);
        case VK_STRUCTURE_TYPE_RENDER_PASS_SAMPLE_LOCATIONS_BEGIN_INFO_EXT: return sizeof(VkRenderPassSampleLocationsBeginInfoEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_BEGIN_INFO_ARM: return sizeof(VkRenderPassStripeBeginInfoARM);
        case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_INFO_ARM: return sizeof(VkRenderPassStripeInfoARM);
        case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_SUBMIT_INFO_ARM: return sizeof(VkRenderPassStripeSubmitInfoARM);
        case VK_STRUCTURE_TYPE_RENDER_PASS_SUBPASS_FEEDBACK_CREATE_INFO_EXT: return sizeof(VkRenderPassSubpassFeedbackCreateInfoEXT);
        case VK_STRUCTURE_TYPE_RENDER_PASS_TRANSFORM_BEGIN_INFO_QCOM: return sizeof(VkRenderPassTransformBeginInfoQCOM);
        case VK_STRUCTURE_TYPE_RENDERING_AREA_INFO_KHR: return sizeof(VkRenderingAreaInfoKHR);
        case VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO: return sizeof(VkRenderingAttachmentInfo);
        case VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_INFO_EXT: return sizeof(VkRenderingFragmentDensityMapAttachmentInfoEXT);
        case VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR: return sizeof(VkRenderingFragmentShadingRateAttachmentInfoKHR);
        case VK_STRUCTURE_TYPE_RENDERING_INFO: return sizeof(VkRenderingInfo);
        case VK_STRUCTURE_TYPE_RESOLVE_IMAGE_INFO_2: return sizeof(VkResolveImageInfo2);
        case VK_STRUCTURE_TYPE_SAMPLE_LOCATIONS_INFO_EXT: return sizeof(VkSampleLocationsInfoEXT);
        case VK_STRUCTURE_TYPE_SAMPLER_BLOCK_MATCH_WINDOW_CREATE_INFO_QCOM: return sizeof(VkSamplerBlockMatchWindowCreateInfoQCOM);
        case VK_STRUCTURE_TYPE_SAMPLER_BORDER_COLOR_COMPONENT_MAPPING_CREATE_INFO_EXT: return sizeof(VkSamplerBorderColorComponentMappingCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SAMPLER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT: return sizeof(VkSamplerCaptureDescriptorDataInfoEXT);
        case VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO: return sizeof(VkSamplerCreateInfo);
        case VK_STRUCTURE_TYPE_SAMPLER_CUBIC_WEIGHTS_CREATE_INFO_QCOM: return sizeof(VkSamplerCubicWeightsCreateInfoQCOM);
        case VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_CREATE_INFO_EXT: return sizeof(VkSamplerCustomBorderColorCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SAMPLER_REDUCTION_MODE_CREATE_INFO: return sizeof(VkSamplerReductionModeCreateInfo);
        case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_CREATE_INFO: return sizeof(VkSamplerYcbcrConversionCreateInfo);
        case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_IMAGE_FORMAT_PROPERTIES: return sizeof(VkSamplerYcbcrConversionImageFormatProperties);
        case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_INFO: return sizeof(VkSamplerYcbcrConversionInfo);
        case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_YCBCR_DEGAMMA_CREATE_INFO_QCOM: return sizeof(VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM);
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_SCREEN_BUFFER_FORMAT_PROPERTIES_QNX: return sizeof(VkScreenBufferFormatPropertiesQNX);
    #endif
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_SCREEN_BUFFER_PROPERTIES_QNX: return sizeof(VkScreenBufferPropertiesQNX);
    #endif
    #ifdef VK_USE_PLATFORM_SCREEN_QNX
        case VK_STRUCTURE_TYPE_SCREEN_SURFACE_CREATE_INFO_QNX: return sizeof(VkScreenSurfaceCreateInfoQNX);
    #endif
        case VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO: return sizeof(VkSemaphoreCreateInfo);
        case VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR: return sizeof(VkSemaphoreGetFdInfoKHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR: return sizeof(VkSemaphoreGetWin32HandleInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_SEMAPHORE_GET_ZIRCON_HANDLE_INFO_FUCHSIA: return sizeof(VkSemaphoreGetZirconHandleInfoFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_SEMAPHORE_SIGNAL_INFO: return sizeof(VkSemaphoreSignalInfo);
        case VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO: return sizeof(VkSemaphoreSubmitInfo);
        case VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO: return sizeof(VkSemaphoreTypeCreateInfo);
        case VK_STRUCTURE_TYPE_SEMAPHORE_WAIT_INFO: return sizeof(VkSemaphoreWaitInfo);
        case VK_STRUCTURE_TYPE_SET_DESCRIPTOR_BUFFER_OFFSETS_INFO_EXT: return sizeof(VkSetDescriptorBufferOffsetsInfoEXT);
        case VK_STRUCTURE_TYPE_SET_LATENCY_MARKER_INFO_NV: return sizeof(VkSetLatencyMarkerInfoNV);
        case VK_STRUCTURE_TYPE_SHADER_CREATE_INFO_EXT: return sizeof(VkShaderCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO: return sizeof(VkShaderModuleCreateInfo);
        case VK_STRUCTURE_TYPE_SHADER_MODULE_IDENTIFIER_EXT: return sizeof(VkShaderModuleIdentifierEXT);
        case VK_STRUCTURE_TYPE_SHADER_MODULE_VALIDATION_CACHE_CREATE_INFO_EXT: return sizeof(VkShaderModuleValidationCacheCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_KHR: return sizeof(VkSharedPresentSurfaceCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_SPARSE_IMAGE_FORMAT_PROPERTIES_2: return sizeof(VkSparseImageFormatProperties2);
        case VK_STRUCTURE_TYPE_SPARSE_IMAGE_MEMORY_REQUIREMENTS_2: return sizeof(VkSparseImageMemoryRequirements2);
    #ifdef VK_USE_PLATFORM_GGP
        case VK_STRUCTURE_TYPE_STREAM_DESCRIPTOR_SURFACE_CREATE_INFO_GGP: return sizeof(VkStreamDescriptorSurfaceCreateInfoGGP);
    #endif
        case VK_STRUCTURE_TYPE_SUBMIT_INFO: return sizeof(VkSubmitInfo);
        case VK_STRUCTURE_TYPE_SUBMIT_INFO_2: return sizeof(VkSubmitInfo2);
        case VK_STRUCTURE_TYPE_SUBPASS_BEGIN_INFO: return sizeof(VkSubpassBeginInfo);
        case VK_STRUCTURE_TYPE_SUBPASS_DEPENDENCY_2: return sizeof(VkSubpassDependency2);
        case VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_2: return sizeof(VkSubpassDescription2);
        case VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_DEPTH_STENCIL_RESOLVE: return sizeof(VkSubpassDescriptionDepthStencilResolve);
        case VK_STRUCTURE_TYPE_SUBPASS_END_INFO: return sizeof(VkSubpassEndInfo);
        case VK_STRUCTURE_TYPE_SUBPASS_FRAGMENT_DENSITY_MAP_OFFSET_END_INFO_QCOM: return sizeof(VkSubpassFragmentDensityMapOffsetEndInfoQCOM);
        case VK_STRUCTURE_TYPE_SUBPASS_RESOLVE_PERFORMANCE_QUERY_EXT: return sizeof(VkSubpassResolvePerformanceQueryEXT);
        case VK_STRUCTURE_TYPE_SUBPASS_SHADING_PIPELINE_CREATE_INFO_HUAWEI: return sizeof(VkSubpassShadingPipelineCreateInfoHUAWEI);
        case VK_STRUCTURE_TYPE_SUBRESOURCE_HOST_MEMCPY_SIZE_EXT: return sizeof(VkSubresourceHostMemcpySizeEXT);
        case VK_STRUCTURE_TYPE_SUBRESOURCE_LAYOUT_2_KHR: return sizeof(VkSubresourceLayout2KHR);
        case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_EXT: return sizeof(VkSurfaceCapabilities2EXT);
        case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_KHR: return sizeof(VkSurfaceCapabilities2KHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_FULL_SCREEN_EXCLUSIVE_EXT: return sizeof(VkSurfaceCapabilitiesFullScreenExclusiveEXT);
    #endif
        case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_BARRIER_NV: return sizeof(VkSurfaceCapabilitiesPresentBarrierNV);
        case VK_STRUCTURE_TYPE_SURFACE_FORMAT_2_KHR: return sizeof(VkSurfaceFormat2KHR);
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_INFO_EXT: return sizeof(VkSurfaceFullScreenExclusiveInfoEXT);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_WIN32_INFO_EXT: return sizeof(VkSurfaceFullScreenExclusiveWin32InfoEXT);
    #endif
        case VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_COMPATIBILITY_EXT: return sizeof(VkSurfacePresentModeCompatibilityEXT);
        case VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_EXT: return sizeof(VkSurfacePresentModeEXT);
        case VK_STRUCTURE_TYPE_SURFACE_PRESENT_SCALING_CAPABILITIES_EXT: return sizeof(VkSurfacePresentScalingCapabilitiesEXT);
        case VK_STRUCTURE_TYPE_SURFACE_PROTECTED_CAPABILITIES_KHR: return sizeof(VkSurfaceProtectedCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_COUNTER_CREATE_INFO_EXT: return sizeof(VkSwapchainCounterCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR: return sizeof(VkSwapchainCreateInfoKHR);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_DISPLAY_NATIVE_HDR_CREATE_INFO_AMD: return sizeof(VkSwapchainDisplayNativeHdrCreateInfoAMD);
    #ifdef VK_USE_PLATFORM_ANDROID_KHR
        case VK_STRUCTURE_TYPE_SWAPCHAIN_IMAGE_CREATE_INFO_ANDROID: return sizeof(VkSwapchainImageCreateInfoANDROID);
    #endif
        case VK_STRUCTURE_TYPE_SWAPCHAIN_LATENCY_CREATE_INFO_NV: return sizeof(VkSwapchainLatencyCreateInfoNV);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_BARRIER_CREATE_INFO_NV: return sizeof(VkSwapchainPresentBarrierCreateInfoNV);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_FENCE_INFO_EXT: return sizeof(VkSwapchainPresentFenceInfoEXT);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODE_INFO_EXT: return sizeof(VkSwapchainPresentModeInfoEXT);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODES_CREATE_INFO_EXT: return sizeof(VkSwapchainPresentModesCreateInfoEXT);
        case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_SCALING_CREATE_INFO_EXT: return sizeof(VkSwapchainPresentScalingCreateInfoEXT);
    #ifdef VK_USE_PLATFORM_FUCHSIA
        case VK_STRUCTURE_TYPE_SYSMEM_COLOR_SPACE_FUCHSIA: return sizeof(VkSysmemColorSpaceFUCHSIA);
    #endif
        case VK_STRUCTURE_TYPE_TEXTURE_LOD_GATHER_FORMAT_PROPERTIES_AMD: return sizeof(VkTextureLODGatherFormatPropertiesAMD);
        case VK_STRUCTURE_TYPE_TILE_PROPERTIES_QCOM: return sizeof(VkTilePropertiesQCOM);
        case VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO: return sizeof(VkTimelineSemaphoreSubmitInfo);
        case VK_STRUCTURE_TYPE_VALIDATION_CACHE_CREATE_INFO_EXT: return sizeof(VkValidationCacheCreateInfoEXT);
        case VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT: return sizeof(VkValidationFeaturesEXT);
        case VK_STRUCTURE_TYPE_VALIDATION_FLAGS_EXT: return sizeof(VkValidationFlagsEXT);
        case VK_STRUCTURE_TYPE_VERTEX_INPUT_ATTRIBUTE_DESCRIPTION_2_EXT: return sizeof(VkVertexInputAttributeDescription2EXT);
        case VK_STRUCTURE_TYPE_VERTEX_INPUT_BINDING_DESCRIPTION_2_EXT: return sizeof(VkVertexInputBindingDescription2EXT);
    #ifdef VK_USE_PLATFORM_VI_NN
        case VK_STRUCTURE_TYPE_VI_SURFACE_CREATE_INFO_NN: return sizeof(VkViSurfaceCreateInfoNN);
    #endif
        case VK_STRUCTURE_TYPE_VIDEO_BEGIN_CODING_INFO_KHR: return sizeof(VkVideoBeginCodingInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_CAPABILITIES_KHR: return sizeof(VkVideoCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_CODING_CONTROL_INFO_KHR: return sizeof(VkVideoCodingControlInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_CAPABILITIES_KHR: return sizeof(VkVideoDecodeCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_CAPABILITIES_KHR: return sizeof(VkVideoDecodeH264CapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_DPB_SLOT_INFO_KHR: return sizeof(VkVideoDecodeH264DpbSlotInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PICTURE_INFO_KHR: return sizeof(VkVideoDecodeH264PictureInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PROFILE_INFO_KHR: return sizeof(VkVideoDecodeH264ProfileInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR: return sizeof(VkVideoDecodeH264SessionParametersAddInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR: return sizeof(VkVideoDecodeH264SessionParametersCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_CAPABILITIES_KHR: return sizeof(VkVideoDecodeH265CapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_DPB_SLOT_INFO_KHR: return sizeof(VkVideoDecodeH265DpbSlotInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PICTURE_INFO_KHR: return sizeof(VkVideoDecodeH265PictureInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PROFILE_INFO_KHR: return sizeof(VkVideoDecodeH265ProfileInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR: return sizeof(VkVideoDecodeH265SessionParametersAddInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR: return sizeof(VkVideoDecodeH265SessionParametersCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_INFO_KHR: return sizeof(VkVideoDecodeInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_DECODE_USAGE_INFO_KHR: return sizeof(VkVideoDecodeUsageInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_CAPABILITIES_KHR: return sizeof(VkVideoEncodeCapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_CAPABILITIES_KHR: return sizeof(VkVideoEncodeH264CapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_DPB_SLOT_INFO_KHR: return sizeof(VkVideoEncodeH264DpbSlotInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_GOP_REMAINING_FRAME_INFO_KHR: return sizeof(VkVideoEncodeH264GopRemainingFrameInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_NALU_SLICE_INFO_KHR: return sizeof(VkVideoEncodeH264NaluSliceInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PICTURE_INFO_KHR: return sizeof(VkVideoEncodeH264PictureInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PROFILE_INFO_KHR: return sizeof(VkVideoEncodeH264ProfileInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUALITY_LEVEL_PROPERTIES_KHR: return sizeof(VkVideoEncodeH264QualityLevelPropertiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_INFO_KHR: return sizeof(VkVideoEncodeH264RateControlInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_LAYER_INFO_KHR: return sizeof(VkVideoEncodeH264RateControlLayerInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_CREATE_INFO_KHR: return sizeof(VkVideoEncodeH264SessionCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR: return sizeof(VkVideoEncodeH264SessionParametersAddInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR: return sizeof(VkVideoEncodeH264SessionParametersCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_FEEDBACK_INFO_KHR: return sizeof(VkVideoEncodeH264SessionParametersFeedbackInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_GET_INFO_KHR: return sizeof(VkVideoEncodeH264SessionParametersGetInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_CAPABILITIES_KHR: return sizeof(VkVideoEncodeH265CapabilitiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_DPB_SLOT_INFO_KHR: return sizeof(VkVideoEncodeH265DpbSlotInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_GOP_REMAINING_FRAME_INFO_KHR: return sizeof(VkVideoEncodeH265GopRemainingFrameInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_NALU_SLICE_SEGMENT_INFO_KHR: return sizeof(VkVideoEncodeH265NaluSliceSegmentInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PICTURE_INFO_KHR: return sizeof(VkVideoEncodeH265PictureInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PROFILE_INFO_KHR: return sizeof(VkVideoEncodeH265ProfileInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUALITY_LEVEL_PROPERTIES_KHR: return sizeof(VkVideoEncodeH265QualityLevelPropertiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_INFO_KHR: return sizeof(VkVideoEncodeH265RateControlInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_LAYER_INFO_KHR: return sizeof(VkVideoEncodeH265RateControlLayerInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_CREATE_INFO_KHR: return sizeof(VkVideoEncodeH265SessionCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR: return sizeof(VkVideoEncodeH265SessionParametersAddInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR: return sizeof(VkVideoEncodeH265SessionParametersCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_FEEDBACK_INFO_KHR: return sizeof(VkVideoEncodeH265SessionParametersFeedbackInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_GET_INFO_KHR: return sizeof(VkVideoEncodeH265SessionParametersGetInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_INFO_KHR: return sizeof(VkVideoEncodeInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR: return sizeof(VkVideoEncodeQualityLevelInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_PROPERTIES_KHR: return sizeof(VkVideoEncodeQualityLevelPropertiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_INFO_KHR: return sizeof(VkVideoEncodeRateControlInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_LAYER_INFO_KHR: return sizeof(VkVideoEncodeRateControlLayerInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_FEEDBACK_INFO_KHR: return sizeof(VkVideoEncodeSessionParametersFeedbackInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_GET_INFO_KHR: return sizeof(VkVideoEncodeSessionParametersGetInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_ENCODE_USAGE_INFO_KHR: return sizeof(VkVideoEncodeUsageInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_END_CODING_INFO_KHR: return sizeof(VkVideoEndCodingInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_FORMAT_PROPERTIES_KHR: return sizeof(VkVideoFormatPropertiesKHR);
        case VK_STRUCTURE_TYPE_VIDEO_INLINE_QUERY_INFO_KHR: return sizeof(VkVideoInlineQueryInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_INFO_KHR: return sizeof(VkVideoPictureResourceInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_PROFILE_INFO_KHR: return sizeof(VkVideoProfileInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_PROFILE_LIST_INFO_KHR: return sizeof(VkVideoProfileListInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_INFO_KHR: return sizeof(VkVideoReferenceSlotInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_SESSION_CREATE_INFO_KHR: return sizeof(VkVideoSessionCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_SESSION_MEMORY_REQUIREMENTS_KHR: return sizeof(VkVideoSessionMemoryRequirementsKHR);
        case VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_CREATE_INFO_KHR: return sizeof(VkVideoSessionParametersCreateInfoKHR);
        case VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_UPDATE_INFO_KHR: return sizeof(VkVideoSessionParametersUpdateInfoKHR);
    #ifdef VK_USE_PLATFORM_WAYLAND_KHR
        case VK_STRUCTURE_TYPE_WAYLAND_SURFACE_CREATE_INFO_KHR: return sizeof(VkWaylandSurfaceCreateInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_KHR: return sizeof(VkWin32KeyedMutexAcquireReleaseInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_NV: return sizeof(VkWin32KeyedMutexAcquireReleaseInfoNV);
    #endif
    #ifdef VK_USE_PLATFORM_WIN32_KHR
        case VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR: return sizeof(VkWin32SurfaceCreateInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET: return sizeof(VkWriteDescriptorSet);
        case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_KHR: return sizeof(VkWriteDescriptorSetAccelerationStructureKHR);
        case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_NV: return sizeof(VkWriteDescriptorSetAccelerationStructureNV);
        case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_INLINE_UNIFORM_BLOCK: return sizeof(VkWriteDescriptorSetInlineUniformBlock);
    #ifdef VK_USE_PLATFORM_XCB_KHR
        case VK_STRUCTURE_TYPE_XCB_SURFACE_CREATE_INFO_KHR: return sizeof(VkXcbSurfaceCreateInfoKHR);
    #endif
    #ifdef VK_USE_PLATFORM_XLIB_KHR
        case VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR: return sizeof(VkXlibSurfaceCreateInfoKHR);
    #endif
        case VK_STRUCTURE_TYPE_LOADER_INSTANCE_CREATE_INFO: return sizeof(VkLayerInstanceCreateInfo);
        case VK_STRUCTURE_TYPE_LOADER_DEVICE_CREATE_INFO: return sizeof(VkLayerDeviceCreateInfo);
        default:
            unreachable("Undefined struct type.");
        }
    }

    const char *
    vk_ObjectType_to_ObjectName(VkObjectType type)
    {
        switch((int)type) {
        case 1:
            return "VkInstance";
        case 2:
            return "VkPhysicalDevice";
        case 3:
            return "VkDevice";
        case 4:
            return "VkQueue";
        case 5:
            return "VkSemaphore";
        case 6:
            return "VkCommandBuffer";
        case 7:
            return "VkFence";
        case 8:
            return "VkDeviceMemory";
        case 9:
            return "VkBuffer";
        case 10:
            return "VkImage";
        case 11:
            return "VkEvent";
        case 12:
            return "VkQueryPool";
        case 13:
            return "VkBufferView";
        case 14:
            return "VkImageView";
        case 15:
            return "VkShaderModule";
        case 16:
            return "VkPipelineCache";
        case 17:
            return "VkPipelineLayout";
        case 18:
            return "VkRenderPass";
        case 19:
            return "VkPipeline";
        case 20:
            return "VkDescriptorSetLayout";
        case 21:
            return "VkSampler";
        case 22:
            return "VkDescriptorPool";
        case 23:
            return "VkDescriptorSet";
        case 24:
            return "VkFramebuffer";
        case 25:
            return "VkCommandPool";
        case 1000000000:
            return "VkSurfaceKHR";
        case 1000001000:
            return "VkSwapchainKHR";
        case 1000002000:
            return "VkDisplayKHR";
        case 1000002001:
            return "VkDisplayModeKHR";
        case 1000011000:
            return "VkDebugReportCallbackEXT";
        case 1000023000:
            return "VkVideoSessionKHR";
        case 1000023001:
            return "VkVideoSessionParametersKHR";
        case 1000029000:
            return "VkCuModuleNVX";
        case 1000029001:
            return "VkCuFunctionNVX";
        case 1000085000:
            return "VkDescriptorUpdateTemplate";
        case 1000128000:
            return "VkDebugUtilsMessengerEXT";
        case 1000150000:
            return "VkAccelerationStructureKHR";
        case 1000156000:
            return "VkSamplerYcbcrConversion";
        case 1000160000:
            return "VkValidationCacheEXT";
        case 1000165000:
            return "VkAccelerationStructureNV";
        case 1000210000:
            return "VkPerformanceConfigurationINTEL";
        case 1000268000:
            return "VkDeferredOperationKHR";
        case 1000277000:
            return "VkIndirectCommandsLayoutNV";
        case 1000295000:
            return "VkPrivateDataSlot";
        case 1000307000:
            return "VkCudaModuleNV";
        case 1000307001:
            return "VkCudaFunctionNV";
        case 1000366000:
            return "VkBufferCollectionFUCHSIA";
        case 1000396000:
            return "VkMicromapEXT";
        case 1000464000:
            return "VkOpticalFlowSessionNV";
        case 1000482000:
            return "VkShaderEXT";
        default:
            return "Unknown VkObjectType value.";
        }
    }
